#!/usr/bin/env perl
#Generate makefiles for FastTranslateChart,MER,MergeRules
use strict;
use Getopt::Long;


#hardcoded content about the package structure
#sub directories
my @dirs = ('Share', 'LM', 'BleuSearch', 'Lattice', 'SRI', 'FORTFY22');

#sub packages
my $lattice_files = "FastLatticeChartParser.o ParserInfo.o ChartParserHyp.o ChartParserHypBeam.o UniqueCompleteHyps.o Rule.o BTree.o NBestList.o Vocabulary.o CompactTSequence.o CombinationBeam.o ";

my $lm_files = "CachedSriLM.o LanguageModelSAMT.o ";

my $share_files = "Globals.o Parameter.o";

my $sri_files = "Array.o Debug.o Discount.o File.o LHash.o LHashTrie.o LM.o LMStats.o Map.o Map2.o MemStats.o NBest.o NBestSet.o NgramLM.o NgramStats.o NgramStatsDouble.o NgramStatsFloat.o NgramStatsInt.o Prob.o RefList.o SArray.o SArrayTrie.o SubVocab.o TextStats.o Trie.o Vocab.o VocabDistance.o VocabMap.o VocabMultiMap.o WordAlign.o zio.o ";

my $bleu_files = "MetricSearch.o MetricRaw.o BleuSearch.o TERSearch.o EvaluatorTER.o Optimize.o OptimizeMER.o ";

#this is the location of the source files
my $srcdir = "./src";
#location of the Berkeledy DB header files that you have installed on the system
my $bdb = "/usr/local/BerkeleyDB.4.4/";

#system options to allow g++ to verify that your code is compilable based on dependancies
my $system = " -DSTDC_HEADERS=1 -DHAVE_SYS_TYPES_H=1 -DHAVE_SYS_STAT_H=1 -DHAVE_STDLIB_H=1 -DHAVE_STRING_H=1 -DHAVE_MEMORY_H=1 -DHAVE_STRINGS_H=1 -DHAVE_INTTYPES_H=1 -DHAVE_STDINT_H=1 -DHAVE_UNISTD_H=1 -DHAVE_DLFCN_H=1 -DSTDC_HEADERS=1 -DHAVE_STACK=1 -DHAVE_STRING=1 -DHAVE_LIST=1 -DHAVE_VECTOR=1 -DHAVE_MAP=1 -DHAVE_EXT_HASH_MAP=1 -DHAVE_EXT_HASH_SET=1 -DHAVE_ZOPEN";

my $cxx = "g++";
my $cxxflags = "-DNDEBUG -O2";
die if !GetOptions("srcdir=s" => \$srcdir, 'bdb=s' => \$bdb, 'system=s' => \$system, 'cxx=s' => \$cxx, 'cxxflags=s' => \$cxxflags);

my $bdb_include = "$bdb/include/";
my $bdb_lib = "$bdb/lib/";

#we are going to use g++ to find the header dependencies and
#generate the makefile. Make sure the include path is set
#for gcc to each of the subdirs
my @includedirs = map { "-I".$srcdir."/".$_; } @dirs;
my $include = join(" ",@includedirs);
$include = $include." -I$bdb_include";

#find all the source files and run the g++ tool on them - but make sure
#we dont include the temporary emacs ~ files
my $line = "find $srcdir\/ -name \"*.cc\" | fgrep -v '~'  | xargs g++ $include $system -MM";

print "CXX = $cxx\n";
print "CXXFLAGS = $cxxflags $include $system -DINSTANTIATE_TEMPLATES\n";
print "LD_FLAGS = -L$bdb_lib -ldb_cxx\n";

#the binaries that we are going to produce

print "all : FastTranslateChart MER MergeRules\n\n";

print "FastTranslateChart : FastTranslateChart.o $lattice_files $bleu_files $share_files $lm_files $sri_files \n\t \$(CXX) \$(CXXFLAGS) \$(LD_FLAGS) FastTranslateChart.o $lattice_files $bleu_files $share_files $lm_files $sri_files -o FastTranslateChart \n\n";

print "MER : MER.o $lattice_files $bleu_files $share_files $lm_files $sri_files \n\t \$(CXX) \$(CXXFLAGS) \$(LD_FLAGS) MER.o $lattice_files $bleu_files $share_files $lm_files $sri_files -o MER \n\n";

print "MergeRules: MergeRules.o \n\t \$(CXX) \$(CXXFLAGS) \$(LD_FLAGS) MergeRules.o -o MergeRules \n\n";

#generic rules to build o files from cc and c files
print "%.o: %.cc \n\t \$(CXX) -c \$(CXXFLAGS) -o \$@ \$<\n";
print "%.o: %.c \n\t \$(CXX) -c \$(CXXFLAGS) -o \$@ \$<\n";

#figure out which files are in the dependency structures
my @lines =  `$line`;
$line = "find $srcdir\/ -name \"*.c\" | fgrep -v '~'  | xargs g++ $include $system -MM";
my @lines2 =  `$line`;
foreach my $l2 (@lines2){
    push(@lines, $l2);
}

#command to compile
my $cmd = "\t \$(CXX) -c \$(CXXFLAGS)  -o \$@ \$<";

my $count = 0;
foreach my $e (@lines){

    if (($e =~ m/\.o: /)){
	if ($count>0){
	    print $cmd,"\n";
	}
	$count++;
    }
    print $e;
}
#the last command
print $cmd,"\n";




