#!/usr/bin/env perl
#Analyze the log file to get a perspective of pruning decisions made
#Re-written to directly read logged parameters, instead of having to specify them here every time we change them
use Getopt::Long;


my %sentenceLevelStats = ();
my %corpusLevelStats = ();

my $firstDone = 0;
while (<STDIN>){
  
  #get the first element of the n-best list
  if (m/^ScaledTMLMCost: ([0-9\.\-]+)/){
    if ($firstDone==0){
      my $hypCost = $1;
      print "BestHypCost: $hypCost\n";
      if (exists $corpusLevelStats{"modelCost:"}){
	$corpusLevelStats{"modelCost:"} += $hypCost;
      }else{
	$corpusLevelStats{"modelCost:"} = $hypCost;
      }
      $firstDone = 1;
    }
  }
  
  #get stats for each cell and accumulate them
  if (m/Log: numHyps/){
    #print "Line: $_\n";
    UpdateStats($_, \%corpusLevelStats);
  }
  #get nbest stats for each sentence and accumulate them
  if (m/Log: NBest: (.*)/){
      UpdateStats($1, \%corpusLevelStats);
  }

  #to know when each sentence is done
  if (/^Retrieving/){
    $firstDone = 0;
  }

  #one run complete display statistics
  if (/^TrainTranslationScore: ([0-9e\.\-]+) ([0-9e\.\-]+)/){
    print " ********** Set Summary ************* \n";
    print $_;

    my $score = $1;
    $score =~ s/\s+$//g;

    my $lpen = $2;
    $lpen =~ s/\s+$//g;


    $corpusLevelStats{"bleuScore:"} = $score;
    $corpusLevelStats{"lpen:"} = $lpen;
    PrintStats(\%corpusLevelStats);
    ClearStats(\%corpusLevelStats);

  }
  
  #TIME_CPU one translation loop:NumSecsDecoding_CPU: 91.85 NumSecsNBest_CPU: 0.23 NumSecsSuppNBest_CPU: 0 Spawn: 0
  if (/^TIME_CPU one translation loop:NumSecsDecoding_CPU: ([0-9\-\.]+) NumSecsNBest_CPU: ([0-9\-\.]+) NumSecsSuppNBest_CPU: ([0-9\-\.]+) Spawn: ([0-9\-\.]+) NumLMCalls: ([0-9\-\.]+) NumLMCacheHit: ([0-9\-\.]+)/){
      
      $corpusLevelStats{"numSecDecoding:"} = abs($1);
      $corpusLevelStats{"numSecNBest:"} = abs($2);
      $corpusLevelStats{"spawn:"} = abs($4);
      $corpusLevelStats{"lmcall:"} = abs($5);
      $corpusLevelStats{"lmhit:"} = abs($6);
      print $_;
  }
  
}

sub ClearStats{
  my $hash_ref = shift;
  while (my($k, $v) = each %$hash_ref){
    $hash_ref->{$key} = 0;
  }
}

sub PrintStats{
  my $hash_ref = shift;
  while (my($k, $v) = each %$hash_ref){
    print "key: $k value: $v\n";
  }
}

sub PrintStats{
  my $hash_ref = shift;
  my @statsOrder = ("bleuScore:", "lpen:", "modelCost:", "numSecDecoding:", "numSecNBest:","numHyps:", "m_hypsAdded:", "m_numTopCalls:", "m_hypsDeleted:", "m_hypsDeletedHistogram:", "m_hypsDeletedThreshold:", "numAddedPostFailure:", "m_numTuplesProposed:", "m_allPossibleExpansions:", "m_topCallsKBest:", "lmcall:", "lmhit:");
  foreach my $key  (@statsOrder){
    print "$key ", $hash_ref->{$key}, " "; 
  }
  print "\n";
}

sub UpdateStats{
  my $line = shift;
  my $hash_ref = shift;
  my @parts = split(/\s+/, $line);
  my $lastStat = "";

  foreach my $part (@parts){
    if ($part =~ m/([0-9]+)/){
      if ($lastStat eq ""){
	die "lastStat was still blank";
      }
      if (exists $hash_ref->{$lastStat}){
	$hash_ref->{$lastStat} += $1;
      }else{
	$hash_ref->{$lastStat} = $1;
      }
    }else{
      $lastStat = $part;
    }
  }
}


sub InitArr{
  my $arrRef = shift;
  my $numSrcWords = shift;
  @{$arrRef} = ();
  for (my $i = 0; $i<$numSrcWords; $i++){
    my @oneRow = ();
    for (my $j=0; $j<$numSrcWords; $j++){
      push(@oneRow, "0");
    }
    $arrRef->[$i] = \@oneRow;
  }
}

# args: reference to the array
sub PrintArr{
    my $arrRef = shift;
    for (my $i = 0; $i<scalar(@{$arrRef}); $i++){     
      #print $i+1, "\t| ", join("\t", @{$arrRef->[$i]}), "\n";
      print $i+1, " ";
      if ($i+1<10){
	print " ";
      }
      PrintFix($arrRef->[$i]);
      print "\n";
    }
    print "\n";
    
}

#fixed width print
sub PrintFix{
  my $arrRef = shift;
  my $sz = scalar(@{$arrRef});
  for (my $i=0; $i<$sz; $i++){
    my $num = $arrRef->[$i];
    if ($num<10){
      print $num, "  ";
    }elsif ($num<100){
      print $num, " ";      
    }elsif ($num<1000){
      print $num;    
    }elsif ($num<10000){
      my $dec = $num/1000;
      printf("%1.1f", $dec);
    }else{
      print $num;
    }
    print " ";    
  }
}
