#!/bin/tcsh -f 
#Build models from the source provided and run the mer experiment on this data
#STTK binaries have to be in the path! (TranslateTS is needed by this script and GraphAlignO is needed by CreateSplitTM.pl)
set translatets =  TranslateTS
#set translatets = /nfs/progress2/Users/Ashish/working/CVTesting/CheckOut/STTK2K4/bin/TranslateTS
#set translatets = /nfs/dish8/Andreas/STTK2K4/bin/TranslateTS
set single_score_testing = 0

set SOURCE_TM = $1;
echo 'SOURCE_TM' $SOURCE_TM
set TARGET_TM = $2;
echo 'TARGET_TM' $TARGET_TM

set TARGET_LM = $3;
echo 'TARGET_LM' $TARGET_LM

set SOURCE_HELDOUT = $4;
echo 'SOURCE_HELDOUT' $SOURCE_HELDOUT

set TARGET_HELDOUT = $5;
echo 'TARGET_HELDOUT' $TARGET_HELDOUT

set PHRASE_LIST = $6;
echo 'PHRASE_LIST' ${PHRASE_LIST}

set STAGING_DIR = $7;
echo 'STAGING_DIR' ${STAGING_DIR}


#runtime options
set DoMakeStagingDirAndCreateVoc = $8

set DoBuildModels = $9

set DoRunMER = $10

set DoGenPhrases = $11

set DoTestGenPhrases = $12

set DoTestGenPhrasesMER = $13

set ScoringMetric = $14 # if this option is empty, we do NIST optimization and testing
if ($ScoringMetric != ) then
    # set ScoringMetric = IBMBLEU
else
    # default: NIST
    set ScoringMetric = NIST
endif

# Note: input the params in ARG15 in quotes so that they all go into $15 !
set translatetsparams = `echo "$15"|replace _ ' '` 
# e.g., --NumReferences=16_--LengthOfReorderingWindow=3_--ReorderingsPerWord=0.2 
#(also multiple params allowed, in that case separate them by _)
echo "Translatetsparams: $translatetsparams"

set phrasetablefile = $16

#set referencecount = $16
#if ($referencecount != ) then
#    #
#else
#    # default: 1
#    set referencecount = 1
#endif



#setenv LC_ALL UTF8
unsetenv DISPLAY # important to avoid the startup slow-down of Matlab

    
set VOC = ${STAGING_DIR}/All.voc

if ($DoMakeStagingDirAndCreateVoc) then

    #rm -rf ${STAGING_DIR}
    mkdir -p ${STAGING_DIR}

    echo "Building voc All.voc";
    cat ${SOURCE_TM} ${TARGET_TM} ${TARGET_LM} ${SOURCE_HELDOUT} ${TARGET_HELDOUT} | ~vogel/bin/Tools/MakeVoc > ${STAGING_DIR}/All.voc

endif

if ($DoBuildModels) then
    echo "Building LM with training target direction";
    perl -S CreateSplitLM.pl ${TARGET_LM} ${TARGET_LM}.srilm
    
#    if ($DoBuildTM) then
	echo "Building TMs: perl -S CreateSplitTM.pl ${SOURCE_TM} ${TARGET_TM} ${VOC} ${SOURCE_TM}.T"
	perl -S CreateSplitTM.pl ${SOURCE_TM} ${TARGET_TM} ${VOC} ${SOURCE_TM}.T
	
	grep '^10'  ${SOURCE_TM}.T
	grep '^10'  ${SOURCE_TM}.T.rev
#    endif
    
    echo "Building SA with $SOURCE_TM";
    ~joy/Release/SuffixArray/IndexSA ${SOURCE_TM}

endif	
    
    
set DoGenerateParamFile=1
#generate the params file and run the mer
if ($DoGenerateParamFile) then
    echo "Building parameter file for mer training and testing"
    
    set MERPARAMS =  "${STAGING_DIR}/mer.param"
    rm -f ${MERPARAMS}

    set MERFORWARD = "${STAGING_DIR}/mer-arabicToEnglish.param"
    rm -f ${MERFORWARD}
    #    set MERREVERSE = "${STAGING_DIR}/mer-englishToArabic.param"
    #    rm -f ${MERREVERSE}
    
    #basic parameters
    echo "SplitCharacters #" >> $MERPARAMS
    echo "LM SriLM " >> ${MERPARAMS}
    echo "TraceLevel 0.0 " >> ${MERPARAMS}
    echo "DisplayLatticeStatistics 0 " >> ${MERPARAMS}
    echo "DisplayNBestData 0 " >> ${MERPARAMS}
    
    #ordering parameters
	# more scores prefixed: target-cond. rel.freq, source-cond. rf, ...
#	set TMLIST = 2.817130_0.986065_-0.134580_0.294462_0.296005_0.986820_-0.419237_0.023931_0.036294_0.039189_0.090000_0.027000
	set TMLIST = 0_0.30082_-0.287502_-0.0151103_0.0189121_0.000174896_-0.00724393
    echo 'TMScaleList' $TMLIST >> ${MERPARAMS}
    echo "LMScale 2.2" >> ${MERPARAMS}
#    echo "ALScale 0.0" >> ${MERPARAMS} # no reordering
    echo "ALScale 0.1" >> ${MERPARAMS} # reordering allowed 
    echo "SLScale -2.6" >> ${MERPARAMS} 
    echo "PLScale 0.25" >> ${MERPARAMS}
    ##--LMScale 2.6516 --ALScale 0.0001 --SLScale -1.76328 --PLScale 1.02184 --TMScaleList -0.032271_2.3316_0.377178_-0.226234_0.805255_1.67474

    echo "NormalizingScript /afs/cs/usr/ashishv/DoNothing.pl " >> ${MERPARAMS}
#    echo "NormalizingScript /afs/cs.cmu.edu/user/zollmann/transtacbin/BTEC-EnglishNormalizeHyps.pl " >> ${MERPARAMS} # make sure you have Perl 5.8.7 or higher for this; otherwise it will run but give false results!

    echo "PhraseAlignBeamSize                 3.0" >> ${MERPARAMS}
#    echo "PhraseAlignBeamSize                 1.3" >> ${MERPARAMS}
    echo "PhraseAlignScaleList                2.0_2.0_0.5_0.5_0.1_0.1" >> ${MERPARAMS}
    
    #search parameters
    echo "CompareHypFeatures 7.0 " >> ${MERPARAMS}
    echo "PruneHypFeatures 1.0 " >> ${MERPARAMS}
    echo "LengthOfReorderingWindow 1 " >> ${MERPARAMS}
    echo "ReorderingsPerWord 0.0 " >> ${MERPARAMS}
    echo "BeamSize 20.0 " >> ${MERPARAMS}

    
    #mer parameters
    echo "MEROptimize 1 " >> ${MERPARAMS}
    echo "NBest 2000.0 " >> ${MERPARAMS}
    echo "IterationLimit 7 " >> ${MERPARAMS}
    echo "ParameterRestrictionFile ${STAGING_DIR}/AllOpen-6Scores.opt " >> ${MERPARAMS}
    #create the all open 6 param file
    rm -f ${STAGING_DIR}/AllOpen-6Scores.opt

    # last 7 params are for phrase table
	echo "0 0 -6 -6    0.0 0.010 -30 -30 -30 -30 -30" >> ${STAGING_DIR}/AllOpen-6Scores.opt
	echo "30 6  6  6  0.0 100.0 +30 +30 +30 +30 +30" >> ${STAGING_DIR}/AllOpen-6Scores.opt # reordering

    cp ${STAGING_DIR}/AllOpen-6Scores.opt .
    
    echo "Vocabulary $VOC " >> ${MERPARAMS}
    echo "WordLM $TARGET_LM.srilm" >> ${MERPARAMS}


    #copy over the base stuff to $MERFORWARD, to be used for phrase translation
    cp ${MERPARAMS} ${MERFORWARD}

    #pesa parameters
    echo "LambdaLexicon 0.5 " >> ${MERFORWARD}
    echo "SA_HighestFrequency 10000000.0 " >> ${MERFORWARD}
    echo "SA_MaxOccurence    10.0 " >> ${MERFORWARD}
#    echo "SA_ShortesPhraseLength 1.0 " >> ${MERFORWARD}
    echo "SA_ShortestPhraseLength 1.0 " >> ${MERFORWARD}
    echo "SourceFileSA $SOURCE_TM" >> ${MERFORWARD}
    echo "SourceFile $SOURCE_TM" >> ${MERFORWARD}
    echo "TargetFile $TARGET_TM" >> ${MERFORWARD}
    
    echo "Lexicon $SOURCE_TM.T" >> ${MERFORWARD}
    echo "ReverseLexicon $SOURCE_TM.T.rev" >> ${MERFORWARD}

    #echo "ReferenceList $TARGET_HELDOUT" >> ${MERFORWARD}

    #add some more parameters to MERPARAMS, which will be used for MER training and testing
    #    echo "RemoveUnt 1" >> $MERPARAMS # remove untranslated words
    echo "RemoveUnt 0" >> $MERPARAMS # remove untranslated words
    
    
    
endif #DoGenerateParamFile






if ($DoGenPhrases) then
    echo "DoGenPhrases"


    #use translate phrase here (represented by parameter --SA_ShortesPhraseLength=-1 [sic!] - corrected now to SA_ShortestPhraseLength), since we just want the complete phrase,not the sub phrases

    echo "$translatets -f ${MERFORWARD} --SA_ShortestPhraseLength=-1 --SentenceList=${PHRASE_LIST}  --DisplayLatticeStatistics=1 --BeamSize=1 --MEROptimize=0 --NBest=1 $translatetsparams > ${STAGING_DIR}/forwards.tdlog"
    
    $translatets -f ${MERFORWARD} --SA_ShortestPhraseLength=-1 --SentenceList=${PHRASE_LIST}  --DisplayLatticeStatistics=1 --BeamSize=1 --MEROptimize=0 --NBest=1 $translatetsparams > ${STAGING_DIR}/forwards.tdlog

#    $translatets -f ${MERFORWARD} --SA_ShortesPhraseLength=-1 --SentenceList=${PHRASE_LIST}  --DisplayLatticeStatistics=1 --BeamSize=1 --MEROptimize=0 --NBest=1 --LMScale 2.6516 --ALScale 0.0001 --SLScale -1.76328 --PLScale 1.02184 --TMScaleList -0.032271_2.3316_0.377178_-0.226234_0.805255_1.67474 > ${STAGING_DIR}/forwards.tdlog

    Log2TD-MS-fast.sh ${STAGING_DIR}/forwards.tdlog > ${STAGING_DIR}/forwards.table.raw

    endif
endif








#this just tests the impact of using the fixed td
if ($DoTestGenPhrases) then 
    echo "DoTestGenPhrases"
    #create the param files for the phrases
    #rm -f ${STAGING_DIR}/*TD*
    echo "TransducerList ${STAGING_DIR}/TD.List" >  ${STAGING_DIR}/TD.Param

    echo "${STAGING_DIR}/TD.bi" >  ${STAGING_DIR}/TD.List
    echo "Name PESA" >   ${STAGING_DIR}/TD.bi
    echo "File $phrasetablefile" >>  ${STAGING_DIR}/TD.bi
    echo "LambdaSmooth 0.0" >>  ${STAGING_DIR}/TD.bi
    echo "Renormalize 0.0" >>  ${STAGING_DIR}/TD.bi

    #run the optimization once

    echo "perl /afs/cs.cmu.edu/usr/ashishv/www/GenerateNISTInfoGains.pl ${TARGET_HELDOUT} 1 4 > ${TARGET_HELDOUT}.gains"
    perl /afs/cs.cmu.edu/usr/ashishv/www/GenerateNISTInfoGains.pl ${TARGET_HELDOUT} 1 4 > ${TARGET_HELDOUT}.gains


    if ($DoTestGenPhrasesMER) then 
	echo "DoTestGenPhrasesMER"
	echo "Using optimizeV5$ScoringMetric.m"
	cp -f ~ashishv/www/optimizeV5IBMBLEU.m ./optimizeV2IBMBLEU.m
	cp -f ~ashishv/www/optimizeV5NIST.m ./optimizeV2NIST.m

	echo "$translatets -f ${MERPARAMS} --TransMemParams=${STAGING_DIR}/TD.Param --ScoringMetric=$ScoringMetric --GainFile ${TARGET_HELDOUT}.gains --SentenceList=$SOURCE_HELDOUT --ReferenceList $TARGET_HELDOUT --MEROptimize 1 $translatetsparams > ${STAGING_DIR}/fixed-$ScoringMetric.log"

    	$translatets -f ${MERPARAMS} --TransMemParams=${STAGING_DIR}/TD.Param --ScoringMetric=$ScoringMetric --GainFile ${TARGET_HELDOUT}.gains --SentenceList=$SOURCE_HELDOUT --ReferenceList $TARGET_HELDOUT --MEROptimize 1 $translatetsparams | tee ${STAGING_DIR}/fixed-$ScoringMetric.log | egrep '^TrainTranslationScore|^Final'

	#	grep '^TrainTransl' $STAGING_DIR/fixed-$ScoringMetric.log

	#grep '^Got' ${STAGING_DIR}/fixed-$ScoringMetric.log | tail -n 1 | perl -S ExtractParams2.pl 1 > ${STAGING_DIR}/fixed-$ScoringMetric.params
	grep '^Got' ${STAGING_DIR}/fixed-$ScoringMetric.log | perl -Sw ExtractParams2.pl 1 > ${STAGING_DIR}/fixed-$ScoringMetric.alliterationparams
	tail -n 1 < ${STAGING_DIR}/fixed-$ScoringMetric.alliterationparams > ${STAGING_DIR}/fixed-$ScoringMetric.params 

	
	# for single score mode:
	grep '^Got' ${STAGING_DIR}/fixed-$ScoringMetric.log | tail -n 1 | perl -S ExtractParams2-RestrictPDA.pl 0.2 > ${STAGING_DIR}/fixed-PDA-$ScoringMetric.params
    endif
	

    echo "\n\n\n\nDoing $ScoringMetric-testing (translating $SOURCE_HELDOUT, comparing to references in $TARGET_HELDOUT)"
    
    perl -e "print join ' ',(1..`wc -l <$STAGING_DIR/fixed-$ScoringMetric.alliterationparams`)" > $STAGING_DIR/tmp
    #uncomment this if you want to test scores for all iterations:
    #    foreach iteration_number (`cat $STAGING_DIR/tmp`)
    foreach iteration_number (`cat $STAGING_DIR/tmp | replace '^.* ' ''`)
	set tmp_score_params = `selectline $iteration_number < $STAGING_DIR/fixed-$ScoringMetric.alliterationparams`
	echo "\n\n\nIteration $iteration_number, params $tmp_score_params"
	echo "$translatets -f ${MERPARAMS} --TransMemParams=${STAGING_DIR}/TD.Param --ScoringMetric=$ScoringMetric --GainFile ${TARGET_HELDOUT}.gains --SentenceList=$SOURCE_HELDOUT --ReferenceList $TARGET_HELDOUT --NBest=1000 --MEROptimize=0 $tmp_score_params $translatetsparams --GenOracleScores=1 > ${STAGING_DIR}/fixed-base-$ScoringMetric.1Kbest.log"

	$translatets -f ${MERPARAMS} --TransMemParams=${STAGING_DIR}/TD.Param --ScoringMetric=$ScoringMetric --GainFile ${TARGET_HELDOUT}.gains --SentenceList=$SOURCE_HELDOUT --ReferenceList $TARGET_HELDOUT --NBest=1000 --MEROptimize=0 $tmp_score_params $translatetsparams --GenOracleScores=1 > ${STAGING_DIR}/fixed-base-$ScoringMetric.1Kbest.log

	grep '@PESA' < ${STAGING_DIR}/fixed-base-$ScoringMetric.1Kbest.log | perl -S AnalyseTranslations.pl
	echo -n "$ScoringMetric "
	grep 'TranslationScore' ${STAGING_DIR}/fixed-base-$ScoringMetric.1Kbest.log
    end

    if ($single_score_testing) then
	echo convert the phrase table into a single score
	grep '^Got' ${STAGING_DIR}/fixed-$ScoringMetric.log | tail -n 1  > ${STAGING_DIR}/fixed-$ScoringMetric.params.raw
	cat ${STAGING_DIR}/forwards.table.raw | perl -S ConvertMultiScoreAddPhrase2.pl ${STAGING_DIR}/fixed-$ScoringMetric.params.raw 0.2 >! ${STAGING_DIR}/forwards.table.$ScoringMetric.singleScore
	cp ${STAGING_DIR}/forwards.table.$ScoringMetric.singleScore ${STAGING_DIR}/forwards.table.td

	echo "$translatets -f ${MERPARAMS} --TransMemParams=${STAGING_DIR}/TD.Param --ScoringMetric=$ScoringMetric --GainFile ${TARGET_HELDOUT}.gains --SentenceList=$SOURCE_HELDOUT --ReferenceList $TARGET_HELDOUT --NBest=1 --MEROptimize=0 `cat ${STAGING_DIR}/fixed-PDA-$ScoringMetric.params` $translatetsparams --GenOracleScores=1 > ${STAGING_DIR}/fixed-PDA-$ScoringMetric.1best.log"

	$translatets -f ${MERPARAMS} --TransMemParams=${STAGING_DIR}/TD.Param --ScoringMetric=$ScoringMetric --GainFile ${TARGET_HELDOUT}.gains --SentenceList=$SOURCE_HELDOUT --ReferenceList $TARGET_HELDOUT --NBest=1 --MEROptimize=0 `cat ${STAGING_DIR}/fixed-PDA-$ScoringMetric.params` $translatetsparams --GenOracleScores=1 > ${STAGING_DIR}/fixed-PDA-$ScoringMetric.1best.log

	grep '@PESA' < ${STAGING_DIR}/fixed-PDA-$ScoringMetric.1best.log | perl -S AnalyseTranslations.pl
	echo -n "PDA $ScoringMetric "
	grep 'Score' ${STAGING_DIR}/fixed-PDA-$ScoringMetric.1best.log
    endif

endif


# if ($DoTesting)
#     echo Now things are getting exciting: Testing on the test set....
#     echo 'SOURCE_TEST' $SOURCE_TEST
#     echo 'TARGET_TEST' $TARGET_TEST

#     echo "perl /afs/cs.cmu.edu/usr/ashishv/www/GenerateNISTInfoGains.pl ${TARGET_TEST} 1 4 > ${TARGET_TEST}.gains"
#     perl /afs/cs.cmu.edu/usr/ashishv/www/GenerateNISTInfoGains.pl ${TARGET_TEST} 1 4 > ${TARGET_TEST}.gains

#     $translatets -f ${MERPARAMS} --TransMemParams=${STAGING_DIR}/TD.Param --ScoringMetric=$ScoringMetric --GainFile ${TARGET_TEST}.gains --SentenceList=$SOURCE_TEST --ReferenceList $TARGET_TEST --NBest=1000 --MEROptimize=0 `cat ${STAGING_DIR}/fixed-$ScoringMetric.params` > ${STAGING_DIR}/$ScoringMetric-testing.1Kbest.log

#     grep '@PESA' < ${STAGING_DIR}/$ScoringMetric-testing.1Kbest.log | perl -S AnalyseTranslations.pl
#     echo -n "$ScoringMetric "
#     grep 'Score' ${STAGING_DIR}/$ScoringMetric-testing.1Kbest.log

# endif
