#!/bin/tcsh -f
#The full segmentation, pre-processing and training for BTEC

# SRC = /project/IWSLT/IWSLT-2006/fullBTEC/chn/fullBTEC.gb
# TGT =  /project/IWSLT/IWSLT-2006/fullBTEC/usa/fullBTEC.en
set SRC = `expandfilename $1`
set TGT = `expandfilename $2`
set DEVSRC = `expandfilename $3`
set DEVREF = `expandfilename $4`
set TESTSRC = `expandfilename $5`
set TESTREF = `expandfilename $6`
set MODEL = `expandfilename $7`

#match by utterance ID
#if you need to do any utt id matching
set UTTID = 0
if ($UTTID) then 
echo "perl ExtractByID.pl $SRC $TGT  > /tmp/tmp.both"
echo 'cat /tmp/tmp.both | awk -F"#"' "'{print "\$"1}'" "> $SRC.clean"
echo 'cat /tmp/tmp.both | awk -F"#"' "'{print "\$"2}'" "> $TGT.clean"
set SRC = $SRC.clean
set TGT = $TGT.clean
endif 

#produces: $SRC.seg.uni.utf8 (the untagged version) and $SRC.seg.uni.utf8.class.norm.trans (the tagged version)
echo "./FinalSrcSegmentAndTag.sh $SRC"

#produces: $TGT.ibm.pp
echo "./FinalTgtTokenizeAndTag.sh $TGT 1" # 1 says: move punct. marks

#produces: $TGT.ibm.num.pp
echo "./FinalTgtTokenizeOnly.sh $TGT 1" # 1 says: move punct. marks

#create training versions of these files, which means that reserved symbols are removed and tagging is collapsed
#to a single tag so that alignment works correctly
echo "cat $SRC.seg.uni.utf8 | ./PrepareTraining.sh > $SRC.seg.uni.utf8.training"
echo "cat $SRC.seg.uni.utf8.class.norm.trans | ./PrepareTraining.sh > $SRC.seg.uni.utf8.class.norm.trans.training"
echo "cat $TGT.ibm.pp  | ./PrepareTraining.sh > $TGT.ibm.pp.training"
echo "cat $TGT.ibm.num.pp | ./PrepareTraining.sh > $TGT.ibm.num.pp.training"

#now you have src and tgt, tagged and untagged, for those elements that have matching num tags, lets align them too
echo "./BothTrain.sh $SRC.seg.uni.utf8.training $SRC.seg.uni.utf8.class.norm.trans.training $TGT.ibm.pp.training $TGT.ibm.num.pp.training /tmp/fullBTEC-final.both.training"
echo 'cat /tmp/fullBTEC-final.both.training | awk -F"#"' "'{print "\$"1}'" "> $SRC.final"
echo 'cat /tmp/fullBTEC-final.both.training | awk -F"#"' "'{print "\$"2}'" "> $TGT.final"

#now build the dev set - and rule files that can be derived directly from DEV
echo "cat $DEVSRC | sed -n 's/ //g; p' > $DEVSRC.clean" #remove existing segmentation
set DEVSRC = $DEVSRC.clean
echo "./FinalSrcSegmentAndTag.sh $DEVSRC"
echo "cat $DEVSRC.seg.uni.utf8.class.norm.trans | perl TaggedToRules.pl > $DEVSRC.seg.uni.utf8.class.norm.trans.rules"
echo "cat $DEVSRC.seg.uni.utf8.class.norm.trans | perl TaggedToPharaoh.pl > $DEVSRC.seg.uni.utf8.class.norm.trans.pretrans"

#same for test set
echo "cat $TESTSRC | sed -n 's/ //g; p' > $TESTSRC.clean" #remove existing segmentation
set TESTSRC = $TESTSRC.clean
echo "./FinalSrcSegmentAndTag.sh $TESTSRC"
echo "cat $TESTSRC.seg.uni.utf8.class.norm.trans | perl TaggedToRules.pl > $TESTSRC.seg.uni.utf8.class.norm.trans.rules"
echo "cat $TESTSRC.seg.uni.utf8.class.norm.trans | perl TaggedToPharaoh.pl > $TESTSRC.seg.uni.utf8.class.norm.trans.pretrans"


#move into the model dir, and symlink in the files that are needed for training
#if (-e $MODEL) then
#    echo "WARNING Model dir: $MODEL already exists"
#endif 

echo "mkdir -p $MODEL"
set CURRENT = `expandfilename`
echo "cd $MODEL"
echo "ln -s $SRC.final training.utf8"
echo "ln -s $TGT.final training.en"
echo "ln -s $DEVSRC.seg.uni.utf8 devset.utf8"
echo "ln -s $TESTSRC.seg.uni.utf8 testset.utf8"
echo "ln -s $DEVREF devset.en"
echo "ln -s $TESTREF testset.en"
echo "ln -s $DEVSRC.seg.uni.utf8.class.norm.trans.rules devset.numrules"
echo "ln -s $TESTSRC.seg.uni.utf8.class.norm.trans.rules testset.numrules"
echo "cat devset.utf8 testset.utf8 > devset-testset.utf8"
#run the pharaoh training
set EXT = /nfs/islpc7_2/ashishv/external-tools/
set PHARAOHDIR = $EXT/training-release-1.3/
echo "cd $PHARAOHDIR"
#remove the last-step=2 if you want to build all the way up to a phrase model with Pharaoh
echo "./train-phrase-model.perl --root-dir $MODEL --f utf8 --e en --corpus $MODEL/training --parts 3" # " --last_step=2"
#build the LM - with then numtag replaced with @CD so that the decoder can use this symbol
echo "cd $CURRENT"
echo 'cat '"$MODEL"'/training.en | sed -n "s/numtag/@CD/g; p" | '"$EXT/sri/bin/i686-m64/ngram-count -lm $MODEL"'/training.en.CD.srilm -interpolate -kndiscount -text -'
echo "cat $MODEL/training.en | sed -n 's/numtag/@CD/g; p' | reversewords.sh | $EXT/sri/bin/i686-m64/ngram-count -lm $MODEL/training.en.CD.srilmbackwards -interpolate -kndiscount -text -"


#=========================================================
# run MER traing in Pharaoh
#=========================================================
echo "cd $PHARAOHDIR"
#generate the pre-translated number file for pharaoh? or just use UTF8 file
echo './minimum-error-rate-training.perl ' "$MODEL"'/mert' "$DEVSRC"'.seg.uni.utf8' "$DEVREF" '100 /nfs/islpc7_2/ashishv//external-tools/training-release-1.3/pharaoh-v1.2.3/pharaoh "-f '"$MODEL"'/model/pharaoh.ini -dl 4 -b 0.03 -s 100 -lmodel-file '"$MODEL"'/training.en.CD.srilm" "d:1,0.5-1.5 lm:1,0.5-1.5 tm:0.3,0.25-0.75;0.3,0.25-0.75;0.3,0.25-0.75;0.3,0.25-0.75;0,-0.5-0.5 w:0,-0.5-0.5" > & ' "$MODEL"'/log-mert-avg &'

#=========================================================
#run SBMT style
#=========================================================
cd $CURRENT
#models are built, unzip them and send them for refinement
echo "gunzip $MODEL/giza.utf8-en/utf8-en.A3.final.gz"
echo "gunzip $MODEL/giza.en-utf8/en-utf8.A3.final.gz"

#generate refined alignment and phrase table file
set HIERODIR = /nfs/islpc7_2/ashishv/external-tools/downloads/hiero-i2.0/
set PYTHON = /nfs/islpc7_2/ashishv/external-tools/Python-2.4.3/bin/python
echo "cd $HIERODIR"
echo "$PYTHON refiner.py $MODEL/giza.utf8-en/utf8-en.A3.final $MODEL/giza.en-utf8/en-utf8.A3.final > $MODEL/refined"
echo "$PYTHON lexweights.py $MODEL/refined -w $MODEL/lex.n2f $MODEL/lex.f2n"
echo "$PYTHON extractor.py $MODEL/refined -o $MODEL/rules | fgrep 'LOG:' > $MODEL/extractor.log"

# extract rules
