#!/bin/env perl

use warnings;
use Getopt::Long;
#Takes in files that are already tagged $NUM_(chinese||english) and create a direct rule database
#Which will look like chinese#english#@CD#features...
#You use this rule database when translating the .utf8 (non-tagged) dev data
my %rules = ();
my $productionTAG = "\@_FORCE\@CD";
my $productionTAGNP = "\@_FORCE\@NP";
my $productionTAGPP = "\@_FORCE\@PP";

my $SimpleCount = 1;
my $SpokenNumbers = 1;
die if !GetOptions("SimpleCount!" => \$SimpleCount, "SpokenNumbers!" => \$SpokenNumbers);
my $featureVector = "0 1 1 0 0 0 -5 -5";
$featureVector = "0 1 1 0 0 0 -5" if (!$SimpleCount);

# the colon in 6:30 pm
my $timeColon = $SpokenNumbers? " " : ":";
# the point in 13.2 US dollars
my $point = $SpokenNumbers? " point " : ".";

while (<STDIN>) {
  my $line = $_;
  chomp($line);

  #some top level data management
  while ($line =~ m/(.*) \$([A-Z]+)_\(([^)]+)\) (\xE6\x9C\x88) \$([A-Z]+)_\(([^)]+)\) (\xE5\x8f\xb7) (.*)/) {
    #print "Month: $3 between $4 Day $6 after $7\n";
    printDateRule($3,$4,$6,$7);
    $line = $1." ".$8;
  }


  #some top level time management
  #CHINESEAM NUM_(7||seven) PTCHINESE NUM(31||31) MINCHINESE
  while ($line =~ m/(.*) (\xE4\xB8\x8A\xE5\x8D\x88|\xE6\x99\x9A\xE4\xB8\x8A|\xE6\x97\xA9\xE4\xB8\x8A>) \$([A-Z]+)_\(([^)]+)\) (\xE7\x82\xB9) \$([A-Z]+)_\(([^)]+)\) (\xE5\x88\x86) (.*)/) {
    #print "Month: $3 between $4 Day $6 after $7\n";

    my $before = $1;
    my $timeday = $2;
    my $hour = $4;
    my ($hourSrc, $hourTgt) = split(/\|\|/, $hour);
    my $pt = $5;
    my $min = $7;
    my ($minSrc, $minTgt) = split(/\|\|/, $min);
    my $minmark = $8;
    my $after = $9;

    my $timedayTgt; 
    $timedayTgt = "am" if ($timeday =~ m/\xE4\xB8\x8A\xE5\x8D\x88/);
    $timedayTgt = "pm" if ($timeday =~ m/\xE6\x99\x9A\xE4\xB8\x8A/);
    $timedayTgt = "am" if ($timeday =~ m/\xE6\x97\xA9\xE4\xB8\x8A/);

    my $tmpTime = $timeColon;
    #print it out in english form
    if ($hourTgt =~ m/[0-9]+/) {
	$hourTgt = num2word($hourTgt) if ($SpokenNumbers);
#
if ($minTgt =~ m/[0-9]+/) {
    $minTgt = num2word($minTgt) if ($SpokenNumbers);
}
}else{
    #the hr is in word form, make the minutes in word form too!
    #print "Hr is in words: converting $minTgt\n";
    my $old = $SpokenNumbers;
    $SpokenNumbers = 1;
    $minTgt = num2word($minTgt);
$SpokenNumbers = $old;
$tmpTime = " ";
}


    print "$timeday $hourSrc $pt $minSrc $minmark \# $hourTgt$tmpTime$minTgt $timedayTgt \# $productionTAGNP \#", $featureVector, "\n";
    print "$timeday $hourSrc $pt $minSrc $minmark \# at $hourTgt$tmpTime$minTgt $timedayTgt \# $productionTAGPP \#", $featureVector, "\n";
    #print "$timeday $hourSrc $pt $minSrc $minmark \# $hourTgt $minTgt $timedayTgt \# $productionTAG \#", $featureVector, "\n";
    #print "$timeday $hourSrc $pt $minSrc $minmark \# at $hourTgt $minTgt $timedayTgt \# $productionTAG \#", $featureVector, "\n";
    $line = $before." ".$after;
  }

  #am pm NUM_{8||8} PT $NUM{ch|zero}
  while ($line =~ m/(.*) (\xE4\xB8\x8A\xE5\x8D\x88|\xE6\x99\x9A\xE4\xB8\x8A|\xE6\x97\xA9\xE4\xB8\x8A>) \$([A-Z]+)_\(([^)]+)\) (\xE7\x82\xB9) \$([A-Z]+)_\(([^)]+)\|\|zero\) (.*)/) {


    my $before = $1;
    my $timeday = $2;
    my $hour = $4;
    my ($hourSrc, $hourTgt) = split(/\|\|/, $hour);

    if ($hourTgt =~/[0-9]+/) {
	$hourTgt = dealWithPT(num2word($hourTgt)) if ($SpokenNumbers);
    }

    my $pt = $5;
    my $srczero = $7;
    my $after = $8;

    my $timedayTgt; 
    $timedayTgt = "am" if ($timeday =~ m/\xE4\xB8\x8A\xE5\x8D\x88/);
    $timedayTgt = "pm" if ($timeday =~ m/\xE6\x99\x9A\xE4\xB8\x8A/);
    $timedayTgt = "am" if ($timeday =~ m/\xE6\x97\xA9\xE4\xB8\x8A/);
    #    print "$timeday $hourSrc $pt $srczero\# $hourTgt $timedayTgt \# $productionTAG \#", $featureVector, "\n";
    #    print "$timeday $hourSrc $pt $srczero\# at $hourTgt $timedayTgt \# $productionTAG \#", $featureVector, "\n";

    print "$timeday $hourSrc $pt $srczero\# $hourTgt $timedayTgt \# $productionTAGNP \#", $featureVector, "\n";
    print "$timeday $hourSrc $pt $srczero\# at $hourTgt $timedayTgt \# $productionTAGPP \#", $featureVector, "\n";
    $line = $before." ".$after;
  }

  #deal with am/pm NUM_{8:30||8:30} seconds
  while ($line =~ m/(.*) (\xE4\xB8\x8A\xE5\x8D\x88|\xE6\x99\x9A\xE4\xB8\x8A|\xE6\x97\xA9\xE4\xB8\x8A) \$([A-Z]+)_\(([^)]+)\) (\xE7\xA7\x92)/) {
    my $before = $1;
    my $timeday = $2;
    my $time = $4;
    my $secmark = $5;

    my $after = $6;
    my ($timeSrc, $timeTgt) = split(/\|\|/,$time);
    my @timeParts = split(/\./,$timeTgt);
    if (scalar(@timeParts)==2) {
	$timeTgt = ($SpokenNumbers) ? num2word($timeParts[0])." ".num2word($timeParts[1]) : $timeParts[0]." ".$timeParts[1];
    } else {
      if ($timeTgt =~ m/[0-9]+/) {
	$timeTgt = dealWithPT(num2word($timeTgt)) if ($SpokenNumbers);
      }
    }

    my $timedayTgt; 
    $timedayTgt = "am" if ($timeday =~ m/\xE4\xB8\x8A\xE5\x8D\x88/);
    $timedayTgt = "pm" if ($timeday =~ m/\xE6\x99\x9A\xE4\xB8\x8A/);
    $timedayTgt = "am" if ($timeday =~ m/\xE6\x97\xA9\xE4\xB8\x8A/);

    #    print "$timeday $timeSrc $secmark $ \# $timeTgt $timedayTgt \# $productionTAG \#", $featureVector, "\n";
    #    print "$timeday $timeSrc $secmark $ \# at $timeTgt $timedayTgt \# $productionTAG \#", $featureVector, "\n";

    print "$timeday $timeSrc $secmark $ \# $timeTgt $timedayTgt \# $productionTAGNP \#", $featureVector, "\n";
    print "$timeday $timeSrc $secmark $ \# at $timeTgt $timedayTgt \# $productionTAGPP \#", $featureVector, "\n";
    $line = $before." ".$after;
  }

  #deal with the seconds annotation very last, we treat this as the least likely to be correct case
  while ($line =~ m/(.*) \$([A-Z]+)_\(([^)]+)\) (\xe7\xa7\x92) (.*)/) {
    my $before = $1;
    my $secs = $3;
    my $secsmarker = $4;
    my $after = $5;

    my ($secsSrc, $secsTgt) = split(/\|\|/, $secs);
    if ($secsTgt =~ m/[0-9]+/) {
      $secsTgt = num2word($secsTgt);
    }

    #    print "$secsSrc $secsmarker \# $secsTgt seconds \# $productionTAG \#",$featureVector, "\n";
    print "$secsSrc $secsmarker \# $secsTgt seconds \# $productionTAGNP \#",$featureVector, "\n";

    $line = $before." ".$after;
  }




  #this is to handle individual rules
  while ($line =~ m/(.*)\$([A-Z]+)_\(([^)]+)\)(.*)/) {
    #print "before $1 tag $2 payload $3 after $4\n";
    #print "tag $2 payload $3\n";
    my $before = $1;
    my $tag = $2;
    my $after = $4;
    my $payload = $3;

    $after =~ s/^\s+//g;
    if ($after =~ m/(PT[A-Z]+) (.*)/) {
      #print "PTTAG $line\n";
      $payload = "$1 $payload";
      $after = $2;
    } else {
      if ($after =~ m/(PT[A-Z]+)$/) {
	#print "<FOUNDEND> PTTAG </FOUND> $line \n";
	$payload = "$1 $payload";
      }
      $after = "";
    }

    #save the line as a rule
    $rules{$payload} = 1;
    $line = $before." ".$after;
  }
}


foreach my $key (keys %rules) {
#  print $key . "\n";
  my @parts = split(/\|\|/, $key);

    
  my $src = $parts[0];
  #if there is a pt tag, swap it now
  if ($src =~ m/(PT[A-Z]+) (.*)/) {
    $src = $2." ".$1;
  }

  my $tgt = $parts[1];
    
  $tgt =~ s/^\s+//g;
  $tgt =~ s/\s+$//g;
  if ($tgt =~ m/$([0-9]+)\.([0-9]+)^/) {
    #print "Time: $1 -> $2";
    my $hours = $1;
    my $minutes = $2;
	
    #    #only assume its a time if minutes has mod%5=0
    #    if ($minutes%5==0) {
    #      if ($hours>12) {
    #	#print "Modding: ";
    #	$hours = $hours%12;
    #      }
    #      $tgt = num2word($hours)." ".num2word($minutes);
    #    } else {
    $tgt = num2word($hours). $point . num2word($minutes);
    #    }
  }
    
    
  if ($tgt =~ m/^([0-9,]+)$/) {
    $tgt = num2word($tgt);
  }
    
  if ($tgt =~ m/(.*)%/) {
      $tgt = $1."@@"."percent" if ($SpokenNumbers);
  }

  if ($tgt =~ m/@@/) {
	
    $tgt =~ s/@@/ /go;
    my @tgtp = split(/ /, $tgt);
    my $i =0;
    foreach my $t (@tgtp) {
      #print "checking:$t\n";
      $t =~ s/^\s+//g;
      $t =~ s/\s+$//g;
      if ($t =~ m/[0-9]+/) {
	$t = num2word($t);
	$tgtp[$i] = $t;
      }
      $i++;
    }
    $tgt = join(" ", @tgtp);
  }

    
    
  $tgt =~ s/\s+/ /g;
  $tgt =~ s/^\s+//g;
  $tgt =~ s/\s+$//g;
    
  #fix the stupid spelling by ibm
  $tgt =~ s/ninteen/nineteen/g;
  if ($src =~ m/PTTAG/) {
	
    if ($src =~ m/PTTAGMID/ && $tgt =~ m/zero/) {
      $tgt =~ s/zero/zero point/g;
    } else {
      #print "Dealing with PT $src tgt: $tgt\n";
	$tgt = dealWithPT($tgt) if ($SpokenNumbers);
    }
  }

  my @tgtList = split(/ /, $tgt);
  my $numTargets = @tgtList;
  print "$src # $tgt # ".$productionTAG." # $featureVector\n"
}

sub addth{
  my ($w) = @_;

  return $w if !$SpokenNumbers;

  $w =~ s/\s+/ /g;
  $w =~ s/^\s+//g;
  $w =~ s/\s+$//g;
  if ($w =~ m/(.*)one$/) {
    return $1."first";
  }
  if ($w =~ m/(.*)two$/) {
    return $1."second";
  }
  if ($w =~ m/(.*)three$/) {
    return $1."third";
  }
  if ($w =~ m/(.*)four$/) {
    return $1."fourth";
  }
  if ($w =~ m/(.*)five$/) {
    return $1."fifth";
  }
  if ($w =~ m/(.*)six$/) {
    return $1."sixth";
  }
  if ($w =~ m/(.*)seven$/) {
    return $1."seventh";
  }
  if ($w =~ m/(.*)eight$/) {
    return $1."eighth";
  }
  if ($w =~ m/(.*)nine$/) {
    return $1."ninth";
  }

  if ($w =~ m/(.*)eleven$/) {
    return $1."eleventh";
  }
  if ($w =~ m/(.*)twelve$/) {
    return $1."twelvth";
  }


  #endings
  if ($w =~ m/(.*)teen$/) {
    return $1."teenth";
  }

  #endings
  if ($w =~ m/(.*)ty$/) {
    return $1."tieth";
  }


  return $w
}


sub printDateRule{
  my ($month, $monthMark, $date, $dateMark) = @_;
  #print "month $month date $data\n";
  my ($monthSrc, $monthTgt) = split(/\|\|/,$month);
  my ($dateSrc, $dateTgt) = split(/\|\|/,$date);    
  #print "monthTgt $monthTgt\n";
  my $m = monthNumToWord($monthTgt);
  #print "m is $m\n";
  my $d1;
  if ($dateTgt=~ m/[0-9]+/) {
    $d1 = addth(num2word($dateTgt));
  } else {
    $d1 = addth($dateTgt);
  }
    
  #create on shorter and one longer version

  #  print "$monthSrc $monthMark $dateSrc $dateMark \# $m $d1 \# $productionTAG \#",$featureVector, "\n";
  #  print "$monthSrc $monthMark $dateSrc $dateMark \# on $m $d1 \# $productionTAG \#",$featureVector, "\n";
  #  print "$monthSrc $monthMark $dateSrc $dateMark \# $d1 of $m \# $productionTAG \#",$featureVector, "\n";

  print "$monthSrc $monthMark $dateSrc $dateMark \# $m $d1 \# $productionTAGNP \#",$featureVector, "\n";
  print "$monthSrc $monthMark $dateSrc $dateMark \# on $m $d1 \# $productionTAGPP \#",$featureVector, "\n";
  #  print "$monthSrc $monthMark $dateSrc $dateMark \# $d1 of $m \# $productionTAGNP \#",$featureVector, "\n";



}



sub monthNumToWord{
  my ($dt) = @_;    
  #print "dt is $dt\n";
  if ($dt == 1 || $dt eq "one") {
    return "January";
  }
  if ($dt == 2 || $dt eq "two") {
    return "February";
  }
  if ($dt == 3 || $dt eq "three") {
    return "March";
  }
  if ($dt == 4 || $dt eq "four") {
    return "April";
  }
  if ($dt == 5 || $dt eq "five") {
    return "May";
  }
  if ($dt == 6 || $dt eq "six") {
    return "June";
  }
  if ($dt == 7 || $dt eq "seven") {
    return "July";
  }
  if ($dt == 8 || $dt eq "eight") {
    return "August";
  }
  if ($dt == 9 || $dt eq "nine") {
    return "September";
  }
  if ($dt == 10 || $dt eq "ten") {
    return "October";
  }
  if ($dt == 11 || $dt eq "eleven") {
    return "November";
  }
  if ($dt == 12 || $dt eq "twelve") {
    return "December";
  }
  print STDERR "Warning: Invalid month: $dt \n";
  return undef;
}

sub dealWithPT{
  $words = shift;
  $words =~ s/zero/noon/g;
  $words =~ s/thirteen/one/g;
  $words =~ s/fourteen/two/g;
  $words =~ s/fifteen/three/g;
  $words =~ s/sixteen/four/g;
  $words =~ s/seventeen/five/g;
  $words =~ s/eighteen/six/g;
  $words =~ s/nineteen/seven/g;


  $words =~ s/twenty-one/nine/g;
  $words =~ s/twenty-two/ten/g;
  $words =~ s/twenty-three/eleven/g;
  $words =~ s/twenty-four/midnight/g;

  #only if the longer ones dont match!
  $words =~ s/twenty/eight/g;
  return $words;
}

sub num2word{
  $n=shift;

  if ($SpokenNumbers) {

    #    my $originalN = $n;
    @a=qw(0 one two three four five six seven eight nine ten
	  eleven twelve thir four fif six seven eigh nine twen thir for);
    $w = "";
    for ($c=0xf4240;$c;$c/=1000) {
      if ($j=int$n/$c) {
	$w.=$a[int$j/100].' hundred ',$j%=100if$j>99;$w.=
	  $a[int$j/10+($j>49?10:18)].'ty-',$j%=10if$j>19;$w.=$a[$j].($j>12?'teen ':' ')if
	    $j;$w.=$c>1000?'million ':$c>1?'thousand ':'';$n%=$c;
      }
    }

    #    if (int$originalN<100){
    #print "Splitting $originalN\n";
    $w =~ s/^s\+//g;
    $w =~ s/s\+$//g;
    #	my @wparts = split(/ /, $w);
    #	if (scalar(@wparts)==2){
    #	    $w =~ s/ /\-/g;
    #	    $w =~ s/-$//g;
    #	}
    #    }
    # change e.g. 'twenty-' into 'twenty'
    $w =~ s/-$//g;
    # 19xx is likely going to be a year
    $w =~ s/one thousand nine hundred/nineteen/g; 
    # so is 18xx
    $w =~ s/one thousand eight hundred/eighteen/g; 
    $w =~ s/\s+/ /g;
    $w =~ s/^\s+//g;
    $w =~ s/\s+$//g;
    return $w;

  } else {
#    print STDERR "$n -> ";
    $n =~ s/(\d)(\d\d\d)$/$1,$2/;
    # don't put commas in between years
    $n =~ s/^1,9(\d\d)$/19$1/;
    $n =~ s/^1,8(\d\d)$/18$1/;
    $n =~ s/^2,0(\d\d)$/20$1/;
#    print STDERR "$n\n";
    return $n;
  }
}
