#!/bin/bash
#Authors: zollmann, ashishv initially from vogel
#Example Usage: condorsubmit-stderr-standard.sh 1 gizarun GIZA++ gizargs
#Where GIZA++ is an executable compiled for standard universe
#Output of GIZA++ will go to gizarun.log/err

#Uses the Job as an identifier and runs the Exe with arguments
#The Job.clog is the condor log file that we need to wait for in order to wait for this process to finish: Usage condor_wait Job.clog
#while Job.log is the actual log file

maxdays="$1" #the maximum number of days that this jobs will run
Job="$2"     #the job identifier, keep these unique!
#if you give a full path for this Job argument, you can get in trouble when trying to create files like Condor./project/mylog
#so lets quickly alias this
JobName=`echo $Job | replace "\/" "-"`
echo "JobName $JobName"

CondorParams=""
Exe="$3"
#In other scripts we pipe thru runexp first, but that script is not 
#a standard environment executable! This can be fixed, but not yet
#runexp-stderr.sh
#The RUNEXP deal does not work for standard universe jobs

#Remove the first 3 args, ie exec, maxdays, log
shift
shift
shift

#  Machine="islr0s0 islr0s1 islr0s2 islr0s3 islr0s4 islr0s5 islr0s6 islr0s7 islr0s8 islpc9"

if [ $maxdays -le 21 ]; then  m=islpc9 ; fi
if [ $maxdays -le 14 ]; then  m="islr0s8 $m" ; fi
if [ $maxdays -le 7 ]; then  m="islr0s7 $m" ; fi
if [ $maxdays -le 6 ]; then  m="islr0s6 $m" ; fi
if [ $maxdays -le 5 ]; then  m="islr0s5 $m" ; fi
if [ $maxdays -le 4 ]; then  m="islr0s4 $m" ; fi
if [ $maxdays -le 3 ]; then  m="islr0s3 $m" ; fi
if [ $maxdays -le 2 ]; then  m="islr0s2 $m" ; fi
if [ $maxdays -le 1 ]; then  m="islr0s0 islr0s1 $m" ; fi

Machineline=`echo $m|perl -pe 's/ /.is.cs.cmu.edu\" \|\| Machine == \"/g , s/^/\( Machine == \"/g , s/\n/.is.cs.cmu.edu\" \)/g ;'`

echo $Machineline
#exit

#set Args = "$@"
#echo "with    " $Args

Dir=`pwd`
#echo "in      " $Dir
unset DISPLAY

echo "########################################"  > Condor.$JobName
echo "# Job: Condor."$JobName     >> Condor.$JobName
echo "# Submitted on:" `date` >> Condor.$JobName
echo "# In directory:" `pwd`  >> Condor.$JobName
echo "########################################" >> Condor.$JobName
echo ""                     >> Condor.$JobName
echo "Executable = " $Exe   >> Condor.$JobName
echo "Arguments  = " "$@"  >> Condor.$JobName
echo "Initialdir = " $Dir   >> Condor.$JobName

#make sure you use the real job here to make sure it goes in the right place
echo "Output     = " $Job.log >> Condor.$JobName
echo "Error      = " $Job.err >> Condor.$JobName
echo "Log        = " $Job.clog >> Condor.$JobName
echo "Universe   = standard" >> Condor.$JobName
echo "notification = never" >> Condor.$JobName
echo "Getenv     = True"    >> Condor.$JobName
echo "Rank       = KFLOPS + Memory"    >> Condor.$JobName
echo 'Requirements = Arch == "x86_64" && Memory > 0' >> Condor.$JobName
echo "WantCheckpoint = TRUE"                     >> Condor.$JobName
echo ""                     >> Condor.$JobName
echo "Queue"                >> Condor.$JobName

cat Condor.$JobName

condor_submit $CondorParams Condor.$JobName

