#!/usr/bin/env perl

my $from = "";
my $to = "";
my $infile = "";
my $outfile = "";
my $link_type = "all";
my $start_sentence = 1;

for(my $i=0;$i<=$#ARGV;$i+=2) {
   usage() if (!$ARGV[$i+1]);

   if ($ARGV[$i] eq "--from"){
       $from = $ARGV[$i+1];
   }
   elsif ($ARGV[$i] eq "--to"){
       $to = $ARGV[$i+1];
   }
   elsif ($ARGV[$i] eq "--in"){
       $infile = $ARGV[$i+1];
   }
   elsif ($ARGV[$i] eq "--out"){
       $outfile = $ARGV[$i+1];
   }
   elsif ($ARGV[$i] eq "--link-type"){
       $link_type = $ARGV[$i+1];
   }
   elsif ($ARGV[$i] eq "--start-sentence"){
       $start_sentence = $ARGV[$i+1];
   }
   elsif ($ARGV[$i] eq "--e"){
       $eng_corpus = $ARGV[$i+1];
   }
   elsif ($ARGV[$i] eq "--f"){
       $fl_corpus = $ARGV[$i+1];
   }
   elsif ($ARGV[$i] eq "--bbn"){
       $bbn_corpus = $ARGV[$i+1];
   }
}

if (!$from || !$to || !$infile || !$outfile){
   usage();
}
print STDERR "LINK TYPE: $link_type\n";
print STDERR "Reading alignment\n";
@alignments = ();
open(IN,$infile);
if ($from eq "sp"){
   while ($aline = <IN>){
       chomp $aline;
       ($sn, $e1, $f1, $type) = split " ", $aline;
       if (!$sn || !$e1 || !$f1){
           die "ERROR in SP format: $aline\n";
       }
       if ($link_type eq "all"){
           if ($alignments[$sn]{"$e1\-$f1"} == 1){
               print "Duplicate alignment: $aline\n";
           }
           $alignments[$sn]{"$e1\-$f1"} = 1;
       }
       elsif ($link_type eq $type){
           $alignments[$sn]{"$e1\-$f1"} = 1;
       }
   }
}
elsif ($from eq "idg"){
   $sn = 0;
   while ($aline = <IN>){
       chomp $aline;
       if ($aline =~ /begin/){
           $sn++;
       }
       elsif ($aline =~ /^(\d+)\s+(.*?)\s*$/){
           $e1 = $1;
           $fs = $2;
           next if ($e1 == -1);
           @a1 = split " " , $fs;
           foreach $f1 (sort ({$a <=> $b} @a1)){
               if ($f1 != -1){
                   $alignments[$sn]{"$e1\-$f1"} = 1;
               }
           }
       }
   }
}
elsif ($from eq "pharaoh_f2e"){
   $sn = 0;
   while ($aline = <IN>){
       chomp $aline;
       $sn++;
       @links = split /\s+/ , $aline;
       foreach $link (@links){
           ($f1,$e1) =split "-" , $link;
           $f1++;$e1++;
           $alignments[$sn]{"$e1\-$f1"} = 1;
       }
   }
}
elsif ($from eq "pharaoh_e2f"){
   $sn = 0;
   while ($aline = <IN>){
       chomp $aline;
       $sn++;
       @links = split /\s+/ , $aline;
       foreach $link (@links){
           ($e1,$f1) =split "-" , $link;
           $alignments[$sn]{"$e1\-$f1"} = 1;
       }
   }
}
elsif ($from eq "giza_f2e"){
   $sn = 0;
   while ($aline = <IN>){
       $sn++;
       $aline = <IN>; $aline = <IN>; chomp($aline);
       my @words = split(/\}\)/,$aline);
       # ignore NULL word
       for(my $e1=1;$e1<$#words;$e1++) {
           my @links= split(/ /,$words[$e1]);
           for(my $a=3; $a<=$#links; $a++) {
               if ($links[$a] !~ /^\d+$/){
                   print STDERR "Error: Sentence $sn: ($e1,$links[$a]) \n";
               }
               else{
                   $f1 = $links[$a];
                   $alignments[$sn]{"$e1\-$f1"} = 1;
               }
           }
       }
   }
}
elsif ($from eq "giza_e2f"){
   $sn = 0;
   while ($aline = <IN>){
       $sn++;
       $aline = <IN>; $aline = <IN>; chomp($aline);
       my @words = split(/\}\)/,$aline);
       for(my $f1=1;$f1<$#words;$f1++) {
           my @links= split(/ /,$words[$f1]);
           for(my $a=3; $a<=$#links; $a++) {
               if ($links[$a] !~ /^\d+$/){
                   print STDERR "Error: Sentence $sn: [$words[$f1]] \n .... $aline\n";
               }
               else{
                   $e1 = $links[$a];
                   $alignments[$sn]{"$e1\-$f1"} = 1;
               }
           }
       }
   }
}
elsif  ($from =~ /ibm/){
   $sn = 0;
   while ($aline = <IN>){
       chomp $aline;
       $aline =~ s/^\s+//;
       next if ($aline =~ /^\s*$/);
       next if ($aline =~ /^\s*\]b\s*$/);
       if ($aline =~ /^\s*\[b/){
           $sn++;
           $aline = <IN>; chomp $aline;
           if ($aline =~ /^\s*\/\/\s+(.*?)\s*$/){
               $src_sentences[$sn] = $1;
               #for arabic
               $src_sentences[$sn] =~ s/\s+\(\)\s*$//;
           }
           $aline = <IN>; chomp $aline;
           if ($aline =~ /^\s*\/\/\s+(.*?)\s*$/){
               $trg_sentences[$sn] = $1;
               # for arabic
               $trg_sentences[$sn] =~ s/\s+\(\)\s*$//;
           }
       }
       else{
           if ($from eq "ibm_f2e"){
               ($fl_str,$en_str) = split /\s+/ , $aline;
           }
           else{
               ($en_str,$fl_str) = split /\s+/ , $aline;
           }
           @f = split /\,/ , $fl_str;
           @e = split /\,/ , $en_str;
           foreach $f1 (@f){
               foreach $e1 (@e){
                   if (($f1 >0 ) && ($e1 > 0)){
                       $alignments[$sn]{"$e1\-$f1"} = 1;
                   }
               }
           }
       }
   }
}
else{
   usage();
}
close(IN);

$maxSent = scalar(@alignments) - 1;
print STDERR "Maximum Sentence Number = $maxSent\n";

if ($to eq "giza_f2e"){
   if ($from !~ /ibm/){
       print STDERR "Reading corpus\n";
       open(SOURCE, $eng_corpus) || die "Error: Can not open $eng_corpus\n";
       open(TARGET, $fl_corpus) || die "Error: Can not open $fl_corpus\n";
   }
}
elsif ($to eq "giza_e2f"){
  if ($from !~ /ibm/){
      print STDERR "Reading corpus\n";
      open(SOURCE, $fl_corpus) || die "Error: Can not open $fl_corpus\n";
      open(TARGET, $eng_corpus) || die "Error: Can not open $eng_corpus\n";
  }
}
elsif ($to eq "bbn"){
   print STDERR "Reading corpus\n";
   open(BBN, $bbn_corpus) || die "Error: Can not open $bbn_corpus\n";
}

print STDERR "Writing alignment\n";
open(OUT, ">$outfile");
for (my $i=$start_sentence;$i <= $maxSent;$i++){
   if ($to eq "bbn"){
       $bbn_line = <BBN>; chomp $bbn_line;
       while ($bbn_line !~ /^ALIGNMENT\: /){
           print OUT "$bbn_line\n";
           $bbn_line = <BBN>; chomp $bbn_line;
       }
       $lineResult = "";
   }
   elsif (($to eq "giza_f2e") || ($to eq "giza_e2f")){
       if ($from !~ /ibm/){
           $src_sent = <SOURCE>; chomp $src_sent; $trg_sent = <TARGET>; chomp $trg_sent;
       }
       else{
           $src_sent = $src_sentences[$i];
           $trg_sent = $trg_sentences[$i];
       }
       @src_words = split " " , $src_sent;     @trg_words = split " " , $trg_sent;
       $src_length = $#src_words+1;
       $trg_length = $#trg_words+1;
       print OUT "# Sentence pair ($i) source length $src_length target length $trg_length alignment score : 1.0\n";
       print OUT "$trg_sent \n";

       @giza_links = ();
       $giza_links[0] = "NULL ({ })";
       for (my $lnk = 1; $lnk <=$src_length; $lnk++){
           $giza_links[$lnk] = "$src_words[$lnk-1] ({ })";
       }
       @null_links = ();
   }
   elsif ($to eq "idg"){
       print OUT "begin $i\n";
   }
   elsif ($to =~ /pharaoh/){
       $lineResult = "";
   }
   # sort according to $e1
   @sorted_keys = ();
   if (($to eq "bbn") || ($to eq "pharaoh_f2e")){
       @sorted_keys = (sort {
           ($a1,$a2) = split "-" , $a;
           ($b1,$b2) = split "-" , $b;
           if ($a2 < $b2) { return -1; }
           elsif ($a2 > $b2) { return 1; }
           else { return $a1 <=> $b1; }
       } keys %{$alignments[$i]});
   }
   else{
       @sorted_keys = (sort {
           ($a1,$a2) = split "-" , $a;
           ($b1,$b2) = split "-" , $b;
           if ($a1 < $b1) { return -1; }
           elsif ($a1 > $b1) { return 1; }
           else { return $a2 <=> $b2; }
       } keys %{$alignments[$i]});
   }
   foreach $key (@sorted_keys){
       $pair = $key;
       if ($to eq "idg"){
           $pair =~ s/\-/\t/;
           print OUT "$pair\n";
       }
       elsif ($to eq "sp"){
           $pair =~ s/\-/ /;
           print OUT "$i $pair\n";
       }
       elsif ($to eq "bbn"){
           ($e1,$f1) = split "-" , $pair;
           $e1--;$f1--;
           $lineResult .= "$f1\:$e1 ";
       }
       elsif ($to eq "pharaoh_f2e"){
           ($e1,$f1) = split "-" , $pair;
           $e1--;$f1--;
           $lineResult .= "$f1\-$e1 ";
       }
       elsif ($to eq "pharaoh_e2f"){
           $lineResult .= "$pair ";
       }
       elsif ($to eq "giza_f2e"){
           ($e1,$f1) = split "-" , $pair;
           $giza_links[$e1] =~ s/\}\)\s*$/$f1 \}\)/;
           $null_links[$f1] = "NO";
       }
       elsif ($to eq "giza_e2f"){
           ($e1,$f1) = split "-" , $pair;
           $giza_links[$f1] =~ s/\}\)\s*$/$e1 \}\)/;
           $null_links[$e1] = "NO";
       }
   }
   if ($to eq "idg"){
       print OUT "end\n";
   }
   elsif ($to =~ /pharaoh/){
       $lineResult =~ s/ $//;
       print OUT "$lineResult\n";
   }
   elsif ($to eq "bbn"){
       $lineResult =~ s/ $//;
       print OUT "ALIGNMENT: $lineResult\n";
   }
   elsif ($to =~ /giza/){
       for (my $lnk=1; $lnk<=$trg_length;$lnk++){
           if (!defined($null_links[$lnk])){
               $giza_links[0] =~ s/\}\s*\)$/$lnk \}\)/;
           }
       }
       print OUT "@giza_links \n";
   }
}

close(OUT);


sub usage{
   die "Usage: convert-alignment-format.pl --from <FORMAT_1> --to <FORMAT_2> --in <INPUT-FILE> --out <OUTPUT-FILE> [--start-sentence <SS> --link-type <LINK-TYPE> --e <ENGLISH_CORPUS> --f <FOREIGN_CORPUS> --bbn <BBN_ALIGNMENT_FILE>\n  Possible format names are: idg sp pharaoh_e2f pharaoh_f2e giza_e2f giza_f2e bbn\n";
}
