#!/usr/bin/env perl
#Extract translations from a bee directory - look into each file, extract the relevant information and then process it
use strict;
use Getopt::Long;

my $outdir = ".";
die if !GetOptions("outdir=s" => \$outdir);

if (!(-e $outdir)){
    print STDERR "Creating dir $outdir\n";
    print STDERR `mkdir $outdir\n`;
}

#my $generateLogs;
#if ($nbest){
#    $generateLogs  = "cat $expdir/bee*.log |  egrep '(==Sent)|(ScaledTMLMCost)'";
#}else{
#    $generateLogs  = "cat $expdir/bee*.log |  egrep -A2 '(==Sent)|(fullSent)'";
#}
#print STDERR $generateLogs, "\n";
#my @lines = `$generateLogs`;

my @translations = ();
my %found = ();
my %printed = ();

my $currentSentence = -1;
my $numprocessed = 0;
my $numduplicates = 0;
#foreach my $line (@lines){
while (<STDIN>){
    
    my $line = $_;
    if ($line =~ m/.*===Sentence ([0-9]+).*/){
	#print the data from the last sentence 
	#print STDERR "BEFORE $currentSentence numprocessed $numprocessed\n";
	checkAndPrint();

	@translations = ();
	$currentSentence = $1;
	$numprocessed++;
	print STDERR "Processing sentence $currentSentence numprocessed $numprocessed numduplicates $numduplicates\n";
    }
    if ($line =~ m/ScaledTMLM/){
	chomp($line);
	die "currentSentence not defined for $line " if (!defined($currentSentence));	
	push (@translations, $line);
	++$found{$currentSentence};
	#print "Setting: $currentSentence = $line\n";
    }
}
#now you might have encountered the very last sentence
checkAndPrint();

sub checkAndPrint{
	if ($currentSentence>-1){
	    #if there is a current sentence defined, then print it
	    if (defined($printed{$currentSentence})){
		my $lastprint = $printed{$currentSentence};
		#if you found more valid translations now... then overwrite the file
		if ($found{$currentSentence} > $lastprint){
		    printLast($currentSentence);
		    $printed{$currentSentence} = $found{$currentSentence};
		    $found{$currentSentence} = 0;
		    print STDERR "INFO: Overwriting duplicate for $currentSentence\n";
		}else{
		    print STDERR "INFO: Rejecting duplicate for $currentSentence\n";
		}
		$numduplicates++;
	    }else{
		#go ahead and print, since you have never seen anything before
		printLast($currentSentence);
		#indicate how many you printed
		$printed{$currentSentence} = $found{$currentSentence};
		#and reset the counter for the next one
		$found{$currentSentence} = 0;
	    }
	}
}

sub printLast{
    my ($sentenceNumber) = @_;
    
    open OUT, ">$outdir/sent.$sentenceNumber" or die "Unable to open $outdir/sent.$sentenceNumber for writing\n";
    foreach my $line (@translations){
	print OUT "$line\n";
    }
    close(OUT);
}

#print $sentencenumber . ' ||| ' . translation . ' ||| ' . costs . "\n";

my $sequential = 1;
my $count = 0;
foreach my $key (sort{$a <=> $b}(keys %found)){
    #print "$key => ", $translations{$key}, "\n";
    if ($count != $key){
	print STDERR "WARNING: Extracted output is not sequential: missing $count\n";
	$count = $key;
    }
    $count++;
}

print STDERR "INFO: Num sentences extracted: $count\n";

