#! perl

#####################################################################
# Purpose: 
# call example:
# perl -Sw extractrules.pl -outputdbfile=extractedrules -m=1 europarl.fr-en.fr europarl.fr-en.en.preprocessed targetparsetrees sttk.log 5 5 5 5 5 --outputallrules --onlyuptosentence 1 > extractedrules.txt
#
#################################################################
# without DBM on transtac data: 8:37 for first 50 sentences according to top on dish (i.e., 10.34 seconds per sentence)


use warnings;
use strict;
#use Encode::Arabic;

#use Encode qw/encode decode/;
use Storable;

#Warning: Tree:R has the weird peculiarity of assuming that the to-be-changed variable \@objects passed to its methods points to an empty array (if not that array is just appended to)
use Tree::R;			# obtained at http://users.tkk.fi/~jolma/pub/snapshots/Tree-R-0.05.tar.gz, extended to deal with our needs, source provided with this package
use List::Util qw(first min max);
use Fcntl;
use File::Temp qw/ tempfile tempdir /;
#use DB_File; # uncomment this line if you don't have Berkeley DB installed on your unix system
use BerkeleyDB; # uncomment this line if you don't have Berkeley DB installed on your unix system
use Getopt::Long;


#binmode STDIN, ":utf8";
#binmode STDERR, ":utf8";
#binmode STDOUT, ":utf8";

# settings

my $BERKELEY_DB = 1; # set to zero if Berkeley DB is not available

# input params


my $MaxLengthSingleCountSourcePhrases = 1;
my $AbstractOverX = 0;
my $GluePhrasePairs = 1;
my $outputallrules = 0;
my $exitgluemoderules = 0;
my $keepunbalancedrules = 0;
my $tmpdir;
my $outputdbfile;
my $VERBOSE = 0;
my $resume = 0; # set corresponding parameter to true if you want to resume rule extraction at sentence where it previously got interrupted
my $parsing = 1; # if zero, we don't read parses from the parse tree file and merely write out lexical non-syntactical rules (basically, phrase pairs)
my $createtmpsubdir = 1;
my $onlyuptosentence = 1000000000000; # the corresponding option allows you to process only the first N sentences
my $onlystartwithsentence = 1; # the corresponding option allows you to process only the sentence from sentence number N onwards

die if !GetOptions('tmpdir=s' => \$tmpdir, 'createtmpsubdir!' => \$createtmpsubdir, 'verbose!' => \$VERBOSE, 'outputdbfile=s' => \$outputdbfile, 'parsing!' => \$parsing, 'MaxLengthSingleCountSourcePhrases=s' => \$MaxLengthSingleCountSourcePhrases, 'outputallrules!' => \$outputallrules, 'AbstractOverX!' => \$AbstractOverX, 'GluePhrasePairs!' => \$GluePhrasePairs, 'exitgluemoderules!' => \$exitgluemoderules, 'keepunbalancedrules!' => \$keepunbalancedrules, 'onlystartwithsentence=s' => \$onlystartwithsentence, 'onlyuptosentence=s' => \$onlyuptosentence, 'resume!' => \$resume);

# $tmpdir = File::Spec->tmpdir() if (!$tmpdir);
$tmpdir = "." if (!$tmpdir);


my $sourcesentencefile = $ARGV[0];
my $targetsentencefile = $ARGV[1];
my $targetparsetreefile = $ARGV[2];
my $translationlogfile = $ARGV[3];
my $MAXSOURCEWORDCOUNT = $ARGV[4]; # max n.o. source words in created rules
my $MAXTARGETWORDCOUNT = $ARGV[5]; # same for target
my $MAXSOURCESYMBOLCOUNT = $ARGV[6]; # max n.o. source symbols (source words + substitutions) in created rules
my $MAXTARGETSYMBOLCOUNT = $ARGV[7]; # same for target
my $MAXSUBSTITUTIONCOUNT = $ARGV[8]; # max n.o. substitutions allowed in created rules
# note: $MAXSOURCESYMBOLCOUNT and $MAXSUBSTITUTIONCOUNT are the two parameters that can significantly influence execution speed,
# as they allow for omission of recursive subroutine calls
if ($MAXSOURCESYMBOLCOUNT > $MAXSUBSTITUTIONCOUNT+$MAXSOURCEWORDCOUNT) {
  print STDERR "Warning: MAXSOURCESYMBOLCOUNT ($MAXSOURCESYMBOLCOUNT) > MAXSUBSTITUTIONCOUNT ($MAXSUBSTITUTIONCOUNT) + MAXSOURCEWORDCOUNT ($MAXSOURCEWORDCOUNT). ";
  $MAXSOURCESYMBOLCOUNT = $MAXSUBSTITUTIONCOUNT + $MAXSOURCEWORDCOUNT;
  print STDERR "Setting MAXSOURCESYMBOLCOUNT=$MAXSOURCESYMBOLCOUNT\n";
}
if ($MAXTARGETSYMBOLCOUNT > $MAXSUBSTITUTIONCOUNT+$MAXTARGETWORDCOUNT) {
  print STDERR "Warning: MAXTARGETSYMBOLCOUNT ($MAXTARGETSYMBOLCOUNT) > MAXSUBSTITUTIONCOUNT ($MAXSUBSTITUTIONCOUNT) + MAXTARGETWORDCOUNT ($MAXTARGETWORDCOUNT). ";
  $MAXTARGETSYMBOLCOUNT = $MAXSUBSTITUTIONCOUNT + $MAXTARGETWORDCOUNT;
  print STDERR "Setting MAXTARGETSYMBOLCOUNT=$MAXTARGETSYMBOLCOUNT\n";
}
#my $VERBOSE = $ARGV[9];
#my $tmpdir = $ARGV[10];
#$tmpdir = File::Spec->tmpdir() if (!$tmpdir);


# global variables

# create my own temp dir as subdir of $tmpdir, it and its contents will 
# automatically removed at program end
my $mytmpdir = $createtmpsubdir?
  tempdir( "tmp_extractrules_XXXXXXXXX", DIR => $tmpdir, CLEANUP => 1)
  :
  $tmpdir;
#my $mytmpdir = tempdir( "tmp_extractrules_XXXXXXXXX", DIR => $tmpdir, CLEANUP => 0);

my $sentencenumber = 0;	# counts the sentences processed
my $phrasepairnumber = 0; # counts the phrase pairs processed
my $ruleinstancecount = 0; # counts the number of instances of rules extracted (sum of all %resulttypecounts values)
my $writtenoutrulenumber = 0; # number of rules written out to STDOUT / database
my %complexrules = ();		# will contain the complex rules, with $key= "$source\#$target\#$result" and $value=\@costvector
my %phrasecounts = ();          # will contain occurrence counts of source side and target side of the rules and the occurrence counts of the source/target pair
my $outputdb; # the output rule database - a B-tree
my $db1; my $db2; my $db3; # the other DBs used
my %resulttypecounts = ();	# counts the total n.o. rules in %complexrules that have the same result type (keys are categories, values are numbers); this will enable relative-frequency calculations when writing out the rules

#for use within the while loop, but using subroutine, thus defined before main block
my %currentparsespans = ();
my @currentparsetokens = ();
my $currentparsetokenspos = 0;
my @sourcewords;
my @targetwords;

{				# main block
  print STDERR `date`."\nRunning extracrules.pl. Source sentence file: $sourcesentencefile, target sentence file: $targetsentencefile\ntarget parse tree file: $targetparsetreefile, translation log file: $translationlogfile, temp dir: $tmpdir\nMAXSOURCEWORDCOUNT=$MAXSOURCEWORDCOUNT, MAXTARGETWORDCOUNT=$MAXTARGETWORDCOUNT, MAXSOURCESYMBOLCOUNT=$MAXSOURCESYMBOLCOUNT, MAXTARGETSYMBOLCOUNT=$MAXTARGETSYMBOLCOUNT, MAXSUBSTITUTIONCOUNT=$MAXSUBSTITUTIONCOUNT\n";
  print STDERR "Ignoring all sentences before sentence number $onlystartwithsentence\n" if $onlystartwithsentence > 1;
  print STDERR "Ignoring all sentences after sentence number $onlyuptosentence\n" if $onlyuptosentence != -1;
  print STDERR ($outputdbfile ?
		"Outputting rules into newly created Berkeley DB file $outputdbfile\n" :
		"Outputting rules in text-form to STDOUT\n");
  print STDERR ($AbstractOverX? "" : "NOT") . "Abstracting over _X rules\n";
  print STDERR "Glue phrase pairs together to yield longer phrase pairs that span a syntactic category.\n" if $GluePhrasePairs;
  if ($BERKELEY_DB) {
    # old DB_File version
    #my $a = new DB_File::HASHINFO ; $a->{'cachesize'} = 2**30 ;
    #tie %complexrules, "DB_File", "$mytmpdir/complexrules", O_CREAT|O_RDWR, 0666, $a || die "Can't create temp database complexrules in $mytmpdir! Try setting \$Berkeley_DB=0 at top of extractrules.pl";
    #tie %phrasecounts, "DB_File", "$mytmpdir/phrasecounts", O_CREAT|O_RDWR, 0666, $a || die "Can't create temp database phrasecounts in $mytmpdir!";

    $db1 = tie %complexrules, "BerkeleyDB::Hash", 
      -Cachesize => 2**29, 
	-Filename => "$mytmpdir/complexrules", 
	  -Flags => DB_CREATE # don't truncate it if it already exists# |DB_TRUNCATE
	    || die "Can't create temp database complexrules in $mytmpdir! Try setting \$Berkeley_DB=0 at top of extractrules.pl";
    $db2 = tie %phrasecounts, 
      "BerkeleyDB::Hash", 
	-Cachesize => 2**29, 
	  -Filename => "$mytmpdir/phrasecounts", 
	    -Flags => DB_CREATE # |DB_TRUNCATE
	      || die "Can't create temp database phrasecounts in $mytmpdir!";
    $db3 = tie %resulttypecounts, 
      "BerkeleyDB::Hash", 
	-Cachesize => 2**29, 
	  -Filename => "$mytmpdir/resulttypecounts", 
	    -Flags => DB_CREATE # |DB_TRUNCATE
	      || die "Can't create temp database resulttypecounts in $mytmpdir!";
  } else {
    dbmopen(%phrasecounts, "$mytmpdir/phrasecounts", 0666) || die "Can't create temp database phrasecounts in $mytmpdir!";
    dbmopen(%complexrules, "$mytmpdir/complexrules", 0666) || die "Can't create temp database complexrules in $mytmpdir!";
    dbmopen(%resulttypecounts, "$mytmpdir/resulttypecounts", 0666) || die "Can't create temp database resulttypecounts in $mytmpdir!";
  }
  if ($outputdbfile) {
    if ($BERKELEY_DB) {
      $outputdb = new BerkeleyDB::Btree 
	-Cachesize => 2**26, 
	  -Filename => $outputdbfile, 
	    -Flags => DB_CREATE|DB_TRUNCATE,
	      -Property => DB_DUP
	      || die "Can't create output rule Berkeley DB database $outputdbfile! Bye.";
    } else {
      die "Variable \$BERKELEY_DB is set to 0, therefore the rules cannot be output as database file but just as plain text! Bye.\n";
    }
  }

  if ($resume) {
    # resume mode: resume from 1+sentence number that has been successfully processed
    $onlystartwithsentence = 1 + $phrasecounts{'_SID'};
    die "No resumption info found in database!" if !defined($onlystartwithsentence);
  }
  open(SOURCE, "< $sourcesentencefile");
  open(TARGET, "< $targetsentencefile");
  open(PARSES, "< $targetparsetreefile") if $parsing;
  open(PHRASEPAIRS, "perl -S Log2Phrasetable.pl $translationlogfile |") || die "can't fork: $!"; # the PESA phrase costs in the phrase table are still costs, not probs! Use perl -S ConvertLogTDtoProbs.pl if you want them to be probs

  # read and discard first line from PHRASEPAIRS pipe, which should start with _SentenceNumber
  die "Wrong format of log file $translationlogfile! Byebye.\n" if !(<PHRASEPAIRS> =~ /^_SentenceNumber/);

  #print `cat $targetparsetreefile`;
  #print `pwd`;
  #print scalar <PARSES>; die;
  while (my $sourceline = <SOURCE>) {
    last if $sentencenumber >= $onlyuptosentence;
    $sentencenumber++;
    print STDERR "Processing sentence number $sentencenumber...\r" if ($sentencenumber-1) % 100 == 0;

    my $parserline = $parsing ? <PARSES> : undef;
    @sourcewords = str2list($sourceline);
    @targetwords = str2list(scalar <TARGET>);
    print STDERR "Processing $sentencenumber: @sourcewords \# @targetwords\n" if $VERBOSE;

    my %currentphrasepairs = (); # indexes the phrase pairs of the current sentence by target spans
    %currentparsespans = ();	#reset parse span hash for current sentence
    # analyze the parser output
    if ($parsing) {
      chomp($parserline);
      if ($parserline =~ /^_fail/) {
	print STDERR "Parse failure at sentence number $sentencenumber: $parserline; target sentence: ".join(" ",@targetwords). "\n";
      } else {
	#expand parens to have spaces around them
	$parserline =~ s/\(/\( /g;
	$parserline =~ s/\)/ \) /g;
	#   print "$parserline\n";
	#    @currentparsetokens = split(/\s*(\()|(\))|(\s)/, $parserline); # split on (,), or space, where ( might be preceded by spaces that we ignore
	@currentparsetokens = split(' ', $parserline);
	#print join ("," , @currentparsetokens)."\n";
	$currentparsetokenspos = 0;
	my $sentencelengthtmp = findparsespans(0); #create the %currentparsespans entries, have span indices start from 0, not 1
	if ($sentencelengthtmp != scalar @targetwords) {
	  print STDERR "Sentence $sentencenumber: Sentence length mismatch parse tree yield vs. target sentence!\nParse tree: ".join(" ",@currentparsetokens)."\nTarget sentence: ".join(" ",@targetwords)."\n";
	  %currentparsespans = ();	#reset parse span hash for current sentence
	} else {
	  print STDERR "$parserline\n" if $VERBOSE>=2;
	  (print STDERR "Target parse spans:\n"),printhash(\%currentparsespans) if $VERBOSE;
	}
      }
    }

    my $rtree = new Tree::R;	# Create R-tree, which will store all elementary rules in such a way as to efficiently allow for sub-rule relationship testing


    # now retrieve phrase pairs for current sentence
    print STDERR "Phrase pairs retrieved:\n" if $VERBOSE;
    my $nophrasepairsfound = 1;
    {
      my %alreadyaddedphrases = (); # hash set to avoid processing multiple identical phrase pairs
      
      while (my $phrasepairline = <PHRASEPAIRS>) {
	last if ($phrasepairline =~ /^_SentenceNumber/);
	next if ($onlystartwithsentence > $sentencenumber); # ignore phrase pair if we're not supposed to extract yet
	# Note: only outgoing edges are assumed to be in PHRASEPAIRS!

	$nophrasepairsfound = 0;
	$phrasepairnumber++;
	print STDERR $phrasepairline if $VERBOSE;

	# retrieve source phrase, target phrase and phrase cost vector
	chomp($phrasepairline);
	my ($sourcephrasestr, $targetphrasestr, $coststr) = split(/ *\# */, $phrasepairline);
	my @sourcephrasewords = split(' ', $sourcephrasestr);
	my @targetphrasewords = split(' ', $targetphrasestr);

	#	  # sanity test: check if every target word occurs in the parse tree yield
	#	  for my $w (@targetphrasewords) {
	#	    if (!exists $currentparsespans{$w}) {
	#	      print STDERR "Sentence $sentencenumber: Couldn't find word '$w' in parse tree yield.\nParse tree: ".join(" ",@currentparsetokens)."\nTarget sentence: ".join(" ",@targetwords)."\nPhrase pair line: $phrasepairline\n";
	#	    }
	#	  }

	# check for special case: source span spans whole source sentence
	if (scalar(@sourcephrasewords) == scalar(@sourcewords)) {
	  # we wanna make sure that all phrase pairs whose source side is the whole source sentence also have
	  # as target side the whole target sentence
	  if (scalar(@targetphrasewords) != scalar(@targetwords)) {
	    print STDERR "Warning: Phrase pair whose source side is the whole source sentence but whose target side ".
	      "is not the whole target sentence. Correcting target side to whole target sentence. Sentence pair $sentencenumber (".
		join(" ",@sourcewords). "#".join(" ",@targetwords)."), phrase pair (" .
		  join(" ",@sourcephrasewords). "#".join(" ",@targetphrasewords).")\n" if $VERBOSE;
	    @targetphrasewords = @targetwords;
	  }
	}

	# now determine the source/target spans of the current phrase pair
	my @tmpsourcespans = lookupspans(\@sourcephrasewords,\@sourcewords);
	my @tmptargetspans = lookupspans(\@targetphrasewords,\@targetwords);
	for my $tmpsourcespan (@tmpsourcespans) {
	  for my $tmptargetspan (@tmptargetspans) {
	    my ($tmpminx,$tmpsizex,$tmpminy,$tmpsizey) = split(' ', "$tmpsourcespan $tmptargetspan");

	    # Sanity test: check for forbidden case: source phrase spans whole source sentence but target phrase doesn't
	    if ($tmpsizex == scalar(@sourcewords) && $tmpsizey != scalar(@targetwords)) {
	      die "Failed sanity test. Kill AZ. Bye. Sentence pair $sentencenumber (".
		join(@sourcewords). "#".join(@targetwords)."), phrase pair (" .
		  join(@sourcephrasewords). "#".join(@targetphrasewords).")\n";
	    }

	    # look whether the target span has at least one syntactic category
	    my $tmpcategories = $currentparsespans{$tmptargetspan};
	    my $tmpcategories = $tmpcategories->[0] if defined($tmpcategories);
	    print STDERR "... corresponding to sourcespan (start length): $tmpsourcespan, targetspan (start length): $tmptargetspan, categories: ".
	      (defined($tmpcategories)? $tmpcategories : "(none)") ."\n" if $VERBOSE;
	    if (!defined $tmpcategories) {
	      # no syntactic category for the target span found
	      # just assign it the _X category
	      $tmpcategories = "_X";
	    }
	    # make sure we haven't processed this phrase pair yet (might happen when a word occurs multiple times - in that case we hopefully have the same cost vectors for the identical phrase pairs)
	    if (!exists $alreadyaddedphrases{"$tmpsourcespan $tmptargetspan"}) { #!! this is a bug, have to avoid double-counting differently
	      $alreadyaddedphrases{"$tmpsourcespan $tmptargetspan"} = 1;
		  
	      #		for my $tmpcategory (split(' ', $tmpcategories)) {
	      # store elementary rule with result categories $tmpcategories
	      my @tmpphrasecosts = split(' ', $coststr);
	      $rtree->insert([\@tmpphrasecosts,$tmpcategories],($tmpminx,$tmpminy,$tmpminx+$tmpsizex-1,$tmpminy+$tmpsizey-1));
	      if ($GluePhrasePairs) {
		# also store phrase pair in hash %currentphrasepairs to allow for gluing together of phrase pairs
		my $refcpp = $currentphrasepairs{$tmptargetspan};
		if (
		    defined($refcpp) 
		    # oops, we have already a phrase pair with this target span
		    # use the one with the bigger source span
		    && 
		    $refcpp->[3] >= $tmpsizex
		    # the source span of the currently stored phrase pair is not smaller, so let's stick with that phrase pair
		   ) {
		  # do nothing
		} else {
		  # store current phrase pair under the targetspan key
		  $currentphrasepairs{$tmptargetspan} = [\@tmpphrasecosts,$tmpcategories,$tmpminx,$tmpsizex];
		}
	      }
	    }
	  }
	}
      }				#while ... <PHRASEPAIRS>
      
      if ($onlystartwithsentence <= $sentencenumber) { # if we're supposed to extract phrase pairs for this sentence
	print STDERR "Warning: STTK returned no phrase pairs for sentence $sentencenumber!\n" if $nophrasepairsfound;

	if ($GluePhrasePairs && $parsing) {
	  # iterate through the parse spans of length > 1 bottom up (important!) and see which syntactical 
	  # span doesn't have a corresponding phrase pair yet
	  for (my $spansize = 2; $spansize <= scalar(@targetwords); $spansize++) {
	    for (my $spanstart = 0; $spanstart + $spansize <= scalar(@targetwords); $spanstart++) {
	      my $targetspan = "$spanstart $spansize";
#		print STDERR "\nSpan $targetspan: ". "\n";
	      my $parsespaninfo = $currentparsespans{$targetspan};
	      if (defined($parsespaninfo)) {
		print STDERR "\nDefined: $parsespaninfo";
		print STDERR "Span $targetspan: ". $parsespaninfo->[0] . "\n";
	      }
	      if (defined($parsespaninfo) && !exists $currentphrasepairs{$targetspan}) {
		print STDERR "Considering $targetspan: ".$parsespaninfo->[0] ;
		# this target span corresponds to a syntactic category but we don't have a phrase pair for it
		# try to glue phrase pairs of the child nodes together to create one
		my $failure = 0;

		# iterate through the child nodes' parse spans (starting with 1 b/c 0 is this span's category field)
		# and determine their corresponding phrase pairs if existing
		my @childsourcespanstarts = ();
		my @childsourcespansizes = ();
		my @childtargetspanstarts = ();
		my @childtargetspansizes = ();
		my @childphrasecostvectors = ();
		for (my $i = 1; !$failure && $i < scalar(@$parsespaninfo); $i += 2) {
		  my $tmpspanstart = $parsespaninfo->[$i];
		  my $tmpspansize = $parsespaninfo->[$i+1];
		  my $phrasepairinfo = $currentphrasepairs{"$tmpspanstart $tmpspansize"};
		  if (!defined($phrasepairinfo)) {
		    # child parse span also doesn't have a corresponding phrase pair
		    $failure = 1;
		  } else {
		    push @childtargetspanstarts, $tmpspanstart;
		    push @childtargetspansizes, $tmpspansize;
		    push @childsourcespanstarts, $phrasepairinfo->[2]; 
		    push @childsourcespansizes, $phrasepairinfo->[3]; 
		    push @childphrasecostvectors, $phrasepairinfo->[0]; 
		  }
		}
		$failure = 1 if scalar(@childphrasecostvectors) == 0; # there are no child nodes which we could glue together
		if (!$failure) {
		  print STDERR " Has child nodes and they all have phrase pairs.";
		  # syntactic category/ies of this span and thus of the to-be-created phrase pair
		  my $resultcategories = $parsespaninfo->[0];
		  # to-be source span of the created phrase pair
		  my $sourcespanstart;
		  my $sourcespansize;
		  my @costvector; # to-be phrase cost vector

		  # first sort the child nodes by source span start; return an array of indices to the sorted children
		  my @indices = (0..(scalar(@childtargetspanstarts)-1));
		  my @sortedindices = sort {$childsourcespanstarts[$a] <=> $childsourcespanstarts[$b]} @indices;
		  my $firstchild = 1;
		  for my $i (@sortedindices) {
		    if ($firstchild) {
		      # first child: set our source span to child's source span
		      $sourcespanstart = $childsourcespanstarts[$i];
		      $sourcespansize = $childsourcespansizes[$i];
		      # initialize also the cost vector to this child's cost vector
		      @costvector = @{$childphrasecostvectors[$i]};
		      $firstchild = 0;
		    } else {
		      # check that next child's source span either borders on or overlaps with current source span
		      if ($childsourcespanstarts[$i] > $sourcespanstart+$sourcespansize) {
			# there's a hole... forget it
			$failure = 1;
		      } else {
			# extend our span
			$sourcespansize = $childsourcespanstarts[$i] + $childsourcespansizes[$i] - $sourcespanstart;
			for (my $j = 0; $j < scalar @costvector; $j++) {
			  $costvector[$j] += $childphrasecostvectors[$i]->[$j];
			}
		      }
		    }
		    print STDERR "child $i: childsourcespanstart $childsourcespanstarts[$i], childsourcespanlength $childsourcespansizes[$i]. "
		  }
		  if (!$failure) {
		    # gluing succeeded - store the new phrase pair
		    print STDERR "Sucess!\n";

		    # check for special case: target span spans whole target sentence but source span doesn't span whole source sentence
		    if ($spansize == scalar(@targetwords) && $sourcespansize != scalar(@sourcewords)) {
		      # correct source span
		      $sourcespanstart = 0;
		      $sourcespansize = scalar(@sourcewords);
		    }

		    $rtree->insert([\@costvector,$resultcategories],($sourcespanstart,$spanstart,$sourcespanstart+$sourcespansize-1,$spanstart+$spansize-1));
		    # also store phrase pair in hash %currentphrasepairs to allow for glue-creation of super-span phrase pairs
		    $currentphrasepairs{$targetspan} = [\@costvector,$resultcategories,$sourcespanstart,$sourcespansize];
		  }
		}
	      }
	    }
	  }
	}


	# it follows some special handling of the complete-sentence spans:
	# for each category that spans the whole target-sentence (e.g.: S), we add a unary abstract rule for the start symbol _S
	# (e.g.: @1 / @1 / S->_S)

	my ($minx,$sizex,$miny,$sizey) = (0, scalar @sourcewords, 0, scalar @targetwords);
	my $tmpsourcespan = "$minx $sizex";
	my $tmptargetspan = "$miny $sizey";	

	# look whether the complete target span has at least one syntactic category
	my $tmpcategories = $currentparsespans{$tmptargetspan};
	my $tmpcategories = $tmpcategories->[0] if defined($tmpcategories);
	if (!defined $tmpcategories) {
	  print STDERR "Error: Target side sentence $sentencenumber has no syntactic category spanning the whole sentence assigned!\n" if $parsing;
	  $tmpcategories = "_X";
	}
	if (!exists $alreadyaddedphrases{"$tmpsourcespan $tmptargetspan"}) {
	  # we have to worry about the problem of not getting the whole target sentence as target-phrase of the whole source-sentence - therefore, we now add that phrase pair manually unless it exists already
	  #this also means other reasonable phrase pairs are not found (preference tends to go to shorter ones)
	  #!!suggestion1: create additional phrases by attaching unaligned target words from left and from right to target-sides of phrases
	  #!!suggestion2: symmetrize: also do PESA extraction target->source, which will get us more phrase pairs (this should perhaps only be done for the source/target category version of the system, b/c currently we kind of 'detect' junk phrase pairs by their target side not corresponding to any syntactic target category, which wouldn't be the case four target->source PESA extracted phrase pairs
	  #Changed 2006/01/10: Because we don't know what kind of IBM1 probs we should put here, we now don't add the whole pair anymore and merely hope that the grammar derived from other sentence pairs can make for not having a "_S" rule for this sentence.
	  #this hope is no good, so better make sure to have the complete-sentence spanning phrase pair in...
	  
	  print STDERR "Warning: no complete-sentence spanning phrase pair for sentence $sentencenumber!\n";
	  
	  # $alreadyaddedphrases{"$tmpsourcespan $tmptargetspan"} = 1;
	  # # store elementary rule with result categories $tmpcategories
	  # $rtree->insert(["0 0 0 0 0 0",$tmpcategories],($minx,$miny,$minx+$sizex-1,$miny+$sizey-1));
	}
	
	# now add unary rules for the start symbol
	my $sourcestr = "\@1"; 
	my $targetstr = "\@1";
	my @startrule_type = ($tmpcategories, "_S");
	my @startrule_costvector = (1, 0, 0, 0, 0, 0, 0);
      print STDERR "Adding rule: $sourcestr\#$targetstr\#".join("->",@startrule_type)."...\n" if $VERBOSE;
	addrules(\$sourcestr,\$targetstr,\@startrule_costvector,\@startrule_type,scalar(@startrule_type),[]);
	
	
      } # end: if ($onlystartwithsentence <= $sentencenumber)
    }	# end of retrieving phrase pairs for current sentence

      
    # Write out the non-start-symbol elementary and complex rules
      
    #first obtain an array of all elementary rules stored 
    #(don't worry, the rules are stored by reference, so that won't take much memory)
    my @elementaryrules = ();
    # we have to make sure that we have at least one element in the R-tree 
    # b/c of a bug in the R tree implementation that returns the root of the tree even if it is empty
    if (!$nophrasepairsfound) {
     $rtree->object_and_bbox_refs(\@elementaryrules);
      
     # now iterate through these rules
     for my $elemrule (@elementaryrules) {
      #print join(",",@$elemrule) ."\n";die;
      my ($sourcespanstart,$targetspanstart,$sourcespanend,$targetspanend) = @$elemrule[2..5]; # span info

      # r is the current rule to be written out and consists of four parts:

      # source part of rule (array of length equal to source span, whose elements are: 
      #  0 (write out corresponding sentence word at this position), 
      #  pos. number N (write out "#N" at this position), or 
      #  -1 (ignore this position when writing out the rule)
      my @r_source = ();
      # initialize array with zeros
      for (my $i=0; $i<=$sourcespanend-$sourcespanstart; $i++) {
	push @r_source, 0;
      }

      # target part of rule (analogous to source part)
      my @r_target = ();
      # initialize array with zeros
      for (my $i=0; $i<=$targetspanend-$targetspanstart; $i++) {
	push @r_target, 0;
      }

      # cost vector part of the rule
      my $refphrasecostvector = $elemrule->[1]->[0];
      my @r_costvector = @$refphrasecostvector;
      unshift @r_costvector, 1; # first element of the array will be the n.o. occurrences of the rule, the rest is the $coststr part returned by the PESA phrase extractor

      #  type signature part of the rule
      my @r_type = ($elemrule->[1]->[1]); # starting of with single element: the result categories



      # get array of (references to) elementary rules whose source and target spans are contained in $elemrule's source and target spans
      my @subrules = ();
      # if this is a nonsyntactical rule and $AbstractOverX is false, we won't retrieve any subrules to abstract over
      # otherwise, retrieve all subrules
      if ($AbstractOverX || $r_type[0] ne '_X') {
	$rtree->object_and_bbox_refs_completely_within_rect(@$elemrule[2..5],\@subrules);
      }
	
      # index the array by relative source span start (rel. to source span start of $elemrule
      my @indexedsubrules = (); # array of arrays of rule references
      # initialize the array with empty arrays
      for (my $i=0; $i<=$#r_source; $i++) {
	push @indexedsubrules, [];
      }
      for my $rule (@subrules) {
	my ($subrulesourcespanstart,$subruletargetspanstart,$subrulesourcespanend,$subruletargetspanend) = @$rule[2..5]; # span info
	if ($subrulesourcespanstart < $sourcespanstart || $subruletargetspanstart < $targetspanstart || $subrulesourcespanend > $sourcespanend || $subruletargetspanend > $targetspanend) {
	  print STDERR "Something's wrong with the R-tree module!\n";
	}
	if ($rule == $elemrule) {
	  # ignore $elemrule itself as a subrule
	  # Note that this also causes rules such as @1#@1#PRP->NP not to be written out
	  # if PRP and NP both span target word 'we'
	  # we don't want to increase spurious ambiguity in the translate-parser unnecessarily
	} elsif (!$keepunbalancedrules && 
		 (
		  ($subrulesourcespanstart == $sourcespanstart && $subrulesourcespanend == $sourcespanend) ||
		  ($subruletargetspanstart == $targetspanstart && $subruletargetspanend == $targetspanend)
		 ) 
		) {
	  # either source side of subrule is identical to source side of elemrule or target side of subrule is identical to target side
	  # of elemrule
	  # since we don't want unbalanced rules (such as: de @NN # @1 # @NP) , ignore this subrule
	} elsif (!$AbstractOverX && $rule->[1]->[1] eq '_X') {
	  # don't abstract over _X rules - ignore this subrule
	} else {
	  my $relsourcespanstart = $rule->[2] - $elemrule->[2];
	  push @{$indexedsubrules[$relsourcespanstart]}, $rule;
	}
      }
	  
	
      # recursively create and write out complex rules from $elemrule into %complexrules
      createcomplexrules($elemrule, \@r_source, \@r_target, \@r_costvector, \@r_type, \@indexedsubrules, scalar(@r_source), scalar(@r_target), 0, 0, 0);
      
     }				#for my $elemrule
    } # if (!$nophrasepairsfound)
    
    if ($sentencenumber % 10 == 0 && $BERKELEY_DB) {
      # store number of sentence processed in case we get interrupted and want to resume later
      $phrasecounts{'_SID'} = $sentencenumber;
      # write the cache of the DBs to disk
      $db1->db_sync();
      $db2->db_sync();
      $db3->db_sync();
    }
  }				#while my $sourceline = <SOURCE>

  print STDERR `date`."$sentencenumber sentences, $phrasepairnumber phrase pairs. Writing out the rules now...\n";
  close SOURCE;
  close TARGET;
  close PARSES if $parsing;
  close PHRASEPAIRS;

  print STDERR `ls -lh $mytmpdir` . "\n" . `du -h $mytmpdir` . "\n";

  # write out the rules in %complexrules
  my $totalrulefreqcount_sanitytest = 0;
  my $rulenumber = 0;
  my ($key, $value);
  while (($key,$value) = each %complexrules) { 
    $rulenumber++;
    #  my $neglogfrequency = - log $value->[0];
    #my $frequency = $value->[0];
    
    my ($sourcestr,$targetstr,$typestr) = split("#", $key);
    my @costvector = split(" ", $value);
    my $frequency = $costvector[0]; # frequency of rule
    $totalrulefreqcount_sanitytest += $frequency;

    my $sourcefrequency = $phrasecounts{"$sourcestr\#"}; # frequency of rule's source side
    if (!defined($sourcefrequency)) {
      print STDERR "DB error: Rule $key\#$value: sourcefrequency not contained in phrasecounts DB!\n";
      next;
    }
    my $targetfrequency = $phrasecounts{"\#$targetstr"}; # frequency of rule's target side
    if (!defined($targetfrequency)) {
      print STDERR "DB error: Rule $key\#$value: targetfrequency not contained in phrasecounts DB!\n";
      next;
    }
    my $rulebodyfrequency = $phrasecounts{"$sourcestr\#$targetstr"}; # frequency of rule's body (different from $frequency in that multiple types for identical rule bodies were not multiply counted)
    if (!defined($rulebodyfrequency)) {
      print STDERR "DB error: Rule $key\#$value: rulebodyfrequency not contained in phrasecounts DB!\n";
      next;
    }

    # find the result category in the rule
    $key =~ /[ \#]([^ \#]*)$/;
    #my $resultcategory = $1;
    my $resultcount = $resulttypecounts{$1};
    if (!defined($resultcount)) {
      print STDERR "DB error: Rule $key\#$value: resultcount not contained in resulttypecounts DB!\n";
      next;
    }

    my @ruletargetwords = split(" ", $targetstr);
    my @rulesourcewords = split(" ", $sourcestr);
    my @resultcategories = split(" ", $typestr);

    # count source- and target-terminal symbols
    my $sourcewordcount = scalar(@rulesourcewords) + 1 - scalar(@resultcategories);
    my $targetwordcount = scalar(@ruletargetwords) + 1 - scalar(@resultcategories);

    if ( ( # either: parameter -outputallrules was passed
	  $outputallrules
	 ) || 
	 ( # or: two conditions have to hold
	  # first condition: balancedness
	  ( (($sourcewordcount == 0) == ($targetwordcount == 0))
	    # rule has no source terminals iff rule has no target terminals
	    || scalar(@resultcategories) == 1
	    # always counts as fulfilled if rule is fully lexicalized 
	    || $keepunbalancedrules
	    # always counts as fullfilled if keepunbalancedrules
	    # parameter is set
	  ) && 
	  # second condition: plausibility
	  ( $frequency > 1 
	    # seen phrase pair at least twice
	    || (scalar(@resultcategories) == 1 && scalar(@rulesourcewords) <= $MaxLengthSingleCountSourcePhrases)
	    # or: rule is fully lexicalized and source side has at most
	    # MPLSCP words
	  )
	 )
       ) {
      # write out the rule
      print STDERR "Writing out: $key, rel. freq.: $frequency/$resultcount, target-cond. r.f.: $rulebodyfrequency/$targetfrequency, source-conditioned r.f.: $rulebodyfrequency/$sourcefrequency\n" if $VERBOSE;

      #convert rule format into the new style
      foreach (my $i = 0; $i < scalar @rulesourcewords; $i++) {
	if ($rulesourcewords[$i] =~ /^\@/ ) {
	  # source word starts with @ -- must be an arg marker
	  # replace this by next result category, and prepend a @ to the category name (all nonterminals will start with @ in our parsing system)
	  $rulesourcewords[$i] = "\@" . shift(@resultcategories);
	}
      }
      #!! check somewhere that no word starts with @ because that would create ambiguity problems regarding arg markers

      # @resultcategories should now contain exactly one nonterminal symbol
      die "In rule $key: Mismatch between rule's number of type categories and number of arg markers in source of the rule! Perhaps training corpus contains a word starting with @ symbol? Bye.\n" if scalar(@resultcategories) != 1;
      
      #prepend @ to the result category
      my $resultcategory = "\@" . $resultcategories[0];

      #create new rule source
      my $newsourcestr = join(" ", @rulesourcewords);

      # cost vector
      my $finalcostvector = 
	join(" ", (
		   0, # glue feature
		   1, # rule application counter
		   $targetwordcount, # target word counter
		   -log($frequency/$resultcount), # neglog (base e) result-conditioned RF
		   -log($rulebodyfrequency/$targetfrequency), # neglog target-cond. RF
		   -log($rulebodyfrequency/$sourcefrequency), # neglog source-cond. RF
		   #		   -log($frequency/$ruleinstancecount) #neglog joint RF
		   $frequency, # just for info purposes: frequency of rule
		   @costvector[1..$#costvector] # PESA costs (6 values)
		  ));

      #create rule body
      my $rulebody = "$targetstr\#$resultcategory\#$finalcostvector";

      outputrule($newsourcestr,$rulebody);
    }
  }
  print STDERR "!!!!!!!!!!!! Something's wrong with the underlying complexrules database! (Maybe an network file system issue?) \$ruleinstancecount ($ruleinstancecount) != \$totalrulefreqcount_sanitytest ($totalrulefreqcount_sanitytest)!!! \n" if $ruleinstancecount != $totalrulefreqcount_sanitytest;

  #to get a statistics about how many nonterminals there are
  my $nonterminalcount = 0;

  #add some manual rules now
  my $manualrulecount = 0;

  # glue rules: theory is: any result can be transformed into _X
  # (enter glue mode: e.g. @NP#@1#@_X); _X rules can be joined (@_X
  # @_X#@1 @2#@_X); _X can be transformed into any nonterminal (exit
  # glue mode: e.g. @_X#@1#@NP) in practise, avoid doubling the
  # n.o. hyps in the chart tables by using _S instead of _X 
  # also, the exit-glue mode rules will be inserted only when needed
  # directly by the parser (the parser only applies the self-rule to
  # _S only when the whole chart cell contains no hypergraph nodes
  # with non-"_S" result nonterminals)

  #the joining glue rule
  outputrule("\@_S \@_S", "\@1 \@2\#\@_S\#1 1 0 0 0 0 0 0 0 0 0 0 0");
  $manualrulecount++;
  
  #  print STDERR "Nonterminals X for which a rule \@X\#\@1\#\@_S has been extracted \n(for the other NTs we'll add artificial enter-glue-mode rules):\n";
  print STDERR "Nonterminals occurring as rule results: ";
  # enter and exit glue-mode rules, one for each nonterminal symbol
  while (($key,$value) = each %resulttypecounts) { 
    $nonterminalcount++;
    print STDERR $key . " ";
    if ($key eq "_S") {
      # ignore _S
    } else {
      if ($exitgluemoderules) {
	# add exit-glue-mode rule for $key
	# We want exit- and enter-glue-mode rules to be very competetive
	# regarding the rule-application count.
	# Therefore, we put zero into the rule-application
	# feature.
	outputrule("\@_S", "\@1\#\@$key\#1 0 0 0 0 0 0 0 0 0 0 0 0");
	$manualrulecount++;
      }

      #      if (exists($complexrules{"\@1\#\@1\#$key _S"})) {
      #	# rule has already been extracted in a natural way - don't write
      #	# out extra glue rule
      #	print STDERR $key . " ";
      #      } else {

      # add enter-glue-mode rule for $key
      # We want exit- and enter-glue-mode rules to be very competetive
      # regarding the rule-application count.
      # Therefore, we put zero into the rule-application
      # feature.
      outputrule("\@$key", "\@1\#\@_S\#1 0 0 0 0 0 0 0 0 0 0 0 0");
      $manualrulecount++;

      #      }
    }
  }
  print STDERR "\nTotal number of nonterminal symbols (including _S): $nonterminalcount\n";
  

  #unknown rules: only the relative weights of unknown rules
  #w.r.t. each other matter.  To make sure that their feature weights
  #are positive (i.e., they are costs), we add the same value to all
  #neglog prob features. We still put in 1 for rule application and
  #target word count though, although that slightly distorts the
  #relative weights between the rules. 
  outputrule('_UNKNOWN', '_UNKNOWN#@NNP#0 1 1 0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1'); # proper name, weight 9/10; -ln(9/10)=0.1
  outputrule('_UNKNOWN', '_UNKNOWN#@JJ#0 1 1 3.4 3.4 3.4 3.4 3.4 3.4 3.4 3.4 3.4 3.4'); # adjective, weight 1/30
  outputrule('_UNKNOWN', '_UNKNOWN#@RB#0 1 1 3.4 3.4 3.4 3.4 3.4 3.4 3.4 3.4 3.4 3.4'); # adverb, weight 1/30
  outputrule('_UNKNOWN', '_UNKNOWN#@NN#0 1 1 3.4 3.4 3.4 3.4 3.4 3.4 3.4 3.4 3.4 3.4'); # noun singular, weight 1/30
  $manualrulecount += 4;


  #print STDERR "!!!!!!!!!!!! Something's wrong with the underlying output rules database! (Maybe an network file system issue?) \$ruleinstancecount ($ruleinstancecount) != \$totalrulefreqcount_sanitytest!!! ($totalrulefreqcount_sanitytest)!!! \n" if $ruleinstancecount != $totalrulefreqcount_sanitytest;

  print STDERR `date`.". $ruleinstancecount rule instances ($rulenumber unique rules) extracted, $writtenoutrulenumber unique rules (including $manualrulecount manual rules) written out.\n";

  if ($BERKELEY_DB) {
    untie %complexrules;
    untie %phrasecounts;
    undef $outputdb if $outputdbfile;
    undef $db1; undef $db2; undef $db3;
  }

}				# main



















#################### subs

# output rule to database or standard out
sub outputrule {
  my($newsourcestr, $rulebody) = @_;
  if ($outputdbfile) {
    $outputdb->db_put($newsourcestr . "\0", $rulebody . "\0");
  } else {
    print "$newsourcestr\#$rulebody\n";
  }
  $writtenoutrulenumber++;
}


# recursively create and write out complex rules from $elemrule into %complexrules
# $indexedsubrules: array of array of subrules of $elemrule, indexed by source-span start relative to $elemrule's source span start
# $currentrelsourcewordindex: min source word position (relative to $elemrule's source span start) where subrules to be substituted are allowed to start
# recursion start: elementary rules, i.e., no words replaced yet; $currentrelsourcewordindex, $sourcesymbolcount_lowerbound and $recursiondepth are 0
sub createcomplexrules {
  my ($elemrule, $r_source, $r_target, $r_costvector, $r_type, $indexedsubrules, $countunreplacedsourcewords,$countunreplacedtargetwords, $currentrelsourcewordindex, $sourcesymbolcount_lowerbound, $recursiondepth) = @_;

  #Recurse(R, I, Lastsubrule, Recursiondepth):
  # // In Each Recursion Step, We Will Abstract Over A Sub-Rule Contained In R
  # // This Happens From Left To Right Source-Wise, Ensuring That The #I Symbols
  # // Are In The Order #1, #2, ...

  my ($r_minx,$r_miny,$r_maxx, $r_maxy) = @$elemrule[2..5]; # span of R
  
  if ( $countunreplacedsourcewords <= $MAXSOURCEWORDCOUNT && $countunreplacedtargetwords <= $MAXTARGETWORDCOUNT 
       && $sourcesymbolcount_lowerbound + ($r_maxx+1-$r_minx)-$currentrelsourcewordindex <= $MAXSOURCESYMBOLCOUNT
     ) {
    #   Write Out Rule R (unless $MAXTARGETSYMBOLCOUNT condition isn't fulfilled)

    my ($minx,$miny,$maxx, $maxy) = @$elemrule[2..5]; # span of R
    
    # first produce source string
    my $sourcestr = "";
    my $r_source_symbolcount = 0;
    for (my $i = $minx; $i <= $maxx; $i++) {
      my $w = $r_source->[$i-$minx];
      if ($w == -1) {
	# do nothing
      } elsif ($w == 0) {
	# write out source sentence word $i
	$sourcestr .= $sourcestr eq ""? "": " ";
	$sourcestr .= $sourcewords[$i];
	++$r_source_symbolcount;
      } else {
	# write out "@$w"
	$sourcestr .= $sourcestr eq ""? "": " ";
	$sourcestr .= "\@$w";
	++$r_source_symbolcount;
      }
    }

    # now produce target string
    my $targetstr = "";
    my $r_target_symbolcount = 0;
    for (my $i = $miny; $i <= $maxy; $i++) {
      my $w = $r_target->[$i-$miny];
      if ($w == -1) {
	# do nothing
      } elsif ($w == 0) {
	# write out target sentence word $i
	$targetstr .= $targetstr eq ""? "": " ";
	$targetstr .= $targetwords[$i];
	++$r_target_symbolcount;
      } else {
	# write out "@$w"
	$targetstr .= $targetstr eq ""? "": " ";
	$targetstr .= "\@$w";
	++$r_target_symbolcount;
      }
    }

    die "Sanity check \$r_source_symbolcount != \$sourcesymbolcount_lowerbound + (\$r_maxx+1-\$r_minx)-\$currentrelsourcewordindex ($r_source_symbolcount == $sourcesymbolcount_lowerbound + ($r_maxx+1-$r_minx)-$currentrelsourcewordindex) failed; sourcestr=$sourcestr; targetstr=$targetstr; r=".join(" ",@$r_source)."#".join(" ",@$r_target)."#".join(" ",@$r_type)."\n" if $r_source_symbolcount != $sourcesymbolcount_lowerbound + ($r_maxx+1-$r_minx)-$currentrelsourcewordindex;

    if ($r_target_symbolcount <= $MAXTARGETSYMBOLCOUNT) {
      # iterate over the category alternatives, adding each possible resulting rule to %complexrules
      print STDERR "Adding rule: $sourcestr\#$targetstr\#".join("->",@$r_type)."...\n" if $VERBOSE;
      addrules(\$sourcestr,\$targetstr,$r_costvector,$r_type,scalar(@$r_type),[]);
    }
  }				#if $countunreplaced...

  if ($recursiondepth < $MAXSUBSTITUTIONCOUNT) {
    # iterate through all subrules whose rel. source span starts at or after $currentrelsourcewordindex
    # and abstract over each of those (separately)
    for (my $j = $currentrelsourcewordindex; $j < scalar @$indexedsubrules; $j++) {
    LABEL320: 
      for my $subrule (@{$indexedsubrules->[$j]}) {
	my ($sub_minx,$sub_miny,$sub_maxx, $sub_maxy) = @$subrule[2..5]; # span of $subrule
	my $new_countunreplacedsourcewords = $countunreplacedsourcewords-($sub_maxx+1-$sub_minx);
	my $new_countunreplacedtargetwords = $countunreplacedtargetwords-($sub_maxy+1-$sub_miny);
	if (!$keepunbalancedrules &&
	    ( ($new_countunreplacedsourcewords == 0) != ($new_countunreplacedtargetwords == 0) )
	   ) {
	  # unbalanced rule because source still contains words but target doesn't or vice versa
	  # ignore this rule
	  next LABEL320;
	}
	      
	my $new_sourcesymbolcount_lowerbound = $sourcesymbolcount_lowerbound + 1 + ($sub_minx-$r_minx)-$currentrelsourcewordindex ; # the new n.o. source symbols is the old one plus one (new substitution symbol @N) plus n.o. skipped (non-substituted) words
	# check that our lower bound hasn't exceeded the limit $MAXSOURCESYMBOLCOUNT yet
	if ($r_maxx == $sub_maxx) {
	  # case a) we will have processed the whole source span of r; no further substitution will be necessary
	  if ($new_sourcesymbolcount_lowerbound > $MAXSOURCESYMBOLCOUNT) {
	    next LABEL320; # maximum exceeded -> try next subrule
	  }
	} else {
	  # case b) we won't have processed the whole source span of r yet; at least one further substitution will be necessary
	  if ($new_sourcesymbolcount_lowerbound + 1 > $MAXSOURCESYMBOLCOUNT) {
	    next LABEL320; # maximum exceeded -> try next subrule
	  }
	}

	my @new_r_source = @$r_source;
	my @new_r_target = @$r_target;
	# now loop over target positions of subrule $subrule and replace positions in @new_r_target accordingly
	# still have to check if $subrule's target span is still lexicalized within R 
	# (i.e., hasn't been partially replaced yet), do that on-the-fly in the following loop
	for (my $k = $sub_miny-$r_miny; $k<=$sub_maxy-$r_miny; $k++) {
	  next LABEL320 if $r_target->[$k] != 0;
	  $new_r_target[$k] = 
	    $k==$sub_miny-$r_miny? # first position to be changed?
	      $recursiondepth+1 : -1;
	}
	
	# do similar loop over source positions of subrule $subrule and replace positions in @new_r_source accordingly
	for (my $k = $sub_minx-$r_minx; $k<=$sub_maxx-$r_minx; $k++) {
	  print STDERR "Assertion error: \$r_source->[$k] != 0\n" if $r_source->[$k] != 0;
	  $new_r_source[$k] = 
	    $k==$sub_minx-$r_minx? # first position to be changed?
	      $recursiondepth+1 : -1;
	}

	# new cost vector: keep first element at 1; for the other elements, subtract $subrule's value from R's value
	# initialize new cost vector with values from $subrule
	my $refphrasecosts = $subrule->[1]->[0];
	my @new_r_costvector  = @$refphrasecosts;
	unshift @new_r_costvector, 1; # first element of the array will be the n.o. occurrences of the rule, the rest is the $coststr part returned by the PESA phrase extractor
	for (my $k = 1; $k < scalar(@new_r_costvector); $k++) {
	  $new_r_costvector[$k] = $r_costvector->[$k] - $new_r_costvector[$k];
	}

	# create the new type signature
	my $subrule_resultcat = $subrule->[1]->[1]; # result categories of $subrule
	my @new_r_type = (@$r_type[0..scalar(@$r_type)-2], $subrule_resultcat, $$r_type[$#$r_type]);
	
#	print STDERR "Recursing to next level with ".'($sub_minx,$sub_miny,$sub_maxx, $sub_maxy)='."($sub_minx,$sub_miny,$sub_maxx, $sub_maxy)\n";

	createcomplexrules($elemrule, \@new_r_source, \@new_r_target, \@new_r_costvector, \@new_r_type, $indexedsubrules, $new_countunreplacedsourcewords,$new_countunreplacedtargetwords, $sub_maxx+1-$r_minx, $new_sourcesymbolcount_lowerbound, $recursiondepth+1)
      }
    }
  }				# if $recursiondepth
}


# recursively add rules to the hash %complexrules, forking on category alternatives in the r_type array components
sub addrules {
  my ($refsourcestr,$reftargetstr,$r_costvector,$r_type,$typestogocount,$reftypeswritten) = @_;

  #check if we are at the first recursive call
  if (scalar @$reftypeswritten == 0) {
      my $tmpkey; 
      my $tmpvalue;

      # increase source count for source-conditioned Rel. Freq. scores
      $tmpkey = $$refsourcestr . "\#";
      $tmpvalue = $phrasecounts{$tmpkey};
      if (!defined $tmpvalue) {
	$tmpvalue = 1;
      } else {
	$tmpvalue++;
      }
      $phrasecounts{$tmpkey} = $tmpvalue;

      # increase target count for target-conditioned RF scores
      $tmpkey = "\#" .$$reftargetstr;
      $tmpvalue = $phrasecounts{$tmpkey};
      if (!defined $tmpvalue) {
	$tmpvalue = 1;
      } else {
	$tmpvalue++;
      }
      $phrasecounts{$tmpkey} = $tmpvalue;

      # increase count for source/target pair, later to be utilized by source- and target-conditioned RF scores
      $tmpkey = $$refsourcestr . "\#" . $$reftargetstr;
      $tmpvalue = $phrasecounts{$tmpkey};
      if (!defined $tmpvalue) {
	$tmpvalue = 1;
      } else {
	$tmpvalue++;
      }
      $phrasecounts{$tmpkey} = $tmpvalue;
  }

  #check if we are at the recursion base level
  if ($typestogocount==0) {

    # add rule to %complexrules
    my $key = $$refsourcestr . '#' . $$reftargetstr . '#' . join(' ',@$reftypeswritten);
    my $value = $complexrules{$key};
    $ruleinstancecount++; # update rule counter (this is just a statistics for info purposes)

    # determine rule's cost vector
    my @costvector;
    if (defined($value)) {
      # rule exists already
      # do the correct operation on stored costvector $value and to-be-added costvector $r_costvector
      # for now, stored costvector only contains one value: frequency, so just increment that value
      #++($value->[0]);
      
      @costvector = split(" ", $value);
      # first element is the rule count - increment it
      $costvector[0]++;
      # the other elements are the PESA alignment costs - choose the elementwise minimum
      for (my $i = 1; $i < scalar(@costvector); $i++) {
	if ($$r_costvector[$i] < $costvector[$i]) {
	  $costvector[$i] = $$r_costvector[$i];
	}
      }
    } else {
      # rule doesn't exist yet: store frequency count 1 and the PESA alignment costs
      @costvector = @$r_costvector;
    }
    $complexrules{$key} = join(" ", @costvector);

    # get result category from $typeswritten
    # my @typeswrittenarray = split(" ", $typeswritten);
    my $resultcat = $reftypeswritten->[-1];
    my $resultcatcount = $resulttypecounts{$resultcat};
    if (defined($resultcatcount)) {
      ++$resultcatcount;
    } else {
      $resultcatcount = 1;
    }
    $resulttypecounts{$resultcat} = $resultcatcount;

  } else {
    # we are either at beginning or somewhere in the recursion, but not at the base level

    my @categoryalternatives = split(" ", $r_type->[--$typestogocount]);
    #assert#    print STDERR "Assertion error: \@categoryalternatives is empty!\n" if $#categoryalternatives == -1;
    # $typeswritten = " $typeswritten" unless $typeswritten eq "";
    unshift @$reftypeswritten, ''; # prepend dummy to the front of the typeswritten array
    for my $type (@categoryalternatives) {
      # replace dummy with $type
      $reftypeswritten->[0] = $type;
      # call addrules recursively
      addrules($refsourcestr,$reftargetstr,$r_costvector,$r_type,$typestogocount, $reftypeswritten);
    }
    # we have to drop the first element again from the array, since the array is a reference re-used by the calling
    # instance of addrules
    shift @$reftypeswritten; 
  }
}


sub str2list {
  $_ = shift;
  die "Target/source/parsetree mismatch (nonequal number of sentences) at sentence number $sentencenumber. Byebye." if !defined($_);
  chomp($_);
  s/^\s+|\s+$//g;		# delete leading and trailing spaces
  #  my @words = split(/\s+/, $sourcesentence);
  split;
}


sub findparsespans {
  my ($spanstart) = @_;
  if ($currentparsetokenspos <= $#currentparsetokens) {      # only if we are not beyond the end of the array, we can succeed
    my $token = $currentparsetokens[$currentparsetokenspos++];
    if ($token eq "(") {
      #new nonterminal root
      my $root = $currentparsetokens[$currentparsetokenspos++];
      #print "root $root\n";
      my $spanpos = $spanstart;
      my @backpointerinfo = (); # list of spanstart/spanlength info pairs of the child nodes
      while ($currentparsetokens[$currentparsetokenspos] ne ")") { # iterate over all child nodes
	my $childspanlength = findparsespans($spanpos);
	push @backpointerinfo, ($spanpos, $childspanlength); # store the child's span
	$spanpos += $childspanlength;
      }
      # arrived at the ")" telling us that this span is completed
      $currentparsetokenspos++;
      my $spanlength = $spanpos-$spanstart;
      if ($spanlength == 0) {
	# parsing error - do nothing, below we will fail
      } else {
	my $erroroccurred = 0;
	if ($root ne "S1") {	# exclude the case of the start symbol
	  my $hashindex = "$spanstart $spanlength";
	  #print "$spanstart $spanlength\n";
	  
	  my $hashvalue = $currentparsespans{$hashindex};
	  if (defined $hashvalue) {
	    # this span has already been indexed - this must be a unary rule node (one child only)
	    $erroroccurred = 1 if (scalar(@backpointerinfo) != 2); # assert that we only have one child
	    # add nonterminal $root to the list indexed by our span
	    #	    $hashvalue .= " $root";
	    $hashvalue->[0] .= " $root"; # add our category to category field
	  } else {
	    if (scalar(@backpointerinfo) == 2) {
	      # unary rule (only one child node) - don't want backpointer info
	      # this case occurs if and only if this node is a POS tag since it's child node is not in the $currentparsespans hash
	      @backpointerinfo = (); # remove backpointer to the terminal symbol's span, which is identical to this span
	    }
	    unshift @backpointerinfo, $root; # insert category as first element
	    $hashvalue = \@backpointerinfo; # store reference to array whose first element is the category and whose other elements are children's span info pairs
	  }
	  $currentparsespans{$hashindex} = $hashvalue;
	}
	return $spanlength if !$erroroccurred;
      }
    } else {			#if ($token eq " ") {
      # we've arrived at the word level
      #print "word ".$token."\n";
      if ($currentparsetokens[$currentparsetokenspos] eq ")") {
	return 1;		# length of span is one (the word we just encountered)
      } else {
	# parsing error - do nothing, below we will fail
      }
    }
  }
  die "Parsing error in parsing log file, sentence number $sentencenumber around token number $currentparsetokenspos+1: ".join(",",@currentparsetokens)."\nByebye!\n";
}


# look up array of words referenced by $phrasewords as subarray of array referenced 
# by $sentencewords and return the matching spans in from of a list each of whose
# entries is of the form "$spanstart $spanlength" ($spanlength==1 means spanning one word)
sub lookupspans {
  my ($phrasewords,$sentencewords) = @_;
  my @matchingspans = ();
  my $spanlength = scalar @$phrasewords;
  # efficiency is not an issue in this sub, so do stupid search
 LABEL438:
  for (my $i=0;$i<=scalar(@$sentencewords)-scalar(@$phrasewords);$i++) {
    for (my $j=0;$j<scalar(@$phrasewords);$j++) {
      next LABEL438 if $phrasewords->[$j] ne $sentencewords->[$i+$j];
    }
    # haven gotten so far means we found a match
    push @matchingspans,"$i $spanlength";
  }
  return @matchingspans;
}


sub printhash {
  my ($hashref) = @_;

  my ($key, $value);
  while (($key,$value) = each %$hashref) { 
    #  my $neglogfrequency = - log $value->[0];
    print STDERR "$key --> $value\n";
  }
}
