
use warnings;
use strict;
#use Encode::Arabic;

#use Encode qw/encode decode/;
use Storable;

#Warning: Tree:R has the weird peculiarity of assuming that the to-be-changed variable \@objects passed to its methods points to an empty array (if not that array is just appended to)
use Tree::R;			# obtained at http://users.tkk.fi/~jolma/pub/snapshots/Tree-R-0.05.tar.gz
use List::Util qw(first min max);
use Fcntl;
use File::Temp qw/ tempfile tempdir /;
#use DB_File; # uncomment this line if you don't have Berkeley DB installed on your unix system
use BerkeleyDB;			# uncomment this line if you don't have Berkeley DB installed on your unix system
use Getopt::Long;


#binmode STDIN, ":utf8";
#binmode STDERR, ":utf8";
#binmode STDOUT, ":utf8";

# settings

my $BERKELEY_DB = 1;		# set to zero if Berkeley DB is not available

# options and input params

# puts everything into probability space - use for STTK
#my $neglogprobs2probs
#use: HtdToTD.sh | ConvertLogTDtoProbs.pl for that

my $originalfeaturecount = 11; #now 11 b/c of Pharaoh phrases #13
# n.o. relevant features from input rule files (all features beyond this count are cut off)
#0		   0,		# glue feature
#1		   1,		# rule application counter
#2		   $targetwordcount, # target word counter
#3		   -log($frequency/$resultcount), # neglog (base e) result-conditioned RF
#4		   -log($rulebodyfrequency/$targetfrequency), # neglog target-cond. RF
#5		   -log($rulebodyfrequency/$sourcefrequency), # neglog source-cond. RF
#6   		   -log($frequency) #neglog frequency of rule ($NegLogOccurrenceCountFeature == true)
#	or:	   $frequency,	# frequency of rule (NegLogOccurrenceCountFeature == false)
#     (filterrules.pl will always put this feature into neg-log space)
#7-12		   @costvector[1..$#costvector]  	 # PESA costs (6 values)

# n.o. additional features being created by filterrules.pl
my $additionalfeaturecount = 10;
my $lexicalfeatureoffset = $originalfeaturecount;
#0  purely-lexical feature (rule body consists only of terminals)
#1  purely-abstract feature (rule body consists only of nonterminals)
#2  _X rule feature
#3  unbalancedness penalty
#4  rareness penalty
#5  squared neg-log result-cond. RF
#6-9 repetition of phrase costs but zeroed out for rules that are not lexical

# tells us that we have to take exp(-x) to get occurrence count from 7th feature
my $NegLogOccurrenceCountFeature = 1;

# by default, we will recompute the result-conditioned RF feature based on the cumulative result-category counts of the rules.
# This way, rule bases can be created independently from training corpora split into several parts and then joined
# via filterrules.pl
my $RecomputeResultCondRF = 1;

# set this less than inf if you want to restrict the number of
# abstraction sites in the rules
my $maxabstractioncount = 30;

my $keepunbalancedpunctuationrules = 1; #keep rules that have unbalanced source-target punctuation / word behavior 

# set this to true to keep natural _S rules (e.g. @S#@1#@_S#0.5 ....) in the rule DB (this will cause double-rules creating _S because if glue rules are enabled)
my $KeepNaturalSentenceRules = 0;

# set this to false if you want to filter out unbalanced rules
my $keepunbalancedrules = 0;
my $bal_limit = 3;		# max. ratio between source word count and target word count and vice versa
my $MeanTargetSourceRatio = 1;
# if set to 0, removes all rules containing _X in the rule body, but keeps lexical rules with result category _X
my $AbstractOverX = 1;

my $crazyclusterrules = 0;	# 1 for crazy (only keep first two letters from categories), 2 for very crazy (only keep first letter from categories), 3 for insane (everything becomes _X - this together with $maxabstractioncount=0 plus some reordering rule simulates STTK mode )

# set to true what you want to be written out
my $gluerules = 1;
#my $exitgluemoderules = 0;
#my $entergluemoderules = 1;
#my $joininggluerule = 1;
my $unknownrules = 1;

my $tmpdir;
my $outputdbfile;
my $VERBOSE = 0;
my $vocabFile = '';

# binary vector B1_B2_..._Bn indicating by Bi=1 that i-th feature should be written out, and by Bi=0 that it shouldn't
my $RestrictToFeatures = '';
my $sttk = 0; # set crazyclusterrules = 3, maxabstractioncount = 0, add reordering rule, change glue feature into reordering feature
#my $KeepOccurrenceCountFeature = 0;
#my $LostOccurrenceCountFeature = 0;
my $maxsourcephraselength = 0;	# upper bound for test-sentence filtering

# removes a rule source/target/type if its occurrence count is less than $minoccurrencecount and there is another rule source/target2/type2 with higher occurrence count
my $minoccurrencecount = 2;
# in case such a low-occurrence-count rule DOES have the highest occurrence count amongst all rules with same source,
# it will be added if its source length is less or equal $MaxSourceLengthLowOccurrenceRules
my $MaxSourceLengthLowOccurrenceRules = 1;

# set this greater than zero if you want to remove a rule if it has a smaller
# lexical-source conditioned relative frequency than $minlexsourcecondrf and
# there is another rule with the same source with higher occurrence count
my $minlexsourcecondrf = 0.05;

my $cachesize = 512; # Berkeley DB cache size in MB

die if !GetOptions('tmpdir=s' => \$tmpdir, 'cachesize=s' => \$cachesize, 'verbose' => \$VERBOSE, 'outputdbfile=s' => \$outputdbfile, 'r=s' => \$vocabFile, 'KeepNaturalSentenceRules!' => \$KeepNaturalSentenceRules, 'AbstractOverX!' => \$AbstractOverX, 'maxsourcephraselength=s' => \$maxsourcephraselength, 'minoccurrencecount=s' => \$minoccurrencecount, 'MaxSourceLengthLowOccurrenceRules=s' => \$MaxSourceLengthLowOccurrenceRules, 'MeanTargetSourceRatio=s' => \$MeanTargetSourceRatio, 'minlexsourcecondrf=s' => \$minlexsourcecondrf, 'RecomputeResultCondRF!' => \$RecomputeResultCondRF, 'NegLogOccurrenceCountFeature!' => \$NegLogOccurrenceCountFeature, 'STTK' => \$sttk, 'RestrictToFeatures=s' => \$RestrictToFeatures, 'CrazyClusterRules=s' => \$crazyclusterrules);

# $tmpdir = File::Spec->tmpdir() if (!$tmpdir);
$tmpdir = "." if (!$tmpdir);

if ($sttk) {
  $crazyclusterrules = 3;
  $maxabstractioncount = 0;
}

my @featureRestrictionVector;
if ($RestrictToFeatures) {
  @featureRestrictionVector = split('_', $RestrictToFeatures);
  die "Wrong length of feature restriction vector $RestrictToFeatures" if scalar(@featureRestrictionVector) != $originalfeaturecount+$additionalfeaturecount;
}

my %vocCheck = ();
if ($vocabFile) {
  die "Parameter --maxsourcephraselength must be specified!\n" if !$maxsourcephraselength;
  open(SOURCE, "< $vocabFile");
  while (<SOURCE>) {
    chomp($_);
    my @sentence = split(' ',$_);
    for (my $phraselength = 1; $phraselength <= $maxsourcephraselength; $phraselength++) {
      for (my $phrasestart = 0; $phrasestart+$phraselength <= scalar(@sentence); $phrasestart++) {
	$vocCheck{join(' ',@sentence[$phrasestart..$phrasestart+$phraselength-1])} = 1;
      }
    }
  }
  close(SOURCE);
}



# global variables

# create my own temp dir as subdir of $tmpdir, it and its contents will 
# automatically removed at program end
my $mytmpdir = tempdir( "tmp_extractrules_XXXXXXXXX", DIR => $tmpdir, CLEANUP => 1);
#my $mytmpdir = tempdir( "tmp_extractrules_XXXXXXXXX", DIR => $tmpdir, CLEANUP => 0);

my $sentencenumber = 0;		# counts the sentences processed
my $phrasepairnumber = 0;	# counts the phrase pairs processed
my $ruleinstancecount = 0;	# counts the number of instances of rules extracted
my $writtenoutrulenumber = 0;	# number of rules written out to STDOUT / database

# will contain the complex rules, with $key= "$source\#$target\#$result" and $value=\@costvector
# also stores for each rule source side the maximum occurrence count achieved by a rule with that source side
my %complexrules = ();		
#my %phrasecounts = ();          # will map each rule source side to the maximum occurrence count achieved by a rule with that source side
my $outputdb = ();		# the output rule database - a B-tree
my %resulttypecounts = ();	# counts the total n.o. rules in %complexrules that have the same result type (keys are categories, values are numbers); this will enable relative-frequency calculations when writing out the rules




{				# main block
  print STDERR 'filterrules.pl - started @ '. `date`;
  print STDERR ($outputdbfile ?
		"Outputting rules into newly created Berkeley DB file $outputdbfile\n" :
		"Outputting rules in text-form to STDOUT\n");

  if ($outputdbfile) {
    if ($BERKELEY_DB) {
      $outputdb = new BerkeleyDB::Btree 
	-Cachesize => 2**20 * $cachesize/4, 
	  -Filename => $outputdbfile, 
	    -Flags => DB_CREATE|DB_TRUNCATE,
	      -Property => DB_DUP
		|| die "Can't create output rule Berkeley DB database $outputdbfile! Bye.";
    } else {
      die "Variable \$BERKELEY_DB is set to 0, therefore the rules cannot be output as database file but just as plain text! Bye.\n";
    }
  }

  my $db1 = tie %complexrules,
    #    "BerkeleyDB::Hash", 
    "BerkeleyDB::Btree", 
      -Cachesize => 2**20 * $cachesize *3/4, 
	-Filename => "$mytmpdir/complexrules", 
	  -Flags => DB_CREATE | DB_TRUNCATE
	    || die "Can't create temp database complexrules in $mytmpdir!";
#  my  $db2 = tie %phrasecounts, 
##      "BerkeleyDB::Hash", 
#      "BerkeleyDB::Btree", 
#	-Cachesize => 2**29, 
#	  -Filename => "$mytmpdir/phrasecounts", 
#	    -Flags => DB_CREATE # |DB_TRUNCATE
#	      || die "Can't create temp database phrasecounts in $mytmpdir!";
   

  while (<>) {
    $sentencenumber++;
    print STDERR "Processing rule number $sentencenumber...\r" if $sentencenumber % 10000 == 0;

    chomp;
    my @rule = split("#",$_);

    my $skiprule = 0;
    if (!$AbstractOverX && ($rule[0] =~ /\@_X/)) {
      # _X occurs in rule's body - skip rule
      $skiprule = 1;
    }
#    if (!$AbstractOverX && ($rule[2] eq '@_X')) {
#      # rule has result _X - change it to _S
#      $rule[2] = '@_S';
#    }

    if ($rule[2] eq '@_G' || $rule[0] =~ /^\@_G/ || $rule[0] eq '_UNKNOWN' || 
	(!$KeepNaturalSentenceRules && $rule[2] eq '@_S')) {
      # skip enter / exit glue rules as well as unknown-rules
      # in case $KeepOccurrenceCountFeature is false, skip also natural _S result rules
      $skiprule = 1;
    }
    
    if (!$skiprule) {

      if ($crazyclusterrules == 1) {
	$rule[0] =~ s/(\@_*\w\w)\S+/$1/g; # substitute e.g. @NNS to @NN in sourceside
	$rule[2] =~ s/(\@_*\w\w)\S+/$1/g; # substitute e.g. @NNS to @NN in resultcategory
      } elsif ($crazyclusterrules == 2) {
	$rule[0] =~ s/(\@_*\w)\S+/$1/g; # substitute e.g. @NP to @N in sourceside
	$rule[2] =~ s/(\@_*\w)\S+/$1/g; # substitute e.g. @NP to @N in resultcategory
      } elsif ($crazyclusterrules == 3) {
	$rule[0] =~ s/\@\S+/\@_X/g; # substitute all @something to @_X
	$rule[2] =~ s/\@\S+/\@_X/g; # substitute @something to @_X in resultcategory
      }

      my @rulesourcewords = split(" ", $rule[0]);
      my @ruletargetwords = split(" ", $rule[1]);
      my @rulecostvector  = split(" ", $rule[3]);
      my $rulefreq = $NegLogOccurrenceCountFeature? sprintf("%.2f", exp(-$rulecostvector[6])) : $rulecostvector[6]; # rule frequency count


      #restrict rules to those that have sources that occur in the source data + abstract elememnts
      my $vocMatched = 1;	#by default it passes
      my $phraseToCheck = '';	#used for the voc test
      #checking for mixed punctuation and words
      my $hasPunct_src =0;
      my $hasWords_src =0;
      #compute the abstraction count
      my $abstractioncount = 0;
      foreach (my $i = 0; $i < scalar @rulesourcewords; $i++) {
	if ($rulesourcewords[$i] =~ /^\@/ ) {
	  # source word starts with @ -- must be an arg marker
	  $abstractioncount++;
	  # this abstraction stops a consecutive phrase, so check the current phrase
	  $vocMatched = 0 unless $phraseToCheck eq '' || $vocCheck{$phraseToCheck};
	  $phraseToCheck = '';	# reset phraseToCheck
	  #print "Reject:",join(' ',@rulesourcewords), " DUE: $rulesourcewords[$i]\n";
	} else {
	  if ($vocabFile) {
	    # append the current word to the phrase-to-check
	    $phraseToCheck .= ($phraseToCheck eq '' ? '' : ' ') . $rulesourcewords[$i];
	  }
	  if (!$keepunbalancedpunctuationrules) {
	    if ($rulesourcewords[$i] =~ /^[,\!\"\'\`\%\(\)\+\-\/\:\;]+$/) {
	      $hasPunct_src=1;
	    } else {
	      $hasWords_src=1;
	    }
	  }
	}
      }
      # have to check one more time, for last phrase of the rule
      $vocMatched = 0 unless $phraseToCheck eq '' || $vocCheck{$phraseToCheck};
      #print "Reject:",join(' ',@rulesourcewords), " DUE: $rulesourcewords[$i]\n";

      # target side punctuation check
      my $hasPunct_tgt =0;
      my $hasWords_tgt =0;
      if (!$keepunbalancedpunctuationrules) {
	foreach (my $i = 0; $i < scalar @ruletargetwords; $i++) {
	  if ($ruletargetwords[$i] =~ /^\@/ ) {
	    # is always same as source abstraction count#	    $abstractioncount_tgt++;
	  } else {
	    if ($ruletargetwords[$i] =~ /^[,\!\"\'\`\%\(\)\+\-\/\:\;]+$/) {
	      $hasPunct_tgt=1;
	    } else {
	      $hasWords_tgt=1;
	    }
	  }
	}
      }

      my $punctPassed = $keepunbalancedpunctuationrules || 
	($hasWords_src == $hasWords_tgt && $hasPunct_src == $hasPunct_tgt);
      if (!$punctPassed) {
#	print STDERR "Unbalanced punctuation for rule: $_ (hasPunct_tgt=$hasPunct_tgt, hasPunct_src=$hasPunct_src, hasWords_tgt=$hasWords_tgt, hasWords_src=$hasWords_src)\n"       if ($vocMatched && $rulefreq >= $minoccurrencecount);
      }
	
      # count source- and target-terminal symbols
      my $sourcewordcount = scalar(@rulesourcewords) - $abstractioncount;
      my $targetwordcount = scalar(@ruletargetwords) - $abstractioncount;
      die "In rule $_: sourcewordcount ($sourcewordcount) longer than maxsourcephraselength ($maxsourcephraselength)!" if $sourcewordcount > $maxsourcephraselength && $maxsourcephraselength;
	
      # check whether rule is balanced
      my $isbalanced = 
	( (($sourcewordcount == 0) == ($targetwordcount == 0))
	  # rule has no source terminals iff rule has no target terminals
	  || $abstractioncount == 0 # or rule is lexical (in that case we want to allow n-to-0 phrase pairs)
	);
      if ($targetwordcount>0 && $sourcewordcount > 0) {
	# check of source / target ratio is okay
	my $ratioSrcTgt = $sourcewordcount / $targetwordcount;
	if ( $ratioSrcTgt>$bal_limit || $ratioSrcTgt<1.0/$bal_limit) {
	  $isbalanced = 0;
	}
      }
#      print STDERR "Bad source/target length ratio for rule: $_\n" if !$isbalanced && ($vocMatched && $rulefreq >= $minoccurrencecount);
	
      # caution: don't exclude rules whose vocab doesn't match yet b/c that will mess up the frequency count
      if ( 
	  (
	   # first condition: balancedness
	   ( $isbalanced
	     || $keepunbalancedrules
	   )
	  ) &&
	  (
	   # third condition: should be no glue rule
	   $rulecostvector[0] == 0
	  ) &&
	  (
	   # condition: abstraction count
	   $abstractioncount <= $maxabstractioncount
	  ) &&
	  (
	   # condition: punctuation balanced (or $keepunbalancedpunctuationrules == 1)
	   $punctPassed
	  )
	 ) {

	# update result category occurrence frequency
	my $rc = $resulttypecounts{$rule[2]};
	$rc = 0 if (!defined($rc));
	$resulttypecounts{$rule[2]} = $rc + $rulefreq;
    
	if ($vocMatched) {

	  # update total rule instance counter
	  $ruleinstancecount += $rulefreq;

	  # set lexical-feature ($lexicalfeatureoffset == $originalfeaturecount)
	  $rulecostvector[$lexicalfeatureoffset] = ($abstractioncount == 0? 1 : 0);

	  # set abstract-feature
	  $rulecostvector[$originalfeaturecount+1] = ($sourcewordcount == 0 && $targetwordcount == 0? 1 : 0);

	  # set _X rule feature
	  $rulecostvector[$originalfeaturecount+2] = ($rule[2] eq '@_X'? 1 : 0);

	  # set unbalancedness penalty
	  $rulecostvector[$originalfeaturecount+3] = abs($MeanTargetSourceRatio * $sourcewordcount - $targetwordcount) ;
	  

	  my $keytmp = join('#', @rule[0..2]);
	  my $valuetmp = $complexrules{$keytmp};
	  if (!defined($valuetmp)) {
	    # put rule occurrence count feature into frequency space if still in neg-log space
	    $rulecostvector[6] = $rulefreq;
	    $valuetmp = join(' ', @rulecostvector);
	    #	    print "Adding $valuetmp...\n";
	  } else {
	    # this case occurs when we're combining rule files
	    # add the rule frequency counts and max all costs to minimize impact of inconsistency problems
	    my @tmpcostvector = split(' ', $valuetmp);
	    $rulefreq += $tmpcostvector[6]; # revise this rule's frequency (this info will be used below for the max-occurrence count statistics)
	    $tmpcostvector[6] = $rulefreq; # store revised rule freq
	    for (my $i = 0; $i < scalar(@tmpcostvector); $i++) {
	      if ($i != 6 && $rulecostvector[$i] > $tmpcostvector[$i]) {
		$tmpcostvector[$i] = $rulecostvector[$i];
	      }
	    }
	    $valuetmp = join(' ', @tmpcostvector);
	  }
	  $complexrules{$keytmp} = $valuetmp;

	  # check if we have set a new record with the occurrence count of this rule amongst all rules with the same source
	  my $MaxOccCount = $complexrules{$rule[0]};
	  if (!defined($MaxOccCount) || $rulefreq > $MaxOccCount) {
	    # new record - store occurrence count
	    $complexrules{$rule[0]} = $rulefreq;
	  }
	}
      }
    }
  }

  {
    # write out the rules in %complexrules
    print STDERR "\nWriting out the rules...\n";
    my $totalrulefreqcount_sanitytest = 0;
    my $rulenumber = 0;
    my ($key, $value);
    while (($key,$value) = each %complexrules) { 
      my ($sourcestr,$targetstr,$result) = split("#", $key);
      if (!defined($targetstr)) {
	# this is just the max-occurrence count statistics - ignore this entry
      } else {
	$rulenumber++;
	print STDERR "Processing rule number $rulenumber...\r" if $rulenumber % 10000 == 0;
    
	my @rulecostvector = split(" ", $value);
#	print "Retrieved $value\n"; 
	my $rulefreq = $rulecostvector[6]; # frequency of rule
	$totalrulefreqcount_sanitytest += $rulefreq;

	if ($RecomputeResultCondRF) {
	  # re-compute result-conditioned neg log prob
	  $rulecostvector[3] = -log($rulefreq/$resulttypecounts{$result});
	}

	if (
	    (
	     # condition: minimum lexical-source conditioned RF
	     exp(-$rulecostvector[5]) >= $minlexsourcecondrf
	     && 
	     # condition: occurrence count >= $minoccurrencecount
	     $rulefreq >= $minoccurrencecount-0.1 # -0.1 to avoid rounding issues
	    )
	    || # or: highest occurrence count amongst all rules with the same source side and maximum source length $MaxSourceLengthLowOccurrenceRules
	    ($rulefreq >= $complexrules{$sourcestr}-0.1 && scalar(my @rulesourcewords=split(' ',$sourcestr)) <= $MaxSourceLengthLowOccurrenceRules)
	    || # sanity test
	    ($rulefreq > $complexrules{$sourcestr} && die 'Sanity check failed!')
	   ) {
	  
	  # add some more features

	  # rareness penalty
	  # 1 for freq-one rules, then levelling off with slope determined 
	  # by feature weight, with limit 0 for $rulefreq --> infinity
	  $rulecostvector[$originalfeaturecount+4] = exp(1-$rulefreq);

	  # squared neg log result-cond. RF
	  $rulecostvector[$originalfeaturecount+5] = $rulecostvector[3] ** 2;

	  # repitition of phrase costs, zeroed out for non-lexical rules
	  $rulecostvector[$originalfeaturecount+6] = $rulecostvector[$lexicalfeatureoffset]? $rulecostvector[7] : 0;
	  $rulecostvector[$originalfeaturecount+7] = $rulecostvector[$lexicalfeatureoffset]? $rulecostvector[8] : 0;
	  $rulecostvector[$originalfeaturecount+8] = $rulecostvector[$lexicalfeatureoffset]? $rulecostvector[9] : 0;
	  $rulecostvector[$originalfeaturecount+9] = $rulecostvector[$lexicalfeatureoffset]? $rulecostvector[10] : 0;

	  # change occurrence count feature back into neg-log space
	  $rulecostvector[6] = -log($rulefreq);

	  # output the rule
#	  for (my $i = 0; $i<=scalar @rulecostvector; $i++) 
#	    {
#	      print $rulecostvector[$i] . ' ';
#	    } 
#	  print "\n";
	  my $coststr = join(' ', @rulecostvector);
	  outputrule($sourcestr,join("#",($targetstr,$result,$coststr)));
	}
      }
    }

    print STDERR "!!!!!!!!!!!! Something's wrong with the underlying complexrules database! (Maybe an network file system issue?) \$ruleinstancecount ($ruleinstancecount) != \$totalrulefreqcount_sanitytest ($totalrulefreqcount_sanitytest)!!! \n" if $ruleinstancecount != $totalrulefreqcount_sanitytest;
  }


  
  #to get a statistics about how many nonterminals there are
  my $nonterminalcount = 0;

  #add some manual rules now
  my $manualrulecount = 0;

  # glue rules: theory is: any result can be transformed into _X
  # (enter glue mode: e.g. @NP#@1#@_X); _X rules can be joined (@_X
  # @_X#@1 @2#@_X); _X can be transformed into any nonterminal (exit
  # glue mode: e.g. @_X#@1#@NP) in practise, avoid doubling the
  # n.o. hyps in the chart tables by using _S instead of _X 
  # also, the exit-glue mode rules will be inserted only when needed
  # directly by the parser (the parser only applies the self-rule to
  # _S only when the whole chart cell contains no hypergraph nodes
  # with non-"_S" result nonterminals)

  #the joining glue rule
  if ($gluerules) {
    if (!$sttk) {
      outputrule("\@_S \@_G", "\@1 \@2\#\@_S\#1 1 0" . concatenatenumber(0, $originalfeaturecount+$additionalfeaturecount-3));
      $manualrulecount++;
    } else {
      # phrase decoder simulation: add a regular glue rule without glue feature...
      outputrule("\@_S \@_G", "\@1 \@2\#\@_S\#0 1 0" . concatenatenumber(0, $originalfeaturecount+$additionalfeaturecount-3));
      # and a binary reodering rule (reordering two _S rules) with glue (now meaning: reordering) feature
      outputrule("\@_S \@_S", "\@2 \@1\#\@_S\#1 1 0" . concatenatenumber(0, $originalfeaturecount+$additionalfeaturecount-3));
      $manualrulecount += 2;
    }
#    my $source = '@_G';
#    my $target = '@1';
#    for (my $i=2; $i < 100; $i++) {
#      $source .= ' @_G';
#      $target .= ' @' . $i;
#      # write out rule $source#$target#@_S#costvector, where costvector's first value
#      # (glue feature counting n.o. pieces that were glued to produce the translation output) is $i 
#      # and rule app count is $i-1 (to compete with but not out-compete regular binary rules)
#      outputrule($source, 
#		 #		 $target . '#@_S#' . $i . ' ' . ($i-1) . ' 0 0 0 0 0 0 0 0 0 0');
#		 $target . '#@_S#' . 0 . ' ' . 0 . ' 0 0 0 0 0 0 0 0 0 0');
#      $manualrulecount++;
#    }
  }

#  if ($sttk) {
#    # to enable STTK simulation, also add a reordering rule that will be the only rule without the GLUE feature
#    # (need to allow negative feature weight for glue feature, so that reordering can be penalized!)
#    outputrule("\@_S \@_S", "\@2 \@1\#\@_S\#1 1 0 0 0 0 0 0 0 0 0 0");
#  }
  

  # enter and exit glue-mode rules, one for each nonterminal symbol
  print STDERR "Nonterminals:\n";
  my $key;
  my $value;
  while (($key,$value) = each %resulttypecounts) { 
    print STDERR $key . " ";
    if ($key eq '@_S' || $key eq '@_G') {
      # ignore _S and _G
    } else {
    $nonterminalcount++;
#      if ($exitgluemoderules) {
#	# add exit-glue-mode rule for $key
#	outputrule('@_S', "\@1\#$key\#1 0 0 0 0 0 0 0 0 0 0 0");
#	$manualrulecount++;
#      }
      # unary glue rules are next 
      if ($gluerules) {
	# let each category directly finish up as a full sentence _S)
	#	# if rule has already been extracted in a natural way - don't write
	#	# out extra unary glue rule
	#	if (!exists($complexrules{"$key\#\@1\#\@_S"})) {
	outputrule($key, 
		   "\@1\#\@_S\#" . ($sttk? 0 : 1) . " 1 0" . 
		   concatenatenumber(-log($value), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );
	$manualrulecount++;
	  #	}
	# now add enter-glue-mode rule for $key
	outputrule($key, "\@1\#\@_G\#0 0 0" . 
		   concatenatenumber(-log($value), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );
	$manualrulecount++;
      }
    }
  }
  print STDERR "\nTotal number of nonterminal symbols (excluding _S and _G): $nonterminalcount\n";

  if ($unknownrules) {
    #unknown rules: here only the relative weights of unknown rules w.r.t. each other matter
    #to make sure that their feature weights are positive (i.e., they are costs), we add the same value to all neglog prob features
    if ($crazyclusterrules == 0) {
      outputrule('_UNKNOWN', '_UNKNOWN#@NNP#0 1 1' . 
		   concatenatenumber(-log(9/1000), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );	 # proper name, weight 9/10; -ln(9/10)=0.1
      outputrule('_UNKNOWN', '_UNKNOWN#@NP#0 1 1' . 
		   concatenatenumber(-log(9/1000), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );  # az 06/03/09: allow also directly NP to cope with the new unary-rule handling of extractrules.pl
      outputrule('_UNKNOWN', '_UNKNOWN#@JJ#0 1 1' . 
		   concatenatenumber(-log(1/3000), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );   # adjective, weight 1/30
      outputrule('_UNKNOWN', '_UNKNOWN#@RB#0 1 1' . 
		   concatenatenumber(-log(1/3000), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );  # adverb, weight 1/30
      outputrule('_UNKNOWN', '_UNKNOWN#@NN#0 1 1' . 
		   concatenatenumber(-log(1/3000), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );   # noun singular, weight 1/30
      $manualrulecount += 5;
    } elsif ($crazyclusterrules == 1) {
      # substitute e.g. @NNS to @NN in resultcategory
      outputrule('_UNKNOWN', '_UNKNOWN#@NN#0 1 1' . 
		   concatenatenumber(-log(9/1000), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );
      outputrule('_UNKNOWN', '_UNKNOWN#@NP#0 1 1' . 
		   concatenatenumber(-log(9/1000), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );	 # az 06/03/09: allow also directly NP to cope with the new unary-rule handling of extractrules.pl
      outputrule('_UNKNOWN', '_UNKNOWN#@JJ#0 1 1' . 
		   concatenatenumber(-log(1/3000), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );
      outputrule('_UNKNOWN', '_UNKNOWN#@RB#0 1 1' . 
		   concatenatenumber(-log(1/3000), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );
      $manualrulecount += 4;
    } elsif ($crazyclusterrules == 2) {
      # substitute e.g. @NP to @N in resultcategory
      outputrule('_UNKNOWN', '_UNKNOWN#@N#0 1 1' . 
		   concatenatenumber(-log(9/1000), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );
      outputrule('_UNKNOWN', '_UNKNOWN#@J#0 1 1' . 
		   concatenatenumber(-log(1/3000), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );
      outputrule('_UNKNOWN', '_UNKNOWN#@R#0 1 1' . 
		   concatenatenumber(-log(1/3000), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );
      $manualrulecount += 3;
    } elsif ($crazyclusterrules == 3) {
      # substitute @something to @_X in resultcategory
      outputrule('_UNKNOWN', '_UNKNOWN#@_X#0 1 1' . 
		   concatenatenumber(-log(9/1000), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );
      $manualrulecount += 1;
    }

    if ($gluerules) {
      # also output a copy-glue-rule for '.' for the case that '.' was removed during training
      outputrule('.', '.#@_G#0 1 1' . 
		   concatenatenumber(-log(1/10), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );
      $manualrulecount += 1;
      # and a full-sentence rule in the very unlikely case that '.' was removed during 
      # training AND the whole sentence consists only of a single '.'
      outputrule('.', '.#@_S#1 1 1' . 
		   concatenatenumber(-log(1/10), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );
      $manualrulecount += 1;
    }
  }

  print STDERR "$writtenoutrulenumber rules written out. ";
  print STDERR 'filterrules.pl - finished @ '. `date`;

  if ($BERKELEY_DB) {
    undef $outputdb if $outputdbfile;
  }

}				# main






#################### subs

# output rule to database or standard out
sub outputrule {
  my($newsourcestr, $rulebody) = @_;
  if ($RestrictToFeatures) {
    my @tmp = split('#', $rulebody);
    my @costvector = split(' ', $tmp[2]);
    my @newcostvector = ();
    die "Assertion error" if scalar(@costvector) != $originalfeaturecount + $additionalfeaturecount;
    for (my $i=0; $i < scalar(@costvector); $i++) {
      if ($featureRestrictionVector[$i]) {
	push @newcostvector, $costvector[$i];
      }
    }
    die "All values of feature restriction vector $RestrictToFeatures set to false!" if scalar(@newcostvector) == 0;
    $tmp[2] = join(' ', @newcostvector);
    $rulebody = join('#', @tmp);
  }
      
  if ($outputdbfile) {
    $outputdb->db_put($newsourcestr . "\0", $rulebody . "\0");
  } else {
    print "$newsourcestr\#$rulebody\n";
  }
  $writtenoutrulenumber++;
}

# concatenates number $num $times times, each $num occurrence preceded by ' '.
sub concatenatenumber {
  my ($num, $times) = @_;
  my $result = '';
  while (--$times >= 0) {
    $result .= ' ' . $num;
  }
  return $result;
}
