#!/usr/bin/env perl

# STDIN: rule file
# ARG1,...,ARGn: source language corpus/corpora (typically a test set)
# STDOUT: all source words w from the test set(s) (separated by \n) which don't have a direct translation rule w#...#...

use strict;
use warnings;


use Getopt::Long;

my $OutputEachInstance=0;
die if !GetOptions('OutputEachInstance!' => \$OutputEachInstance);

# hash in the source sides of the rules
my %covered_words = ();
while (<STDIN>){
  my $line = $_;
  chomp($line);
  my @parts = split(/( *\# *| *\|\|\| *)/, $line);
  my $src = $parts[0];
  $covered_words{$src}++;
}

# iterate through test set(s) and output all unknown words
foreach my $testset (@ARGV) {
  open(TESTSET, "< $testset");
  while (<TESTSET>) {
    chomp($_);
    my @words = split(' ',$_);
    foreach my $w (@words) {
      # note the ++ below, which ensures that we output each unknown only once
      if (!$covered_words{$w}) {
	print $w . "\n";
	if (!$OutputEachInstance) {
	  $covered_words{$w}++;
	}
      }
    }
  }
  close(TESTSET);
}
