#!/usr/bin/env perl

#==========================================================================
# Can be used in a pipe as pre-filtering step to filter out
# low-frequency n-grams using a different processor
# See also convertgooglelm.pl




use warnings;
use strict;

use List::Util qw(first min max);
use Fcntl;
use File::Temp qw/ tempfile tempdir /;
#use DB_File; # uncomment this line if you don't have Berkeley DB installed on your unix system
# Note: Be careful with post-increment / decrement when tying a hash into Berkeley DB. This doesn't quite work in certain cases!
use Getopt::Long;
#use if $BERKELEY_DB, BerkeleyDB =>;

#binmode STDIN, ":utf8";
#binmode STDERR, ":utf8";
#binmode STDOUT, ":utf8";

# settings

# Command line parameters


my $tmpdir;
my $outputdbfile;
my $VocabularyFile;
my $KeepTruecaseFile;
my $verbose = 0;
my $vocabRestrictionFile;
my $cachesize = 3400;		# Berkeley DB cache size in MB
my $OnlyUseBerkeleyDBForDumping = 0; # set this to 1 to do all calculations with a regular hash and only convert to Berkeley DB in the end
my $MaxMarkovOrder=4;		# max. Markov order allowed
my $MinHistoryFrequency=1;
my $MinFrequency=1;



die if !GetOptions('MinHistoryFrequency=s' => \$MinHistoryFrequency, 'MinFrequency=s' => \$MinFrequency);

#my $TrainingCorpusFile = shift(@ARGV);
die "Too many arguments (expected: 0). Superfluous arguments: " . join(' ',@ARGV) if scalar(@ARGV) != 0;
# if we decide not to use a Berkeley DB, then we have to set $OnlyUseBerkeleyDBForDumping to true


{				# main block
  
  my $lineNumber = 0;
 LINELOOP:  while (<STDIN>) {
#    $lineNumber++;
#    print STDERR "At line number $lineNumber: $_" if $lineNumber % 1000000 == 1;

    my @sentence_str = split;	# split on whitespace (important: can
                                # be space but also tab, and ignored
                                # the trailing NL)
      
    # last element is the frequency count - pop it
    # ignore it if the frequency count is too low
    next LINELOOP if $sentence_str[-1] < $MinFrequency;
    print $_;
  }
}
