# settings
my $BERKELEY_DB = 0;		# set to zero if Berkeley DB is not available

use warnings;
use strict;
#use Encode::Arabic;

#use Encode qw/encode decode/;
use Storable;

use List::Util qw(first min max);
use Fcntl;
use File::Temp qw/ tempfile tempdir /;
#use DB_File; # uncomment this line if you don't have Berkeley DB installed on your unix system
use Getopt::Long;
use if $BERKELEY_DB, 'BerkeleyDB';			# uncomment this line if you don't have Berkeley DB installed on your unix system

#binmode STDIN, ":utf8";
#binmode STDERR, ":utf8";
#binmode STDOUT, ":utf8";

# settings



# input params


my $tmpdir;
my $outputdbfile;
my $VERBOSE = 0;
my $cachesize = 512; # Berkeley DB cache size in MB
my $minsourcecondrf = 0;
my $switchsourcetarget = 0;

GetOptions('switchsourcetarget' => \$switchsourcetarget, 'tmpdir=s' => \$tmpdir, 'cachesize=s' => \$cachesize, 'verbose' => \$VERBOSE, 'minsourcecondrf=s' => \$minsourcecondrf, 'outputdbfile=s' => \$outputdbfile);

# $tmpdir = File::Spec->tmpdir() if (!$tmpdir);
$tmpdir = "." if (!$tmpdir);




# global variables

# create my own temp dir as subdir of $tmpdir, it and its contents will 
# automatically removed at program end
#!!my $mytmpdir = tempdir( "tmp_extractrules_XXXXXXXXX", DIR => $tmpdir, CLEANUP => 1);
#my $mytmpdir = tempdir( "tmp_extractrules_XXXXXXXXX", DIR => $tmpdir, CLEANUP => 0);

my $sentencenumber = 0;	# counts the sentences processed
my $phrasepairnumber = 0; # counts the phrase pairs processed
my $ruleinstancecount = 0; # counts the number of instances of rules extracted
my $writtenoutrulenumber = 0; # number of rules written out to STDOUT / database
my %complexrules = ();		# will contain the complex rules, with $key= "$source\#$target\#$result" and $value=\@costvector
my %phrasecounts = ();          # will contain occurrence counts of source side and target side of the rules
my $outputdb = (); # the output rule database - a B-tree
my %resulttypecounts = ();	# counts the total n.o. rules in %complexrules that have the same result type (keys are categories, values are numbers); this will enable relative-frequency calculations when writing out the rules

#for use within the while loop, but using subroutine, thus defined before main block
my %currentparsespans = ();
my @currentparsetokens = ();
my $currentparsetokenspos = 0;
my @sourcewords;
my @targetwords;
my $sourceposition = $switchsourcetarget? 1 : 0;
my $targetposition = $switchsourcetarget? 0 : 1;

{				# main block
#  die "No output file specified" if !$outputdbfile;
  print STDERR ($outputdbfile ?
		"Outputting phrases into newly created Berkeley DB file $outputdbfile\n" :
		"Outputting phrases in text-form to STDOUT\n");


  if ($outputdbfile) {
    die "BERKELEY_DB set to 0!" if !$BERKELEY_DB;
    no warnings; no strict;
    my $db1 = tie %complexrules,
      #    "BerkeleyDB::Hash", 
      "BerkeleyDB::Btree", 
	-Cachesize => 2**20 * $cachesize, 
	  -Filename => $outputdbfile, 
	    -Flags => DB_CREATE | DB_TRUNCATE
	      || die "Can't create output rule Berkeley DB database $outputdbfile! Bye.";
  }
  my $lastcostvectorlength = -1;
  while (<>) {
    $sentencenumber++;
    print STDERR "Processing phrase pair number $sentencenumber...\r" if $sentencenumber % 10000 == 0;

    chomp;
    # something like: voulons essayer ||| are to try ||| 1 0.000831281 1 0.00404281 2.718
    my @rule = split(/ \|\|\| /,$_);
    my @costvector = split(' ', $rule[2]);
    #    print STDERR join('#',@rule); print STDERR scalar(@rule) . "\n";
    if (scalar @rule != 3 || ($lastcostvectorlength != -1 && $lastcostvectorlength != scalar(@costvector))) {
      print STDERR "Warning! Skipping invalid phrase pair $sentencenumber: $_\n";
    } elsif ($costvector[2] < $minsourcecondrf) {
#      print STDERR "Skipping too-low-source-cond. rel. freq. phrase pair $sentencenumber: $_\n";
    } else {
      $lastcostvectorlength = scalar(@costvector);
      my $currententries = $complexrules{$rule[$targetposition]};
      if (!defined($currententries)) {
	# first entry for this English phrase
	$currententries = '';
      } else {
	# entries exist - append
	$currententries .= '#';
      }
      # convert probs into costs and remove the phrase penalty
      pop @costvector; # remove phrase penalty
      # put all probs into neg-log space and round them to save space
      foreach my $item (@costvector) {
#	$item = sprintf("%.2f", -log($item));
	$item = sprintf("%.3g", -log($item));
      }
      #      # push in two zeros to match cost vector size to PESA costvector size
      #      push @costvector, (0, 0);
      # append French phrase and cost vector
      $currententries .= $rule[$sourceposition] . '#' . join(' ', @costvector);

      # store the entries under the English phrase as key
      $complexrules{$rule[$targetposition]} = $currententries;
    }
  }

  print STDERR "$sentencenumber phrase pairs.\n";

  if (!$outputdbfile) {
    print STDERR "Outputting phrase rules to standard out...\n";
    my $rulenumber = 0;
    my $rulesdeleted = 0;
    my $sourcephrasesdeleted = 0;
    my $targetphrasesdeleted = 0;
    my $rulebodiesdeleted = 0;
    my ($key, $value);
    while (($key,$value) = each %complexrules) { 
      $rulenumber++;
      print STDERR "Processing rule $rulenumber...\r" if ($rulenumber-1) % 50000 == 0;
      print "$key\#$value\n";
    }
    print STDERR "Done.\n";
  }
}				# main



















#################### subs



