#!/nfs/islpc7_1/Andreas/ActivePerl-5.8/bin/perl
###!/usr/bin/env perl

#use strict;
use warnings;
use Getopt::Long;

# set to 1 to remove non-ASCII symbols (likely to be Chinese/Arabic/etc. source symbols)
my $RemoveNonAscii = 1;

# set to 1 to lower-case the translations
# if 0, we will upper-case sentences first words (also handling multiple sentences per line) and preserve
# the case of all other words
my $Lowercasing = 1;

# If set to 1, we will move end-of-sentence marks ('.','!','?') from
# the end to the beginning of each sentence. This also works for lines
# with multiple sentences
my $ShiftSentenceMarks = 0;

# when there is no end of sentence mark at the end of an input line,
# then add a dot (if this option is 1)
my $AddDotForMissingSentenceMarks = 0;

# set to 1 translation output comes from Hiero
my $HieroFormat = 0;

my %vocab = ();

die if !GetOptions('RemoveNonAscii!' => \$RemoveNonAscii, 'HieroFormat!' => \$HieroFormat, 'Lowercasing!' => \$Lowercasing, 'ShiftSentenceMarks!' => \$ShiftSentenceMarks, 'AddDotForMissingSentenceMarks!' => \$AddDotForMissingSentenceMarks);

die "--ShiftSentenceMarks only works if --AddDotForMissingSentenceMarks is set" if $ShiftSentenceMarks && !$AddDotForMissingSentenceMarks;

# die "No text file with truecasing data specified (option --TruecasingData)" if (!$Lowercasing && !$TruecasingData);

#fire up output buffering
$|=1;
while (<>) {
  chomp;			# get rid of the newline symbol (will otherwise disturb us when removing multiple whitespace below)
  s/^\s+//g;			# handle possible \r after \n

  my @hiero_parts;
  if ($HieroFormat) {
    # temporarily chop off the sentence number and the cost vector -- will be reassembled below after normalization is done
    @hiero_parts = split(/ *\|\|\| */, $_);
    $_ = $hiero_parts[1];
  }
  my $line = PostProcessBTEC($_);
  print "$line\n";		# print out the modified line, followed by a newline symbol

}

sub PostProcessBTEC{
  #Punctuation and case handling ################################################################################################
  $_ = $_[0];

  # get rid of numtag / _NULL artifacts
  s/\b(_NULL|numtag)\b *//g;

  if ($RemoveNonAscii) {
    my $ascii = "\x20-\x7F";
    s/([^$ascii]+)//g;		# remove non-ascii
  }

  $_ = lc($_) if $Lowercasing;

  if ($ShiftSentenceMarks) {
    my @tokens = split(' ',$_);

    # if there's no end-of-sentence mark as first token, prepend a dot
    unshift @tokens, '.' unless $tokens[0] =~ /^(\.|\!|\?)$/ ;

    my $sentence_start = 0;
    for (my $i = 1; $i < scalar(@tokens); $i++) {
      
      if ($i == $#tokens || $tokens[$i] =~ /^(\.|\!|\?)$/) {
	$i++ if $i == $#tokens;
	#      print $i . join(' ',@tokens[$sentence_start..$i]
	# the current token is an end-of-sentence mark (for the next sentence on the line) or the last token of the line
	# move previous end-of-sentence mark to the end of this chunk
	@tokens[$sentence_start..$i-1] = (@tokens[$sentence_start+1..$i-1], $tokens[$sentence_start]);

	# remove comma(s) at the start of a sentence
	while ($tokens[$sentence_start] eq ',') {
	  shift @tokens;
	}

	if (!$Lowercasing) {
	  # uppercase first letter of first word of sentence
	  $tokens[$sentence_start] = ucfirst($tokens[$sentence_start]);
	}

	# update sentence_start marker
	$sentence_start = $i;
      }
    }
  
    $_ = join(' ',@tokens);

  } else {			# else of $ShiftSentenceMarks

    # if $AddDotForMissingSentenceMarks and there's no end-of-sentence mark at the very end, then add a dot to the end
    $_ .= ' .' if $AddDotForMissingSentenceMarks && !( /[.?!\"]$/ );

    # remove beginning of line commas, dots, !, or ?
    s/^(?:[;:,.?!] *)+//;

    # remove comma & co. after full-stop, !, or ? (but not after e.g. 'etc.')
    s/( [.!?])(?: *[,;:])+/$1/g;

    if (!$Lowercasing) {
      # uppercase first letter of first word of each sentence
      s/^(.)/\u$1/g;
      s/( [.!?]) (.)/$1 \u$2/g;
    }
  }
    
  #degenerate cases where we see things like  ", ." which we'll change into simply "."
  s/(?:[,;:] )+(\.|\?|\!)/$1/g; 

  # English quotation marks at the end of the sentence have the weirdness of 
  # always coming before the dot
  s/\" \./\. \"/g;

  # collapse multiple identical subsequent EOS marks or commas etc. into one
  s/([.!?,;:])( \1)+/$1/g;

  # for an inconsistent sequence of EOS marks, collapse it into a single dot
  # but require space before the tokens to avoid collapsing 'p.m. ?' into p.m.
  s/( \.| \!| \?)( \.| \!| \?)+/ \./g;

  # now collapse double dots such as in 'p.m. .'
  s/\.( \.)+/\./g;
    



  #=================================================================================
  # speech-specific stuff (writing out numbers and certain symbols)

  s/%/ percent /g;


#  #need to convert any numbers that havent been converted yet
#  my @wordParts = split(/ /, $_);
#  my $count = 0;
#  foreach $w (@wordParts) {
      
#    if ($w =~ m/^[0-9]+$/) {
#      if ($w =~ m/19([0-9][0-9])/) {
#	$w = "nineteen ".num2word($1);
#      } else {
#	$w = num2word($w);
#      }
#    }
#    $wordParts[$count] = $w;
#    $count++;
#  }
#  $_ = join(' ', @wordParts);

  #kill left over @@ from numtagging
  s/@ *@//g;

  # remove '- .'
  s/- *([\.,;:?!])/$1/g;
  # remove ' . -'
  s/([\.,;:?!]) *-/$1/g;

  # collapse word - word into word-word
  s/ - /-/g;

  # replace a followed by vowal by an
  s/ a ([aeiouAEIOU])/ an $1/g;

  # replace '$' / 'US$' / 'US $' by 'US dollars'
  #  s/(US ?)?\$ ?(dollars?)?/ US dollars /g;
  s/\bdollars\b/\$/gi;
  s/(U\.? ?S\.? ?)(\$ ?)+/ U.S. \$ /g;
  s/\$/ dollars /g;

#  #Standard BTEC handling ################################################################################################
#  #from /project/IWSLT/IWSLT-2006/Official-Preprocessing
#  s/\./ \. /g;
#  s/\?/ \? /g;
#  s/\!/ \! /g;
#  s/,/ , /g;
#  s/:/ : /g;
#  s/;/ ; /g;
#  s/"/ " /g;
#  s/\|/ /g;
#  s/[ ][ ]*/ /g;

#  #but now join back some chars
#  s/a \. m \./a\.m\./g;
#  s/p \. m \./p\.m\./g;


  #############################################################
  #De tokenization
  s/(\w)\s+\'s/$1\'s/g;		#attach 's to the leading word
  s/(I)\s+\'m/$1\'m/g;		#attach 'm to I
  s/(I)\s+\'d/$1\'d/g;		#attach 'd to I
  s/(I)\s+\'ll/$1\'ll/g;	#attach 'll to I
  s/([W|w]e)\s+\'ll/$1\'ll/g;	#attach 'll to We 
  s/([W|w]e)\s+\'d/$1\'d/g;	#attach 'd to We 
  s/([W|w]e)\s+\'re/$1\'re/g;	#attach 're to We 

  s/([Y|y]ou)\s+\'d/$1\'d/g;	#attach 'd to You
  s/([Y|y]ou)\s+\'re/$1\'re/g;	#attach 're to You
  s/([Y|y]ou)\s+\'ll/$1\'ll/g;	#attach 'd to You

  s/([D|d]o)\s+n\'t/$1n\'t/g;	#attach n't to do
  s/([D|d]id)\s+n\'t/$1n\'t/g;	#attach n't to do
  s/([H|h]ave)\s+n\'t/haven\'t/g; #attach n't to do
  s/([C|c]a)\s+n\'t/$1n\'t/g;	#attach n't to do
  s/([C|c]an)\s+n\'t/$1n\'t/g;	#attach can n't 
  s/(is)\s+n\'t/isn\'t/g;	#attach n't to do
  s/(wo)\s+n\'t/won\'t/g;	#attach n't to do
  
  s/ \'ll/ will/g;		#backup case
  s/ \'ve/ have/g;		#backup case
  s/ \'m/ is/g;			#backup case
  s/ n\'t/ not/g;		#backup case
  s/ \'re/ are/g;		#backup case
  s/ \'s/ is/g;			#backup case
  s/ \'d/ I\'d/g;		#backup case
  s/should\'s /should be /g;	#backup case

  s/PTT[A-Z]+//g;		#remove leftover tags

  s/(\w)\s+\'/$1\'/g;		#attach any remaining freestanding ' to the leading word




  #match nist normalization
  $_ = NormalizeText_NIST($_);
  #  $_ = NormalizeText_BLEU($_);



  s/\s+/ /g;			# one space only between words
  s/^\s+//;			# no leading space
  s/\s+$//;			# no trailing space


  if ($HieroFormat) {
    # reassemble the Hiero n-best list parts
    $hiero_parts[1] = $_;
    $_ = join(' ||| ', @hiero_parts);
  }

  return $_;
}


#####################################################################3

sub num2word{
  $n=shift;
  my $originalN = $n;
  @a=qw(0 one two three four five six seven eight nine ten
	eleven twelve thir four fif six seven eigh nine twen thir for);
  $w = "";
  for ($c=0xf4240;
       $c;$c/=1000) {
    if ($j=int$n/$c) {
      $w.=$a[int$j/100].' hundred ',$j%=100if$j>99;$w.=
	$a[int$j/10+($j>49?10:18)].'ty ',$j%=10if$j>19;$w.=$a[$j].($j>12?'teen ':' ')if
	  $j;$w.=$c>1000?'million ':$c>1?'thousand ':'';$n%=$c;
    }
  }
  if (int$originalN<100) {
    #print "Splitting $originalN\n";
    $w =~ s/^s\+//g;
    $w =~ s/s\+$//g;
    my @wparts = split(/ /, $w);
    if (scalar(@wparts)==2) {
      $w =~ s/ /\-/g;
      $w =~ s/-$//g;
    }
  }
  $w
}

#################################
 

sub NormalizeText_BLEU{
  my $strPtr = $_[0];

  # language-independent part:
  $strPtr =~ s/^\s+//;
  $strPtr =~ s/\n/ /g;		# join lines
  $strPtr =~ s/(\d)\s+(\d)/$1$2/g; #join digits

  # language-dependent part (assuming Western languages):
  #$strPtr =~ tr/[A-Z]/[a-z]/ unless $preserve_case;
  $strPtr =~ s/([^A-Za-z0-9\-\'\.,])/ $1 /g; # tokenize punctuation (except for alphanumerics, "-", "'", ".", ",")
  $strPtr =~ s/([^0-9])([\.,])/$1 $2 /g; # tokenize period and comma unless preceded by a digit
  $strPtr =~ s/([\.,])([^0-9])/ $1 $2/g; # tokenize period and comma unless followed by a digit
  $strPtr =~ s/([0-9])(-)/$1 $2 /g; # tokenize dash when preceded by a digit
  $strPtr =~ s/\s+/ /g;		# one space only between words
  $strPtr =~ s/^\s+//;		# no leading space
  $strPtr =~ s/\s+$//;		# no trailing space
  my $ascii = "\x20-\x7F";
  $strPtr =~ s/([^$ascii])\s+([^$ascii])/$1$2/g; # combine sequences of non-ASCII characters into single words

  return $strPtr;

}

sub NormalizeText_NIST{

  my ($norm_text) = @_;
 
  # language-independent part:
  $norm_text =~ s/<skipped>//g; # strip "skipped" tags
  $norm_text =~ s/-\n//g;	# strip end-of-line hyphenation and join lines
  $norm_text =~ s/\n/ /g;	# join lines
  $norm_text =~ s/(\d)\s+(?=\d)/$1/g; #join digits
  $norm_text =~ s/&quot;/"/g;  # convert SGML tag for quote to "
    $norm_text =~ s/&amp;/&/g;	# convert SGML tag for ampersand to &
  $norm_text =~ s/&lt;/</g;	# convert SGML tag for less-than to >
  $norm_text =~ s/&gt;/>/g;	# convert SGML tag for greater-than to <
 
  # language-dependent part (assuming Western languages):
  $norm_text = " $norm_text ";
  #$norm_text =~ tr/[A-Z]/[a-z]/ unless $preserve_case;
  $norm_text =~ s/([\{-\~\[-\` -\&\(-\+\:-\@\/])/ $1 /g; # tokenize punctuation
  $norm_text =~ s/([^0-9])([\.,])/$1 $2 /g; # tokenize period and comma unless preceded by a digit
  $norm_text =~ s/([\.,])([^0-9])/ $1 $2/g; # tokenize period and comma unless followed by a digit
  $norm_text =~ s/([0-9])(-)/$1 $2 /g; # tokenize dash when preceded by a digit
  $norm_text =~ s/\s+/ /g;	# one space only between words
  $norm_text =~ s/^\s+//;	# no leading space
  $norm_text =~ s/\s+$//;	# no trailing space
 
  return $norm_text;
}
 
#################################



