#!/usr/bin/env perl

use strict;
use warnings;
use Getopt::Long;

my $Lowercasing = 0;
my $TruecasingData = '';

my %vocab = ();

die if !GetOptions('Lowercasing!' => \$Lowercasing, 'TruecasingData=s' => \$TruecasingData);

# die "No text file with truecasing data specified (option --TruecasingData)" if (!$Lowercasing && !$TruecasingData);


while (<>) {
  chomp; # get rid of the newline symbol (will otherwise disturb us when removing multiple whitespace below)

  # now we really start

  my @tokens = split(' ',$_);

  # if there's no end-of-sentence mark as first token, prepend a dot
  unshift @tokens, '.' unless $tokens[0] =~ /^(\.|\!|\?)$/ ;

  my $sentence_start = 0;
  for (my $i = 1; $i < scalar(@tokens); $i++) {

    if ($i == $#tokens || $tokens[$i] =~ /^(\.|\!|\?)$/) {
      $i++ if $i == $#tokens;
      #      print $i . join(' ',@tokens[$sentence_start..$i]
      # the current token is an end-of-sentence mark (for the next sentence on the line) or the last token of the line
      # move previous end-of-sentence mark to the end of this chunk
      @tokens[$sentence_start..$i-1] = (@tokens[$sentence_start+1..$i-1], $tokens[$sentence_start]);

      if (!$Lowercasing) {
	# uppercase first letter of first word of sentence
	$tokens[$sentence_start] = ucfirst($tokens[$sentence_start]);
      }

      # update sentence_start marker
      $sentence_start = $i;
    }
  }
  
  $_ = join(' ',@tokens);
  
  # collapse multiple identical subsequent EOS marks into one
  s/(\.|\!|\?)( \1)+/$1/g;

  # for an inconsistent sequence of EOS marks, collapse it into a single dot
  # but require space before the tokens to avoid collapsing 'p.m. ?' into p.m.
  s/( \.| \!| \?)( \.| \!| \?)+/ \./g;

  # now collapse double dots such as in 'p.m. .'
  s/\.( \.)+/\./g;

  print "$_\n"; # print out the modified line, followed by a newline symbol
}
