# perl -S restrictparses.pl < diff.log parses.en
# output: subset of parses.en's lines that fit aligned.en according to the diff logfile

use strict;
use warnings;

my %removelines = ();
# my $deletions_to_ignore = 0;
open(PARSES, "< $ARGV[0]") or die "Can't open parse tree file $ARGV[0]";
my @delete_history = ();
while (<STDIN>) {
  if ( /^(d|a)(\d+) (\d+)$/ ) {
    my $diffcommand = $_;
    # if $1 eq 'd': $2 is starting point of line numbers to delete, $3 is number of lines to delete from that line on
    # if $1 eq 'a': $2 is starting point of line numbers to add, $3 is number of lines to add from that line on
    
    for (my $i = $2; $i < $2+$3; $i++) {
      if ($1 eq 'd') {
	#	if ($deletions_to_ignore) {
	#	  $deletions_to_ignore--;
	#	} else {
	$removelines{$i} = 1;
	push @delete_history, $i;
	#	}
      } else {
	# skip over the actual add-line
	$_ = <STDIN>;
	if (scalar(@delete_history) > 0) {
	  my $x = pop @delete_history;
	  delete $removelines{$x};
	} else {
	  print STDERR "Warning: add-line without previous delete. Diff-command: $diffcommand" . "Line-to-add: $_";
	}
      }
    }
  } else {
    print STDERR "Warning: Non-understood diff output line: $_";
  }
}

my $parseline = 0;
my $lines_removed_count = 0;
while (<PARSES>) {
  $parseline++;
  if ($removelines{$parseline}) {
    $lines_removed_count++;
  } else {
    print $_;
  }
}
    
print STDERR "$lines_removed_count parses removed.\n"
