#!/usr/bin/env perl

use warnings;
use strict;
#use Encode::Arabic;

#use Encode qw/encode decode/;
use Storable;

#Warning: Tree:R has the weird peculiarity of assuming that the to-be-changed variable \@objects passed to its methods points to an empty array (if not that array is just appended to)
use Tree::R;			# obtained at http://users.tkk.fi/~jolma/pub/snapshots/Tree-R-0.05.tar.gz
use List::Util qw(first min max);
use Fcntl;
use File::Temp qw/ tempfile tempdir /;
#use DB_File; # uncomment this line if you don't have Berkeley DB installed on your unix system
use BerkeleyDB; # uncomment this line if you don't have Berkeley DB installed on your unix system
use Getopt::Long;


#binmode STDIN, ":utf8";
#binmode STDERR, ":utf8";
#binmode STDOUT, ":utf8";

# settings

my $BERKELEY_DB = 1; # set to zero if Berkeley DB is not available

# input params


my $tmpdir;
my $outputdbfile;
my $VERBOSE = 0;
my $cachesize = 1024; # in MB
my $RestrictToFeatures = '';

GetOptions('tmpdir=s' => \$tmpdir, 'cachesize=s' => \$cachesize, 'verbose' => \$VERBOSE, 'outputdbfile=s' => \$outputdbfile, 'RestrictToFeatures=s' => \$RestrictToFeatures);

# $tmpdir = File::Spec->tmpdir() if (!$tmpdir);
$tmpdir = "." if (!$tmpdir);




# global variables

# create my own temp dir as subdir of $tmpdir, it and its contents will 
# automatically removed at program end
#!!my $mytmpdir = tempdir( "tmp_extractrules_XXXXXXXXX", DIR => $tmpdir, CLEANUP => 1);
#my $mytmpdir = tempdir( "tmp_extractrules_XXXXXXXXX", DIR => $tmpdir, CLEANUP => 0);

my $sentencenumber = 0;	# counts the sentences processed
my $phrasepairnumber = 0; # counts the phrase pairs processed
my $ruleinstancecount = 0; # counts the number of instances of rules extracted
my $writtenoutrulenumber = 0; # number of rules written out to STDOUT / database
my %complexrules = ();		# will contain the complex rules, with $key= "$source\#$target\#$result" and $value=\@costvector
my %phrasecounts = ();          # will contain occurrence counts of source side and target side of the rules
my $outputdb = (); # the output rule database - a B-tree
my %resulttypecounts = ();	# counts the total n.o. rules in %complexrules that have the same result type (keys are categories, values are numbers); this will enable relative-frequency calculations when writing out the rules

#for use within the while loop, but using subroutine, thus defined before main block
my %currentparsespans = ();
my @currentparsetokens = ();
my $currentparsetokenspos = 0;
my @sourcewords;
my @targetwords;

my @featureRestrictionVector;

{				# main block
  print STDERR ($outputdbfile ?
		"Outputting rules into newly created Berkeley DB file $outputdbfile\n" :
		"Outputting rules in text-form to STDOUT\n");

  if ($RestrictToFeatures) {
    # first remove comments from Restriction vector
    $RestrictToFeatures =~ s/\#[^_]*//g;
    @featureRestrictionVector = split('_', $RestrictToFeatures);
  }


  if ($outputdbfile) {
    if ($BERKELEY_DB) {
      $outputdb = new BerkeleyDB::Btree 
	-Cachesize => 2**20 * $cachesize, 
	  -Filename => $outputdbfile, 
	    -Flags => DB_CREATE|DB_TRUNCATE,
	      -Property => DB_DUP
	      || die "Can't create output rule Berkeley DB database $outputdbfile! Bye.";
    } else {
      die "Variable \$BERKELEY_DB is set to 0, therefore the rules cannot be output as database file but just as plain text! Bye.\n";
    }
  }



  while (<>) {
    $sentencenumber++;
    print STDERR "Processing rule number $sentencenumber...\r" if $sentencenumber % 10000 == 0;

    chomp;

    my @rule = split(/ *\# */,$_);
    if (scalar(@rule) == 4) {
      outputrule($rule[0],join("#",@rule[1..$#rule]));
    } elsif (scalar(@rule) == 3) {
      outputrule($rule[0],join("#",($rule[1], '@_X', $rule[2])));
    } else {
    print "Processing rule: ", $_, "\n";
    die "rules2db: wrong rule file format";
    }
  }

  print STDERR "$writtenoutrulenumber rules written out.\n";

  if ($BERKELEY_DB) {
    undef $outputdb if $outputdbfile;
  }

}				# main






#################### subs


# output rule to database or standard out
sub outputrule {
  my($newsourcestr, $rulebody) = @_;
  if ($RestrictToFeatures && $newsourcestr ne '@_COUNT') {
    my @tmp = split('#', $rulebody);
    my @costvector = split(' ', $tmp[2]);
    my @newcostvector = ();
    die "Error: mismatch length of costvector " . $tmp[2] . "with length of RestrictToFeatures vector $RestrictToFeatures in rule $newsourcestr\#$rulebody" if scalar(@costvector) != scalar(@featureRestrictionVector);
    for (my $i=0; $i < scalar(@costvector); $i++) {
      if ($featureRestrictionVector[$i] != 0) {
	push @newcostvector, $costvector[$i];
      }
    }
    die "All values of feature restriction vector $RestrictToFeatures set to false!" if scalar(@newcostvector) == 0;
    $tmp[2] = join(' ', @newcostvector);
    $rulebody = join('#', @tmp);
  }
      
  if ($outputdbfile) {
    $outputdb->db_put($newsourcestr . "\0", $rulebody . "\0");
  } else {
    print "$newsourcestr\#$rulebody\n";
  }
  $writtenoutrulenumber++;
}
