#!/usr/bin/env perl

use warnings;
use strict;
#use Encode::Arabic;

#use Encode qw/encode decode/;
use Storable;

use List::Util qw(first min max);
use Fcntl;
use File::Temp qw/ tempfile tempdir /;
#use DB_File; # uncomment this line if you don't have Berkeley DB installed on your unix system
#use BerkeleyDB;			# uncomment this line if you don't have Berkeley DB installed on your unix system
use Getopt::Long;


#binmode STDIN, ":utf8";
#binmode STDERR, ":utf8";
#binmode STDOUT, ":utf8";

# settings

#my $BERKELEY_DB = 1;		# set to zero if Berkeley DB is not available

# options and input params

my $RetainAll=0;
my $Lexical=0;
my $Abstract=0;
my $Mixed=0;
my $XRule=0;
my $ExtendedCat=0;
my $Syntactic=0;
my $NTCount=-1;
my $BooleanFeature=-1;
my $FrequencyOffset = 6;

my $OutputCounts = 0;

my $cachesize = 512;		# Berkeley DB cache size in MB

die if !GetOptions('RetainAll' => \$RetainAll, 'Lexical' => \$Lexical, 'Abstract' => \$Abstract, 'Mixed' => \$Mixed, 'XRule' => \$XRule, 'ExtendedCat' => \$ExtendedCat, 'Syntactic' => \$Syntactic, 'BooleanFeature=s' => \$BooleanFeature, 'NTCount=s' => \$NTCount, 'OutputCounts!' => \$OutputCounts, 'FrequencyOffset=s' => \$FrequencyOffset);

my $totalrulecount = 0;
my $totalruleinstancecount = 0;
my $filteredrulecount = 0;
my $filteredruleinstancecount = 0;

my @filteredrules = ();

while (<>) {
  chomp;
  my ($src,$tgt,$result,$costs) = split('#', $_);

  next if $src eq '@_COUNT';
  my @rulecostvector = split(' ',$costs);
  next if ($result eq '@_G' || $src =~ /^\@_G/ || $src eq '_UNKNOWN' || $rulecostvector[0] > 0);
  
  my $rulefreq = sprintf("%.2f", exp(-$rulecostvector[$FrequencyOffset]));

  $totalrulecount++;
  $totalruleinstancecount += $rulefreq;

  my $islexical = !($tgt =~ /(^| )\@/);
  my $isabstract = !($tgt =~ /^[^\@]| [^\@]/);
  my $ismixed = !$islexical && !$isabstract;
  my $isxrule = ($src . ' ' . $result) =~ /(^| )\@[^ ]*_X/;
  my $isextendedcat = !$isxrule && 
    (($src . ' ' . $result) =~ /(^| )\@[^ ]*[\/\\\+]/);
  my $issyntactic = !$isxrule && !$isextendedcat;

  my $pass_booleanfeature = $BooleanFeature != -1 && ($rulecostvector[$BooleanFeature] > 0);
  my $ntcountplus1 = $NTCount+1;
  my $pass_ntcount = $NTCount != -1 && ($tgt =~ /(^| )\@$NTCount/) && !($tgt =~ /(^| )\@$ntcountplus1/);

  if ($RetainAll || $Lexical && $islexical || $Syntactic && $issyntactic || $Abstract && $isabstract || $Mixed && $ismixed || $XRule && $isxrule || $ExtendedCat && $isextendedcat || $pass_booleanfeature || $pass_ntcount) {

    $filteredrulecount++;
    $filteredruleinstancecount += $rulefreq;

    push @filteredrules, [$rulefreq, $_];
    
  }
}

print STDERR "Sorting rules...\n";
# now sort filtered rules descendingly by frequency
my @sortedrules = sort {$b->[0] <=> $a->[0]} @filteredrules;

print STDERR "Writing out rules...\n";
foreach my $ruleref (@sortedrules) {
  print $ruleref->[0] . "\t" if $OutputCounts;
  print $ruleref->[1] . "\n";
}

print STDERR "$totalrulecount rules ($totalruleinstancecount instances) total; $filteredrulecount rules ($filteredruleinstancecount instances) after filtering.\n";
