#!/usr/bin/env perl
#Script that starts multiple worker bees to run MER training
#Takes in runline - that gets passed directly to the binary
use Getopt::Long;
use strict;

#assumes the the binary - FastTranslateChart and MER is in the path
#submission scripts
#my $condor_bin = "condorsubmit-stderr-malbecall.sh";
#my $condorlarge_bin = "condorsubmit-stderr-malbeclarge.sh";

my $condor_bin = $ENV{'CONDOR_SMALL'};
my $condormerge_bin = $ENV{'CONDOR_MERGE'};
my $condorlarge_bin = $ENV{'CONDOR_LARGE'};
#binaries - lets ensure that they are on the path
my $mer_bin = `which MER`;
my $decode_bin = `which FastTranslateChart`;
my $merger_bin = `which mergeopti.pl`; # expdir/merge0/ expdir/iter0/ expdir/merge1
chomp($mer_bin); 
chomp($decode_bin);
chomp($merger_bin);

#error checking
die "Condor submits scripts not defined in CONDOR_LARGE and CONDOR_SMALL" if (!defined($condor_bin) || !defined($condorlarge_bin) || !defined($condormerge_bin));
die "Unable to find MER " if ($mer_bin !~ m/MER/);
die "Unable to find FastTranslateChart " if ($decode_bin !~ m/FastTranslateChart/);
die "Unable to find mergeopti.pl " if ($merger_bin !~ m/mergeopti/);


my $params; #params for decoding each section
my $merparams; #params for running mer on each section
my $numbees; #number of parallel runners you want to send out
my $expdir; #directory to write all output to - passed onto decoder
my $timeout = 480; #in seconds - this is specific to the operaton of runbees

# these are mer specific parameters, and thus explicitely called 
# for in this script - these params are written out to init.opt
my $scaling;
my $minscaling = undef;
my $maxscaling = undef;

my $iterstart = 0;
my $iterlimit = 10;
my $nummodify = 8; #this parameter is moved into runbees since we have to start which variables to start blanking at
my $first = 0;
my $test = 0; #if this is set to one, iterlimit it set to one, and no mer happens

die if !GetOptions("params=s" => \$params, "merparams=s" => \$merparams, "numbees=s" => \$numbees, "expdir=s" => \$expdir, "scaling=s" => \$scaling, "minscaling=s" => \$minscaling, "maxscaling=s" => \$maxscaling, "timeout=s" => \$timeout, "iterstart=s" => \$iterstart, "iterlimit=s" => \$iterlimit, "nummodify=s" => \$nummodify, 'test' => \$test);


die "--scaling must be defined " if (!defined($scaling));
#create an iteration directory for the  bees to work in
my $iter = 0;

#make sure the experimental dir exists
if (!(-e $expdir)){
    mkdir("$expdir", 0755) || die "Cannot mkdir $expdir: $!";
}
my $mergedir = "$expdir/merge";
if (!(-e $mergedir)){
    mkdir("$mergedir", 0755) || die "Cannot mkdir $mergedir: $!";
}

my $numfeatures = split(/_/,$scaling);
print "numfeatures: $numfeatures\n";

#termination condition is when you generate no new translations
for (my $iter=0; $iter<$iterlimit; $iter++){
    
    #only regen when you need to
    if ($iterstart<=$iter){
	my $iterdir = "$expdir/iter$iter";
	if (!(-e $iterdir)){
	    mkdir("$iterdir", 0755) || die "Cannot mkdir $iterdir: $!";
	}
#files in this directory would be overwritten anyway
	my $regenlogs = 1;
	if ($regenlogs){
	    for (my $i=0; $i<$numbees; $i++){
		my $beeline = "$condor_bin $iterdir/bee$i $decode_bin $params --FeatureWeightsParsing $scaling --WorkerBeeDir $iterdir --MaxSentenceProcessingTime $timeout '>>' $iterdir/bee$i.out";
		print "$beeline\n";
		print `$beeline\n`;
	    }
	    
#wait for all the bee's to finish
	    my $waitline = "ls $iterdir/bee*.clog | condorwaitforlogs.sh";
	    print "$waitline\n";
	    print `$waitline\n`;
	    print "Moving onto MER\n";
	    
	}
	if ($test){
	    print "Testing mode: no MER performed\n";
	    last;
	}
	my $regenmer = 1;
	my $nextiter = $iter+1;
	if ($regenmer){
#run the merger to collect up all the files from the oldmerge directory, and the files from this new directoy
#and write the files out to a newmerge directory - also creating a feats.opt and cands.opt file that can be used for opti
#these files will be in the merge/$nextiter directory

	    #first run a simple merge operation to calculate TrainTranslationScore
	    #this basically just creats the .opt files
	    my $domergeSimple =  "$merger_bin $expdir/iter$iter/ $expdir/iter$iter/ ";
	    print "$domergeSimple\n"; #creates the required feats.opt and cands.opt
	    print `$condormerge_bin $expdir/iter$iter/mergesimple $domergeSimple`;
	    my $runwaitmergeSimple = "echo $expdir/iter$iter/mergesimple.clog | condorwaitforlogs.sh";
	    
	    #this is the more complex merge
	    my $domerge =  "$merger_bin $expdir/merge/ $expdir/iter$iter/ ";
	    print "$domerge\n"; #creates the required feats.opt and cands.opt
	    print `$condormerge_bin $expdir/iter$iter/merge $domerge`;
	    my $runwaitmerge = "echo $expdir/iter$iter/merge.clog | condorwaitforlogs.sh";
	    print "Waiting for merge: $runwaitmerge";
	    `$runwaitmerge`;
	    print "Waiting for mergeSimple: $runwaitmergeSimple";
	    `$runwaitmergeSimple`;

	    my $numnew = `tail -n 1 $expdir/iter$iter/merge.log`;
	    #use this numnew only after reporting the TrainTranslationScore	    
	    print "New Translations Iter: $iter = $numnew\n";

#create the init file
	    if (defined($maxscaling) && defined($minscaling)){
		print "echo $minscaling > $expdir/merge/init.opt\n";	 
		print `echo $minscaling > $expdir/merge/init.opt`;
		#also write this init file into the iter directory to get TrainTr score
		print "echo $minscaling > $expdir/iter$iter/init.opt\n";	 
		print `echo $minscaling > $expdir/iter$iter/init.opt`;


		print "echo $maxscaling >> $expdir/merge/init.opt\n";
		print `echo $maxscaling >> $expdir/merge/init.opt`;
		print "echo $scaling >> $expdir/merge/init.opt\n";	 
		print `echo $scaling >> $expdir/merge/init.opt`;

		print "echo $maxscaling >> $expdir/iter$iter/init.opt\n";
		print `echo $maxscaling >> $expdir/iter$iter/init.opt`;
		print "echo $scaling >> $expdir/iter$iter/init.opt\n";	 
		print `echo $scaling >> $expdir/iter$iter/init.opt`;
	    }else{
		print "No min and max scaling specified\n";
		print "echo $scaling > $expdir/merge/init.opt\n";	 
		print `echo $scaling > $expdir/merge/init.opt`;

		print "echo $scaling > $expdir/iter$iter/init.opt\n";	 
		print `echo $scaling > $expdir/iter$iter/init.opt`;
	    }
	    
	    #run a simple MER run just for scoring purposes. We want it to terminate
	    #immediately after the InitialScore has been seen, we can use sed to do this
	    

	    if ($numnew==0 && $iter!=$iterstart){
		#break out of iterations
		print "Ending at Iter: $iter\n";
		last;
	    }
	    
	    my $runmerSimple = "$condorlarge_bin $expdir/iter$iter/score '$mer_bin --InitFile $expdir/iter$iter/init.opt --CandsFile $expdir/iter$iter/cands.opt --FeatsFile $expdir/iter$iter/feats.opt $merparams --Opti_NumModify $nummodify --Opti_First $first | fgrep InitialScore | head -n 4'";
	    print "$runmerSimple\n";
	    print `$runmerSimple`;
	    print `cat $expdir/iter$iter/score.log`;

	    my $runwaitSimple = "echo $expdir/iter$iter/score.clog | condorwaitforlogs.sh";
	    print "$runwaitSimple\n";
	    print `$runwaitSimple`;


#run mer thru condor with prio 2, writing the mer log file into the merg dir as well
	    my $runmer = "$condorlarge_bin $expdir/merge/mer$iter $mer_bin --InitFile $expdir/merge/init.opt --CandsFile $expdir/merge/cands.opt --FeatsFile $expdir/merge/feats.opt $merparams --Opti_NumModify $nummodify --Opti_First $first";
	    print "$runmer\n";
	    print `$runmer`;
#wait for it
	    my $runwait = "echo $expdir/merge/mer$iter.clog | condorwaitforlogs.sh";
	    print "$runwait\n";
	    print `$runwait`;
	}
    
    }
#grab the new scaling line
    my $scalingLine = `tail -n 1  $expdir/merge/mer$iter.log`;
    chomp($scalingLine);
    if ($scalingLine =~ m/Abort/){
	die "MER aborted in iter $iter\n";
    }
    $scalingLine =~ s/Got Params: //g;
    $scaling = $scalingLine;
    print "Got Params: ", $scaling, "\n"; 
    #move the first variable forward
    $first = ($first + $nummodify) % $numfeatures;
    print "Setting first: ", $first, "\n"; 
}
