#!/usr/bin/env perl
#Separate a stream like STDIN based on a separator pattern and write a series of files
#Must indicate the expected number of parts that each line will generate
use IO::File;
use strict;

my $sep = $ARGV[0]; #separator
my $prefix = $ARGV[1]; #separator
my $numparts = $ARGV[2]; #num of expected parts

my @fds = ();
for (my $i=0; $i<$numparts; $i++){
    my $fd =  new  IO::File("$prefix.$i", "w") or die "Unable to open file $prefix.$i\n";
    push(@fds, $fd);
}
#go thru the input and write to each stream
while (<STDIN>){
    my $line = $_;
    chomp($line);
    my @parts = split(/$sep/o, $line);
    die "Wrong number of parts (expected $numparts) in $line " if (scalar(@parts)!=$numparts);
    for (my $i=0; $i<$numparts; $i++){
	$fds[$i]->print($parts[$i], "\n");
    }
}

#close down all the files
for (my $i=0; $i<$numparts; $i++){
    $fds[$i]->close();
}
