#!/usr/bin/env perl
#Take a log file (STDIN) generated by the python extractor script as well as a parse file and split this task up
#into chunks of sentences, where nsplits is the number of sentences in each split. Put the output files (0 indexed)
#into directory "dir" with the .log and .parse extensions

use IO::File;
use strict;
use Getopt::Long;

my $parseFile;
my $nsplits;
my $dir = ".";

die if !GetOptions("parse=s" => \$parseFile, "lines=s" => \$nsplits, "dir=s" => \$dir);

die "Usage: cat phrase.log | splitlogandparses.pl --lines=X [--parse parses.log --dir workingdir]" if (!(defined($nsplits)));

if (!(-e $dir)){
    system("mkdir $dir");
}

my $logfd = *STDIN;
#my $logfd = new IO::File("$logFile", "r") or die "Unable to open file $logFile\n";
my $parsefd;
if ($parseFile){
    $parsefd = new IO::File("$parseFile", "r") or die "Unable to open file $parseFile\n";
}
my $count=0;
my $id = 0; #
my $dataID = 0; #data set specific id

my $outlogfd = new IO::File("$dir/$id.log", "w") or die "Unable to open file $dir/$id.log\n";
my $outparsefd;
if ($parseFile){
    $outparsefd = new IO::File("$dir/$id.parse", "w") or die "Unable to open file $dir/$id.parse\n";
}
print "Processing STDIN";
if ($parseFile){ print " and $parseFile\n"; }

my $currentDataLog = ""; #this has all the info extracted for the data identifier
my $currentDataParse = ""; #the parse for the data identifier
my $currentSeen = 0; #indicates whether you have seen a data set identifier in this block
while ((my $logline = <$logfd>)){

    my $parseline;
    #if there are parses, then grab a line
    if ($parseFile){
	$parseline= <$parsefd>;
	if (!$parseline){
	    last;
	}
    }
    
    #print "READ $logline\n";
    #print "READ $parseline\n";

    my $headerline = $logline;
    my $srcline = <$logfd>; 	
    my $tgtline = <$logfd>; 	

    #check the tgtline to see if this is a data line
    if ($tgtline =~ m/(DATA[A-Z0-9]+.*-.*tgt).*/){
	print "Data set: $tgtline\n";
	my $ptgtline = "LOG: TGT: ".$1;
	
	$currentDataLog = $headerline.$srcline.$ptgtline." - $dataID\n"."LOG: ALT: 0-0 1-1 2-2 \nLOG: PHRASES_BEGIN:\nLOG: PHRASES_END:\n";
	$currentDataParse = $parseline;
	$currentSeen = 1;
	$dataID++;
    }
    
    #if you have processed the first line and you have yet to see the data tag, then add it now
    if ($count==0 && $currentSeen ==0){
	$outlogfd->print($currentDataLog);
	$outparsefd->print($currentDataParse) if ($parseFile);
    }

    #now print out the first line of the segment that you were supposed to be on
    #if this was the actual tgt line, then print it with our tag identifier
    if ($tgtline =~ m/DATA[A-Z0-9]+.*-.*tgt/){
	$outlogfd->print($currentDataLog);
	$outparsefd->print($currentDataParse) if ($parseFile);
	#discard the rest of the lines till PHRASES_END
	while (!($logline =~ /PHRASES_END/)){
	    $logline = <$logfd>;
	}
    }else{
	$outlogfd->print($headerline); #NUM    
	$outlogfd->print($srcline); #SRC
	$outlogfd->print($tgtline); #TGT        
	while (!($logline =~ /PHRASES_END/)){
	    $logline = <$logfd>;
	    $outlogfd->print($logline);
	}
	$outparsefd->print($parseline) if ($parseFile);
    }
    #done printing one sentence data

    $count++;
    if ($count==$nsplits){
	print "Count $count ID=$id\n";
	print "Parse: $parseline\n"; #the parse that is going to be the last in the file
	print "Target: $tgtline\n";
	
	$outlogfd->close();
	$outparsefd->close() if ($parseFile);

	$id++;
	$outlogfd = new IO::File("$dir/$id.log", "w") or die "Unable to open file $dir/$id.log\n";
	$outparsefd = new IO::File("$dir/$id.parse", "w") or die "Unable to open file $dir/$id.parse\n" if ($parseFile);
	$count=0;
	$currentSeen = 0;
    }
}

$outlogfd->close();
$outparsefd->close() if ($parseFile);
