#!/bin/tcsh -f
#Script that uses the irstlm tools to build a LM in parts to avoid memory problems
#during creation. Smoothing options are limited to Witten-Bell and KNey.
#Authors: nimishg, ashishv

set WORKINGDIR = $1; #directory to create parts and build final LM
set PARTS = $2; #number of parts to use
set NGRAM_ORDER = $3;
set DATAFILE = $4;

set MERGER = `which merge-sublm.pl`

if (! -e ${WORKINGDIR} ) then
    mkdir ${WORKINGDIR}
endif

if (! -e ${DATAFILE} ) then
    echo "DATAFILE ${DATAFILE} not found"
    exit
endif

pushd ${WORKINGDIR}
shift #removes the working dir
shift #removes the number of parts
shift #get ngram order
shift #get datafile
set RUNLINE = "$*"; #cmdline options --kneser-ney --prune-singletons


echo "working dir is: $WORKINGDIR"
echo "parts are: $PARTS"
echo "ngram order: $NGRAM_ORDER"
echo "datafile: $DATAFILE"


#create the dictionary file
if (! -e dictfile ) then
    echo "Generating DICTFILE"
    echo "${CONDOR} conddictfile " "dict -i=${DATAFILE} -o=dictfile -f=y -sort=n"
    ${CONDOR} conddictfile "dict -i=${DATAFILE} -o=dictfile -f=y -sort=n"
    ls conddictfile.clog | condorwaitforlogs.sh
else
    echo "Using existing DICTFILE: dictfile"
endif



#now the splitting happens in this working directory
echo "${CONDOR} condsplitdict " "split-dict.pl --input dictfile --output df. --parts $PARTS"
${CONDOR} condsplitdict split-dict.pl --input dictfile --output df. --parts $PARTS
ls condsplitdict.clog | condorwaitforlogs.sh 

#the input data is ready now
    
    #use an automatically generate log file name
    #foreach p (x[0-9][0-9][0-9][0-9])
     foreach p (df.*)

	 #run ngt to collect n-gram counts on each part
         #run build-sublm.pl to build an n-gram on these parts
         echo "${CONDOR} cond.${p} " " ngt -i=$DATAFILE -n=$NGRAM_ORDER -gooout=y -o=ngram.$p -fd=$p; build-sublm.pl --size $NGRAM_ORDER --ngrams ngram.$p --sublm lm.$p ${RUNLINE}"
         ${CONDOR} cond.${p} " ngt -i=$DATAFILE -n=$NGRAM_ORDER -gooout=y -o=ngram.$p -fd=$p; build-sublm.pl --size $NGRAM_ORDER --ngrams ngram.$p --sublm lm.$p ${RUNLINE}" | stdin2stderr

     end
    #the filenames are available...


    #now.. wait for all the log files that this submission has generated
    #the script may or may not have actually generated any clog files
    sleep 10
    echo "WAITING for runs with: $DATAFILE"
    ls cond*clog | condorwaitforlogs.sh
    
    #run the merger
   echo "${CONDOR} condmerge ${MERGER} --size $NGRAM_ORDER --sublm lm.df.0 -lm lm.gz"
   ${CONDOR} condmerge ${MERGER} --size $NGRAM_ORDER --sublm lm.df.0 -lm lm.gz
   ls condmerge.clog | condorwaitforlogs.sh

popd #get out of that directory
