#!/usr/bin/env perl
#Convert the UMD format phrase table to the SAMT format

use strict;
use warnings;
use Getopt::Long;
my $addglue = 0;
my $addreorder = 0;
my $removejoint = 0;
my $asciifromsource = "";
my $fixus =0;
my $ballimit = '';
my $addislexical = 0;
my $zerooutruleappcount = 0;
#addglue if we want to generate NTs
#addreorder if we want to generate a count on the rule for monotonicity
#removejoint if you want to strip the first prob from the UMD phrase table - which is a joint prob
#asciifromsource to import all  the ascii wordsthe source data
#fixus converts 'u s' to 'us'

#model list
# (glue rulecount) islexical reordering wordcount balanced (joint) 4-translation-models
die if !GetOptions('zerooutruleappcount!' => \$zerooutruleappcount, 'addglue' => \$addglue, 'addreorder' => \$addreorder, 'removejoint' => \$removejoint, 'asciifromsource=s' => \$asciifromsource, 'fixus' => \$fixus, 'ballimit=s' => \$ballimit, 'addislexical' => \$addislexical);

print STDERR "zerooutruleappcount: $zerooutruleappcount\n";
my $numScoresFinal = -1;
my %allNTs = ();
while (<STDIN>){
    my $line = $_;
    chomp($line);
    my @parts = split(/ \|\|\| /,$line);
    #print join("#", @parts), "\n";
    my $result = $parts[0];
    my $src = $parts[1];
    my $tgt = $parts[2];
    my $scores = $parts[3];
    #print "SRC: $src TGT: $tgt SCORES: $scores\n";
    $result =~ s/\[([A-Z]+)\]/\@$1/g;
    #$result = "\@X";
    $allNTs{$result} = 1;

    #replace sacred character
    $src =~ s/\#/__HASH__/g;
    $tgt =~ s/\#/__HASH__/g;
    $src =~ s/@/__AMP__/g;
    $tgt =~ s/@/__AMP__/g;

    $src =~ s/\[([A-Z]+),[0-9]+\]/\@$1/g;
    $tgt =~ s/\[([A-Z]+),([0-9]+)\]/\@$2/g;


    #fix this to match the LM which seems to know about "us"
    if ($fixus){
	$tgt =~ s/\bu s\b/ us /g;	
	$tgt =~ s/\s+/ /g;
	$tgt =~ s/^\s+//g;
	$tgt =~ s/\s+$//g;
    }

    #remove the first log probin the score list, its a joint prob
    if ($removejoint){
	#print "Scores was: $scores\n";
	$scores =~ s/^\s+//g;
	$scores =~ s/([0-9\.e\-]+) (.*)/$2/g;
	#print "Scores after: $scores\n";
    }

    #add a feature that says if this is a purely lexical rule
    if ($addislexical){
	my @src_parts = split(/ /, $src);
	if (scalar(@src_parts)==countTerminals($src)){
	    $scores = "1 $scores"; #is purely lexical
	    #print "PURELEX $line\n";
	}else{
	    $scores = "0 $scores"; #is mixed
	}
    }

    #add a feature for how whether on not there was a swap
    if ($addreorder){
	my $refeat = hasReordering($tgt);
	$scores = "$refeat $scores";
    }


    #adding the count of number of terminals
    my $numTerminals = countTerminals($tgt);
    $scores = "$numTerminals $scores";

    #checks whether the number  on terminals is balanced
    if ($ballimit){
	my $numSourceTerminals = countTerminals($src);
	my $ratio = $numSourceTerminals / $numTerminals;
	if ($ratio>$ballimit || $ratio<1.0/$ballimit){
	    $scores = "1 $scores"; #indicates that its unbalanced
	    #print "Unbalanced: $line\n";
	}else{
	    $scores = "0 $scores"; #indicates that its unbalanced
	}
    }

    my @numScores = split(/ /, $scores);
    if ($numScoresFinal==-1){
	$numScoresFinal = scalar(@numScores);
    }else{
	if (scalar(@numScores)!=$numScoresFinal){
	    print "Different number of scores for Rule: $line\n";
	    exit(-1);
	}
    }

    #add a glue rule count (0), and a rule app count    
    if ($addglue){
	$scores = "0 " . ($zerooutruleappcount? 0 : 1) ." ".$scores;
    }

    print "$src # $tgt # $result # $scores\n";
	
}

if ($addglue){
#add all the @X -> @_G
  print STDERR "Nonterminals: ";
    foreach my $nt (keys %allNTs){
      print STDERR $nt . ' ';
#!!!	print "$nt # \@1 # \@_G # "; printZeros($numScoresFinal + 2);		print "\n";
	print "$nt # \@1 # \@_S # 1 " . ($zerooutruleappcount? 0 : 1) ." "; printZeros($numScoresFinal);		print "\n";    
    }
  print STDERR "\n";
    #will print out a 0 for the num terminals too
    print "\@_S \@_G# \@1 \@2 # \@_S #"; print " 1 " . ($zerooutruleappcount? 0 : 1) ." "; printZeros($numScoresFinal);		print "\n";
    print "_UNKNOWN # _UNKNOWN # \@PHRASE # 5 " . ($zerooutruleappcount? 0 : 1) ." 1 "; printZeros($numScoresFinal-1); print "\n"; 	
}

#if we need to extract ascii from a source file
if (!($asciifromsource eq "")){
    
    open (SRC, "$asciifromsource") or die "Unable to open source file to extract ascii rules\n";
    my %asciiVoc = ();
    my $ascii = "\x20-\x7F";
    while (<SRC>){
	my $line = $_;
	chomp($line);
	my @parts = split(/ /, $line);
	foreach my $part (@parts){
	    if ($part =~ m/[$ascii]+/){

		if (!(exists $asciiVoc{$part})){
		    print "$part # $part # \@X # ";
			if ($addglue){
			    print "0 0 1 ";
			}
		    #since we have  already added the length
		    printZeros($numScoresFinal-1);
		    print "\n";
		    $asciiVoc{$part} = 1;
		}
	    }
	}
    }
}



#assume that there are at most  2 non-terminals
sub hasReordering{
    my ($str)=@_;
    #print "CHECK $str ";
    if ($str =~ m/^[^\@]*\@(.*)/){
	#non-terminals exist
	#print "REST$1\n";
	my $rest = $1;
	if ($rest=~ m/^2(.*)/){
	    #print "RE\n";
	    return 1.0;
	}else{
	    #print "MONO\n";
	    return 0.0;
	}
	
    }else{
	#if nt's exist
	if ($str =~ m/\@/){
	    print "ERROR: regexp failed to match here: $str\n";
	    print STDERR "ERROR: regexp failed to match here: $str\n";
	    exit(-1);
	}
	#print "NONE\n";
	return 0;
    }
}


sub countTerminals{
    my ($str) =@_;
    my @parts = split(/ /, $str);
    my $count = 0;
    foreach my $p (@parts){
	if ($p =~ m/^@/){
	    
	}else{
	    $count++;
	}
    }
    return $count;
}
sub printZeros{
    my $num = shift(@_);
    for (my $i=0; $i<$num; $i++){
	print "0";
	if ($i!=$num-1){
	    print " ";
	}
    }
}

