//#include <config.h>
//#include <fortify.h>

#include <BleuSearch.hh>
//#include <iostream>
//#include <fstream>
//#include <istream>
//#include <sstream>
//#include <assert.h>
//#include <utils.hh>

//#include <Text.hh>
#include <math.h>

using namespace std;

BleuSearch::BleuSearch(string scoringType){
  m_scoringType = scoringType;
  if (m_scoringType=="NIST"){
    m_maxScore = 20.0;
    m_minEpsilon = 0.001;
    BETA = -log(0.5)/log(1.5)/log(1.5);
  }

  if (m_scoringType=="IBMBLEU"){
    m_maxScore = 1.0;
    m_minEpsilon = 0.0001;
  }

  if (m_scoringType=="NISTBLEU"){
    m_maxScore = 1.0;
    m_minEpsilon = 0.0001;
  }
  mp_gainsHash=NULL;
}

BleuSearch::BleuSearch(){
  m_scoringType = "IBMBLEU";
  m_maxScore = 1.0;
  m_minEpsilon = 0.0001;
  mp_gainsHash=NULL;
}
//version that can be used from within the translation
//program to avoid reading and writing to files
BleuSearch::BleuSearch(const char* referenceListFile, redi::pstream& pbuf, bool normalizeReference, const char* gainFile, string scoringType, const int numReferences, const int maxGram,CVocabulary* pVoc){
  

  m_scoringType = scoringType;
  m_pVoc = pVoc;
  m_scoringBackup = false;
  
  BETA = -log(0.5)/log(1.5)/log(1.5);
  
  if (m_scoringType=="NIST"){
    m_maxScore = 20.0;
    m_minEpsilon = 0.001;
    BETA = -log(0.5)/log(1.5)/log(1.5);
  }

  if (m_scoringType=="IBMBLEU"){
    m_maxScore = 1.0;
    m_minEpsilon = 0.0001;
  }

  if (m_scoringType=="NISTBLEU"){
    m_maxScore = 1.0;
    m_minEpsilon = 0.0001;
  }

  if (gainFile!=NULL && strcmp(gainFile,"")!=0 ){
    BuildInfoGains(gainFile);

    m_maxScore = 20.0;
    m_minEpsilon = 0.001;
  }

  if ((gainFile==NULL) && (scoringType=="NIST")){
    cerr<<"No gains file provided, NIST scoring requires a gain file"<<endl;
    cerr<<"Gains files can be generated with the GenerateNISTInfoGains.pl script"<<endl;
    exit(-1);
  }

  s_ScoringOnly = false;
  m_nGramConsidered = maxGram;
  BuildReferences(referenceListFile, pbuf, normalizeReference,  numReferences, pVoc);
  m_numSentences = m_referenceList.size();


  m_config = getInitialConfiguration();
  //cout<<"Setting max:"<<m_maxNGramSize<<endl;
  //m_raw = initRawCounts(m_nGramConsidered);
  //RawCounts r; //just to make sure it got initialized
  //m_raw = r;
}

//since the bleu scoring is so computationally expensive, there is an alternative form which directly
//takes the CCompactTSequence's of the set hyp and the reference directly

void BleuSearch::ScorePairwise( CCompactTSequence<int>* hypNGrams,  CCompactTSequence<int>* refNGrams, int refLength, RawCounts& raw){

  //for (int i=hypNGrams->m_TransducerItems.MinIndex(); i<=hypNGrams->m_TransducerItems.MaxIndex(); i++){
  for (int i=0; i<hypNGrams->m_TransducerItems.size(); i++){
    CCompactTSequenceItem<int>* lookFor = hypNGrams->m_TransducerItems[i];
    CCompactTSequenceItem<int>* existing = refNGrams->GetCompactTSequence(lookFor->GetSequence());
    int numSuggested = lookFor->GetValue();
    //lookFor->Display();
    
    if (existing!=NULL){
      int numFound = existing->GetValue();
      
      //cout<<"numFound: "<<numFound<<" adjust: "<<MIN(numSuggested, numFound)<<" set existing: "<<MAX(numFound-numSuggested, 0)<<endl;
      int min=0;
      //MIN(min, numSuggested, numFound);
      numFound = MIN(min, numSuggested);
      raw.numCorrect[lookFor->GetSequence().size()-1] += min;
      
      //dont change the value of the reference
      //existing->SetValue(MAX(numFound-numSuggested, 0));
    }else{
      //cout<<"no match:"<<endl;
    }
    raw.numSuggested[lookFor->GetSequence().size()-1] +=numSuggested;
  }
  raw.referenceLengths += refLength;
}

/*
void BleuSearch::ScorePairwise(CTranslationItem& hyp, CTranslationItem& ref, RawCounts& raw){

  CCompactTSequence<int>* refNGrams = CollectNGramsBase(ref.m_translation, m_nGramConsidered);
  CCompactTSequence<int>* hypNGrams = CollectNGramsBase(hyp.m_translation, m_nGramConsidered);

  //cout<<"Ref ========="<<endl;
  //refNGrams->DisplayAllItems();
  //cout<<"Hyp ========="<<endl;
  //hypNGrams->DisplayAllItems();

  for (int i=hypNGrams->m_TransducerItems.MinIndex(); i<=hypNGrams->m_TransducerItems.MaxIndex(); i++){
    CCompactTSequenceItem<int>* lookFor = hypNGrams->m_TransducerItems[i];
    CCompactTSequenceItem<int>* existing = refNGrams->GetCompactTSequence(lookFor->GetSequence());
    int numSuggested = lookFor->GetValue();
    //lookFor->Display();

    if (existing!=NULL){
      int numFound = existing->GetValue();

      //cout<<"numFound: "<<numFound<<" adjust: "<<MIN(numSuggested, numFound)<<" set existing: "<<MAX(numFound-numSuggested, 0)<<endl;
      int min=0;
      MIN(min, numSuggested, numFound);
      raw.numCorrect[lookFor->GetSequence().size()-1] += min;

      //dont change the value of the reference
      //existing->SetValue(MAX(numFound-numSuggested, 0));
    }else{
      //cout<<"no match:"<<endl;
    }
    raw.numSuggested[lookFor->GetSequence().size()-1] +=numSuggested;
  }
  raw.referenceLengths += ref.m_translation.size();
  
  refNGrams->ClearAll();
  hypNGrams->ClearAll();
  
  delete refNGrams;
  delete hypNGrams;
  
}
*/

//takes a complete nbest list and reference the data in a way that we can run the optimization
/*
void BleuSearch::ProcessNBestList(CNBest* oneSourceList, int sentenceNumber){
  
  int numCands = oneSourceList->m_TransducerItems.Size();

  //m_numCands.push_back(numCands);
  //cout<<"Setting up arrays:"<<oneSourceList->getNumberCorrectList().size()<<endl;
  //m_numberCorrect.push_back( oneSourceList->getNumberCorrectList());
  //m_numberSuggested.push_back(oneSourceList->getNumberSuggestedList());
  //m_closestRefLength.push_back( oneSourceList->getReferenceLengthsList());
  
  // get the translations from the transducer list
  int candidateNumber =0;
  
  //cout<<"Reference====================="<<endl;
  //PrintHashElements(sentenceNumber);
  //cout<<"====================="<<endl;
  
  for (int i=oneSourceList->m_TransducerItems.MinIndex(); i<=oneSourceList->m_TransducerItems.MaxIndex(); i++){

    // only recalculate the counts if this item has not been calculated yet
    if (!(oneSourceList->m_TransducerItems[i]->isCalculated())){
      ProcessCounts(sentenceNumber, oneSourceList->m_TransducerItems[i], m_numReferences);
      oneSourceList->m_TransducerItems[i]->setCalculated(true);
    }
    candidateNumber++;
  }

}
*/

//score a translation run after at least one iteration has been run thru the system
/*
AbstractScore BleuSearch::ScoreTranslationRun(vector<string>& translationSet, vector<CNBest*>& nBestLists){
  int sentenceNumber = 0;
  
  //the counts object for this scoring
  RawCounts raw = initRawCounts(NSZ);
  for (vector<string>::iterator iter = translationSet.begin(); iter!= translationSet.end(); ++iter){

    CNBestItem* oneItem = nBestLists[sentenceNumber]->GetNBestItem(*iter);

    // add its counts and references to the raw counts object
    Counter* correctCNT = oneItem->getNumberCorrect();
    Counter* suggestedCNT = oneItem->getNumberSuggested();
    for (int n=0; n<NSZ; n++){
      raw.numCorrect[n] += (*correctCNT)[n];
      raw.numSuggested[n] += (*suggestedCNT)[n];
    }
    raw.referenceLengths += *(oneItem->getClosestRefLength());
  
    sentenceNumber++;
  }
  double finalScore;
  // score the raw counts associated with this translation bank
  if (m_scoringType == "NIST"){
    double myscore = getNISTScore(raw);    
    finalScore= myscore;
  }else{
    double myscore = getBleuScore(raw);    
    finalScore = myscore;
  }
  
  AbstractScore toRet;
  toRet.push_back(finalScore);
  return toRet;
}

*/

/*
pair<double, RawCounts> BleuSearch::singleSentenceOracle(RawCounts raw, Configuration config, vector<CNBest*>& nBestLists, int sentenceNumber){

  int currentIndex = config[sentenceNumber];
  int numCands = nBestLists[sentenceNumber]->m_TransducerItems.Size();
  vector<Counter*> numberCorrect = nBestLists[sentenceNumber]->getNumberCorrectList();
  vector<Counter*> numberSuggested = nBestLists[sentenceNumber]->getNumberSuggestedList();
  vector<double*> closestRefLength = nBestLists[sentenceNumber]->getReferenceLengthsList();

  
  // start with the current score, see if it goes up
  double bestScore = getBleuScore(raw);
  int bestIndex = currentIndex;
  RawCounts bestCounts = raw;

  for (int i=0; i<numCands; i++){
    if (i!=currentIndex){
      
      Counter correctDelta;
      
      getCounterDifference((*(numberCorrect[currentIndex])), (*(numberCorrect[i])), correctDelta, NSZ);
      Counter suggestedDelta; 
      getCounterDifference((*(numberSuggested[currentIndex])), (*(numberSuggested[i])), suggestedDelta, NSZ);
      double closestRefLengthDelta = (*(closestRefLength[i]))- (*(closestRefLength[currentIndex]));
      pair<double, RawCounts> scoreRaw = CalculateBleuNew(raw, correctDelta, suggestedDelta, closestRefLengthDelta);

      if (scoreRaw.first>bestScore){
	bestScore = scoreRaw.first;
	bestIndex = i;
	bestCounts = scoreRaw.second;
      }
    }
  }
  pair<double, RawCounts> retPair;
  retPair.first = bestScore;
  retPair.second = bestCounts;

  // see if there is any change
  //cout<<"Initial:"<<currentIndex<<" Final:"<<bestIndex<<endl;
  config[sentenceNumber] = bestIndex;
  return retPair;
}
*/

/*
Configuration BleuSearch::GetOracleConfiguration(vector<CNBest*>& nBestLists){

  bool madeChanges = true;  

  Configuration initialConfiguration = getInitialConfiguration();
  RawCounts rawHolder = buildRawCounts(nBestLists, initialConfiguration,NSZ);
  double scoreHolder = getBleuScore(rawHolder);
  
  cout<<"InitialScore:"<<scoreHolder<<" for "<<m_numSentences<<" sentences"<<endl;
  int iterationNumber = 0;
  while (madeChanges){
    madeChanges = false;
    for (int i=0; i<m_numSentences; i++){
      //PrintConfiguration(config);
      pair<double, RawCounts> retPair = singleSentenceOracle(rawHolder, initialConfiguration, nBestLists,i);
      rawHolder = retPair.second;
      if (retPair.first!=scoreHolder){
	scoreHolder = retPair.first;
	madeChanges = true;
      }
    } // one complete iteration done
    cout<<"Iteration:"<<iterationNumber<<" NumCorrect:"<<rawHolder.numCorrect[0]<<" NumSuggested:"<<rawHolder.numSuggested[0]<<" Score:"<<getBleuScore(rawHolder)<<endl;
    iterationNumber++;
  }

  return initialConfiguration;
}



AbstractScore BleuSearch::ScoreConfiguration(vector<CNBest*>& nBestLists, Configuration config){
  RawCounts raw = buildRawCounts(nBestLists, config, NSZ);
  AbstractScore toRet;
  toRet.push_back(getBleuScore(raw));
  return toRet;
}

ScoreSort BleuSearch::ScoreLocalAlternatives(vector<CNBest*>& nBestLists, Configuration conf, int sentenceNumber){
  
  int currentIndex = conf[sentenceNumber];
  int numCands = nBestLists[sentenceNumber]->m_TransducerItems.Size();
  vector<Counter*> numberCorrect = nBestLists[sentenceNumber]->getNumberCorrectList();
  vector<Counter*> numberSuggested = nBestLists[sentenceNumber]->getNumberSuggestedList();
  vector<double*> closestRefLength = nBestLists[sentenceNumber]->getReferenceLengthsList();
  RawCounts raw = buildRawCounts(nBestLists, conf, NSZ);

  ScoreSort sorter;

  for (int i=0; i<numCands; i++){
    Counter correctDelta;
    getCounterDifference((*(numberCorrect[currentIndex])), (*(numberCorrect[i])), correctDelta, NSZ);
    Counter suggestedDelta; 
    getCounterDifference((*(numberSuggested[currentIndex])), (*(numberSuggested[i])), suggestedDelta, NSZ);
    double closestRefLengthDelta = (*(closestRefLength[i]))- (*(closestRefLength[currentIndex]));
    
    //pair<double, RawCounts> scoreRaw = CalculateBleuNew(raw, correctDelta, suggestedDelta, closestRefLengthDelta);
    pair<double, RawCounts> scoreRaw = CalculateMetricNew(raw, correctDelta, suggestedDelta, closestRefLengthDelta);

    //insert the score into the return map
    sorter.insert(pair<double, int>(scoreRaw.first, i));  
  }
  return sorter;

}
*/

// received the output from the translation system
// and adds it to the bank of processed counts
//OLD






BleuSearch::~BleuSearch(){

  if (!s_ScoringOnly){

    delete [] m_config;
  }
  
  if (mp_gainsHash!=NULL){
    mp_gainsHash->clear();
    delete mp_gainsHash;
  }

}

void BleuSearch::ScoreSingleHyp(const int sentenceNumber, string candidate, RawCounts& countObject){
  //initialize the count object
  for (int n=0; n<m_nGramConsidered; n++){
    countObject.numCorrect[n]=0;
    countObject.numSuggested[n]=0;
  }
  countObject.referenceLengths = 0;

  // build up the source sentence from the raw string
  istringstream candStream(candidate);
  NGram sourceSentence;
  while (candStream.good()){
    string oneWord;
    candStream>>oneWord;
    //cout<<"SentWord:"<<oneWord<<endl;
    sourceSentence.push_back(oneWord);
    
  }
  
  pair<NGram, std::vector< pair<int, int> > > gramsAndSizes = CollectNGrams(sourceSentence, m_nGramConsidered ); 
  
  // we have to copy it so that other sentences can use this data as well
  // since the SentenceHash is consolidated, we don't have to load each on separately
  SentenceHash hashCopy(m_referenceList[sentenceNumber]);    
  
  //used to get the ngram size (1,2,3 or 4 etc) of a given gram from this sentence
  int trackGram = 0;
  //collecting the grams
  
  int countTotal = 0;
  for (NGram::iterator iter = gramsAndSizes.first.begin(); iter!= gramsAndSizes.first.end(); ++iter){
    std::string gram = (*iter);
        
    //int ngramIndex = gramsAndSizes.second[trackGram]-1;
    int ngramIndex = gramsAndSizes.second[trackGram].second - gramsAndSizes.second[trackGram].first;
    //cout<<"gram:"<<gram<<" ngramIndex:"<<ngramIndex<<endl;
    
    if (hashCopy.count(gram) != 0){
      NgramDetails payload = hashCopy[gram];
      int numLeft = payload.numOccurrences;
      int ngramSz = payload.gramSize;
      
      
      if (numLeft>0){
	hashCopy[gram].numOccurrences = numLeft-1;

	//got an n-gram correct
	if ((m_scoringType=="NIST") && (mp_gainsHash!=NULL)){
	  countObject.numCorrect[ngramIndex] = countObject.numCorrect[ngramIndex]+ (*mp_gainsHash)[gram];
	}else{
	  countObject.numCorrect[ngramIndex] = countObject.numCorrect[ngramIndex]+ 1;
	}
      }else{

      }
    }else{
      //suggested an n-gram but it was wrong
      //cout<<"wrong"<<endl;
    }

    countObject.numSuggested[ngramIndex] = countObject.numSuggested[ngramIndex]+ 1;
    trackGram++;
  }  


  if (m_scoringType=="IBMBLEU"){

    m_maxScore = 1.0;
    m_minEpsilon = 0.0001;
    mp_gainsHash=NULL;

    //trackgram indicated how many words there are in this candidate sentence
    int minDiffw = 100000;
    int closestRefW = 0;
    
    int sourceSize = sourceSentence.size();
    for (int r=0; r<m_numReferences; r++){
      int thisRefW = m_referenceLengths[sentenceNumber*m_numReferences + r];

      // in update mode
      if (abs((int)(sourceSize-thisRefW)) <= minDiffw){
	// if its equals
	if (abs((int)(sourceSize-thisRefW)) == minDiffw){
	  // only update it in this condition
	  if (closestRefW > thisRefW){
	    closestRefW = thisRefW;
	  }else{
	    // don't do anything
	  }       
	}else{
	  closestRefW = thisRefW;
	}
	minDiffw = abs((int)(sourceSize -thisRefW));
      }
    }// closest length has been picked

    //(*(m_closestRefLength[sentenceNumber][candidateNumber])) = closestRefW;
    //nBestItem->setRefLengthData(closestRefW);
    countObject.referenceLengths  = closestRefW;
  }

  if (m_scoringType== "NISTBLEU"){

    m_maxScore = 1.0;
    m_minEpsilon = 0.0001;
    mp_gainsHash=NULL;

    int smallestRef = 10000;
    //use the smallest reference
    for (int r=0; r<m_numReferences; r++){
      int thisRefW = m_referenceLengths[sentenceNumber*m_numReferences + r];
      if (thisRefW<smallestRef){
	smallestRef=thisRefW;	
      }
    }

    //nBestItem->setRefLengthData(smallestRef);
    countObject.referenceLengths  = smallestRef;
  }

  if (m_scoringType=="NIST"){

    int sum=0;

    //hold the sum of all references to be used for an average later on
    for (int r=0; r<m_numReferences; r++){
      int thisRefW = m_referenceLengths[sentenceNumber*m_numReferences + r];
      sum += thisRefW;
    }
    double averageLength=(double)sum/(double)m_numReferences;
    //nBestItem->setRefLengthData(averageLength);
    countObject.referenceLengths  = averageLength;
  }
}


/*
void BleuSearch::ProcessCounts(const int sentenceNumber, CNBestItem* nBestItem, const int numReferences){

  string candidate = nBestItem->getCandidateString();
  //cout<<"candidate:"<<candidate<<endl;
  
  //initialize the metric statitics
  for (int n=0; n<m_nGramConsidered; n++){
    nBestItem->setNGramData(n, 0, 0);
  }

  // build up the source sentence from the raw string
  istringstream candStream(candidate);
  NGram sourceSentence;
  while (candStream.good()){
    string oneWord;
    candStream>>oneWord;
    //cout<<"SentWord:"<<oneWord<<endl;
    sourceSentence.push_back(oneWord);
    
  }

  //sentence is loaded
  //collect all the n-grams in the sentence
  pair<NGram, std::vector< pair<int, int> > > gramsAndSizes = CollectNGrams(sourceSentence, m_nGramConsidered );

  // we have to copy it so that other sentences can use this data as well
  // since the SentenceHash is consolidated, we don't have to load each on separately
  SentenceHash hashCopy(m_referenceList[sentenceNumber]);    
  
  //used to get the ngram size (1,2,3 or 4 etc) of a given gram from this sentence
  int trackGram = 0;
  //collecting the grams
  
  int countTotal = 0;
  for (NGram::iterator iter = gramsAndSizes.first.begin(); iter!= gramsAndSizes.first.end(); ++iter){
    string gram = (*iter);
        
    //int ngramIndex = gramsAndSizes.second[trackGram]-1;
    int ngramIndex = gramsAndSizes.second[trackGram].second - gramsAndSizes.second[trackGram].first;
    //cout<<"gram:"<<gram<<" ngramIndex:"<<ngramIndex<<endl;
    
    if (hashCopy.count(gram) != 0){
      NgramDetails payload = hashCopy[gram];
      int numLeft = payload.numOccurrences;
      int ngramSz = payload.gramSize;
      
      
      if (numLeft>0){
	hashCopy[gram].numOccurrences = numLeft-1;

	//got an n-gram correct
	if ((m_scoringType=="NIST") && (mp_gainsHash!=NULL)){
	  nBestItem->incrementNGramData(ngramIndex, (*mp_gainsHash)[gram], 0) ;
	}else{
	  nBestItem->incrementNGramData(ngramIndex, 1, 0) ;
	}
      }else{

      }
    }else{
      //suggested an n-gram but it was wrong
      //cout<<"wrong"<<endl;
    }

    nBestItem->incrementNGramData(ngramIndex, 0, 1) ;
    trackGram++;
  }  


  if (m_scoringType=="IBMBLEU"){

    //trackgram indicated how many words there are in this candidate sentence
    int minDiffw = 100000;
    int closestRefW = 0;
    
    int sourceSize = sourceSentence.size();
    for (int r=0; r<numReferences; r++){
      int thisRefW = m_referenceLengths[sentenceNumber*numReferences + r];

      // in update mode
      if (abs((int)(sourceSize-thisRefW)) <= minDiffw){
	// if its equals
	if (abs((int)(sourceSize-thisRefW)) == minDiffw){
	  // only update it in this condition
	  if (closestRefW > thisRefW){
	    closestRefW = thisRefW;
	  }else{
	    // don't do anything
	  }       
	}else{
	  closestRefW = thisRefW;
	}
	minDiffw = abs((int)(sourceSize -thisRefW));
      }
    }// closest length has been picked

    //(*(m_closestRefLength[sentenceNumber][candidateNumber])) = closestRefW;
    nBestItem->setRefLengthData(closestRefW);
  }

  if (m_scoringType== "NISTBLEU"){
    int smallestRef = 10000;
    //use the smallest reference
    for (int r=0; r<numReferences; r++){
      int thisRefW = m_referenceLengths[sentenceNumber*numReferences + r];
      if (thisRefW<smallestRef){
	smallestRef=thisRefW;	
      }
    }

    nBestItem->setRefLengthData(smallestRef);
  }

  if (m_scoringType=="NIST"){

    int sum=0;

    //hold the sum of all references to be used for an average later on
    for (int r=0; r<numReferences; r++){
      int thisRefW = m_referenceLengths[sentenceNumber*numReferences + r];
      sum += thisRefW;
    }
    double averageLength=(double)sum/(double)numReferences;
    nBestItem->setRefLengthData(averageLength);
  }

}
*/

/*
Configuration BleuSearch::getInitialConfiguration(){
  cout<<"About to return config of size:"<<m_numSentences<<endl;
  Configuration conf = new int[m_numSentences];
  for (int i=0; i<m_numSentences; i++){
    conf[i] = 0;
  }
  return conf;
}
*/

Configuration BleuSearch::GetFinalConfiguration(){
  return m_config;
}



/*
RawCounts BleuSearch::buildRawCounts(vector<CNBest*>& nBestLists, Configuration config, int N){

  double numCorrect = 0;
  double numSuggested = 0;
  double referenceLengths = 0;
  RawCounts raw;
  for (int n=0; n<NSZ; n++){
    raw.numCorrect[n] = 0;
    raw.numSuggested[n] = 0;
  }
  raw.referenceLengths = 0;


  for (int s=0; s<m_numSentences; s++){
    
    vector<Counter*> numberCorrect = nBestLists[s]->getNumberCorrectList();
    vector<Counter*> numberSuggested = nBestLists[s]->getNumberSuggestedList();
    vector<double*> closestRefLength = nBestLists[s]->getReferenceLengthsList();
    
#ifndef NDEBUG
    cout<<"Counts ";
#endif
    for (int n=0; n<NSZ; n++){
      raw.numCorrect[n] += (*(numberCorrect[config[s]]))[n];
#ifndef NDEBUG
      cout<<n+1<<":"<<(*(numberCorrect[config[s]]))[n]<<"/"<<(*(m_numberSuggested[s][config[s]]))[n]<<" ";
#endif
      raw.numSuggested[n] += (*(numberSuggested[config[s]]))[n];
    }
#ifndef NDEBUG
    cout<<endl;
#endif
    raw.referenceLengths += (*(closestRefLength[config[s]]));
    //cout<<"referenceLengths:"<<raw.referenceLengths<<endl;
  }
  
  return raw;
}
*/

// changing this to directly use the counts instead of the delta table
/*
pair<double, RawCounts> BleuSearch::optimizeForSentence(RawCounts raw, Configuration config, int sentenceNumber){
  
  int numCands = m_numCands[sentenceNumber];
  int currentIndex = config[sentenceNumber];
  
  // start with the current score, see if it goes up
  double bestScore = getBleuScore(raw);
  int bestIndex = currentIndex;
  RawCounts bestCounts = raw;

  for (int i=0; i<numCands; i++){
    if (i!=currentIndex){
      
      Counter correctDelta;
      
      getCounterDifference((*(m_numberCorrect[sentenceNumber][currentIndex])), (*(m_numberCorrect[sentenceNumber][i])), correctDelta, m_nGramConsidered);
      Counter suggestedDelta; 
      getCounterDifference((*(m_numberSuggested[sentenceNumber][currentIndex])), (*(m_numberSuggested[sentenceNumber][i])), suggestedDelta, m_nGramConsidered);
      double closestRefLengthDelta = (*(m_closestRefLength[sentenceNumber][i]))- (*(m_closestRefLength[sentenceNumber][currentIndex]));
      //pair<double, RawCounts> scoreRaw = CalculateBleuNew(raw, correctDelta, suggestedDelta, closestRefLengthDelta);
      pair<double, RawCounts> scoreRaw = CalculateMetricNew(raw, correctDelta, suggestedDelta, closestRefLengthDelta);
      //pair<double, RawCounts> scoreRaw = CalculateBleuNew(raw, m_numberCorrectDelta[sentenceNumber][currentIndex][i], m_numberSuggestedDelta[sentenceNumber][currentIndex][i], m_closestRefLengthDelta[sentenceNumber][currentIndex][i]);
      if (scoreRaw.first>bestScore){
	bestScore = scoreRaw.first;
	bestIndex = i;
	bestCounts = scoreRaw.second;
      }
    }
  }
  pair<double, RawCounts> retPair;
  retPair.first = bestScore;
  retPair.second = bestCounts;

  // see if there is any change
  //cout<<"Initial:"<<currentIndex<<" Final:"<<bestIndex<<endl;
  config[sentenceNumber] = bestIndex;
  return retPair;
}
*/


bool BleuSearch::noDifference(Counter correctDelta, Counter suggestedDelta, double closestRefLengthDelta){
  
  for (int i=0; i<m_nGramConsidered; i++){
    if (correctDelta[i]!=0){
      return false;
    }
    if (suggestedDelta[i]!=0){
      return false;
    }
  }
  
  if (closestRefLengthDelta==0){
    return true;
  }
}


double BleuSearch::CDFNormal(double z){
  if ( z > 6.0) {return 1.0;}
  if (z < -6.0) {return 0.0;}
  
  double b1 = 0.31938153;
  double b2 = -0.356563782;
  double b3 = 1.781477937;
  double b4 = -1.821255978;
  double b5 = 1.330274429;
  double p = 0.2316419;
  double c2 = 0.3989423;

  double a = fabs(z);
  double t = 1.0/(1.0+a*p);
  double b = c2*exp((-z)*(z/2.0));
  double n = ((((b5*t+b4)*t + b3)*t + b2)*t + b1)*t;
  n = 1.0-b*n;
  if (z <0.0) n = 1.0-n;
  return n;
}

/*
std::pair<double, RawCounts> BleuSearch::optimizeBleuScore(){
  bool madeChanges = true;

  //RawCounts rawHolder = raw;
  RawCounts rawHolder = m_raw;
  double scoreHolder = getBleuScore(m_raw);
  
  cout<<"InitialScore:"<<scoreHolder<<" for "<<m_numSentences<<" sentences"<<endl;
  int iterationNumber = 0;
  while (madeChanges){
    madeChanges = false;
    for (int i=0; i<m_numSentences; i++){
      //PrintConfiguration(config);
      pair<double, RawCounts> retPair = optimizeForSentence(rawHolder, m_config, i);
      rawHolder = retPair.second;
      if (retPair.first!=scoreHolder){
	scoreHolder = retPair.first;
	//cout<<"Updated S:"<<i<<" -> "<<scoreHolder<<endl;
	madeChanges = true;
      }
    } // one complete iteration done
    cout<<"Iteration:"<<iterationNumber<<" NumCorrect:"<<rawHolder.numCorrect[0]<<" NumSuggested:"<<rawHolder.numSuggested[0]<<" Score:"<<getBleuScore(rawHolder)<<endl;
    iterationNumber++;
  }
  std::pair<double, RawCounts> opti;
  opti.first = scoreHolder;
  opti.second = rawHolder;

  // output the scale confidences
  for (int i=0; i<m_numSentences; i++){
    getScaledConfidences(i, m_config, rawHolder );
  }


  return opti;
}
*/

double BleuSearch::getNISTScore(RawCounts& raw){
  return getNISTScoreDetailed(raw).first;
}

pair<double, double> BleuSearch::getNISTScoreDetailed(RawCounts& raw){
 
  // n-gram factor blue
  double precisionSum = 0.0;
  for ( int n=0; n<m_nGramConsidered; n++){
    //cout<<n+1<<"-gram NC:"<<raw.numCorrect[n]<<" Sug:"<<raw.numSuggested[n]<<endl;
    double precision = raw.numCorrect[n] / raw.numSuggested[n];
    //cout<<n+1<<"-Gram Precision:"<<precision<<endl;
    precisionSum += precision;
  }
  //cout<<"PrecisionSum:"<<precisionSum<<endl;
  //cout<<"RefLengths:"<<raw.referenceLengths<<endl;
  double ratio = raw.numSuggested[0]/raw.referenceLengths;
  //cout<<"Ratio:"<<ratio<<endl;
  //cout<<"BETA:"<<BETA<<endl;
  double lpen = 1.0;
  if (ratio < 1.0){
    lpen = exp(-BETA*log(ratio)*log(ratio));
  }
  //cout<<"LPen:"<<lpen<<endl;
  
  double nistscore = lpen*precisionSum;
  
  return pair<double, double>(nistscore, lpen);
}

Score BleuSearch::getMetricScore(RawCounts& raw){
  if (m_scoringType == "NIST"){
    return getNISTScore(raw);
  }else{
    return getBleuScore(raw);
  }
}

AbstractScore BleuSearch::getMetricScoreDetailed(RawCounts& raw){
  pair<double, double> s;
  if (m_scoringType == "NIST"){
    s = getNISTScoreDetailed(raw);
  }else{
    s= getBleuScoreDetailed(raw);
  }
  AbstractScore toRet;
  toRet.push_back(s.first);
  toRet.push_back(s.second);
  return toRet;
}

double BleuSearch::getBleuScore(RawCounts& raw){
  return getBleuScoreDetailed(raw).first;
}

pair<double, double> BleuSearch::getBleuScoreDetailed(RawCounts& raw){
  

  // n-gram factor blue
  double precisionSum = 0.0;
  
  //cout<<"m_nGramConsidered: "<<m_nGramConsidered<<endl;
  for ( int n=0; n<m_nGramConsidered; n++){
    //cout<<n+1<<"-gram NC:"<<raw.numCorrect[n]<<" Sug:"<<raw.numSuggested[n]<<endl;
    double precision = (double)raw.numCorrect[n] / (double)raw.numSuggested[n];
    if (precision==0){
      //cout<<"Return 0"<<endl;
      return pair<double, double>(0.0, 0.0);
    }
    //cout<<n+1<<"-Gram Precision:"<<precision<<endl;
    ///assert(precision>0);
    double logPrecision = log(precision);
    //cout<<n+1<<"-Gram LogPrecision:"<<logPrecision<<endl;
    //cout<<"Precision Sum:"<<precisionSum<<endl;
    precisionSum += logPrecision;
    
  }
  
  //cout<<"LINE========================"<<endl;
  //cout<<"NSZ:"<<NSZ<<endl;
  //cout<<"PrecisionSum:"<<precisionSum<<endl;
  double logPrecisionFinal = precisionSum/(double)m_nGramConsidered;
  //cout<<"LogPrecisionFinal:"<<logPrecisionFinal<<endl;
  double precisionScore = exp(logPrecisionFinal);
  //cout<<"Precision Score:"<<precisionScore<<endl;
  // the numSuggested[0] is the uni-gram length of the corpus
  
  
  //cout<<"PS:"<<precisionScore<<endl;
  //cout<<"RawRefLens:"<<raw.referenceLengths<<endl;
  assert(raw.referenceLengths>0);
  //cout<<"Suggested:"<<raw.numSuggested[0]<<endl;
  assert(raw.numSuggested[0]>0);

  double lratio = (double)raw.referenceLengths/(double)raw.numSuggested[0];
  //cout<<"Ref2SysLen:"<<lratio<<endl;
  double lPen = 1.0;
  if (lratio <= 1.0){
    //do nothing (ref/c) and c>r
  }else{
    // (ref/c) > 1.0 
    lPen = exp(1.0-lratio);
  }
  //cout<<"Precision:"<<precisionScore<<endl;
  //cout<<"LPen:"<<lPen<<endl;

  return pair<double, double>(lPen*precisionScore, lPen);

}
/*
pair<double, RawCounts> BleuSearch::CalculateBleuNew(RawCounts raw, Counter correctDelta, Counter suggestDelta, double closestLenDelta){
  RawCounts rawNew;
  for (int n=0; n<NSZ; n++){
    rawNew.numCorrect[n] = raw.numCorrect[n] + correctDelta[n];
    rawNew.numSuggested[n] = raw.numSuggested[n] + suggestDelta[n];
  }
  rawNew.referenceLengths = raw.referenceLengths+ closestLenDelta;
  //cout<<"NewRefLen:"<<closestLenDelta<<endl;

  //return all the new data
  pair<double, RawCounts> scoreRaw;
  scoreRaw.first = getBleuScore(rawNew);
  scoreRaw.second = rawNew;
  return scoreRaw;
}

*/




void BleuSearch::PrintHashElements(int sentenceNumber){
  cout<<"HashFor:"<<sentenceNumber<<endl;
  SentenceHash oneHash = m_referenceList[sentenceNumber];
  for (SentenceHash::iterator iter = oneHash.begin(); iter!=oneHash.end(); ++iter){
    cout<<"Key:"<<(*iter).first<<" Details numOccurrences:"<<(*iter).second.numOccurrences<<" gramSize:"<<(*iter).second.gramSize<<endl;
  }
  cout<<endl;
}

void BleuSearch::PrintHashElements(SentenceHash& oneHash){
  for (SentenceHash::iterator iter = oneHash.begin(); iter!=oneHash.end(); ++iter){
    cout<<"Key:"<<(*iter).first<<" Details numOccurrences:"<<(*iter).second.numOccurrences<<" gramSize:"<<(*iter).second.gramSize<<endl;
  }
  cout<<endl;
}

void BleuSearch::BuildInfoGains(const char* fileName){


  mp_gainsHash = new GainsHash();
  
  ifstream fp(fileName);
  if (!fp.good()){
    cout<<"Unable to open gains file listing:"<<fileName<<endl;
    exit(-1);
  }
  cout<<"Loading:"<<fileName<<endl;
  char buf[1024];
  buf[0] = '\0';
  for (; !fp.eof(); ){
    fp.getline(buf, 1024);
    string gainString(buf);
    //cout<<"GainString:"<<gainString<<endl;
    stringstream ss(stringstream::in | stringstream::out);
    ss<<gainString;
    double gain=0;
    string ngram = "";
    int ngramIndex = 0;
    ss>>gain>>ngramIndex;

    for (int i=0; i<ngramIndex; i++){
      string tmp;
      ss>>tmp;
      ngram +=tmp;
      if (i!=ngramIndex-1){
	ngram+=" ";
      }
    }


    if (gain!=0){
      (*mp_gainsHash)[ngram] = gain;
      //cout<<"Gram:"<<ngram<<" -- "<<(*mp_gainsHash)[ngram]<<endl;
    }
    
  }
  

}

void BleuSearch::BuildReferences(const char* fileName, redi::pstream& pbuf, bool normalizeReference, const int numReferences, CVocabulary* pVoc){
  
  if (fileName!=NULL){
    
    ifstream fp(fileName);
    if (!fp.good()){
      cout<<"Unable to open reference file listing:"<<fileName<<endl;
      exit(-1);
    }
    cout<<"Loading:"<<fileName<<endl;
    char buf[10000];
    buf[0] = '\0';

    int sentenceNumber = -1;
    for (; !fp.eof(); ){


      vector<string> referenceSet;

      //prepare for this bank of setences
      vector<VocabVector> v;
      m_referenceBank.push_back(v);
      sentenceNumber++;
      //      cout<<"Ref:"<<sentenceNumber<<"numReferences:"<<numReferences<<endl;

      for (int r=0; r<numReferences; r++){ 

	fp.getline(buf, 10000);
	//cout<<"READ:"<<buf<<" length="<<strlen(buf)<<endl;
	while ((strlen(buf)==0) && !(fp.eof())){
	  fp.getline(buf, 10000);
	}
	if (fp.eof()){
	  break;
	}
	//cout<<"BeforeToEnterRef:"<<buf<<"#"<<endl;
	//added to do normalizing - global function
	string sourceWord=string(buf);
	if (pbuf.good() && normalizeReference){
	  cout<<"WARNING: normalizing reference\n";
	  pbuf<<sourceWord<<"\n\r";
	  pbuf.flush();

	  std::string str;
	  //pbuf.getLine();
	  //cout<<"original: "<<sourceWord<<endl;
	  std::getline(pbuf.out(), str);
	  //cout<<"returnString:"<<str<<endl;
	  sourceWord = str;
	  //cout<<"after: "<<sourceWord<<endl;
	  
	}




	//cout<<"AfterNorm:"<<sourceWord<<"#"<<endl;
	int Pos = 0;
	int NextPos = 0;
	
	VocabVector oneSentV;
	m_referenceBank[sentenceNumber].push_back(oneSentV);
	
	do{
	  NextPos = sourceWord.find(' ', Pos);
	  string Word = sourceWord.substr(Pos, NextPos-Pos);
	  //cout<<"Checking vocabword:"<<Word<<endl;
	  //if (pVoc->GetIndex( CString( Word.c_str()))==pVoc->UNK){
	  //if (pVoc->GetIndex(Word)==pVoc->UNK){
	    //cout<<"Adding Word:"<<Word<<endl;
	  //pVoc->AddWord(string(Word));
	  //}
	  //int wIndex = pVoc->GetIndex(CString( Word.c_str()));
	  //int wIndex = pVoc->GetIndex(Word);
	  int wIndex = pVoc->AddWord(Word);
	  (m_referenceBank[sentenceNumber])[r].push_back(wIndex);
	    
	  Pos = NextPos + 1;
	  
	}while (NextPos >= 0);
	

	//	cout<<"ToEnterRef:"<<sourceWord<<"#"<<endl;
	referenceSet.push_back(sourceWord);

      }
      
      //this is specific to BleuSearch
      if (referenceSet.size()!=0){
	//cout<<"about to process refs of size="<<referenceSet.size()<<endl;
	m_referenceList.push_back(ProcessReferences(referenceSet));
	//PrintHashElements(m_referenceList.size()-1);
	
      }
      
    
    }
    fp.close();
  }
  m_numReferences = numReferences;
  
}


//takes the 4 references (in string form) and consolidates them into one combined hashed object
SentenceHash BleuSearch::ProcessReferences(std::vector<string> referenceSet){
  
  // extract all the individual sentences
  vector<SentenceHash> individualSentences;

  for(vector<string>::iterator iter = referenceSet.begin(); iter!= referenceSet.end(); ++iter){
    stringstream sentStream((*iter));
    NGram oneSentence;
    while(sentStream.good()){
      string oneWord;
      sentStream>>oneWord;
      //#ifndef NDEBUG
      //cout<<"OneWord:"<<oneWord<<endl;
      //#endif
      oneSentence.push_back(oneWord);
    }// have the sentence in vector form
    // this sentence is a certified reference now
    SentenceHash oneSentenceHash = ProcessOneReference(oneSentence);
    individualSentences.push_back(oneSentenceHash);
    //cout<<"Pushing on one ref ==============="<<endl;
    //PrintHashElements(oneSentenceHash);
  }//all references for one source sentence are processed
  //each element in individualSentences is the hashtable for one sentence

  // sentence bank is completed
  SentenceHash combinedHash;
  for(vector<SentenceHash>::iterator iter = individualSentences.begin(); iter!=individualSentences.end(); ++iter){
    //check out each one now 
    for(SentenceHash::iterator gramIter = iter->begin(); gramIter!=iter->end(); ++gramIter){
      string gram= (*gramIter).first;
      //cout<<"Gram:"<<gram<<endl;
      //      pair<int,int> payload = (*gramIter).second;
      NgramDetails payload = (*gramIter).second;
      // if it exists, just add it in
      if (combinedHash.count(gram)==0){
	//cout<<"FirstGram:"<<gram<<endl;
	combinedHash[gram] = payload;
      }else{
	// see which one has a higher count
	//cout<<"Found:"<<gram<<" Count:"<<combinedHash[gram].first<<" VS:"<< payload.first<<endl;
	if (combinedHash[gram].numOccurrences < payload.numOccurrences){
	  //cout<<"Update:"<<gram<<" -- "<<combinedHash[gram].first<<" -> "<<payload.first<<endl;
	  combinedHash[gram] = payload;
	}
      }
    }//end the loop over each gram
    //cout<<"OneRefProcessed============="<<endl;
  }// all individual sentences are process and their grams places in the combinedHash
  //cout<<"AllRefsProcessed============="<<endl;

  //cout<<"Combined ==============="<<endl;
  //PrintHashElements(combinedHash);
  return combinedHash;
}

//returns the n-gram hash object for one reference
SentenceHash BleuSearch::ProcessOneReference(NGram oneSentence){
  
  //cout<<"Processing ref:"<<m_nGramConsidered<<endl;
  pair<NGram, std::vector< pair<int, int> > > gramsAndSizes = CollectNGrams(oneSentence, m_nGramConsidered);  
  SentenceHash gramHash;
  
  // used to lookup the ngram size in the second part of the gramsAndSizes
  int tracker =0;
  for (NGram::iterator iter = gramsAndSizes.first.begin(); iter!=gramsAndSizes.first.end(); ++iter){
    string gram = (*iter);
    //cout<<"TopGram:"<<gram<<" tracker:"<<tracker<<" sec:"<<gramsAndSizes.second[tracker].second<<" fir:"<<gramsAndSizes.second[tracker].first<<endl;
    if (gramHash.count(gram)==0){
      gramHash[gram].numOccurrences = 1;
      gramHash[gram].gramSize = gramsAndSizes.second[tracker].second - gramsAndSizes.second[tracker].first + 1 ;
      //cout<<"Adding:"<<gram<<" gramSize:"<< gramHash[gram].gramSize<<endl;
    }else{
      //cout<<"Incrementing:"<<gram<<endl;
      gramHash[gram].numOccurrences = gramHash[gram].numOccurrences + 1;
    }
    tracker++;
  }
  m_referenceLengths.push_back(oneSentence.size());
  return gramHash;
}

pair<NGram, std::vector< pair<int, int> > > BleuSearch::CollectNGrams(NGram source, int maxNSize){

  NGram nList;
  //list of sizes to match the n-gram collected 
  std::vector< pair<int, int> > gramSizes;

  int size = source.size();
  //cout<<"maxNSize:"<<maxNSize<<" srcSize="<<size<<endl;
  
  for (int n=1; n<=maxNSize; n++){

    // where to start searching from
    for (int i=0; i<=size-n; i++){

      string accumulated = "";
      for (int j=0; j<n; j++){
	//foundGram.push_back(source[i+j]);
	accumulated = accumulated + source[i+j];
	if (j+1<n){
	  accumulated += " ";
	}
      }
      //cout<<"Acc:"<<accumulated<<" i="<<i<<" i+n-1="<<i+n-1<<" n="<<n<<endl;
      nList.push_back(accumulated);
      gramSizes.push_back(pair<int, int>(i, i+n-1));
      
    }
  }
  
  //pair<NGram, std::vector<int> > ret;
  //ret.first = nList;
  //ret.second = gramSizes;
  return pair<NGram, std::vector< pair<int, int> > >(nList, gramSizes);
}
void BleuSearch::PrintConfiguration(Configuration config){

  cout<<"CurrentConfiguration:";
  for (int i=0; i<m_numSentences; i++){
    cout<<config[i]<<" ";
  }
  cout<<endl;
}

/*
void BleuSearch::PrintSharedElements(int sentenceNumber, int candidateNumber, ostream& st){

   OneSentenceCommonIndices test = m_AllCommonIndices[sentenceNumber];
   for (OneSentenceCommonIndices::iterator blockIter = test.begin(); blockIter!=test.end(); ++blockIter){
     CandidateIndexHash oneGroup = (*blockIter);
     //cout<<"Looking For:"<<candidateNumber<<" Found: "<<oneGroup.count(candidateNumber)<<" Size:"<<oneGroup.size()<<endl;
     if (oneGroup.count(candidateNumber)!=0){
       st<<"Proximity:"<<sentenceNumber<<" "<<oneGroup.size()<<" ";
     for (CandidateIndexHash::iterator checkIter = oneGroup.begin(); checkIter != oneGroup.end(); ++checkIter){
       st<<(*checkIter)<<" ";
     }
     st<<endl;
     st<<flush;
     break;
     }
   }
}
*/

void BleuSearch::GetOneCandidateLine(Configuration config, int sentenceNumber, ifstream& infile, ofstream& ofile){
  stringstream ss (stringstream::in | stringstream::out);
  ss << "Proximity:"<<sentenceNumber<<":"<<config[sentenceNumber]<<":";
  char buf[10000];
  buf[0]='\0';
  string fileLine="";
  cout<<"Searching:"<<ss.str()<<endl;
  while (fileLine.find(ss.str().c_str())!=0){
    //cout<<"   fileLine:"<<fileLine<<endl;
    if (infile.eof()){
      cout<<"UNABLE TO FIND:::::::::"<<sentenceNumber<<":"<<config[sentenceNumber]<<endl;
      break;
    }
    infile.getline(buf, 10000);
    fileLine = buf;
  }
  /*
  char* proxData = strtok((char*)fileLine.c_str(), ":");
  strtok(NULL, ":");
  char* payload = strtok(NULL, ":");
  //cout<<"Outputting:"<<"Proximity:"<<sentenceNumber<<":"<<payload<<endl;
  */

  ofile<<fileLine<<endl;
  //ofile<<buf<<endl;
  // roll up to next line
  
  /*
  
  stringstream ss2 (stringstream::in | stringstream::out);
  ss2<<"SentenceProximity:"<<sentenceNumber+1;
  //cout<<"StrollingForward:"<<ss2.str()<<endl;
  while (fileLine.find(ss2.str().c_str())!=0){    
    if (!infile.eof()){
      infile.getline(buf, 1000);
      fileLine = buf;
    }else{
      break;
    }

  }
  */
  return;
  
}


void BleuSearch::PrintConfigurationToFile(Configuration config, const char* fileName, const char* matchFile){

  ofstream f(fileName);
  if (!f.good()){
    cout<<"Unable to open output file:"<<fileName<<endl;
    exit(-1);
  }
  ifstream infile(matchFile, ifstream::in);
  cout<<"OpeningMatch:"<<matchFile<<endl;
  if (!infile.good()){
    cout<<"Unable to open match file:"<<matchFile<<endl;
    exit(-1);
  }

  for (int i=0; i<m_numSentences; i++){
      GetOneCandidateLine(config, i, infile, f);
      //PrintSharedElements(i, config[i], f);
      
  }
  //  cout<<"Finished printing config files"<<endl;
  
  infile.close();
  f.close();
}












