#ifndef __METRICRAW__
#define __METRICRAW__

#include <iostream>
using namespace std;


//! Size of the n-grams being tracked. Dont use this variable outside MetricRaw, use MetricSearch->GetNGramsConsidered
#define NSZ 5

// a n-gram tracker
typedef double Counter[NSZ];

//holds intermediate metric statistics
class RawCounts;

class RawCounts{

public:
  RawCounts(){
    for (int i=0; i<NSZ; i++){
      numCorrect[i] = 0;
      numSuggested[i] = 0;
    }
    referenceLengths = 0;
  }
  RawCounts(const RawCounts& r){
    for (int i=0; i<NSZ; i++){
      numCorrect[i] = r.numCorrect[i];
      numSuggested[i] = r.numSuggested[i];
    }
    referenceLengths = r.referenceLengths;
  }

  Counter numCorrect;
  Counter numSuggested;
  double referenceLengths;

  RawCounts operator-(const RawCounts& d2) const{
    RawCounts r;
    for (int i=0; i<NSZ; i++){
      r.numCorrect[i] = numCorrect[i]-d2.numCorrect[i];
      r.numSuggested[i] = numSuggested[i]-d2.numSuggested[i];
    }
    r.referenceLengths = referenceLengths-d2.referenceLengths;
    return r;
  }
  
  RawCounts operator + (const RawCounts& d2) const{
    RawCounts r;
    for (int i=0; i<NSZ; i++){
      r.numCorrect[i] = numCorrect[i]+d2.numCorrect[i];
      r.numSuggested[i] = numSuggested[i]+d2.numSuggested[i];
    }
    r.referenceLengths = referenceLengths+d2.referenceLengths;
    return r;
  }
  
  RawCounts& operator = (const RawCounts& r){
    if (this == &r){
      return *this;
    }
    for (int i=0; i<NSZ; i++){
      this->numCorrect[i] = r.numCorrect[i];
      this->numSuggested[i] = r.numSuggested[i];
    }
    this->referenceLengths = r.referenceLengths;

    return *this;
  }
  
};

//ostream& operator << (ostream& os, const RawCounts* s);
//istream& operator >> ( istream& in, RawCounts &s );

#endif
