#include <OptimizeMER.hh>
#include <math.h>
#include <algorithm>
#include <functional>

using namespace std;

//define first dimension
int COptimizeMER::m_startBlank = 0;

//add 2 error counts together ONLY if they are at the same point
ErrorAt operator+(const ErrorAt& left, const ErrorAt& right){
  RawCounts r = left.second+right.second;
  assert(left.first==right.first);
  return ErrorAt(left.first, r);
}

void COptimizeMER::Init(Lambdas& leftRange, Lambdas& rightRange, int numPermutes, int numModify, int numStarts, bool twoStage, bool doNorm, double permutationEpsilon, double epsilonScale){
  
  SetRangeLimits(leftRange, rightRange);
  
  m_numPermutes = numPermutes;
  m_numModify = numModify;
  m_numStarts = numStarts;
  m_doNorm = doNorm;
  m_twoStage = twoStage;
  m_epsilonScale = epsilonScale;
  m_permutationEpsilon = permutationEpsilon;
  
  cout<<m_printTag<<" Init: ";
  copy(m_leftRange.begin(), m_leftRange.end(), ostream_iterator<double>(cout, " "));
  cout<<endl;

  cout<<m_printTag<<" Init: ";
  copy(m_rightRange.begin(), m_rightRange.end(), ostream_iterator<double>(cout, " "));
  cout<<endl;

  cout<<m_printTag<<" Init: "<<" m_numPermutes: "<<m_numPermutes<<"m_numModify: "<<m_numModify<<" m_numStarts: "<<m_numStarts<<" m_doNorm: "<<m_doNorm<<" m_twoStage: "<<m_twoStage<<" twoStage:"<<twoStage<<" m_permutationEpsilon: "<<m_permutationEpsilon<<" m_epsilonScale: "<<m_epsilonScale<<endl;
}


Lambdas COptimizeMER::minModelCostMaxMetricScore(const vector<Features*>& costs, const vector<RawCounts*>&  scores, const vector<int>& startEnd, const Lambdas& initialValues, int firstXsentences){
  //values used over multiple function calls
  //m_costs = costs;
  //m_scores = scores;

  //fill a candidates - with just iterators, shouldnt be a large memory move
  int numSentences = startEnd.size();
  cout<<"numSentences in opti: "<<numSentences<<endl;
  vector<Instance> allSentences;
  allSentences.reserve(numSentences);
  setupSentenceRanges(costs.begin(), scores.begin(), costs.end(), scores.end(), startEnd, allSentences, initialValues.size());


  //get the same Initial Score reading that comes out of the optimizer before any ranking
  double initialError = VerifyCurrentRanks(allSentences, initialValues);
  cout<<"ranges set: initial error:"<<initialError<<endl;
  //run the optimization using the strategy specified here
  //Lambdas returnLambdas = optimizeInOrder(allSentences, initialValues);
  //double JumpPerc = 0.1;
  //pair<Lambdas, float> returnLambdas = optimizeByGainWithPermutations(allSentences, initialValues, JumpPerc, m_numPermutes);


  double JumpPerc = 0.0;
  cout<<"about to start opti\n";
  pair<Lambdas, double> returnLambdas = optimizeByGainWithPermutationsWithRandomStarts(allSentences, initialValues, JumpPerc, m_numPermutes, m_numStarts);
  if (m_twoStage){
    returnLambdas = optimizeByGainTwoStage(allSentences, returnLambdas.first,  returnLambdas.second);
  }

  cout<<"Best Lambdas: ";
  copy(returnLambdas.first.begin(), returnLambdas.first.end(), ostream_iterator<double>(cout, " "));
  cout<<"Error="<<returnLambdas.second<<" Score="<<m_pMetric->maxScore()-returnLambdas.second<<endl;

  //sanity test to make sure you sent back the right values
  RawCounts r1 = ScoreWithLambdas(allSentences, allSentences.size(), returnLambdas.first);
  cout<<"FinalScoreAllNBest(ranked) Raw: ";
  for (int i=0; i<NSZ; i++){
    cout<<r1.numCorrect[i]<<" "<<r1.numSuggested[i]<<" ";
  }
  cout<<r1.referenceLengths<<endl;
  cout<<"FinalScoreAllNBest(ranked): "<<m_pMetric->getMetricScore(r1)<<endl;

  //what was the final score on the real full sentence data (when using sub sentences data in the optimization)
  RawCounts rfull = ScoreWithLambdas(allSentences, firstXsentences, returnLambdas.first);
  cout<<"FinalScore(ranked-firstX) Raw: ";
  for (int i=0; i<NSZ; i++){
    cout<<rfull.numCorrect[i]<<" "<<rfull.numSuggested[i]<<" ";
  }
  cout<<rfull.referenceLengths<<endl;
  cout<<"FinalScore(ranked-firstX): "<<m_pMetric->getMetricScore(rfull)<<endl;

  return returnLambdas.first;
}

void COptimizeMER::SetStart(int start){
  COptimizeMER::m_startBlank = start;
}
//the two stage process is as follow
// foreach each dim, set, fix the m_leftRange[dim] = m_rightRange[dim] = 0
//    run optimization
//    release the lambda ranges
//    run optimization
//    if error is lower than best performing, replace the scaling factors
pair<Lambdas, double> COptimizeMER::optimizeByGainTwoStage(vector<Instance>& allSentences, const Lambdas& initialValues, double initialError){
  
  int numDimensions = initialValues.size(); 
  //represents the working lambda set - ie the best so far
  pair<Lambdas, double> bestLambdasAndError;
  bestLambdasAndError.second = initialError;
  bestLambdasAndError.first = initialValues;

  //left and right ranges across iterations
  Lambdas savedLeftRange = m_leftRange;
  Lambdas savedRightRange = m_rightRange;
  //number of drops that have happenned in a row
  //  int dropSequence = 0;

  //the original scheme involved blanking out the whole 0<p<numDimensions at each iteration
  //now we will maintain a static variable to indicate which parameter to start with, and a
  //user parameter NumModify that indicates how many elements are blanked in this iteration
  //always do NumModify blankings by doing it modulo numDimensions
  int p = COptimizeMER::m_startBlank;
  int numModified = 0;

  cout<<"Log: Opti: m_startBlank: "<<p<<endl;
  //  while (dropSequence!=numDimensions){
  while (true) {

    Lambdas initialValuesCopy = bestLambdasAndError.first;
    
    assert(m_leftRange.size()>p);
    if (p>=m_leftRange.size()){
      cerr<<"ERROR: Starting blanking requested at p:"<<p<<" but numfeatures: "<<m_leftRange.size()<<endl;
    }
    if (m_leftRange[p] <= 0 && m_rightRange[p] >= 0) {
      //blank out the p'th field
      initialValuesCopy[p] = 0.0;
      m_leftRange[p] = 0.0;
      m_rightRange[p] = 0.0;
      
      //track what is going on 
      cout<<"======== Blanking (zeroing-out) "<<p<<" ============="<<endl;
      cout<<"Blanking: "<<p<<" Starting with= ";
      copy(initialValuesCopy.begin(), initialValuesCopy.end(), ostream_iterator<double>(cout, " "));
      cout<<endl;
      cout<<"Blanking: "<<p<<" LeftRange with= ";
      copy(m_leftRange.begin(), m_leftRange.end(), ostream_iterator<double>(cout, " "));
      cout<<endl;
      cout<<"Blanking: "<<p<<" RightRange with= ";
      copy(m_rightRange.begin(), m_rightRange.end(), ostream_iterator<double>(cout, " "));
      cout<<endl;
      
      //normalization is already done in the Permutations layer against the real initialValues as passed into this function
      cout<<"======== Restricted "<<p<<" ============="<<endl;
      pair<Lambdas, double> paramAndError = optimizeByGainWithPermutations(allSentences, initialValuesCopy, initialValues);
      cout<<"Restricted: "<<p<<" Ending with= ";
      copy(paramAndError.first.begin(), paramAndError.first.end(), ostream_iterator<double>(cout, " "));
      cout<<" Error="<<paramAndError.second;
      cout<<endl;
      // if we got better or stayed equal, save these params as best-params
      // note that the inequality is non-strict (do better or equal) in the line below to zero out weights
      // whenever that doesn't hurt - you get a little leeway to even be worse since we prefer zeroed out params
      if (paramAndError.second <= bestLambdasAndError.second + m_permutationEpsilon*m_epsilonScale){ 
	bestLambdasAndError.second = paramAndError.second;
	bestLambdasAndError.first = paramAndError.first;
	cout<<"Zeroed-param opti bettered or is almost equal - saving the params as best-params"<<endl;
	//	initialValuesCopy = bestLambdasAndError.first;
	//	dropSequence=0;
      }else{
	cout<<"Zeroed-param opti worsened the error."<<endl;
	//	initialValuesCopy = bestLambdasAndError.first;
	//	dropSequence++;
      }
      
      //now open the ranges back up, and do optimization again
      m_leftRange = savedLeftRange;
      m_rightRange = savedRightRange;
      //pass in paramAndError.first which are the optimized values from the restricted opti
      cout<<"======== UnRestricted "<<p<<" (i.e., unrestricting the param, but starting opti from current zeroed-out opti result)============="<<endl;
      paramAndError = optimizeByGainWithPermutations(allSentences, paramAndError.first, initialValues);
      cout<<"UnRestricted: "<<p<<" Ending with= ";
      copy(paramAndError.first.begin(), paramAndError.first.end(), ostream_iterator<double>(cout, " "));
      cout<<" Error="<<paramAndError.second;
      cout<<endl;
      
      // note that the inequality is strict (do-better) in the line
      // below to keep the zero out weights
      // whenever that doesn't hurt
      if (paramAndError.second < bestLambdasAndError.second){ 
	bestLambdasAndError.second = paramAndError.second;
	bestLambdasAndError.first = paramAndError.first;
	cout<<"Unrestricting zeroed-out param bettered"<<endl;
	initialValuesCopy = bestLambdasAndError.first;
	//	dropSequence=0;
      }else{
	cout<<"Unrestricting zeroed-out param didn't better - dopping these params"<<endl;
	//	cout<<"Rand Permutation dropped: dropSequence="<<dropSequence<<endl;
	initialValuesCopy = bestLambdasAndError.first;
	//	dropSequence++;
      }
    
    }else{
      cout<<"Blanking: "<<p<<" denied "<<endl;
    }

    //increment the counter
    p++;
    //keep within the range
    p = p % numDimensions;
    numModified++;

    if (numModified==m_numModify){
      break;
    }
  }
  //for the next MER round, start at the ending p
  COptimizeMER::m_startBlank = p;
  
  return bestLambdasAndError;

}

double COptimizeMER::VerifyCurrentRanks(vector<Instance>& allSentences, const Lambdas& lambdas){
  RawCounts r0;
  for (int s=0; s<allSentences.size(); s++){
    Instance& oneSentence = allSentences[s];
    const RawCounts* score = (*(oneSentence.m_scoreRange.first));
    r0 = r0 + (*score);
  }
  cout<<"InitialScore(unranked) Raw: ";
  for (int i=0; i<NSZ; i++){
    cout<<r0.numCorrect[i]<<" "<<r0.numSuggested[i]<<" ";
  }
  cout<<r0.referenceLengths<<endl;
  cout<<"InitialScore(unranked): "<<m_pMetric->getMetricScore(r0)<<endl;
  
  RawCounts r1 = ScoreWithLambdas(allSentences, allSentences.size(),  lambdas);
  cout<<"InitialScore(ranked) Raw: ";
  for (int i=0; i<NSZ; i++){
    cout<<r1.numCorrect[i]<<" "<<r1.numSuggested[i]<<" ";
  }
  cout<<r1.referenceLengths<<endl;
  cout<<"InitialScore(ranked): "<<m_pMetric->getMetricScore(r1)<<endl;
  return m_pMetric->maxScore()-m_pMetric->getMetricScore(r1);
}

RawCounts COptimizeMER::ScoreWithLambdas(const vector<Instance>& allSentences, int firstXsentences,  const Lambdas& lambdas){
  //now do the ranked versions
  RawCounts r1;
  for (int s=0; s<allSentences.size() && s<firstXsentences; s++){
    const Instance& oneSentence = allSentences[s];
    //setup one sentence's feature data
    pair<int, RawCounts> indexAndScore = lowestCostCandidate(oneSentence, lambdas);
    //add up the scores data
    r1 = r1 + indexAndScore.second;
  }
  return r1;
}

pair<int, RawCounts> COptimizeMER::lowestCostCandidate(const Instance& oneSentence, const Lambdas& lambdas){
  //pick the feat with the lowest cost
  int candNum=0;
  int lowestCostCand = 0;
  double lowestCost = COptimizeMER::LargeCost;
  for (vector< Features*>::const_iterator iter= oneSentence.m_dataRange.first; iter!=oneSentence.m_dataRange.second; ++iter ){
    const Features* f = (*iter);
    assert(f->size()==lambdas.size());
    double cost = calc_innerProduct(f, lambdas);
    if (cost<lowestCost){
      lowestCost = cost;
      lowestCostCand = candNum;
    }
    candNum++;
  }
  
  vector< RawCounts*>::const_iterator scIter = oneSentence.m_scoreRange.first + lowestCostCand;
  return pair<int, RawCounts>(lowestCostCand, (*(*scIter)));
  
}

//!Simple test to make sure that if we randomly permute one dimension and set it to a random value, it will always come back to the optimized value
void COptimizeMER::StressTest(int dim, double currentError, vector<Instance>& allSentences, const Lambdas& initialValues){

  int numTests = 20;

  Lambdas l = initialValues;
  cout<<"StressStart: \t\t: "; copy(l.begin(), l.end(), ostream_iterator<double>(cout, " ")); cout<<endl;

  srand(time(NULL));/*initialize random number generator */
  for (int n=0; n<numTests; n++){

    //reset to to the initial values
    l = initialValues;
    double r = (   (float)rand() / ((float)(RAND_MAX)+(float)(1)) );
    double newSpot = m_leftRange[dim] + (m_rightRange[dim]-m_leftRange[dim])*r;    
    l[dim] = newSpot;


    cout<<"StressInitial: \t\t: "; copy(l.begin(), l.end(), ostream_iterator<double>(cout, " ")); cout<<endl;
    pair<double, double> lambdaError = optimizeDimension(dim, initialValues, allSentences);    

    //set the value
    l[dim] = lambdaError.first;
    
    cout<<"StressFinal: \t\t: "; copy(l.begin(), l.end(), ostream_iterator<double>(cout, " ")); cout<<endl;
    cout<<" Value: "<<lambdaError.first<<" Error: "<<lambdaError.second<<endl;
  }
}
pair<Lambdas, double> COptimizeMER::optimizeByGainWithPermutationsWithRandomStarts(vector<Instance>& allSentences, const Lambdas& initialValues, double perc, int numPermutes, int numStarts){


  pair<Lambdas, double> bestLambdasAndError;
  bestLambdasAndError.second = COptimizeMER::LargeCost;
  bestLambdasAndError.first = initialValues;

  assert(numStarts>0);

  //the copy is the one that we permute and try again with
  Lambdas initialValuesCopy = initialValues;
  
  for (int p = 0; p<numStarts; p++){
    cout<<"Rand "<<p<<" Starting with= ";
    copy(initialValuesCopy.begin(), initialValuesCopy.end(), ostream_iterator<double>(cout, " "));
    cout<<endl;

    //normalization is already done in the Permutations layer
    pair<Lambdas, double> paramAndError = optimizeByGainWithPermutations(allSentences, initialValuesCopy, initialValues);
    cout<<"Rand "<<p<<" Ending with= ";
    copy(paramAndError.first.begin(), paramAndError.first.end(), ostream_iterator<double>(cout, " "));
    cout<<" Error="<<paramAndError.second;
    cout<<endl;


    //now do a full stress test
    //StressTest(0, paramAndError.second, allSentences, initialValuesCopy);
    
    if (paramAndError.second<bestLambdasAndError.second){
      bestLambdasAndError.second = paramAndError.second;
      bestLambdasAndError.first = paramAndError.first;
      cout<<"Rand Permutation bettered"<<endl;
      initialValuesCopy = bestLambdasAndError.first;
    }else{
      cout<<"Rand Permutation dropped"<<endl;
      initialValuesCopy = bestLambdasAndError.first;
    }


    //permute the best performing one so far, which is initialValuesCopy
    if (p!=numStarts-1){

      vector<bool> allTrue(initialValues.size(), true);
      RandomPermutation(initialValuesCopy, allTrue);
      
      if (m_doNorm){
	NormalizeAgainstInitial(initialValuesCopy, initialValues);
      }
      cout<<"RandFinalNorm: ";
      copy(initialValuesCopy.begin(), initialValuesCopy.end(), ostream_iterator<double>(cout, " "));
      cout<<endl;
      //assert(L2(initialValuesCopy)==L2(initialValues));

    }
      
    
  }
  return bestLambdasAndError;
}

//optimize, then permute, see if you improve
pair<Lambdas, double> COptimizeMER::optimizeByGainWithPermutations(vector<Instance>& allSentences, const Lambdas& initialValues, const Lambdas& constantInitialValues){

  pair<Lambdas, double> bestLambdasAndError;
  bestLambdasAndError.second = COptimizeMER::LargeCost;
  bestLambdasAndError.first = initialValues;

  assert(m_numPermutes>0);
  Lambdas initialValuesCopy = initialValues;

  for (int p = 0; p<m_numPermutes; p++){
    cout<<"Starting permutation iter:"<<p<<" with: ";
    copy(initialValuesCopy.begin(), initialValuesCopy.end(), ostream_iterator<double>(cout, " "));
    cout<<endl;
    pair<Lambdas, double> paramAndError = optimizeByGain(allSentences, initialValuesCopy);
    cout<<"Ending permutations iter:"<<p<<" with: ";
    copy(paramAndError.first.begin(), paramAndError.first.end(), ostream_iterator<double>(cout, " "));
    cout<<" Error="<<paramAndError.second;
    cout<<endl;
    if (m_doNorm){
      //normalize it against the first set of values that come in here
      NormalizeAgainstInitial(paramAndError.first, constantInitialValues);
      //cout<<"Asserting before"<<endl;
      //cout<<"L2(initialValues)="<<L2(initialValues)<<endl;
      //cout<<"L2(paramAndError.first)="<<L2(paramAndError.first)<<endl;
      //assert(L2(paramAndError.first)==L2(initialValues));

      cout<<"EndingNorm permutations iter:"<<p<<" with: ";
      copy(paramAndError.first.begin(), paramAndError.first.end(), ostream_iterator<double>(cout, " "));
      cout<<" Error="<<paramAndError.second;
      cout<<endl;
      
      //Lambdas rawLambdas = paramAndError.first;
      //RawCounts rBefore = ScoreWithLambdas(allSentences, rawLambdas);
      
      //RawCounts rAfter = ScoreWithLambdas(allSentences, paramAndError.first);
      //double scoreBefore = m_pMetric->getMetricScore(rBefore);
      //double scoreAfter = m_pMetric->getMetricScore(rAfter);
      //assert(scoreBefore==scoreAfter);
      //assert(L2(paramAndError.first)==L2(initialValues));

    }
    
    if (paramAndError.second<=bestLambdasAndError.second){
      bestLambdasAndError.second = paramAndError.second;
      bestLambdasAndError.first = paramAndError.first;
      cout<<"Permutation bettered or is equal -> keeping it"<<endl;
      //now permute based on this value, which is new now
      initialValuesCopy = bestLambdasAndError.first;
    }else{
      cout<<"Permutation dropped"<<endl;
      initialValuesCopy = bestLambdasAndError.first;
    }


    //change values of the lambdas randomly
    if (p!=m_numPermutes-1){
      //randomly permute certain elements of the vector
      RandomPermutation(initialValuesCopy, SelectForSelection(initialValues.size(), m_numModify));
      //renormalizing if neccesary
      if (m_doNorm){

	//against the very first set of values
	NormalizeAgainstInitial(initialValuesCopy, constantInitialValues);
	cout<<"PermFinalNorm: ";
	copy(initialValuesCopy.begin(), initialValuesCopy.end(), ostream_iterator<double>(cout, " "));
	cout<<endl;
      }
    }

    
  }

  //cout<<"Perm Final Lambdas: ";
  //copy(bestLambdasAndError.first.begin(), bestLambdasAndError.first.end(), ostream_iterator<double>(cout, " "));
  //cout<<"Error="<<bestLambdasAndError.second<<" Score="<<m_pMetric->maxScore()-bestLambdasAndError.second<<endl;
  return bestLambdasAndError;
}

//!optimize in the order of parameters starting at the param specified
pair<Lambdas,double> COptimizeMER::optimizeByGain(vector<Instance>& allSentences, const Lambdas& initialValues){
  Lambdas returnLambdas = initialValues;
  int numDimensions = initialValues.size();

  //initialize the dot products
  for (int s=0; s<allSentences.size(); s++){
    InitializeDotProducts(allSentences[s], initialValues);
  }

  int ConvergedLimit = 1; //since we are opportunistically selecting dimensions
  //double PermutationEpsilon = 0.0; //find the very best at each step... it will take longer

  double lastError = COptimizeMER::LargeCost;
  const double maxScorePosible = m_pMetric->maxScore();

  int dim =0;
  int converged=0;

  Lambdas savedLambdas = returnLambdas;
  int itersForConvergence = 0;
  while (converged<ConvergedLimit){
    double savedLambda = returnLambdas[dim];
    //it start off unset
    int useDimension = -1;
    double useDimensionValue = COptimizeMER::LargeCost;
    double useDimensionLowError = lastError;

    //cout<<"Search dimensions: ";
    for (int dim=0; dim<numDimensions; dim++){
      pair<double, double> lambdaError = optimizeDimension(dim, returnLambdas, allSentences);
      double errorForDim = lambdaError.second;
      //cout<<"D("<<dim<<")="<<errorForDim<<"=>"<<lambdaError.first<<" ";
      if (errorForDim<useDimensionLowError){
	useDimension = dim;
	useDimensionValue = lambdaError.first;
	useDimensionLowError = errorForDim;
	//cout<<"--- Lowest so far"<<endl;
      }
    }
    cout<<endl;

    //if none of the dimensions were able to lower the error
    if (useDimension==-1){
      cout<<"No further dimensions"<<endl;
      break;
    }else{
      //if they were able
      if (lastError-useDimensionLowError<m_permutationEpsilon){
	cout<<"Incremental convergence on:"<<m_permutationEpsilon<<endl;
	converged++;
      }else{
	converged=0;
      }
      lastError = useDimensionLowError;
      //cout<<"Using dimension = "<<useDimension<<" lowError: "<<useDimensionLowError<<" value="<<useDimensionValue<<endl;
      double oldValue = returnLambdas[useDimension];
      returnLambdas[useDimension] = useDimensionValue;
      //known to be lower
      cout<<"Setting("<<useDimension<<") "; copy(returnLambdas.begin(), returnLambdas.end(), ostream_iterator<double>(cout, " ")); cout<<" Error="<<useDimensionLowError<<endl;
      itersForConvergence++;

      //recompute all the sums
      for (int s=0; s<allSentences.size(); s++){
	RecomputeScaledSums(allSentences[s], useDimension, oldValue, useDimensionValue);
      }
    }
  }
  cout<<"Opti: itersForConvergence: "<<itersForConvergence<<endl;
  return pair<Lambdas, double>(returnLambdas, lastError);

}

vector<bool> COptimizeMER::SelectForSelection(const int totalNumElements, const int numToSelect){

  assert(numToSelect<=totalNumElements);
  vector<bool> toReturn(totalNumElements, false);


  srand(time(NULL));/*initialize random number generator*/

  int index =0;
  double s = (   (double)rand() / ((double)(RAND_MAX)+(double)(1)) );
  index = (int)floor(s*totalNumElements);

  cout<<"Need to select: "<<numToSelect<<" / "<<totalNumElements<<endl;
  cout<<"Initial sel: ";
  copy(toReturn.begin(), toReturn.end(), ostream_iterator<bool>(cout, " "));
  cout<<" staring index: "<<index<<endl;
  int numSelected = 0;
  do{

    //check if index is going to be selected
    double r = (   (double)rand() / ((double)(RAND_MAX)+(double)(1.0)) );
    assert(index<toReturn.size());
    if (r>0.5 && toReturn[index]==false){
      numSelected++;
      toReturn[index] = true;
    }
    //reset
    index++;
    if (index==totalNumElements){
      index=0;
    }
  }while(numSelected<numToSelect);
  assert(numSelected==numToSelect);

  cout<<"Selected: ";
  copy(toReturn.begin(), toReturn.end(), ostream_iterator<bool>(cout, " "));
  cout<<endl;

  return toReturn;
}

void COptimizeMER::NormalizeAgainstInitial(Lambdas& toModify, const Lambdas& reference){


  double l2initial = L2(reference);
  double l2current = L2(toModify);

#ifndef NDEBUG

  copy(reference.begin(), reference.end(), ostream_iterator<double>(cout, " "));
  cout<<endl;
  //cout<<"L2(reference):"<<L2(reference)<<endl;

  //  cout<<"NormalingInital: ";
  copy(toModify.begin(), toModify.end(), ostream_iterator<double>(cout, " "));
  cout<<endl;
#endif

  for (int i=0; i<toModify.size(); i++){
    toModify[i] = (toModify[i]/l2current);
  }

  //cout <<"L2 after div="<<L2(toModify)<<endl;
  for (int i=0; i<toModify.size(); i++){
    toModify[i] *= l2initial;
  }
  //cout <<"L2 after mult="<<L2(toModify)<<endl;


#ifndef NDEBUG
  //cout<<"NormalingFinal: ";
  copy(toModify.begin(), toModify.end(), ostream_iterator<double>(cout, " "));
  cout<<endl;
  //cout<<"L2(toModify):"<<L2(toModify)<<endl;
#endif



}

void COptimizeMER::RandomPermutation(Lambdas& returnLambdas, const vector<bool> selectRandom){
  //if (m_leftRange.size()==returnLambdas.size() && m_rightRange.size()==returnLambdas.size()){
  
  for (int i=0; i<returnLambdas.size(); i++){

    if (selectRandom[i]==true){
      double r = (   (double)rand() / ((double)(RAND_MAX)+(double)(1)) );
      double newSpot = m_leftRange[i] + (m_rightRange[i]-m_leftRange[i])*r;
      
      if (newSpot>m_rightRange[i]){
	newSpot = m_rightRange[i];
      }
      if (newSpot<m_leftRange[i]){
	newSpot = m_leftRange[i];
      }
      returnLambdas[i] = newSpot;
    }//else dont change it 
  }
    //}
  
}

double COptimizeMER::L2(const Lambdas& lambdas){
  double l2initial = 0.0;
  for(Lambdas::const_iterator iter=lambdas.begin(); iter!=lambdas.end(); ++iter){
    double v = (*iter);
    l2initial += v*v;
  }
  l2initial = pow(l2initial, (double)0.5);
  return l2initial;
}
//!optimize in the order of parameters
Lambdas COptimizeMER::optimizeInOrder(vector<Instance>& allSentences, Lambdas& initialValues){

  Lambdas returnLambdas = initialValues;
  
  int numDimensions = initialValues.size();
  //evaluate the gain from each dimension
  int ConvergedLimit = numDimensions;

  double PermutationEpsilon = m_pMetric->minEpsilon();
  PermutationEpsilon=0.0;
  cout<<"PermutationEpsilon: "<<PermutationEpsilon<<endl;
  
  double lastError = COptimizeMER::LargeCost; //you just have to beat this value to stay in the game
  const double maxScorePosible = m_pMetric->maxScore();

  int dim =0;
  int converged=0;

  while (converged<ConvergedLimit){
    double savedLambda = returnLambdas[dim];
    pair<double, double> lambdaError = optimizeDimension(dim, returnLambdas, allSentences);
    //returnLambdas[dim] = lambdaError.first;
    returnLambdas[dim] = lambdaError.first;
    
    cout<<"Dimension: ("<<dim<<") ";
    copy(returnLambdas.begin(), returnLambdas.end(), ostream_iterator<double>(cout, " "));
    cout<<" value= "<<lambdaError.first<<" error="<<lambdaError.second<<" score="<<maxScorePosible-lambdaError.second;
    if (lambdaError.second>lastError){
      returnLambdas[dim]=savedLambda;
      lambdaError.second=lastError;
      cout<<" REJECT"<<endl;
    }else{
      //we have accepted the change
      cout<<endl;
    }
    if (fabs(lambdaError.second-lastError)<=PermutationEpsilon){
      converged++;
      cout<<"Incremental convergence"<<endl;
    }else{
      converged=0;
    }
    //set it for the next round
    lastError = lambdaError.second;
    dim++; 
    //}
    /*
    else{
      cout<<" REJECT "<<endl;
      converged++;
      cout<<"Incremental convergence"<<endl;
    }
    */

    if (dim==numDimensions){
      dim=0;
    }
  }
  
  cout<<"Final Accuracy:"<<maxScorePosible-lastError<<endl;
  
    
}


//returns lambda and the associated error
pair<double, double> COptimizeMER::optimizeDimension(int dim, const Lambdas& initialValues, vector<Instance>& allSentences){

  Lambdas initialValueCopy = initialValues;
  bool debug = false;
  if (debug){
    cout<<"Optimizing dim="<<dim<<" with: ";
    copy(initialValueCopy.begin(), initialValueCopy.end(), ostream_iterator<double>(cout, " "));
    cout<<endl;
  }
  vector<ErrorAt> totalErrorDelta;
  for (int s=0; s<allSentences.size(); s++){
    vector<ErrorAt> oneSentenceError = singleSentencePoints(dim, initialValueCopy, allSentences[s]); 
    
    //add it to the total error delta
    totalErrorDelta.insert(totalErrorDelta.end(), oneSentenceError.begin(), oneSentenceError.end());
    //display it for now
    
    if (debug){
    cout<<"Sentence: "<<s<<" errorDeltas =============="<<endl;
    for (vector<ErrorAt>::iterator iter=oneSentenceError.begin(); iter!=oneSentenceError.end(); ++iter){
      ErrorAt& e = *iter;
      cout<<"Lambda: "<<e.first<<" ";
      cout<<"Counts: ";
      for (int i=0; i<NSZ; i++){
	cout<<i<<"="<<e.second.numCorrect[i]<<"/"<<e.second.numSuggested[i]<<"\t";
      }
      cout<<e.second.referenceLengths;
      cout<<endl;
    }
    cout<<"Done =================== "<<endl;
    }
    
    //cout<<"Printin all errors from sent: "<<s<<endl;
    //ostream_iterator<RawCounts> out(cout, " ");
    //copy( oneSentenceError.begin(), oneSentenceError.end(), out); cout<<endl;
  }

  
  //sort the values
  sort(totalErrorDelta.begin(), totalErrorDelta.end(), SortErrorAt());

  if (debug){
  cout<<"Sorted Deltas ===================== "<<totalErrorDelta.size()<<endl;
  for (vector<ErrorAt>::iterator iter=totalErrorDelta.begin(); iter!=totalErrorDelta.end(); ++iter){
    ErrorAt& e = *iter;
    cout<<"Lambda: "<<e.first<<" ";
    cout<<"Counts: ";
    for (int i=0; i<NSZ; i++){
      cout<<i<<"="<<e.second.numCorrect[i]<<"/"<<e.second.numSuggested[i]<<" ";
    }
    cout<<e.second.referenceLengths;
    cout<<endl;
  }
  cout<<endl;
  }
  


  assert(m_rightRange.size()>dim);
  vector<ErrorAt> uniqueErrorDeltas;
  mergeErrorDeltas(totalErrorDelta, uniqueErrorDeltas );
  
  //clear the memory off the non-unique one
  //vector<ErrorAt>().swap(totalErrorDelta);
  
  // add one last ErrorAt representing the very last line
  //RawCounts r0 = m_pMetric->initRawCounts(m_pMetric->GetNGramConsidered());
  RawCounts r0; //initialized by the constructor
  uniqueErrorDeltas.push_back(ErrorAt(m_rightRange[dim], r0));

  
  
  if (debug){
  cout<<"Unique Error Deltas ===================== "<<endl;
  for (vector<ErrorAt>::iterator iter=uniqueErrorDeltas.begin(); iter!=uniqueErrorDeltas.end(); ++iter){
    ErrorAt& e = *iter;
    cout<<"Lambda: "<<e.first<<" ";
    cout<<"Counts: ";
    for (int i=0; i<NSZ; i++){
      cout<<i<<"="<<e.second.numCorrect[i]<<"/"<<e.second.numSuggested[i]<<" ";
    }
    cout<<e.second.referenceLengths;
    cout<<endl;
  }
  cout<<endl<<endl;
  }

  vector< pair<double, double> > errorSurface;
  pair<double, double> minErrorLambda =  minPointError(uniqueErrorDeltas, m_leftRange[dim], errorSurface, debug);
  
  if (debug){
    cout<<"Surface ===================== "<<endl;
    for (vector< pair<double, double> >::iterator iter=errorSurface.begin(); iter!=errorSurface.end(); ++iter){
      pair<double, double> p = (*iter);
      cout<<"L="<<p.first<<" error="<<p.second<<endl;
    }
  }
  

  if (debug){
    cout<<"minErrorLambda: "<<minErrorLambda.first<<" error: "<<minErrorLambda.second<<endl;
  }
  return minErrorLambda;
  //cout<<"Dim completed: "<<dim<<endl;
  //return pair<double, Lambdas>(0, initialValueCopy);
}

pair<double, double> COptimizeMER::minPointError(vector<ErrorAt>& uniqueErrorDeltas, double lowestLambda, vector< pair<double, double> >& errorSurface, bool fillErrorSurface){
  //RawCounts baseError = m_pMetric->initRawCounts(m_pMetric->GetNGramConsidered()); //start as zero
  RawCounts baseError; //initialized by the constructor
  //go uptil the last element
  double maxScore = m_pMetric->maxScore();
  double lowestError = COptimizeMER::LargeCost;

  //cout<<"Error Surface ============"<<endl;
  
  for (vector<ErrorAt>::iterator iter=uniqueErrorDeltas.begin(); iter!=uniqueErrorDeltas.end()-1; ++iter){
    ErrorAt incrementalErrorAt=(*iter);
    baseError = baseError+incrementalErrorAt.second;

    vector<ErrorAt>::iterator nextBoundary = iter + 1;
    //get the actual error value
    double newError = maxScore - m_pMetric->getMetricScore(baseError);
    double newLambdaLocation = (incrementalErrorAt.first + (*nextBoundary).first)/2.0;

    if (newError<lowestError){
      lowestError = newError;
      lowestLambda = (incrementalErrorAt.first + (*nextBoundary).first)/2.0;
    }
    if (fillErrorSurface){
      errorSurface.push_back(pair<double, double>((incrementalErrorAt.first + (*nextBoundary).first)/2.0, newError));
    }

    /*
    cout<<"L="<<newLambdaLocation<<" E= "<<newError;
    for (int i=0; i<NSZ; i++){
      cout<<baseError.numCorrect[i]<<" "<<baseError.numSuggested[i]<<" ";
    }
    cout<<baseError.referenceLengths<<endl;
    */
  }
  return pair<double, double>(lowestLambda, lowestError);
  
  
}

void COptimizeMER::mergeErrorDeltas(vector<ErrorAt>& totalErrorDelta, vector<ErrorAt>& uniqueErrorDeltas ){
  
  vector<ErrorAt>::iterator currBegin = totalErrorDelta.begin();

  while (currBegin!=totalErrorDelta.end()){

    //accumulate the raw counts in range currBegin until firstDiff
    ErrorAt block;
    block.first =(*currBegin).first; //save the lambda value
    block.second = (*currBegin).second; //save the lambda value
    //look to see if we need to do accumulation
    vector<ErrorAt>::iterator firstDiff = currBegin+1;
    while(firstDiff!=totalErrorDelta.end() && (*firstDiff).first==(*currBegin).first){
      block.second = block.second + (*firstDiff).second;
      ++firstDiff;
    }
    //accumulate the block counts uptil firstDiff
    //ErrorAt block = accumulate(currBegin, firstDiff, r0);
    
    
    /*
    cout<<"Accumulation done:"<<endl;
    cout<<"Lambda: "<<block.first<<" ";
    cout<<"Counts: ";
    for (int i=0; i<NSZ; i++){
      cout<<i<<"="<<block.second.numCorrect[i]<<"/"<<block.second.numSuggested[i]<<"\t";
    }
    cout<<block.second.referenceLengths;
    cout<<endl;
    */
    uniqueErrorDeltas.push_back(block);

    //keep moving currBegin to firstDiff
    currBegin = firstDiff;
  }
}

//pre-compute the dot products for a given sentence at a given lambda set
void COptimizeMER::InitializeDotProducts(Instance& oneSentence, const Lambdas& initial){

  //reserve the right number of fields
  oneSentence.m_scaledFeaturesAndSum = vector< pair<vector<double>, double> >(oneSentence.m_numCands);
  const int numFeatures = initial.size();
  //go to each candidate
  for (int index=0; index<oneSentence.m_numCands; index++){
    //cout<<"index: "<<index<<endl;
    assert(oneSentence.m_dataRange.first+index<oneSentence.m_dataRange.second);
    Features* pOneCandidateFeatures = *(oneSentence.m_dataRange.first+index);

    //reserve space for the scaled features
    oneSentence.m_scaledFeaturesAndSum[index].first = vector<double>(numFeatures);
    
    assert(pOneCandidateFeatures->size()==initial.size());
    for (int feat = 0; feat<pOneCandidateFeatures->size(); feat++){
      oneSentence.m_scaledFeaturesAndSum[index].first[feat]  = (*pOneCandidateFeatures)[feat] * initial[feat];
    }
    //cout<<" oneSentence.m_scaledFeaturesAndSum.size():"<<oneSentence.m_scaledFeaturesAndSum.size()<<endl;
    //cout<<"index: "<<index<<endl;
    //cout<<" oneSentence.m_scaledFeaturesAndSum.first[index].size():"<<oneSentence.m_scaledFeaturesAndSum[index].first.size()<<endl;
    //copy(initialValueCopy.begin(), initialValueCopy.end(), ostream_iterator<double>(cout, " "));
    
    double featureSum = accumulate(oneSentence.m_scaledFeaturesAndSum[index].first.begin(), oneSentence.m_scaledFeaturesAndSum[index].first.end(), 0.0);
    oneSentence.m_scaledFeaturesAndSum[index].second = featureSum;
  }
}

void COptimizeMER::FillIntercepts(Instance& oneSentence, int dim, vector<double>& a){
  a = vector<double>(oneSentence.m_numCands);
  for (int index=0; index<oneSentence.m_numCands; index++){
    pair< vector<double>, double>* scSum =  &(oneSentence.m_scaledFeaturesAndSum[index]);
    //fill the candidate - its the sum minus the scaled value on the dim
    a[index] = scSum->second - scSum->first[dim];
  }
}

void COptimizeMER::RecomputeScaledSums(Instance& oneSentence, int dim, double oldScalingFactor, double newScalingFactor){
   for (int index=0; index<oneSentence.m_numCands; index++){
     pair< vector<double>, double>* scSum =  &(oneSentence.m_scaledFeaturesAndSum[index]);
     double replacement = GetSlope(oneSentence, index, dim)*newScalingFactor; 
     scSum->second = scSum->second - (scSum->first[dim]) + replacement;
     scSum->first[dim] = replacement;
   }
}

vector<ErrorAt> COptimizeMER::singleSentencePoints(const int dim, Lambdas initialValueCopy, Instance& oneSentence){
  //blank out one field
  initialValueCopy[dim]=0;

  //setup one sentence's feature data
  vector<Features*> feats;
  feats.reserve(oneSentence.m_numCands);
  feats.insert(feats.end(), oneSentence.m_dataRange.first, oneSentence.m_dataRange.second);
  //cout<<"1"<<endl;

  //setup one sentences score data
  vector<RawCounts*> scores;
  scores.reserve(oneSentence.m_numCands);
  scores.insert(scores.end(), oneSentence.m_scoreRange.first, oneSentence.m_scoreRange.second);
  //cout<<"2"<<endl;

  //!intercepts: a = feats*lambda'
  //vector<double> a;
  //a.reserve(oneSentence.m_numCands);
  
  vector<double> a;
  FillIntercepts(oneSentence, dim, a);

  /*
  cout<<"initialValueCopy: ";
  copy(initialValueCopy.begin(), initialValueCopy.end(), ostream_iterator<double>(cout, " "));
  cout<<endl;
  */

  /*
  cout<<"m_leftRange: ";
  copy(m_leftRange.begin(), m_leftRange.end(), ostream_iterator<double>(cout, " "));
  cout<<endl;
  */

  /*
  for( vector<Features*>::iterator iter=feats.begin(); iter!= feats.end(); ++iter){
    Features* f = *iter;
    //cout<<"Addresssing: ";
    //copy(f->begin(), f->end(), ostream_iterator<double>(cout, " "));
    //cout<<endl;

    a.push_back(calc_innerProduct(f, initialValueCopy));
  }
  */
  //transform(feats.begin(), feats.end(), a.begin(), bind2nd(ptr_fun(calc_innerProduct), initialValueCopy));
  //cout<<"3"<<endl;

  /*
  cout<<"a: ";
  copy(a.begin(), a.end(), ostream_iterator<double>(cout, " "));
  cout<<endl;
  */
  //!slopes
  /*
  vector<double> slopes;
  slopes.reserve(oneSentence.m_numCands);

  vector< IndexedValue > slopesWithIndices;
  slopesWithIndices.reserve(oneSentence.m_numCands);
  int candNum=0;
  
  for(vector<Features*>::iterator iter=feats.begin(); iter!=feats.end(); ++iter){
    Features* f = *iter;
    slopes.push_back((*f)[dim]);
    slopesWithIndices.push_back(IndexedValue((*f)[dim], candNum));
    candNum++;
  }
  sort(slopesWithIndices.begin(), slopesWithIndices.end(), SortIndexedValueDecreasing());

  //assertion to check if pre-computation code is correct
  assert(oneSentence.m_decreasingSlopes[dim]==slopesWithIndices);
  */

  //cout<<"a: "; copy(a.begin(), a.end(), ostream_iterator<double>(cout, " ")); cout<<endl;
  //cout<<"a1: "; copy(a1.begin(), a1.end(), ostream_iterator<double>(cout, " ")); cout<<endl;

  /*
  assert(a1.size()==a.size());
  for (int c = 0; c < a.size(); c++){
    assert(fabs(a[c] - a1[c])<0.0001);
  }
  */

  vector< IndexedValue > slopesWithIndices = oneSentence.m_decreasingSlopes[dim];

  //cout<<"4"<<endl;
  /*
  cout<<"slopesWith: ";
  copy(slopesWithIndices.begin(), slopesWithIndices.end(), ostream_iterator<double>(cout, " || "));
  for(vector< IndexedValue >::iterator iter=slopesWithIndices.begin(); iter!=slopesWithIndices.begin()+10; ++iter){
    cout<<(*iter).first<<" "<<(*iter).second<<" || ";
  }
  cout<<endl;
  */
  
  //cout<<"leftRange: "<<m_leftRange[dim]<<" ==================== "<<endl;
  //pair<int, double> minLeftCandidateData = getBoundsCandidates(oneSentence, a, slopes, m_leftRange[dim], dim);
  pair<int, double> minLeftCandidateData = getBoundsCandidates(oneSentence, a, m_leftRange[dim], dim);

  //cout<<"rightRange: "<<m_rightRange[dim]<<" ==================== "<<endl;
  //pair<int, double>  minRightCandidateData = getBoundsCandidates(oneSentence, a, slopes, m_rightRange[dim], dim);
  pair<int, double>  minRightCandidateData = getBoundsCandidates(oneSentence, a, m_rightRange[dim], dim);

  
  //if (minLeftCandidateData.first!=minRightCandidateData.first){
  //cout<<" Read ----------------------------------------------------- "<<endl;
  //}
  
  /*
  cout<<"5: "<<" minLeftCandidateData: "<<minLeftCandidateData.first<<" "<<minLeftCandidateData.second<<" slope: "<<slopes[minLeftCandidateData.first]<<endl;
  cout<<"5: "<<" minRightCandidateData: "<<minRightCandidateData.first<<" "<<minRightCandidateData.second<<" slope: "<<slopes[minRightCandidateData.first]<<endl;
  */

  RawCounts workingError = *(scores[minLeftCandidateData.first]);
  vector<ErrorAt> errorDeltas;
  errorDeltas.push_back(ErrorAt(m_leftRange[dim], workingError));
  //FindGreaterEqAndLess gteLess(minLeftCandidateData.second,  minRightCandidateData.second);
  //vector<IndexedValue>::iterator beginOfRelevant = find_if(slopesWithIndices.begin(), slopesWithIndices.end(), ());
  //while we still find such points

  int numRelevant;

  //assert(slopes[minLeftCandidateData.first]==GetSlope(oneSentence, minLeftCandidateData.first, dim));
  vector<IndexedValue>::iterator relStart = findGeqLt(slopesWithIndices.begin(), slopesWithIndices.end(), GetSlope(oneSentence, minLeftCandidateData.first, dim));
  //vector<IndexedValue>::iterator relStart = findGeqLt(slopesWithIndices.begin(), slopesWithIndices.end(), slopes[minLeftCandidateData.first]);
  int nRel = slopesWithIndices.end() - relStart;

  //cout<<"dim="<<dim<<" nRel "<<nRel<<endl;
  
  /*
  if (nRel>0){
    cout<<"slope next="<<(*relStart).first<<endl;
    assert((*relStart).first<slopes[minLeftCandidateData.first]);
    //if (nRel==5){
      for (vector<IndexedValue>::iterator tmp = relStart; tmp<slopesWithIndices.end(); ++tmp){
	cout<<"Cand: "<<(*tmp).second<<" slope: "<<(*tmp).first<<" slopeLookup:"<<slopes[(*tmp).second]<<endl;
      }
      
      //}
  }
  */
  
  
  //cout<<"got back safe:"<<numRelevant<<endl;


  //start it out with the left most candidate
  int nextCandidate = minLeftCandidateData.first;
  while (relStart!=slopesWithIndices.end()){
    //pt contains the next intersection point and the candidate that we switch over to
    //and re-run the relevancy check, only on the remaining range
    //cout<<"Running find ================ "<<endl;

    //assert( slopes[nextCandidate] == GetSlope(oneSentence, nextCandidate, dim));
    //pair<double, int> pt = findIntersection(dim, nextCandidate, slopes[nextCandidate], a, relStart, slopesWithIndices.end());
    pair<double, int> pt = findIntersection(dim, nextCandidate, GetSlope(oneSentence, nextCandidate, dim), a, relStart, slopesWithIndices.end());
    //if you came out just intersecting your self, just quit - there are no more intersections within the range
    if (pt.second==nextCandidate){
      break;
    }

    //if you found a value that was past the right range, then quit... you are already on the right most candidate
    if (pt.first > m_rightRange[dim]){
      assert(pt.second == minRightCandidateData.first);
      break;
    }
    //debuggin in case of crash here
    if (pt.first <= m_leftRange[dim]){
      cout<<"CRASH =========="<<endl;
      //cout<<"Slopes:"; copy(slopes.begin(), slopes.end(), ostream_iterator<double>(cout, " "));
      cout<<"a:"; copy(a.begin(), a.end(), ostream_iterator<double>(cout, " "));

      cout<<"minLeftCandidateData.first: "<<minLeftCandidateData.first;
      cout<<"pt.first:"<<pt.first<<" pt.second: "<<pt.second<<endl;
      cout<<"nextCandidate: "<<nextCandidate;
    }

    assert(pt.first>m_leftRange[dim]);
    assert(pt.first<=m_rightRange[dim]);

    nextCandidate = pt.second;
    //cout<<"Intersection nextCandidate:"<<nextCandidate<<endl;
    RawCounts newErrorDelta = *(scores[nextCandidate]) - workingError;
    errorDeltas.push_back(ErrorAt(pt.first, newErrorDelta));
    workingError = workingError + newErrorDelta;

    //and re-run the relevancy check, only on the remaining range
    //cout<<"Checking again ================ "<<endl;
    //assert( slopes[nextCandidate] == GetSlope(oneSentence, nextCandidate, dim));
    //relStart = findGeqLt(relStart, slopesWithIndices.end(),slopes[nextCandidate]);
    relStart = findGeqLt(relStart, slopesWithIndices.end(),GetSlope(oneSentence, nextCandidate, dim));

    //cout<<"got back safe:"<<numRelevant<<endl;
    nRel = slopesWithIndices.end()-relStart;
    //cout<<"dim="<<dim<<" nRel "<<nRel<<endl;
    
  } 
  //cout<<"dim="<<dim<<" end"<<endl;
  //cout<<"returning errorDeltas"<<endl;
  return errorDeltas;
}

RawCounts* COptimizeMER::GetScore(const Instance& oneSentence, int candidate){
  vector<RawCounts*>::const_iterator indexIt = oneSentence.m_scoreRange.first+candidate;
  assert(indexIt< oneSentence.m_scoreRange.second);
  return *indexIt;
}

double COptimizeMER::GetSlope(const Instance& oneSentence, int candidate, int dim){
  //assertion to verify that we dont need the slopes variable
  vector<Features*>::const_iterator indexIt = oneSentence.m_dataRange.first + candidate;
  assert(indexIt< oneSentence.m_dataRange.second);
  Features* pCandFeatures = *indexIt;
  assert(pCandFeatures->size()>dim);
  double slopeValue = (*pCandFeatures)[dim];
  return slopeValue;
}

//!find first value, greaterEq min
vector<IndexedValue>::iterator COptimizeMER::findGeqLt(vector<IndexedValue>::iterator begin, vector<IndexedValue>::iterator end, double minValue){
  bool foundMax = false;
  bool foundMin = false;
  vector<IndexedValue>::iterator it = begin;
  vector<IndexedValue>::iterator newBegin = begin;

  ++it;
  while (it!=end && (*it).first>=minValue){
    ++it;
  }
  if (it!=end){
    assert((*it).first<minValue);
  }
  return it;
  /*
  //check it the search is even neccesary
  //cout<<"Beginning: "<<minValue<<" --> "<<maxValue<<endl;
  if (minValue==maxValue){
    return pair<vector<IndexedValue>::iterator, vector<IndexedValue>::iterator>(newEnd, newEnd);
  }
 
  while (it!=end){
    //cout<<"Checking: "<<(*it).first<<" >= "<<minValue<<" and < "<<maxValue<<endl;
    if (!foundMax){
      if ((*it).first<maxValue){
	//cout<<"BEGFOUND: "<<(*it).first<< " < "<<maxValue<<endl;
	foundMax=true;
	newBegin = it;
      }//else dont worry about it
    }

    //past the boundary, but you have already found the first
    if (foundMax && !foundMin){
      if ((*it).first<minValue){
	foundMin=true;
	newEnd=it;
      }
    }

    if (foundMin && foundMax){
      //cout<<"BREAK OUT"<<endl;
      break;
    }

    if (foundMax){
      numRelevant++;
    }
    //cout<<"In loop"<<endl;
    ++it;
    
  }
  //cout<<"returning: "<<endl;
  return pair<vector<IndexedValue>::iterator, vector<IndexedValue>::iterator>(newBegin, newEnd);
  */
}

//search between start and end
pair<double, int> COptimizeMER::findIntersection(int dim, int minLeftCandidate, double slope_minLeftCandidate, vector<double>& a, vector< IndexedValue >::iterator start, vector< IndexedValue >::iterator end){

  //the final slope is the one one the far right
  double bestL = m_rightRange[dim];
  double bestSlope = slope_minLeftCandidate;
  int newMinCandidate = minLeftCandidate; //index of the current working value
  
  vector< IndexedValue >::iterator startReal = start;
  int wasted = 0;
  while (start!=end){
    //cout<<"in here"<<endl;
    IndexedValue candData = *start;
    //slope has to be less
    assert(candData.first<slope_minLeftCandidate);
    
    int candidateIndex = candData.second;
    double top = a[candidateIndex]-a[minLeftCandidate];
    double bottom = slope_minLeftCandidate - candData.first;
    if (bottom!=0.0){
      double newL = top/bottom;
      if (newL<bestL){
	bestL = newL;
	bestSlope = candData.first; //!slope of the found candidate
	newMinCandidate = candidateIndex;
	//cout<<"found at: "<<start-startReal<<endl;
	//cout<<"setting: "<<bestL<<" "<<bestSlope<<" "<<newMinCandidate<<endl;
      }else if ((newL==bestL) && (candData.first < bestSlope)){
	//cout<<"multi triggerred"<<endl;
	bestL = newL;
	bestSlope = candData.first;
	newMinCandidate = candidateIndex;
	//cout<<"setting: "<<bestL<<" "<<bestSlope<<" "<<newMinCandidate<<endl;
      }
    }else{
      cout<<"identical slopes"<<endl;
    }
    ++start;
    wasted++;
  }
  //cout<<"wasted: "<<wasted<<endl;
  return pair<double, int>(bestL, newMinCandidate);
}



//pair<int, double> COptimizeMER::getBoundsCandidates(const Instance& oneSentence, vector<double>& a, vector<double>& slopes, double scalingValue, int dim){
pair<int, double> COptimizeMER::getBoundsCandidates(const Instance& oneSentence, vector<double>& a, double scalingValue, int dim){

  double lowestCostSlope = COptimizeMER::LargeCost;
  double lowestCost = COptimizeMER::LargeCost;
  int lowestCostIndex = 0;
  //for (int i=0; i<slopes.size(); i++){
  for (int i=0; i<a.size(); i++){
    //lowest_cost.push_back((slopes[i]*scalingValue)+a[i]);
    //assert(slopes[i]  == GetSlope(oneSentence, i, dim));
    //double cost = (slopes[i]*scalingValue)+a[i];
    double slopei = GetSlope(oneSentence, i, dim);
    double cost = (slopei*scalingValue)+a[i];
    if (cost<lowestCost){
      lowestCost = cost;
      lowestCostIndex = i;
      //lowestCostSlope = slopes[i];
      lowestCostSlope = slopei;
      //cout<<"Updating lowest: "<<lowestCostIndex<<" lowestCost:"<<lowestCost<<endl;
    }else if (cost==lowestCost){
      //if (slopes[i]<lowestCostSlope){
      if (slopei<lowestCostSlope){
	lowestCostIndex = i;
	//lowestCostSlope = slopes[i];
	lowestCostSlope = slopei;
      }else{
	//this would mean that 2 cands had the same slope + same intercept ie the same
	//if the slopes were all the same anyway dont complain
      }
    }
  }

  /*
  cout<<"lowest_cost:";
  copy(lowest_cost.begin(), lowest_cost.end(), ostream_iterator<double>(cout, " "));
  cout<<endl;
  */
  //select the min with the lowest slope

  
  //pair<int, double> leftMinCandidateData = minValueAndIndex(lowest_cost, COptimizeMER::LargeCost);
  return pair<int,double>(lowestCostIndex, lowestCost);
}
       



//!Assigns iterators for easy, no copy access to the feature and scoring data, with these iterators
/*!
 * With these iterators set we can just call functions on the iterators
 */
void COptimizeMER::setupSentenceRanges(vector<Features*>::const_iterator iter_cost, vector<RawCounts*>::const_iterator iter_score, vector<Features*>::const_iterator end_cost, vector<RawCounts*>::const_iterator end_score, const vector<int>& startEnd, vector<Instance>& allSentences, int numDim ){
  //vector<Features*>::iterator iter_score = scores.begin();
  //vector<RawCounts*>::iterator iter_cost =costs.begin();
  for (int i=0; i<startEnd.size(); i++){
    Instance oneSentence;
    cout<<"reserving: for sentence: "<<i<<endl;
    //move numCands steps forward
    int numCands = startEnd[i];
    assert(numCands!=0);

    oneSentence.m_dataRange.first=iter_cost;
    oneSentence.m_scoreRange.first=iter_score;

    //reserve space for each dimension
    for (int d=0; d<numDim; d++){
      oneSentence.m_decreasingSlopes.push_back(vector<IndexedValue>());
    }
    
    //go thru each candidate in this sentence
    for (int index=0; index<numCands; index++){
      Features* pOneCandidateFeatures = *(iter_cost+index);
      for (int d=0; d<pOneCandidateFeatures->size(); d++){
	vector<IndexedValue>& oneDimVector = oneSentence.m_decreasingSlopes[d];
	IndexedValue toAdd((*pOneCandidateFeatures)[d], index);
	oneDimVector.push_back(toAdd); //register the slope, with candidate number
      }
    }

    //now sort each dimension to get decreasing slopes
    //reserve space for each dimension
    for (int d=0; d<numDim; d++){
      sort(oneSentence.m_decreasingSlopes[d].begin(), oneSentence.m_decreasingSlopes[d].end(), SortIndexedValueDecreasing());
    }

    oneSentence.m_dataRange.second = iter_cost + numCands;
    oneSentence.m_scoreRange.second = iter_score + numCands;

    //move iter_cost and iter_score forward
    iter_cost = oneSentence.m_dataRange.second;
    iter_score = oneSentence.m_scoreRange.second;


    oneSentence.m_numCands = numCands;

    allSentences.push_back(oneSentence);
  }
  cout<<"loop done\n";

  assert(iter_cost==end_cost);
  assert(iter_score==end_score);
  cout<<"Setup done optimization beginning"<<endl;
}
