#include <TERSearch.hh>


TERSearch::TERSearch(const char* referenceListFile, redi::pstream& pbuf, bool normalizeReference, const int numReferences, CVocabulary* pVoc){
   m_pVoc = pVoc;
   BuildReferences(referenceListFile, pbuf, normalizeReference,  numReferences, pVoc);
   m_nGramConsidered = 1; //since its just number of edits that we count
   m_numSentences = m_referenceBank.size();
   //RawCounts r; //just to make sure it got initialized
   //m_raw = r;

   m_maxScore = 0.0; //since this is a metric we are looking to minimize
   m_minEpsilon = 0.0001; 
}

TERSearch::TERSearch(){
  m_nGramConsidered = 1;
  m_maxScore = 0.0; //since this is a metric we are looking to minimize
  m_minEpsilon = 0.0001;   
}

void TERSearch::ScoreSingleHyp(const int sentenceNumber, string candidate, RawCounts& countObject){
  //initialize the count object
  //there is only 1 count: num edits
  for (int n=0; n<1; n++){
    countObject.numCorrect[n]= 0;
    countObject.numSuggested[n]= 0;
  }
  countObject.referenceLengths = 0;

  // build up the source sentence from the raw string
  vector<CWordIndex> hypSentence = stringToVocabVector(candidate);

  vector< vector<CWordIndex> > references = GetAllReferences(sentenceNumber);
  CEvaluatorTER ter;
  vector<float> Scores;
  ter.Evaluate(sentenceNumber, hypSentence, references, Scores);
  countObject.numCorrect[0] = Scores[ter.m_EDIT_COUNT_KEY];
  countObject.numSuggested[0] = 0;
  countObject.referenceLengths = Scores[ter.m_REF_WORD_COUNT_KEY];
  
}

Score TERSearch::getMetricScore(RawCounts& raw){
  return -1*((double)raw.numCorrect[0]/(double)raw.referenceLengths);
}

AbstractScore TERSearch::getMetricScoreDetailed(RawCounts& raw){
  AbstractScore a; 
  a.push_back(getMetricScore(raw));
  a.push_back(raw.numCorrect[0]);
  a.push_back(raw.referenceLengths);
  return a;
}


