#include <CachedSriLM.hh>


//bool CacheLT::operator()(const vector<CWordIndex>& l1, const vector<CWordIndex>& l2 ) const{
//  if (l1<l2){
//    return true;
//  }else{
//    return false;
//  }
//}

CachedSriLM::CachedSriLM(const string& paramFile, CLMType lmType) :
  CLanguageModel(lmType)
{
  //  CachedSriLM::s_debugLM = false;
  
  vocab = new Vocab;
  
  // ngram order (Markov order + 1)
  unsigned order = CChartParserInfo::C_SRI_HistoryLength+1;
  ngramLM = new Ngram(*vocab, order);
  File file(paramFile.c_str(), "r");
  
  if (!ngramLM->read(file)) {
    cerr << "Error opening and reading from LM file " << paramFile << endl;
    exit(1);
  }

  //new the context
  m_context = new VocabIndex[CChartParserInfo::C_SRI_HistoryLength+1];
}

CachedSriLM::~CachedSriLM(){
  delete[] m_context;
  delete ngramLM;
  delete vocab;
}


CNegLogProb CachedSriLM::GetSingleWordLMCost(const CWordSequence& lmString, int wordPos){

  //LM DEBUG HERE
#ifndef NDEBUG
  bool s_debug = false;
  if (s_debug){
    cout<<"wordPos:" << wordPos << ", whole Context: ";
    CChartParserInfo::PrintVocabSequence(lmString);
  }
#endif

  assert(wordPos < lmString.size());
  CWordIndex wi = lmString[wordPos];
  if (
      // the following is to exclude LM cost calculation for <s>
      (wi == CChartParserInfo::m_bosWI && m_lmType==FORWARDLM) ||
      // the following is to exclude LM cost calculation for </s> in case of a backward LM
      (wi == CChartParserInfo::m_eosWI && m_lmType==BACKWARDLM) ||
      // the following is to exclude LM cost calculation of </s> after </s> or <s> before <s>
      // (happens since each target sentence starts with a sequence of CChartParserInfo::C_HistoryLength <s> symbols 
      // and ends with a sequence of C_HistoryLength </s> symbols
      // and we only want to calculate EOS for the first of these </s> given the words before or (in case of 
      // BACKWARDLM) BOS for the last of these <s> given the words after)
      (wi == CChartParserInfo::m_eosWI && wordPos-1 >= 0 && lmString[wordPos-1] == CChartParserInfo::m_eosWI) ||
      (wi == CChartParserInfo::m_bosWI && wordPos+1 < lmString.size() && lmString[wordPos+1] == CChartParserInfo::m_bosWI)
      )
    return 0.0;

  // if we're in BACKWARDLM mode, we have to switch <s> and </s> for
  // the SRI calls
  if (m_lmType==BACKWARDLM && wi==CChartParserInfo::m_bosWI) {
    wi = CChartParserInfo::m_eosWI;
  } else if (m_lmType==BACKWARDLM && wi==CChartParserInfo::m_eosWI) {
    wi = CChartParserInfo::m_bosWI;
  }
  // representation of wi|h_1,...,h_C_SRI_HistoryLength in the SRI format (wi,h_C_SRI_HistoryLength,...,h_1)
  // first insert the word whose LM prob should be computed
  vector<CWordIndex> context(1, wi);
  // then insert up to C_SRI_HistoryLength words of refContext in reverse order
  if (m_lmType==FORWARDLM) {
    for (int i=1; i <= CChartParserInfo::C_SRI_HistoryLength && wordPos-i >= 0 && lmString[wordPos-i] != CChartParserInfo::m_lmBoundaryWI; i++) 
      context.push_back(lmString[wordPos-i]);
  } else {
    for (int i=1; i <= CChartParserInfo::C_SRI_HistoryLength && wordPos+i < lmString.size() && lmString[wordPos+i] != CChartParserInfo::m_lmBoundaryWI; i++) {
      CWordIndex w = lmString[wordPos+i];
      // since  we're in BACKWARDLM mode, we have to switch <s> and </s> for
      // the SRI calls
      if (m_lmType==BACKWARDLM && w==CChartParserInfo::m_bosWI) w = CChartParserInfo::m_eosWI;
      if (m_lmType==BACKWARDLM && w==CChartParserInfo::m_eosWI) w = CChartParserInfo::m_bosWI;
      context.push_back(w);
    }
  }

  //negative to make the log(prob) into a cost
  CNegLogProb negLogProb = -(LogProb(context));

#ifndef NDEBUG
    if (s_debug){
      cout<<" LMCost:"<<negLogProb<<endl;
    }
#endif

    return negLogProb;
}


