#include <LanguageModelSAMT.hh>

bool CLanguageModel::s_debugLM = false;
vector<CLanguageModel*> CLanguageModel::m_languageModels;
vector<CDatabaseSAMTLM*> CDatabaseSAMTLM::m_samtlmDatabases;


CDatabaseSAMTLM::CDatabaseSAMTLM(const string& dbFileName, bool storedAsBTree)
{
  m_db = new Db(NULL,0);
  m_db->set_cachesize(1,0,1); // 1 GB Cache (that's the first parameter) + 0 bytes (2nd parameter); using 1 cache (3rd param)
  //open the DB
  //u_int32_t oFlags = DB_CREATE | DB_EXCL; // Open flags;
  u_int32_t oFlags = DB_RDONLY; // Open flags;
  //try {
  // Redirect debugging information to std::cerr
  m_db->set_error_stream(&std::cerr);
  //  cout<<"Opening db==="<<dbFileName.c_str()<<"==="<<endl;
  // Open the database
  if (m_db->open(NULL, // Transaction pointer
	    dbFileName.c_str(), // Database file name
	    NULL, // Optional logical database name
	    storedAsBTree? DB_BTREE : DB_HASH, // Database access method
	    oFlags, // Open flags
		 0) > 0){ // File mode (using defaults)
    cerr<<"Database"<<dbFileName<<" could not be opened!!!"<<endl;
    cout<<"Database"<<dbFileName<<" could not be opened!!!"<<endl;
    m_db->close(0);
    exit(-1);
  }
}

CDatabaseSAMTLM::~CDatabaseSAMTLM() {
    // Close the database
    m_db->close(0);
    delete m_db;
}



//!Utilty method to see exactly what got put into the db
void CLanguageModelSAMT::PrintDB(){
  cerr << "Currently broken." << endl;
  exit(2);
  Dbc *cursorp;
  m_db->cursor(NULL, &cursorp, 0);
  
  Dbt key, data;
  int ret;
  
  // Iterate over the database, retrieving each record in turn.
  while ((ret = cursorp->get(&key, &data, DB_NEXT)) == 0) {
    // Do interesting things with the Dbts here.
    //cout<<"key=="<<*(double*)key.get_data()<<"=="<<endl;
    //cout<<"key (sz:"<<key.get_size()<<" wds: "<<key.get_size()/4<<" ) ";
    cout<<"key("<<key.get_size()/4<<")=";
    for (int i=0; i < key.get_size()/4; i++) {
      int32_t wd = ((int32_t*) key.get_data())[i];
      //cout << ((int32_t*) key.get_data())[i];
      //cout<<"int="<<wd;
      if (CChartParserInfo::m_pVocabulary->IsInRange(wd)){
	cout<<CChartParserInfo::m_pVocabulary->GetWord(wd)<<" ";
      }else{
	cout<<" NotInRange "<<" ";
      }
      //" InRange: "<<
    }
    cout<<endl;
    cout<<" value=="<<*( (CFrequency*) data.get_data())<<endl;
  }
  cout<<"Dump done\n";
  // Cursors must be closed
  if (cursorp != NULL) 
    cursorp->close(); 
}


//! Gets the LM prob of wi, based on context in english order: context[0],context[1], wi
/*!
  If context is longer than we need, use only the last C_HistoryLength words in the context

  In case context is empty, we back off to unigram.
  In case context only has one element, we back off to bigram
*/
CNegLogProb CLanguageModelSAMT::GetSingleWordLMCost(const CWordSequence& lmString, int wordPos){
  bool sanityCheckCache = false; // set this to true to check whether the cache optimization works correctly

  //LM DEBUG HERE
#ifndef NDEBUG
  bool s_debug = s_debugLM;
  if (s_debug) sanityCheckCache = true; // always do sanity cache checking when debugging
  if (s_debug){
    cout << "P_LM( w_" << wordPos << "=" << CChartParserInfo::m_pVocabulary->GetWord(lmString[wordPos])
	<< " | ";
    CChartParserInfo::PrintVocabSequence(lmString);
    cout << " ) = ";
  }
#endif

  assert(wordPos < lmString.size());
  CWordIndex wi = lmString[wordPos];
  if (
      // the following is to exclude LM cost calculation for <s>
      (wi == CChartParserInfo::m_bosWI && m_lmType==FORWARDLM) ||
      // the following is to exclude LM cost calculation for </s> in case of a backward LM
      (wi == CChartParserInfo::m_eosWI && m_lmType==BACKWARDLM) ||
      // the following is to exclude LM cost calculation of </s> after </s> or <s> before <s>
      // (happens since each target sentence starts with a sequence of C_HistoryLength <s> symbols 
      // and ends with a sequence of C_HistoryLength </s> symbols
      // and we only want to calculate EOS for the first of these </s> given the words before or (in case of 
      // BACKWARDLM) BOS for the last of these <s> given the words after)
      (wi == CChartParserInfo::m_eosWI && wordPos-1 >= 0 && lmString[wordPos-1] == CChartParserInfo::m_eosWI) ||
      (wi == CChartParserInfo::m_bosWI && wordPos+1 < lmString.size() && lmString[wordPos+1] == CChartParserInfo::m_bosWI)
      ) {
#ifndef NDEBUG
    if (s_debug) cout << "1" << endl;
#endif
    return 0.0;
  }


  //=====================================================================================================
  // the following is only an optimization: try to find the LM cost from the cache
  pair<CNegLogProb,int> cacheValue(0,SAMTLM_NOT_IN_CACHE);
  // start by looking up the unigram tuple from the cache, then keep expanding
  CWordSequence tuple(1,wi);
  for (int historyLength=0; historyLength <= m_maxMarkovOrder; historyLength++) {
    cacheValue = GetFromCache(tuple);
    if (cacheValue.second == SAMTLM_FINAL_NGRAM) { 
      // tuple was found and history is non-expandable - break loop and return cached value
      break;
    }
    // expand history-context and n-gram tuple
    if (m_lmType==FORWARDLM) {
      if (wordPos-historyLength-1 < 0) {
	// history can't be expanded further - exit the while loop
	break;
      } else {
	// Note: because of the front-insertion, you might be tempted to change tuple and history to be deques 
	// instead of vectors. DON'T DO THAT! A deque is not guaranteed to store its data consecutively,
	// but GetTupleFrequency() makes exactly that assumption for efficiency reasons!
	tuple.insert(tuple.begin(), lmString[wordPos-historyLength-1]);
      }
    } else {
      // BACKWARDLM
      if (wordPos+historyLength+1 >= lmString.size()) {
	// history can't be expanded further - exit the while loop
	break;
      } else {
	tuple.push_back(lmString[wordPos+historyLength+1]);
      }
    }    
  }
  if ( cacheValue.second != SAMTLM_NOT_IN_CACHE    // we got the cost - just return it, unless in sanity test mode
       && !sanityCheckCache ) 
    return cacheValue.first;


  //=================================================================================================
  // now the actual language model cost calculation starts
  // start with history length 0
  int historyLength = 0;
  // ... and the zero-gram tuple as history ...
  vector<CWordIndex> history;
  // .. and the unigram tuple (wi) ...
  tuple = vector<CWordIndex>(1,wi);
  // ... and the zero-gram's frequency count (i.e., the total number of word instances in the training corpus)
  CFrequency historyFrequency = m_totalEventCount;
  // ... zero-gram's unassigned probab mass
  CFrequency historyUnassignedFrequencyMass = m_unassignedUnigramFrequencyMass;
  // ... 
  CFrequency historyUniqueEventCount = m_underlyingVocabSize;
  // ... and a very low probability estimate (will kick in multiplied with m_minBackoffWeight in case the word is unknown)
  double prob = m_initialProb;
#ifndef NDEBUG
  if (s_debug) cout << prob << " " << endl;
#endif
  // In the following while loop, we keep expanding the LM tuple and its history and adjusting the LM probability
  // for loop's external break condition: history's frequency too small
  // for loop's internal break condition: history can't be expanded any further  or MaxMarkovOrder exceeded
  // The flag prematureExit will be set if we should have expanded further but couldn't b/c not enough 
  // history words were available in lmString. 
  bool prematureExit = false;
  while (historyFrequency >= m_minHistoryFrequency) {
    // get the frequency of the currently considered tuple
    CFrequency tupleFrequency;
    CFrequency dummyFreq; // just some dummy variables we don't care about    
    GetTupleStats(tuple, tupleFrequency, dummyFreq, dummyFreq);
    // history is a sub-ngram of tuple, so the frequency of history must always be greater or equal than the fr. of tuple
    assert(tupleFrequency <= historyFrequency);
    // now compute back-off weight, i.e., 1 - confidence weight in [0,1) for the history
    double backoffWeight = 
      MAX(m_minBackoffWeight, 
	  //	  1.0 / (1.0 + pow((double) historyFrequency / pow((double) historyUniqueEventCount, m_uniqueEventCountImpact) * (double) m_inverseBreakEvenHistoryFrequency, m_historyFrequencyConvergenceSpeed))
	  1.0 / (1.0 + 
		 pow(
		     log2(1.0 + 
			  (double) historyFrequency / 
			  pow((double) historyUniqueEventCount, m_uniqueEventCountImpact) *
			  (double) m_inverseBreakEvenHistoryFrequency
			  ), 
		     m_historyFrequencyConvergenceSpeed)
		 )
	  );
    double smoothingMass = 
      (tupleFrequency==0) ?
      // we encountered a zero-frequency event - smooth it
      (1.0 - backoffWeight) * (double) historyUnassignedFrequencyMass / (double) historyFrequency :
      // regular event - no unknown-smoothing
      0;
    assert(backoffWeight <= 1.0);
    assert(backoffWeight >= 0);
    // the new probability estimate is a weighted sum of the current Markov-order prob estimate and the current prob estimate obtained from
    // the previous (lower Markov-order) estimates, which now serves as a back-off estimate
    prob = (1.0-backoffWeight) * (double) tupleFrequency / (double) historyFrequency + (backoffWeight+smoothingMass) * prob;
#ifndef NDEBUG
    if (s_debug) cout << "   * (" 
		      << backoffWeight << "+" << smoothingMass << ")"
		      << " + "
		      << ((double) tupleFrequency / (double) historyFrequency) << " * "
		      << (1.0-backoffWeight)
		      << " [tupleFr=" << tupleFrequency << ",histFr=" << historyFrequency 
		      << ",histUniqEventCount=" << historyUniqueEventCount << "] "
		      << endl;
#endif
    assert(prob <= 1.0001);
    assert(prob > 0);
    if (historyLength == m_maxMarkovOrder) {
      // max. Markov order has been reached - break the loop, but don't count
      // this as 'prematurely'
      break;
    }
    // expand history-context and n-gram tuple
    if (m_lmType==FORWARDLM) {
      if (wordPos-historyLength-1 < 0) {
	// history can't be expanded further - exit the while loop prematurely
	prematureExit = true;
	break;
      } else {
	// Note: because of the front-insertion, you might be tempted to change tuple and history to be deques 
	// instead of vectors. DON'T DO THAT! A deque is not guarranteed to store its data consecutively,
	// but GetTupleFrequency() makes exactly that assumption for efficiency reasons!
	tuple.insert(tuple.begin(), lmString[wordPos-historyLength-1]);
	history.insert(history.begin(), lmString[wordPos-historyLength-1]);
      }
    } else {
      // BACKWARDLM
      if (wordPos+historyLength+1 >= lmString.size()) {
	// history can't be expanded further - exit the while loop prematurely
	prematureExit = true;
	break;
      } else {
	tuple.push_back(lmString[wordPos+historyLength+1]);
	history.push_back(lmString[wordPos+historyLength+1]);
      }
    }
    // history was expanded by one word
    historyLength++;
    // retrieve stats of new history
    GetTupleStats(history, historyFrequency, historyUnassignedFrequencyMass, historyUniqueEventCount);
    // Note that we will break automatically if expanding the history yielded the LM context 
    // separator <*> since in that case historyFrequency will be zero. 
    // The same will happen when encountering double <s> or double </s> sequences in the history.
    // Also, unknown-words should get reasonable (meaning: small but non-zero) estimates, depending on how frequent the context is.
    // Further, if an unknown word (or the _UNKNOWN symbol, which is not specifically recognized by the LM) occurs in the history,
    // its frequency will be zero and the expansion process stops, i.e., only contexts up to (but excluding) that unknown word
    // are considered.
  }

#ifndef NDEBUG
  if (s_debug){
    cout<<"  = "<<prob<<endl;
  }
#endif

  CNegLogProb negLogProb = -log10(prob);
  if (sanityCheckCache && cacheValue.second != SAMTLM_NOT_IN_CACHE) {
    if  (negLogProb != cacheValue.first) {
      cerr << "Cache Sanity check error: cacheValue.first=" << cacheValue.first << " != " << negLogProb << "=negLogProb!\n";
      exit(1);
    } 
  } else {
    StoreInCache(
		 pair<CNegLogProb,int>(negLogProb, prematureExit ? SAMTLM_NONFINAL_NGRAM : SAMTLM_FINAL_NGRAM),
		 tuple
		 );
  }
  return negLogProb;
}


