#ifndef BTree_hh
#define BTree_hh

//make sure you include the db_cxx for the c++ calls
#include <db_cxx.h>

#include <Vocabulary.hh>
#include <Rule.hh>
#include <ParserInfo.hh>

#include <vector>

//! result returned by a B-tree query
/*!
  the first component of the pair has its least-significant bit 1 iff there was an exact match
  (in that case the second component is an iterator over the exact-match results)
  and its second-least significant bit 1 iff there was a strict prefix match
*/
typedef std::pair< short,std::vector<CRule> > CBTreeResult;
//! define bit constants for the type of match in CBTreeResult.first
#define C_ExactMatch 1
#define C_StrictPrefixMatch 2


class CBTree {

public:
  //!database object, needs to be a pointer cause there is no default constructor
  Db* m_db;
  static CVocabulary* pVoc; //!static voc set at top level

  //!Initializes the DB and also loads in the _COUNT information
  CBTree(std::string dbFileName);
  ~CBTree();

  //!Utilty method to see exactly what got put into the db
  void PrintDB();

  //query the B-tree for prefix and return result as described above in definition of CBTreeResult
  CBTreeResult QueryBTree(const std::vector<CWordIndex>& prefix);

};

#endif

