#include <CombinationBeam.hh>

size_t CombinationTuplePrtHash::operator()(const CCombinationTuple* p) const{
  //use the address of the rule
  size_t hashCode = (long)(p->m_pRule) * 31 + long(p->m_pIncomplete);
  for (int i=0; i<p->m_derivationVector.size(); i++){
    hashCode = 31*hashCode + long(*(p->m_derivationVector[i]));
  }
  //+ integer offsets for each instatiated result
//   for (int i=0; i<p->m_productionBackPointers.size(); i++){
//     int offset = (p->m_derivationVector[i])-(p->m_productionBackPointers[i]->m_lowestToHighestCostFineGrained.begin());
//     hashCode = 31*hashCode + offset;
//   }

  // we must NOT consider m_pComplete as this pointer is deleted whenever hyps are added to the chart
  return hashCode;
}

bool CombinationTuplePrtEq::operator()(const CCombinationTuple* p1, const CCombinationTuple* p2) const{
  
  if (p1->m_pRule!=p2->m_pRule){
    return false;
  }
  
  if (p1->m_pIncomplete!=p2->m_pIncomplete){
    return false;
  }

  assert(p1->m_derivationVector.size()==p2->m_derivationVector.size()); // since the incomplete hyps are the same, the dimensions must match, too
  if (p1->m_derivationVector != p2->m_derivationVector) {
      return false;
  }
    
  // we must NOT consider m_pComplete as this pointer is deleted whenever hyps are added to the chart

  //  if (p1->m_productionBackPointers.size()!=p2->m_productionBackPointers.size()){
  //      return false;
  //  }
//   //+ integer offsets for each instatiated result
//   for (int i=0; i<p1->m_productionBackPointers.size(); i++){
//     int offset1 = p1->m_derivationVector[i]-p1->m_productionBackPointers[i]->m_lowestToHighestCostFineGrained.begin();
//     int offset2 = p2->m_derivationVector[i]-p2->m_productionBackPointers[i]->m_lowestToHighestCostFineGrained.begin();
//     if (offset1!=offset2){
//       return false;
//     }
//   }
  return true;
}

bool CombinationTupleMinCompare::operator()(const CCombinationTuple* p1, const CCombinationTuple* p2) const{
  return (p1->Cost() > p2->Cost());
}

CCombinationTuple::CCombinationTuple(CRule* pRule,  CChartParserHypIncomplete* pIncomplete, const vector< std::vector<CChartParserNode*>::iterator >& derivationVector ) : 
  m_pRule(pRule),
  m_pIncomplete(pIncomplete),
  m_pComplete(NULL),
  m_derivationVector(derivationVector) {
}


void CCombinationTuple::Complete() {
  // first determine the fine-grained EC backpointer vector expressed by m_derivationVector
  vector<CChartParserNode*> m_fineGrainedBackPointers;
  for (int i=0; i < m_derivationVector.size(); i++) {
    m_fineGrainedBackPointers.push_back(*(m_derivationVector[i]));
  }
  // now create a complete hyp and calculate its LM cost
  m_pComplete = new CChartParserHypComplete(m_pIncomplete, m_fineGrainedBackPointers, m_pRule);
}

CCombinationTuple::~CCombinationTuple(){
  // might have to remove m_pComplete, unless it has actually been filed somewhere (in which case the pointer will have been set to NULL)
  if (m_pComplete != NULL) {
             delete m_pComplete;
  }
}


/*
CCombinationTuple::CCombinationTuple(const CCombinationTuple& c){
  m_pRule = c.m_pRule;
  m_pIncomplete = c.m_pIncomplete;
  m_productionBackPointers = c.m_productionBackPointers;
  m_fineGrainedBackPointers = c.m_fineGrainedBackPointers;
  m_derivationVector = c.m_derivationVector;
  m_cost = c.m_cost;
  m_firstWords = c.m_firstWords;
  m_lastWords = c.m_lastWords;
}
*/

/*
CCombinationTuple& CCombinationTuple::operator = (const CCombinationTuple& r){
  if (this == &r){
    return *this;
  }
  m_pRule = r.m_pRule;
  m_pIncomplete = r.m_pIncomplete;
  m_productionBackPointers = r.m_productionBackPointers;
  m_fineGrainedBackPointers = r.m_fineGrainedBackPointers;
  m_derivationVector = r.m_derivationVector;
  m_cost = r.m_cost;
  m_firstWords = r.m_firstWords;
  m_lastWords = r.m_lastWords;
  return *this;
}
*/

void CCombinationTuple::Display(){
  cout<<"Tuple: ";
  m_pRule->Display();
  cout<<"COST: "<<Cost()<<" ";
  //production nodes

  
  for (int i=0; i<m_derivationVector.size(); i++){
    //display its size
    cout<<CChartParserInfo::m_pVocabulary->GetWord(m_pIncomplete->m_productionBackPointers[i]->GetResult())<<"-"<<m_pIncomplete->m_productionBackPointers[i]<<"("<<m_pIncomplete->m_productionBackPointers[i]->m_fineGrainedNodes.size()<<")-";
    //and which number down this tuple represents
    cout<<"[ ";
    cout<<m_derivationVector[i]-m_pIncomplete->m_productionBackPointers[i]->m_lowestToHighestCostFineGrained.begin()<<" ("<<*m_derivationVector[i]<<") = ";
    CChartParserNode* pPNode = *m_derivationVector[i];
    CChartParserInfo::PrintVocabSequence(pPNode->m_lmContext);
    cout<<" => bestRuleCost:"<<pPNode->m_bestTotalCost<<" + scaledFutureCost:"<<pPNode->m_scaledFutureCostEstimate
      <<" + boundaryLM:"<<pPNode->m_scaledBoundaryCost;
    cout<<" ] ";
  }
  cout<<endl;
}









CCombinationTupleBeam::CCombinationTupleBeam(){
  //m_k = k;
  //m_percDiff = percDiff;
  m_numTuplesProposed = 0; //no tuples initially
}


/*
CCombinationTuple* CCombinationTupleBeam::RemoveLast(){
  SortedCombinations::iterator iterLastElem = m_candidateSet.end();
  //if begin == end then the list is empty, dont do anything
  if (iterLastElem == m_candidateSet.begin()){
    return NULL;
  }
  iterLastElem--;
  CCombinationTuple* toRemove = (*iterLastElem).second;
  
  //make sure its actually there
  assert(m_uniqeTuples.count(toRemove)!=0);
  m_candidateSet.erase(iterLastElem);
	    
  //we also have to remove it from the unique set
  //cout<<" ******* deleting ********** "<<endl;
  int sizeBefore = m_uniqeTuples.size();
  UniqueCombinations::iterator iterUnique = m_uniqeTuples.find(toRemove);
  assert(iterUnique!=m_uniqeTuples.end());
  m_uniqeTuples.erase(iterUnique);
  int sizeAfter = m_uniqeTuples.size();
  assert(sizeBefore==sizeAfter+1);
  
  //now finally clear the memory
  //delete toRemove;
  return toRemove;
}
*/

//returns true if this element was added to the beam (always)
bool CCombinationTupleBeam::InsertAndCompleteBaseCombination(CCombinationTuple* toAdd){
  
#ifndef NDEBUG
  bool debug = false;
  if (debug){
    cout<<"InsertBaseCombination: Inserting: ";
    toAdd->Display();
  }
#endif

  m_numTuplesProposed++; //track each time we call InsertBaseCombination since this tuple must have been well formed

    // create complete hyp from the tuple and do LM cost calculation etc.
    toAdd->Complete();
    //insert into both the unique set and the candidate set
    m_uniqeTuples.insert(toAdd);
    //AS pq
    m_candidateSet.push(toAdd);
    //AS multimap
    //m_candidateSet.insert(std::pair<double, CCombinationTuple*>(toAdd->m_cost, toAdd));
    assert(m_uniqeTuples.count(toAdd)==1);
    
  return true;
}

//returns the next tuple from the m_finalNBest list - leave all the cutoff logic to the level that uses this CCombinationTupleBeam
CCombinationTuple* CCombinationTupleBeam::NextTuple(bool expandNeighbors){
  if (m_candidateSet.size()==0){
    return NULL;
  }
  CCombinationTuple* topElement = m_candidateSet.top();
  m_candidateSet.pop();

  if (expandNeighbors) {
    //we WILL return topElement, but we also put his neighbors on the m_candidateSet
    bool betteredBelow = ExpandNeighbors(topElement);

    //  for (vector<CCombinationTuple*>::iterator iter = frontier.begin(); iter!= frontier.end(); ++iter){
    //    //m_candidateSet.insert(std::pair<double, CCombinationTuple*>((*iter)->m_cost, *iter));
    //    InsertBaseCombination(*iter);
    //  }
  }

  return topElement;
}

bool CCombinationTupleBeam::ExpandNeighbors(CCombinationTuple* pTuple){

  bool didBetter = false;

  //az: disables as it seems more effective to increase ComboPruningBeamsize instead
  //  int depthDown = 1;
  //  if (CChartParserInfo::m_comboPruningUseLM){
  //    depthDown = 10; //make this a parameter later 
  //  }

  //for each dimension in |e| ie the set of tails of E we need to clone a new CCombinationTuple and increment it forward in one dimension
  for (int i=0; i< pTuple->m_derivationVector.size(); i++){ // iteration through the dimensions
    assert(pTuple->m_derivationVector[i] != pTuple->m_pIncomplete->m_productionBackPointers[i]->m_lowestToHighestCostFineGrained.end());

    //    for (int d=1; d<=depthDown; d++){
      //if there are more options in this dimension
      //this should always hold:      if (m_derivationVector[i] != m_pIncomplete->m_productionBackPointers[i]->m_lowestToHighestCostFineGrained.end()){

    if (pTuple->m_derivationVector[i]+1 != pTuple->m_pIncomplete->m_productionBackPointers[i]->m_lowestToHighestCostFineGrained.end()) {
      // there's still an alternative backpointed fine grained EC available

      //if you are using cmp=0 then there shouldn't be any neighbors to expand
      assert(CChartParserInfo::m_compareTargetWords==1);

      //now basically clone pTuple, but modify its derivationVector in dimension i to point to the next-best element there
      vector< std::vector<CChartParserNode*>::iterator > newDerivationVector = pTuple->m_derivationVector;
      newDerivationVector[i]++;

      CCombinationTuple* cCloned = new CCombinationTuple(pTuple->m_pRule, pTuple->m_pIncomplete, newDerivationVector);
	  
      //we dont want to do the cost calculations especially the LM calculations if this node is just going 
      //to be a duplicate. so we only complete and add the tuple if it doesn't exist yet
      if (m_uniqeTuples.count(cCloned)==0){
	InsertAndCompleteBaseCombination(cCloned);
 
	// just for diagnostics purposes
	if (pTuple->Cost() - cCloned->Cost() > 0.00001){
	  //cout<<"Bettered at depth: "<<d<<" m_cost: "<<m_cost<<" cCloned->m_cost: "<<cCloned->m_cost<<endl;
	  didBetter = true;
	}
      }else{
	//it's a duplicate... it was never added to any data structure
	delete cCloned;
      }
    }
  }

  return didBetter;
}



/*
SortedCombinations* CCombinationTupleBeam::FillNBestReturnPtr(){
  
  bool debug = false;
  
  if (debug){
    cout<<"Starting FillNBestReturnPtr: "<<endl;
    //Display();
  }

  //when using the language model, we will see reshuffling of this top beam
  //sometimes a hyp explored from lower down will rise up and become the top
  //hyp in the final n-best list - we are going to track these shuffles

  float lowestSavedCost, highestProb, cutoffLogProb; 
  int num_shufflesOccurred = 0;
  int num_betteredBelow = 0; //track how many times when expanding neighbors with the LM included you actually got a better score

  //while you havent got a full k yet, and you still have elements to explore
  //and the next best element is not WAY worse than the best element so far
  while (m_finalNBest.size()<m_k && m_candidateSet.size()>0){

    //AS pq
    CCombinationTuple* topElement = m_candidateSet.top();
    m_candidateSet.pop();

    //AS multimap
    //SortedCombinations::iterator topElementIter = m_candidateSet.begin();
    //CCombinationTuple* topElement = topElementIter->second;
    //remove the top element and add it to the m_finalNBest
    //m_candidateSet.erase(topElementIter);
    
    //if you are not using the LM costs, then this beam is proper order, ie 
    //all the hyps you explore further will actually have a lower cost
    //so we can break out if the hyp we add in this iteration is beam worse than the top hyp

    //if we ARE using the LM, this hyp and the ones that it adds on could very
    //easily become the NEW top hyp, and we owe it to its neighbors to expand them 
    //as well. The assumption is that if we have a new top hyp, then its neighbors 
    //can come in and fill the spots in between - we want to effectively reset the 
    //n-best list - and explore hyps that are neighbors to it
   
    if (debug){
      //cout<<"Top Element: "; topElement->Display();
    }


    if (m_finalNBest.size()==0){
      //this is the very first element being added
      lowestSavedCost = topElement->m_cost;

      //highestProb = exp(-lowestSavedCost);
      //cutoffLogProb = -log(highestProb - CChartParserInfo::m_comboPruningBeamPerc*highestProb);
      cutoffLogProb = CChartParserInfo::GetPruningCutoff(lowestSavedCost, CChartParserInfo::m_comboPruningFuzzCostDifference);
    }
    
    m_finalNBest.insert(std::pair<float, CCombinationTuple*>(topElement->m_cost, topElement));

    //do this after insertion
    //if the element that we are about to throw onto the final beam is that bad, break out
    if (topElement->m_cost>cutoffLogProb){
      if (debug){
	cout<<"Breaking out at size: "<<m_finalNBest.size()<<" / "<<m_k<<endl;
      }
      break;
    }

    //if the m_finalNBest has multiple elements in it
    if (m_finalNBest.size()>1 && lowestSavedCost-m_finalNBest.begin()->second->m_cost>0.001){
      if (debug){
	cout<<"NewTop: "; topElement->Display(); cout<<" OldTop: "<<lowestSavedCost<<endl;
      }
      num_shufflesOccurred++;
      lowestSavedCost =  m_finalNBest.begin()->second->m_cost;
      
      //highestProb = exp(-lowestSavedCost);
      //cutoffLogProb = -log(highestProb - CChartParserInfo::m_comboPruningBeamPerc*highestProb);
      cutoffLogProb = CChartParserInfo::GetPruningCutoff(lowestSavedCost, CChartParserInfo::m_comboPruningFuzzCostDifference);
    }
    
    //now put on all its frontier elements, track if a lower element was bettered
    bool betteredBelow = false;
    vector<CCombinationTuple*> frontier = topElement->ExpandNeighbors(m_uniqeTuples,betteredBelow);
    for (vector<CCombinationTuple*>::iterator iter = frontier.begin(); iter!= frontier.end(); ++iter){
      //m_candidateSet.insert(std::pair<double, CCombinationTuple*>((*iter)->m_cost, *iter));
      InsertBaseCombination(*iter);
    }
    if (betteredBelow){
      num_betteredBelow++;
    }
    
    //if (debug){
    //cout<<"AfterExpansion: "<<endl; Display();
    //cout<<" ============================== "<<endl;
    //}
    
    
  }

  //if we arent using the LM - then each successive hyp explored will be worse than 
  //the previous one, and the top hyp will never be shuffled out
  if (!CChartParserInfo::m_comboPruningUseLM){
    assert(num_shufflesOccurred==0);
      assert(num_betteredBelow==0);
  }

  if (CChartParserInfo::m_verbosity>0){
    cout<<"Log: num_shufflesOccurred: "<<num_shufflesOccurred<<" num_betteredBelow: "<<num_betteredBelow<<endl; 
  }

  //if (debug){
  //cout<<"*** FinalNBest *** "<<endl; Display();
  //  
  //}
  return &m_finalNBest;
}
*/

void CCombinationTupleBeam::Display(){
  cout<<"Candidate Set PQ ================: "<<m_candidateSet.size()<<endl;
  
  /*
  for (SortedCombinations::iterator tmpIter = m_candidateSet.begin(); tmpIter!=m_candidateSet.end(); ++tmpIter){
    (*tmpIter).second->Display();
    cout<<endl;
  }
  */
  //  cout<<"NBest Set ================"<<endl;
  //  for (SortedCombinations::iterator tmpIter = m_finalNBest.begin(); tmpIter!=m_finalNBest.end(); ++tmpIter){
  //    (*tmpIter).second->Display();
  //    cout<<endl;
  ///  }

  cout<<"Unique Set ================"<<endl;
  for (UniqueCombinations::iterator tmpIter = m_uniqeTuples.begin(); tmpIter!=m_uniqeTuples.end(); ++tmpIter){

    CCombinationTuple* ptr = (*tmpIter);
    cout<<"Printing: "<<ptr<<" "; (*tmpIter)->Display();
    cout<<endl;
  }
  
}

/*
void CCombinationTupleBeam::ConvertSortedToUnique(SortedCombinations* pFrom, UniqueCombinations* pTo){
  for (SortedCombinations::iterator tmpIter = pFrom->begin(); tmpIter!=pFrom->end(); ++tmpIter){
    //(*tmpIter).second->Display();
    CCombinationTuple* toAdd = (*tmpIter).second;
    pTo->insert(toAdd);
  }
}
*/

//void CCombinationTupleBeam::Clear(){
CCombinationTupleBeam::~CCombinationTupleBeam(){
  // remove all the m_uniqeTuples
  // this will also call the destructors of the tuples, deleting the tuples' complete hyps which didn't get added to the chart

  //cout<<"Unique Set ================"<<endl;

  // we can't delete the pointers in the hash table directly since for some weird reason iterating through
  // the hash table calls the equals function, which dereferences the pointers, causing a crash when the
  // corresponding objects don't exist anymore
  vector<CCombinationTuple*> del;
  for (UniqueCombinations::iterator tmpIter = m_uniqeTuples.begin(); tmpIter!=m_uniqeTuples.end(); ++tmpIter){

    CCombinationTuple* ptr = (*tmpIter);
    //cout<<"Deleting: "<<ptr<<" "; (*tmpIter)->Display();
    //cout<<endl;
    del.push_back(ptr);
    //delete ptr;
  }
    for(vector<CCombinationTuple*>::iterator tmpIter = del.begin(); tmpIter != del.end(); ++tmpIter){
       CCombinationTuple* ptr = (*tmpIter);
       delete ptr;
    }

  //  m_uniqeTuples.clear();
  //m_finalNBest.clear();
  //m_candidateSet.clear();
}

