#include <FastLatticeChartParser.hh>

using namespace std;


void CChartParser::DisplayCompletedHypsInCell(int i, int j){

  CCompleteHypChartCell& chartCell = m_completeHypsChart[i][j];
  for (CCoarseGrainedEq::const_iterator iter = chartCell.m_productionNodes.begin(); 
       iter!= chartCell.m_productionNodes.end();
       ++iter
       ){
    CProductionNode* node = (*iter).second;
    int initialSize = node->m_fineGrainedNodes.size();
    cout<<"Production Result === "<<CChartParserInfo::m_pVocabulary->GetWord(node->GetResult())<<endl;
    //now the equivalence classees
    for (CChartParserNodeContainer::iterator fiter = node->m_fineGrainedNodes.begin(); fiter!= node->m_fineGrainedNodes.end(); ++fiter){
      CChartParserNode* pChartNode = (*fiter).second;
      cout<<"\t"; 
      pChartNode->Display(); 
      cout<<endl;

      //now display all the hyps
      for (CSortedHypContainer::iterator iterHyp = pChartNode->m_hypContainer.begin(); iterHyp!=pChartNode->m_hypContainer.end(); ++iterHyp){
	cout<<"\t\t";
	CChartParserHypComplete* pTopHyp = (*iterHyp).second;
	pTopHyp->Display();
      }


      cout<<endl;
    }
  }
}

void CChartParser::DisplayIncompletedHypsInCell(int sourceCellRow, int sourceCellColumn){

  // the incomplete hyps in the source cell
  CSortedIncompleteHypContainer& incompleteHyps = m_incompleteHypsChart[sourceCellRow][sourceCellColumn];
  int totalIncompleteHypCount = incompleteHyps.size();

  cout <<" totalIncompleteHypCount: "<<totalIncompleteHypCount<<endl;

  //iterate through the incompleted hyps of current IC chart cell
  //(sourceCellRow,sourceCellColumn)
  bool continueIncompleteHypIteration = true;
  int iterationCountIH = 0;
  for (CSortedIncompleteHypContainer::const_iterator iterHyps =  incompleteHyps.begin(); continueIncompleteHypIteration != 0 && iterHyps != incompleteHyps.end(); ++iterHyps, ++iterationCountIH){

    float ruleAppCountIH = (*iterHyps).first;
	      
    CChartParserHypIncomplete* pIncompleteHyp = (*iterHyps).second; // the current incomplete hyp
    pIncompleteHyp->Display();
    //      cout << "        (rule-app-count according to CSortedIncompleteHypContainer: " << ruleAppCountIH << endl;
    cout << endl;
    assert(ruleAppCountIH == pIncompleteHyp->m_totalRuleApplicationCount);
  }
}


void CChartParser::DisplayCell(int i, int j){

  cout<<" Cell i : "<<i<<" j : "<<j<<" ================"<<endl;
  cout<<"   Completed hyps: ========="<<endl;
  DisplayCompletedHypsInCell(i, j);
  cout<<"   Incompleted hyps: ======="<<endl;
  DisplayIncompletedHypsInCell(i, j);
  cout<<" === End of Cell: "<<i<<" "<<j<<endl;

}

//go thru the complete and incomplete hyps and give a full row wise dump of the chart at this point
void CChartParser::Display(){
  for (int i=1; i<=CChartParserInfo::m_sourceLength; i++){
    for (int j=1; j<=CChartParserInfo::m_sourceLength-i+1; j++){
      DisplayCell(i,j);
    }
  }
}


void CChartParser::InitSearch(CWordSequence& pSentence){

  // reset the combination counter
  m_combinationCount = 0;

  assert(m_newlyCreatedNodes.empty());



  //step 1
  //fill in the completed hyps table with all the words in the sentence, this will involve 
  //creating new completed CChartParserHyp elements for each word in the sentence

  CChartParserInfo::m_pSentence = pSentence;
  CChartParserInfo::m_sourceLength = pSentence.size();

  //create the complete-translation node that holds the complete sentence translating hyps
  //az: now a regular node in cell n,1
  //  m_pCompleteTranslationsNode = new CProductionNode(CChartParserInfo::m_sourceLength, 1, CChartParserInfo::m_FullSentenceNT);

  //allocate space for the equivalence class containers
  SetupChartCells();

  for (int w = 0; w<pSentence.size(); w++){

    cout<<"Log: Checking word: "<<w+1<<" = "<<" index: "<<pSentence[w]<<" word: "<<CChartParserInfo::m_pVocabulary->GetWord(CChartParserInfo::m_pSentence[w])<<endl;

    //reset the stack level - which is ONLY used for pretty printing in some cases
    CChartParserInfo::m_stackLevel = 0;

    // check whether this cell should get a special status
    if (pSentence[w] == CChartParserInfo::m_bosWI) {
      //+1 since the parser works on 1 indexing
      GetCompleteCell(1,w+1).m_cellType = 1; // BOS
    } else if (pSentence[w] == CChartParserInfo::m_eosWI) {
      //+1 since the parser works on 1 indexing
      GetCompleteCell(1,w+1).m_cellType = 2; // EOS
    }

    // create Terminal complete hyp
    //+1 since the parser works on 1 indexing
    CChartParserHypComplete* terminalHyp = new CChartParserHypComplete(pSentence[w], 1, w+1);

    // and add it to the CH chart
    bool wasAdded = AddCompletedHyp(terminalHyp);
    assert(wasAdded==true);
    if (!wasAdded){
      cout<<"ERROR: terminalHyp was rejected by the online beam criteria"<<endl;
      exit(-1);
    }
    //Reset the expansion statistics
    CChartParserInfo::ResetCellStatistics();
    
    //+1 since the parser works on 1 indexing
    IndexNodesAccordingToRuleAppCount(1,w+1);
    
  }
  cout<<"Initialization completed"<<endl;
    
}

void CChartParser::SetupChartCells(){

  //note the i<=CChartParserInfo::m_sourceLength (allows us to do all the parsing from 1)
  m_completeHypsChart.reserve(CChartParserInfo::m_sourceLength+1);
  m_incompleteHypsChart.reserve(CChartParserInfo::m_sourceLength+1);
  for (int i=0; i<=CChartParserInfo::m_sourceLength; i++){
    vector<CCompleteHypChartCell> vComplete(CChartParserInfo::m_sourceLength+1);
    m_completeHypsChart.push_back(vComplete);
    vector<CSortedIncompleteHypContainer> vIncomplete(CChartParserInfo::m_sourceLength+1);
    m_incompleteHypsChart.push_back(vIncomplete);
  }

}


CChartParser::CChartParser(CVocabulary *pVocabulary,CWordIndex unknownWI, bool noTranslationOutputOfUnknownWords, float maxRuleAppCountDifference, string ruleDBFile, int maxHypCount, int MaxCombinationCount, double aFutureCostImpact, int aBeamSizeFactor, bool extractUnique) : 
  m_maxRuleAppCountDifference(maxRuleAppCountDifference),
  m_bTree(ruleDBFile)
{
  CChartParserInfo::m_noTranslationOutputOfUnknownWords=noTranslationOutputOfUnknownWords;
  CChartParserInfo::m_pVocabulary = pVocabulary;
  CChartParserInfo::FutureCostImpact = aFutureCostImpact;
  CChartParserInfo::BeamSizeFactor = aBeamSizeFactor;
  CChartParserInfo::m_extractUnique = extractUnique;
  
  CChartParserInfo::m_unknownWI = unknownWI;
  CChartParserInfo::MAX_HYP_COUNT = maxHypCount;
  CChartParserInfo::m_maxCombinationCount = MaxCombinationCount;

  // note: it is important that m_bosWI refer to the string "<s>" because of the LM call to the SRI, which treats
  // the string <s> as beginning of sentence marker, so don't change this!
  CChartParserInfo::m_bosWI = CChartParserInfo::m_pVocabulary->GetIndex("<s>");
  CChartParserInfo::m_eosWI = CChartParserInfo::m_pVocabulary->GetIndex("</s>");
  CChartParserInfo::m_lmBoundaryWI = CChartParserInfo::m_pVocabulary->GetIndex("<*>");

  m_pBeamToDelete = NULL;
}

void CChartParser::ClearCHCCell(int i, int j){

  CCoarseGrainedEq& prodNodes = m_completeHypsChart[i][j].m_productionNodes;

  //CProductionNode(s) are newed, and inside them ChartParserNode(s) are newed and within them the hyps are stored
  for(CCoarseGrainedEq::iterator iter = prodNodes.begin(); iter!=prodNodes.end(); ++iter){
    CProductionNode* pProdNode = (*iter).second;
    delete pProdNode;
    pProdNode = NULL;
  }
}

void CChartParser::ClearIHCCell(int i, int j){
  for (CSortedIncompleteHypContainer::iterator iter =  m_incompleteHypsChart[i][j].begin(); iter!= m_incompleteHypsChart[i][j].end(); ++iter){
    CChartParserHypIncomplete* pIncompleteHyp = (*iter).second;
    delete pIncompleteHyp;
  }
}

void CChartParser::ClearSearch(){
  //first clear the completed hyps chart. when clearing a container, first clear all its elements
  for (int i=1; i<=CChartParserInfo::m_sourceLength; i++){
    for (int j=1; j<=CChartParserInfo::m_sourceLength; j++){
      ClearCHCCell(i, j);
    }
    m_completeHypsChart[i].clear();
  }
  m_completeHypsChart.clear();

  //  delete m_pCompleteTranslationsNode;

  assert(m_newlyCreatedNodes.empty());

  //clear the cache(s) AFTER rescoring
  if (!CChartParserInfo::m_globalLMCache){
    for (int i=0; i<CLanguageModel::m_languageModels.size(); i++)
      CLanguageModel::m_languageModels[i]->ClearCache();
  }

  //clear any hyps that exist as a result of n-best retrieval
  for (vector<CChartParserHypCompleteNBest*>::iterator iter=m_poppedHyps.begin(); iter!= m_poppedHyps.end(); ++iter){
    CChartParserHypCompleteNBest* pHyp = (*iter);
    delete pHyp;
    pHyp=NULL;
  }

  if (m_pBeamToDelete!=NULL){
    delete m_pBeamToDelete;
    m_pBeamToDelete = NULL;
  }
}





void CChartParser::AddIncompleteHyp(CChartParserHypIncomplete* pIncompleteHyp) {
  bool debug = false;
  if (debug) {
    for (int i=0; i<CChartParserInfo::m_stackLevel; i++) cout<<"\t";
    cout<<  "AddIncompleteHyp:";
    pIncompleteHyp->Display();
    cout<<endl;
  }
  //CSortedIncompleteHypContainer& cell = m_incompleteHypsChart.GetRow(pIncompleteHyp->I())[pIncompleteHyp->J()];
  CSortedIncompleteHypContainer& cell = m_incompleteHypsChart[pIncompleteHyp->I()][pIncompleteHyp->J()];
  cell.insert(std::pair<float, CChartParserHypIncomplete*>(pIncompleteHyp->m_totalRuleApplicationCount, pIncompleteHyp));
}



//!Add a hyp to the m_unfiledHyps vector of the respective production node
/*!
 * Returns true iff the hyp was compliant with the beam operating in the production node
 * In any case, it is AddCompletedHyp's responsibility that pHyp gets deleted at some point of time
 */
bool CChartParser::AddCompletedHyp(CChartParserHypComplete* pHyp, bool addAll){

  //MAIN DEBUG
  bool s_debugFileHyp = false;
  if (s_debugFileHyp){
    cout<<"AddCompletedHyp: ";
    pHyp->DisplayRecursive();
  }
  CCompleteHypChartCell& chartCell = m_completeHypsChart[pHyp->I()][pHyp->J()];
  
  CProductionNode* node; // hypergraph node for pHyp
  bool newlyCreated = false;
  //3 cases
  // - if you need to file in the sentence spanning production node - az06/11/01: not considered a special case anymore
  // - if you need to file in a new production node
  // - if you need to file in an existing production node
  //  if ((pHyp->GetResult()==CChartParserInfo::m_FullSentenceNT) && (pHyp->J()==1) && (pHyp->I()==CChartParserInfo::m_sourceLength)){

  // we have a complete translation -- add the hyp to m_pCompleteTranslationsNode and don't apply self rule
  //    if (s_debugFileHyp){
  //      CChartParserInfo::TabStack(); cout<<"Filing it in m_pCompleteTranslationsNode -- NO RECURSION"<<endl;
  //    }
     
  //    node = m_pCompleteTranslationsNode;
  //    if (s_debugFileHyp){cout<<"AddCompletedHyp m_pCompleteTranslationsNode"<<endl;}
  //  } else 

  if (chartCell.m_productionNodes.count(pHyp->GetResult())==0){     //if there are no elements of this pHyp 's equivalence class (i.e., owning hypergraph node), start a new one

    //create a new node right here
    node = new CProductionNode(pHyp->I(), pHyp->J(), pHyp->GetResult()); 
    chartCell.m_productionNodes.insert(std::pair<CWordIndex, CProductionNode*>(pHyp->GetResult(), node));

    if (s_debugFileHyp){
      CChartParserInfo::TabStack();
      cout<<"New node created in this cell - marking it to apply the self rule in the next wave:";
      node->Display();
      cout<<endl;
    }
    //m_newlyCreatedNodes.push_back(node);
    m_newlyCreatedNodes.insert(std::pair<CWordIndex, CProductionNode*>(pHyp->GetResult(), node));
    newlyCreated = true;

  }else{
    //EC exists already and is not the complete-translation node
    //let's add the hyp to the corresponding production node, which will filter down the fine grained nodes
    CCoarseGrainedEq::iterator curr = chartCell.m_productionNodes.find(pHyp->GetResult());
    node = (*curr).second;
    
    if (s_debugFileHyp){
      CChartParserInfo::TabStack(); cout<<"Ex. Node:";
      node->Display();
      cout<<endl;
    }

  }  
  bool wasAdded = node->CheckBeamInsertHyp(pHyp, addAll);
  //if you just created the hyp you MUST have kept the hyp
  if (newlyCreated){
    assert(wasAdded==true);
  }
  return wasAdded;

}



void CChartParser::UnwindHyp_HandleUnknownAndTags(CWordIndex word, CChartParserHypComplete* pNewHyp, CChartParserHypComplete* pBackHyp){
  if (word == CChartParserInfo::m_unknownWI) { // is it an unknown marker?
    // yes - save the corresponding source word
    // since this must come from an unknown-rule of the form _UNKNOWN#_UNKNOWN#@X#weights, we expect this rule to have 
    // no backpointers (because no nonterminals are involved) and a span of one
    if ((pBackHyp->m_backPointers.size() != 0) || (pBackHyp->I() != 1)) {
      cerr << "In UnwindHyp: wrong format of unknown-rule employed. Unknown-rule's hyp is: ";
      pBackHyp->DisplayRecursive(CChartParserInfo::m_pVocabulary, 0);
    }
    // push the corresponding source word into m_targetOrderedUnknownSourceWords
    //pNewHyp->m_targetOrderedUnknownSourceWords.push_back((*CChartParserInfo::m_pSentence)[pBackHyp->J()]);
    //parsing is indexed from 1
    pNewHyp->m_targetOrderedUnknownSourceWords.push_back(CChartParserInfo::m_pSentence[pBackHyp->J()-1]);
  }else{
    
    //if you are not an m_unknownWI, then you are one of two things
    // a) a regular terminal, and you have been added onto the newTranslationOutput and the newParsingOutput
    // b) a non-terminal symbol that occurs in translation output. This is different from regular rules, which get converted into "I gave -1 a -2" form
    //    in this case we see "I gave him @NUM dollars"
    
    //if we have hit a tagged word backpointer (like @NUM) then we ONLY want to convert it into a m_taggedTargets element if it has no back pointers
    //in order to support multiple levels of tagged hierarchy. If its not a tag then process as usual
    if (CChartParserInfo::m_tagsForLM.count(word)!=0){
      //implies pBackHyp was empty AND the its production result was a tag
      //insert into the pNewHyp at the end of its m_taggedTargets since we are going left to right
      //we add ALL the m_taggedTargets since @NUM could represent  to "two and a half" 
      pNewHyp->m_taggedTargets.insert(pNewHyp->m_taggedTargets.end(), pBackHyp->m_taggedTargets.begin(), pBackHyp->m_taggedTargets.end()); 
    }
  }
}

//CChartParserHypCompleteNBest* CChartParser::UnwindHyp(CSortedHypNBestContainer::iterator& iterPHyp, CHypBeam& beam, float lowestCostHyp){

CChartParserHypCompleteNBest* CChartParser::UnwindHyp(CChartParserHypCompleteNBest* pHyp, CHypBeam& beam, float lowestCostHyp){
  
#ifndef NDEBUG
  bool debug = false;
#endif
  //pre-conditions: 
  //pHyp has a m_posNT (0 based index) that points to the current location of the first NT
  //if m_posNT goes past the size of the lists, then we should not be coming back to this point
  //pHyp->m_scaledFutureCostEstimate is the language model probability of the pHyp->m_translationOutput sequence 
  //accounting for first best approximations when using non-terminals
  //CChartParserHypCompleteNBest* pHyp = iterPHyp->second;
  //eraase the beam from the hyp using its iterator
  //beam.m_store.erase(iterPHyp); //this is done before calling UnwindHyp now


  //th next line is not compatible with g++ 4.1
  //iterPHyp = 0; // just to make sure it's not being used again b/c it's now invalid

  //you should NOT be completely unwound - since there is nothing to expand further
  //but you MAY be ready for the m_completelyUnwound designation. The reason for this two part process where you first get filed
  //in the m_completelyUnwound=0 state and then get transfered to m_completelyUnwound=true is to make sure these unwound hyps
  //actually stay on the beam (and then we dont pop them again in RetrieveKBestParses)
  assert(pHyp->m_completelyUnwound!=true);


   //ensure that this is a hyp that we actually want to do calculations for
  if (!pHyp->m_active && !CChartParserInfo::m_dontReloadBeam){
    //should make this a counted variable
#ifndef NDEBUG
    if (debug){
      cout<<"Rejecting inactive: "; pHyp->PrintTranslationOutput(); cout<<" cost: "<<pHyp->m_totalCost<<" unwnd: "<<pHyp->m_completelyUnwound<<endl;
    }
#endif
    //we have removed it, remove the extra buffer that we had provided
    if (CChartParserInfo::m_updateBeamSize){
      beam.m_maxHypCount--;
#ifndef NDEBUG
    if (debug){
      cout<<"Decreasing beam size to: "<<beam.m_maxHypCount<<endl;
    }
#endif
    }
    return NULL;
  }

#ifndef NDEBUG
  if (debug){
    cout<<"\n\nUnwind Hyp: "; pHyp->PrintTranslationOutput(); cout<<" m_posNT: "<<pHyp->m_posNT<<" orig totalcost: "<<pHyp->m_totalCost<<" neg: ";
    copy(pHyp->m_negLogProbs.begin(), pHyp->m_negLogProbs.end(), ostream_iterator<float>(cout, " ")); 
    cout<<" m_taggedTargets: "; CChartParserInfo::PrintVocabSequence(pHyp->m_taggedTargets);
    cout<<endl;
  }
#endif

  //if you meet the qualifications for unwound ness
  if (pHyp->m_posNT==pHyp->m_translationOutput.size()){
    CChartParserHypCompleteNBest* pNewHyp = new CChartParserHypCompleteNBest(*pHyp);
    pNewHyp->m_completelyUnwound = true;
    pNewHyp->m_backPointers.clear();
#ifndef NDEBUG
    if (debug){
      cout<<"Setting completely unwound!!!"<<endl;
      pNewHyp->PrintTranslationOutput(); cout<<" m_taggedTargets: "; CChartParserInfo::PrintVocabSequence(pNewHyp->m_taggedTargets); cout<<endl;
    }
#endif
    pNewHyp->SetBackInfo(beam.LookupProto(pHyp), pHyp);
    beam.InsertHyp(pNewHyp);

    //this is the only time that u return any value
    return pNewHyp;
  }


  assert(pHyp->m_posNT<=pHyp->m_translationOutput.size()-1);
  //do the same with the m_parsingOutput - unexpanded non-terminals point into the same locations as in the m_translationOutput case
  //but there are additional symbols for [, NT, etc in the m_parsingOutput which are real symbols in the vocab
  //calculate the posNTParse
  int posNTParse = 0;   
  for (CWordSequence::iterator itP = pHyp->m_parsingOutput.begin(); itP!= pHyp->m_parsingOutput.end() && *itP>0; ++itP){
    ++posNTParse;
  }

  CWordSequence sofar = CWordSequence(pHyp->m_translationOutput.begin(), pHyp->m_translationOutput.begin()+ pHyp->m_posNT);
  //sofar has no backpointers (its all terminals), while remaining could have several back pointers
  CWordSequence remaining = CWordSequence(pHyp->m_translationOutput.begin()+pHyp->m_posNT+1, pHyp->m_translationOutput.end());  
  //first remove future costs associated with this non-terminal
  CWordIndex currentNT = pHyp->m_translationOutput[pHyp->m_posNT];
  
  //if (debug){
  //cout<<"sofar: "; CChartParserInfo::PrintVocabSequence(sofar); cout<<endl;
  //cout<<"rem: "; CChartParserInfo::PrintVocabSequence(remaining); cout<<endl;
  //cout<<"currentNT: "<<currentNT<<endl;
  //cout<<endl;
  //}

  assert(currentNT<=0);
  // get the non-terminal object 
  CChartParserNode* pEquivalenceClass = pHyp->m_backPointers[-1 * (currentNT)];

  //COSTS: pHyp->m_totalCost has the accumulated costs for pHyp, including all LM decisions
  //       pHyp->m_negLogProbs have ONLY the surface LM costs as well as the surface (single rule only) translation costs
  //             -- these surface costs include real LM costs of a hyp as well as estimates where nt boundaries occured
  //             -- at then end of this, the pNewHyp needs to have accumulated m_negLogProbs costs
  //
  //       new hyps will be put on the beam based on m_totalCost - we need to remove the LM cost due to the replacement operation

  float totalCostBase = pHyp->m_totalCost;

  //now remove language model probs for the left and right boundary around this non-terminal
  CWordSequence sofar_context = CUtilities::ComputeLMContext(sofar);
  //compute the remaining context back on potential non-terminals that exist in the remaining symbols
  CWordSequence remaining_context = CUtilities::ComputeLMContext(CChartParserNode::HarvestLMString(remaining, pHyp->m_backPointers));

  //if (debug){
  //cout<<"sofar_context: "; CChartParserInfo::PrintVocabSequence(sofar_context); cout<<endl;
  //cout<<"remaining_context: "; CChartParserInfo::PrintVocabSequence(remaining_context); cout<<endl;
  //}

  //subtract out these costs
  //left context
  CWordSequence forLeftContextInitial = sofar_context; 
  forLeftContextInitial.insert(forLeftContextInitial.end(), pEquivalenceClass->m_lmContext.begin(), pEquivalenceClass->m_lmContext.end());
  //we care bout saving the actual probs for the lm parts that we are removing, since these surface costs need to be deleteHC05d in each alternative
  vector<float> forLeftContextRemoval;
  for (int i=0; i<CRule::C_NumLMs; i++) {

    float remval = CUtilities::ComputeLMCost(*CLanguageModel::m_languageModels[i], forLeftContextInitial, false, true);
    //if (debug){
    //cout<<"Removing left: "; CChartParserInfo::PrintVocabSequence(forLeftContextInitial); cout<<" val: "<<remval<<endl;
    //}
    forLeftContextRemoval.push_back(remval);
    totalCostBase -= forLeftContextRemoval[i]*CChartParserInfo::m_scalingFactors[i];
  }
  //right context
  CWordSequence forRightContextInitial = CUtilities::ComputeLMContext(forLeftContextInitial); 
  forRightContextInitial.insert(forRightContextInitial.end(), remaining_context.begin(),  remaining_context.end());

  //we care bout saving the actual probs for the lm parts that we are removing, since these surface costs need to be deleted in each alternative
  vector<float> forRightContextRemoval;
  for (int i=0; i<CRule::C_NumLMs; i++) {
    float remval = CUtilities::ComputeLMCost(*CLanguageModel::m_languageModels[i], forRightContextInitial, false, true);
    //if (debug){
    //cout<<"Removing right: "; CChartParserInfo::PrintVocabSequence(forRightContextInitial); cout<<" val: "<<remval<<endl;
    //}
    forRightContextRemoval.push_back(remval);
    totalCostBase -= forRightContextRemoval[i]*CChartParserInfo::m_scalingFactors[i];
  }
  //pick up the alternatives and spawn of new hyps
  CSortedHypContainer* sortedHyps = &(pEquivalenceClass->m_hypContainer);
  //CChartParserHypComplete* pOldFirstBest =  (sortedHyps->begin())->second;

  //======== Go thru each alternative ==========
  for (CSortedHypContainer::const_iterator iter2 = sortedHyps->begin(); iter2 != sortedHyps->end(); iter2++){


    //pNewHyp is the hyp that represents subbing in pBackHyp for the NT as position m_posNT
    //pNewHyp needs to be full configured to participate on the beam. This involves...
    // - setting its new m_totalCost and accumulating m_negLogProbs 
    // - moving the m_posNT to the appopriate position
    // - changing its m_translationOutput to include the subbed in symbols
    // - updates its backpointers to reflect any new non-terminals that were seen
    CChartParserHypCompleteNBest* pNewHyp = new CChartParserHypCompleteNBest(*pHyp);
    //pNewHyp has m_negLogProbs that are ONLY the surface level costs - ie incorrect LM surface costs that need to be deleted out

    //pBackHyp represents the alternative translation that we are considering
    CChartParserHypComplete* pBackHyp = (*iter2).second;
    //if (!(pBackHyp->m_totalCost>=pOldFirstBest->m_totalCost)){
    //cout<<"curr: "<<pBackHyp->m_totalCost<<" best: "<<pOldFirstBest->m_totalCost<<endl;
    //assert(pBackHyp->m_totalCost>=pOldFirstBest->m_totalCost);
    
    //}
    //if (debug){
    //cout<<"Back Hyp: "; pBackHyp->PrintTranslationOutput(); cout<<" m_taggedTargets: "; CChartParserInfo::PrintVocabSequence(pBackHyp->m_taggedTargets); cout<<endl;
    //}

    //this cost has no language boundary costs for the old hyp because we cut them out above - so its the accumulated total cost
    //based on first best decisions for TM and LM, minus the old boundary costs
    float totalCost = totalCostBase; 
    //left context
    CWordSequence forLeftContext = sofar_context;
    forLeftContext.insert(forLeftContext.end(), pBackHyp->m_lmContext.begin(), pBackHyp->m_lmContext.end());
    for (int i=0; i<CRule::C_NumLMs; i++) {
      //update the surface level costs for the left context
      float updatedLM = CUtilities::ComputeLMCost(*CLanguageModel::m_languageModels[i], forLeftContext, false, true);
      //if (debug){
      //cout<<"Addback left: "; CChartParserInfo::PrintVocabSequence(forLeftContext); cout<<" val: "<<updatedLM<<endl;
      //}
      pNewHyp->m_negLogProbs[CRule::C_LMLocation+i] += updatedLM - forLeftContextRemoval[i]; //remove the old boundary costs, add in a new one
      totalCost += updatedLM * CChartParserInfo::m_scalingFactors[CRule::C_LMLocation+i]; 
    }
    //right context
    CWordSequence forRightContext = CUtilities::ComputeLMContext(forLeftContext);
    forRightContext.insert(forRightContext.end(), remaining_context.begin(), remaining_context.end());
    for (int i=0; i<CRule::C_NumLMs; i++) {
      //update the surface level costs for the right context
      float updatedLM = CUtilities::ComputeLMCost(*CLanguageModel::m_languageModels[i], forRightContext, false, true);
      //if (debug){
      //	cout<<"Addback right: "; CChartParserInfo::PrintVocabSequence(forRightContext); cout<<" val: "<<updatedLM<<endl;
      //}
      pNewHyp->m_negLogProbs[CRule::C_LMLocation+i] += updatedLM - forRightContextRemoval[i]; //remove the old boundary costs, add in a new one
      totalCost += updatedLM * CChartParserInfo::m_scalingFactors[CRule::C_LMLocation+i]; 
    }

    //if (debug){
    //cout<<"pHyp->m_totalCost:"<<pHyp->m_totalCost<<endl;
    //}
    //the simple move of switching out the old vs new for all the internal costs
    totalCost += pBackHyp->m_totalCost - pEquivalenceClass->m_bestTotalCost;

    
    //sanity check assersions for the first best hyp. we know that the estimates were calculated based on first best choices
    //so we should have deleted (before the loop) and added back costs to result in the same value as pHyp->m_totalCost
    if (iter2 == sortedHyps->begin()){
      assert(fabs(totalCost-pHyp->m_totalCost)<0.01);
      //assert(pNewHyp->m_negLogProbs == pHyp->m_negLogProbs); //need to check this hear before the accumulation starts
      assert(pNewHyp->m_negLogProbs.size()==pHyp->m_negLogProbs.size());
      for (int i=0; i<pNewHyp->m_negLogProbs.size(); i++){
	assert(fabs(pNewHyp->m_negLogProbs[i]-pHyp->m_negLogProbs[i])<0.01);
      }
    }

    //now the totalCost and pNewHyp->m_negLogProbs looks as they would if this alternative pBackHyp was taken during parsing
    //we still need to perform accumulation on the pNewHyp->m_negLogProbs. 
    //do the final assignment of costs for the new hyp before it gets considered for addition onto the beam
    //add pBackHyp's neg log probs to pNewHyp - NOTE during parsing these m_negLogProbs were never accumulated
    //now they finally get summed to represent decisions made over multiple rules
    for (int m=0; m<CRule::C_NumModelsTotal; m++){
      pNewHyp->m_negLogProbs[m] += pBackHyp->m_negLogProbs[m]; //no need to delete the old hyps costs, since m_negLogProbs are only accumulated now
    }
    
  
    //adjust m_totalCost variable by the difference between pBackHyp and best hyp in pEquivalenceClass
    pNewHyp->m_totalCost = totalCost;
    //at this point the cost is set... you can check to see if its ok

    //set up the parsing output
    CWordSequence newParsingOutput(pNewHyp->m_parsingOutput.begin(), pNewHyp->m_parsingOutput.begin()+posNTParse);
    CChartParserHypComplete::AddParseOutputSubTreeStart(newParsingOutput, pBackHyp);

    //create a new translation, replacing the not terminal negative indicies with their appropriate values
    CWordSequence newTranslationOutput = sofar; //needs to have the corrected back pointer indices here
    for (CWordSequence::iterator iter3=pBackHyp->m_translationOutput.begin(); iter3!=pBackHyp->m_translationOutput.end(); ++iter3){
      if (*iter3 > 0) {
	// *iter3 is a terminal symbol	
	// copy it over into the new translation output
	newTranslationOutput.push_back(*iter3);
	newParsingOutput.push_back(*iter3);	
	//add the unknown src word and the tagged output to appropriate background vectors
	UnwindHyp_HandleUnknownAndTags(*iter3, pNewHyp, pBackHyp);
	//pNewHyp->NoteUnknownAndTags(*iter3, pBackHyp);	
      } else {
	// *iter3 is a backpointer - gotta adjust it to point to its back node's new location
	newTranslationOutput.push_back(*iter3 - pHyp->m_backPointers.size());
	// also store the backpointer in the translation info tree
	newParsingOutput.push_back(*iter3 - pHyp->m_backPointers.size());
      }
    }
    newTranslationOutput.insert(newTranslationOutput.end(), remaining.begin(), remaining.end());
    CChartParserHypComplete::AddParseOutputSubTreeEnd(newParsingOutput);
    //and copy over the rest of the original hyp's translation info tree
    newParsingOutput.insert(newParsingOutput.end(), pNewHyp->m_parsingOutput.begin()+posNTParse+1, pNewHyp->m_parsingOutput.end());
     
    pNewHyp->m_translationOutput = newTranslationOutput;
    pNewHyp->m_parsingOutput = newParsingOutput;
    pNewHyp->m_backPointers.insert(pNewHyp->m_backPointers.end(),pBackHyp->m_backPointers.begin(),pBackHyp->m_backPointers.end());
    //itendify the location of m_posNT
    
    //identify the number of terminal words in pBackHyp's m_translationOutput (it could be all of them)
    int newPosNT = pHyp->m_posNT;
    //terminals are greater than 0
    while (newPosNT < newTranslationOutput.size() && newTranslationOutput[newPosNT]>0){
      newPosNT++;
    }
    pNewHyp->m_posNT = newPosNT;
    //this may indicate that you are at the end
    //cout<<"final posnt: "<<pNewHyp->m_posNT<<endl;

    if (pNewHyp->ScaledTMLMCost() > CChartParserInfo::GetPruningCutoff(lowestCostHyp, CChartParserInfo::m_aStarBeam) ){
      //this hyp will NEVER be expanded (since if you pop it you will terminate search). Dont add it to the beam!
#ifndef NDEBUG
      if (debug){
	cout<<"Deleting hyp: "; pNewHyp->PrintTranslationOutput(); cout<<" m_totalCost: "<<pNewHyp->m_totalCost<<" lowest: "<<lowestCostHyp<<endl;
      }
#endif
      delete pNewHyp;
      //continue; //dont do any more processing of this hyp
      break; //actually stop processing this root hyp
    }else{
#ifndef NDEBUG
      if (debug){
	cout<<" Adding hyp: "; pNewHyp->PrintTranslationOutput(); cout<<" m_totalCost: "<<pNewHyp->m_totalCost<<" lowest: "<<lowestCostHyp<<endl;
      }
#endif
      //set the new lowestCostHyp if its lower...
      if (pNewHyp->ScaledTMLMCost()<lowestCostHyp){
	lowestCostHyp = pNewHyp->ScaledTMLMCost();
      }
    }


    //now we are ready to add to the beam. This newly created hyp could get deleted so dont use pNewHyp anymore
    //set the back trace info to the equivalence class of the original pHyp, and pHyp itself
    pNewHyp->SetBackInfo(beam.LookupProto(pHyp), pHyp);
    bool wasNew = beam.InsertHyp(pNewHyp);
    
    //if (!wasNew){
    //break;
    //}
  }
  return NULL;
  
}


//!sub used by RetrieveKBestParses
/*! 
  Unwinds pHyp - which contains backpointers to cover part of its span - unwinding means instantiating a hyp from the backpointer 
  pHyp has to be on the beam and will be removed from the beam before unwinding unless pHyp turns out to be completely unwound
  auto-delete hyps worse than k best when inserting into CHypContainer
  //  Removes m_translationOutput _UNKNOWN if m_noTranslationOutputOfUnknownWords==true
  //  Replaces m_translationOutput _UNKNOWN by the corresponding source word if m_noTranslationOutputOfUnknownWords==false
  */
/* !!temporarily disabled
void CChartParser::UnwindHyp(CSortedHypNBestContainer::iterator& iterPHyp, CHypBeam& beam, int K){

#ifndef NDEBUG
  bool debug = false;
#endif

  //track the hyp pointer
  CChartParserHypCompleteNBest* pHyp = iterPHyp->second;
#ifndef NDEBUG
  if (debug) {
    cout << "In UnwindHyp; unwinding hyp: ";
    pHyp->DisplayRecursive();
  }
#endif

  assert(!pHyp->m_completelyUnwound); //make sure that the hyp hasn't been declared completely unwound yet
  //remove hyp from the beam, but don't delete it just yet (we will either re-insert it with 
  //modified totalCost and m_scaledFutureCostEstimate or 
  //clone it several times to create new unwound instances and re-insert these new hyps)
  beam.m_store.erase(iterPHyp);
  iterPHyp = 0; // just to make sure it's not being used again b/c it's now invalid


  if (!pHyp->m_active){
    //should make this a counted variable
    if (debug){
      cout<<"Rejecting hyp on the agenda"<<endl;
    }
    return;
  }
  


  // will contain the additional LM costs incurred while traversing translation output
  float pHypLMCost = 0.0;

  //go through the translation output and unwind the first nonterminal encountered
  // also LM-process the words before the first NT
  int posNT = pHyp->m_firstWords.size();
  CWordSequence::iterator iter;
  for (iter=pHyp->m_translationOutput.begin()+posNT; iter!=pHyp->m_translationOutput.end() && *iter>0; ++iter) {
    // *iter is a terminal symbol
    //old way - always allow BOS to get called if the context is not filled yet
    //pHypLMCost += UnwindHyp_LMProcessWord(pHyp->m_firstWords, *iter);

    //new way - only allow the BOS call if the hyp had BOS done - the first words are still cleared before
    //entry into this function, so the boundary will get rebuilt but the BOS call will not be made
    pHypLMCost += CChartParserInfo::UnwindHyp_LMProcessWord(pHyp->m_firstWords, *iter, pHyp->m_BOSDone);
    ++posNT;
  }
  //do the same with the m_parsingOutput - unexpanded non-terminals point into the same locations as in the m_translationOutput case
  //but there are additional symbols for [, NT, etc in the m_parsingOutput which are real symbols in the vocab
  int posNTParse = 0;
  
  for (CWordSequence::iterator itP = pHyp->m_parsingOutput.begin(); itP!= pHyp->m_parsingOutput.end() && *itP>0; ++itP){
    ++posNTParse;
  }

  assert(pHyp->m_firstWords.size() == posNT);

  //did we encounter a nonterminal symbol at all?
  if (iter==pHyp->m_translationOutput.end()) {
    
    //ARV 10/03 make sure you insert a cloned version - since we deleted based on recombined elements
    //this pNewHyp has its m_completelyUnwound set to true and therefor will be in a class of its own
    CChartParserHypCompleteNBest* pNewHyp = new CChartParserHypCompleteNBest(pHyp);

    //no, we didn't: this means we can declare the hyp completely unwound
    pNewHyp->m_completelyUnwound = true;

    // still have to process the end-of-sentence LM costs
    // using m_firstWords here is no bug, but intended since this vector now contains
    // the complete history context up to the last words of the translation output
    // (see also RetrieveKBestParses)
    if (pNewHyp->I()==CChartParserInfo::m_sourceLength && pNewHyp->J()==1){
      pHypLMCost += CChartParserInfo::UnwindHyp_EOS(pNewHyp->m_firstWords);
    }

    //clear m_firstWords 
    pNewHyp->m_firstWords.clear();
    // update the LM cost and total cost
    pNewHyp->m_negLogProbs[CRule::C_LMLocation] += pHypLMCost;
    pNewHyp->m_totalCost += m_scalingFactors[CRule::C_LMLocation] * pHypLMCost;
    // everything is processed, so there are no future cost estimates anymore
    pNewHyp->m_scaledFutureCostEstimate = 0.0;

    // also, empty the backpointer vector, as m_translationOutput
    // contains no references to the backpointer nodes anymore and calls
    // of UpdateTotalCost() should not add the backpointer nodes'
    // costs to m_totalCost
    pNewHyp->m_backPointers.clear();

    //re-insert pNewHyp into the beam, deleting it if its m_totalCost+m_scaledFutureCostEstimate is worse than the K best hyps on the beam. 
    pNewHyp->SetBackInfo(beam.LookupProto(pHyp), pHyp);
    beam.InsertHyp(pNewHyp);
    

  } else {
    // the word index *iter the first unexpanded backpointer, posNT is the corresponding position in the m_pTranslationOutput vector

    //we'll now clone the hyp several times to create new unwound instances based on the different hyps
    //sitting in the encountered nonterminal's corresponding backpointer node

    //used only when compare tgt words is true
    double savedAllCosts = pHyp->m_totalCost + pHyp->m_scaledFutureCostEstimate;
    // first reset the future cost estimate for LM-unprocessed words -- we'll recompute it below
    //resetting it here causes it to be 0 when the new hyps are spawned
    //DONT DO THIS HERE - its messing with the memory IN the pHyp, do it right AFTER spawning
    //pHyp->m_scaledFutureCostEstimate = 0.0;

    // get the hypergraph node pointed to   
    CChartParserNode* pEquivalenceClass = pHyp->m_backPointers[-1 * (*iter)];

#ifndef NDEBUG
    if (debug){
      cout<<"found backpointer node:";
      pEquivalenceClass->Display(CChartParserInfo::m_pVocabulary);
      cout<<" savedAllCosts: "<<savedAllCosts<<endl;
    }
#endif

    // process all hyps of pEquivalenceClass, creating a new forked hyp for each of these hyps
    CSortedHypContainer* sortedHyps = &(pEquivalenceClass->m_hypContainer);
    
    bool isFirst = true;
    
    //============================================================================================================================
    // Pruning conditions. The fine grained node is expanded to reveal all its hyps, in order of ScaledTMLMCost()
    // It *would* be safe to only expand a few hyps here in cmp=1. In cmp=0 however, a lower cost hyp might rise up
    // But under the same beam+fuzz criteria, we say if a hyp is being rejected by the beam, then we trigger a fuzz
    // condition, and only insert upto fuzz more.
    // ============================================================================================================================
    
    for (CSortedHypContainer::const_iterator iter2 = sortedHyps->begin(); iter2 != sortedHyps->end(); iter2++){

      CChartParserHypComplete* pBackHyp = (*iter2).second;
      //cout<<"Expanding pBackHyp: "; pBackHyp->DisplayRecursive(CChartParserInfo::m_pVocabulary, 1); cout<<endl;

      //clone pHyp
      //ARV 10/03 clone as a nbest element - this kind of h
      //CChartParserHypComplete* pNewHyp = new CChartParserHypComplete(pHyp);
      CChartParserHypCompleteNBest* pNewHyp = new CChartParserHypCompleteNBest(pHyp);
      pNewHyp->m_scaledFutureCostEstimate = 0.0;
      // copy over the newly obtained LM costs
      float pNewHypLMCost = pHypLMCost;

      //our new translation output
      // initialize with all the stuff before the first NT b/c that remains the same
      CWordSequence newTranslationOutput(pNewHyp->m_translationOutput.begin(), pNewHyp->m_translationOutput.begin()+posNT);
      CWordSequence newParsingOutput(pNewHyp->m_parsingOutput.begin(), pNewHyp->m_parsingOutput.begin()+posNTParse);

      //add the non-terminal symbol to the tree-format translation info

      // start a new subtree with '#('
      
//	newParsingOutput.push_back(CChartParserInfo::m_pVocabulary->GetIndex("#("));
//	// root of the subtree: our LHS (result) category
//	newParsingOutput.push_back(pEquivalenceClass->GetResult());
//	//add source span start and end (will be converted into actual numbers later and appended to the LHS category via '-')
//	newParsingOutput.push_back(pBackHyp->J());
//	newParsingOutput.push_back(pBackHyp->J()+pBackHyp->I()-1);
//
//	//      if (pBackHyp->m_backPointers.size()==0){
//	//add all the source words of the span (will also be appended to the LHS category via '-')
//	int startZIndex = pBackHyp->J();
//	for (int si=0; si<pBackHyp->I(); si++){
//	//newParsingOutput.push_back((*CChartParserInfo::m_pSentence)[si+startZIndex]);
//	//the parsing is indexed from 1, thus the -1
//	assert(CChartParserInfo::m_pSentence.size()>si+startZIndex-1);
//	newParsingOutput.push_back(CChartParserInfo::m_pSentence[si+startZIndex-1]);
//	}
//	//      }
//	// marks the end of the subtree root (everything up to this symbol is being appended to the LHS category)
//	newParsingOutput.push_back(CChartParserInfo::m_pVocabulary->GetIndex("#"));
      
      CChartParserHypComplete::AddParseOutputSubTreeStart(newParsingOutput, pBackHyp);

      //append backpointers of pBackHyp to pNewHyp
      pNewHyp->m_backPointers.insert(pNewHyp->m_backPointers.end(),pBackHyp->m_backPointers.begin(),pBackHyp->m_backPointers.end());

      //process pBackHyp's translation output
      //this involves adding the m_translationOutput elements into the new hyp and ALSO handling any UNKNOWN words or tagged elements!
      for (CWordSequence::iterator iter3=pBackHyp->m_translationOutput.begin(); iter3!=pBackHyp->m_translationOutput.end(); ++iter3){
	if (*iter3 > 0) {
	  // *iter3 is a terminal symbol

	  // copy it over into the new translation output
	  newTranslationOutput.push_back(*iter3);
	  newParsingOutput.push_back(*iter3);

	  //add the unknown src word and the tagged output to appropriate background vectors
	  UnwindHyp_HandleUnknownAndTags(*iter3, pNewHyp, pBackHyp);
	  //pNewHyp->NoteUnknownAndTags(*iter3, pBackHyp);
       
	} else {
	  // *iter3 is a backpointer - gotta adjust it to point to its back node's new location
	  newTranslationOutput.push_back(*iter3 - pHyp->m_backPointers.size());
	  // also store the backpointer in the translation info tree
	  newParsingOutput.push_back(*iter3 - pHyp->m_backPointers.size());
	}
      }
      // copy the rest of the original hyp's translation output (the stuff after the replaced nonterminal) into the new
      // translation output
      newTranslationOutput.insert(newTranslationOutput.end(),pNewHyp->m_translationOutput.begin()+posNT+1,pNewHyp->m_translationOutput.end());
      
      //store end-of-subtree marker in translation info tree: '#)'
      //newParsingOutput.push_back(CChartParserInfo::m_pVocabulary->GetIndex("#)"));
      CChartParserHypComplete::AddParseOutputSubTreeEnd(newParsingOutput);

      //and copy over the rest of the original hyp's translation info tree
      newParsingOutput.insert(newParsingOutput.end(), pNewHyp->m_parsingOutput.begin()+posNTParse+1, pNewHyp->m_parsingOutput.end());

      // and replace the old translationOutput and translation info tree by the new ones
      pNewHyp->m_translationOutput = newTranslationOutput;
      pNewHyp->m_parsingOutput = newParsingOutput;

      // LM-process the remaining words of the new translation output
      bool noNonterminalsReachedYet = true;
      CWordSequence historyContext;
      float futureLMCost = 0.0;
      for (CWordSequence::iterator iter3=pNewHyp->m_translationOutput.begin()+posNT; iter3!=pNewHyp->m_translationOutput.end(); ++iter3){
	if (*iter3 > 0) {
	  // *iter3 is a terminal symbol

	  if (noNonterminalsReachedYet) {
	    //09/20/06 only do BOS if pHyp had it done (which is to deal with the case where we want sub-sentence n-best)
	    //09/20/06 in the normal n-best retrieval you are dealing with sentence spanning hyps, so you do it by default
	    pNewHypLMCost += CChartParserInfo::UnwindHyp_LMProcessWord(pNewHyp->m_firstWords, *iter3, pHyp->m_BOSDone);  
	  } else {

	    //new way - only do BOS if pHyp had it done
	    futureLMCost += CChartParserInfo::UnwindHyp_LMProcessWord(historyContext, *iter3, pHyp->m_BOSDone);
	  }
	} else {
	  // *iter3 is a backpointer
	  if (noNonterminalsReachedYet) {
	    // this is the first nonterminal encountered
	    // change the noNonterminalsReachedYet flag
	    noNonterminalsReachedYet = false;
	    // copy over m_firstWords as our history context to be used for the now following future cost estimations
	    historyContext = pNewHyp->m_firstWords;
	  }
	  // our nonterminal's hypergraph node
	  CChartParserNode* pNode = pNewHyp->m_backPointers[-*iter3];

	  // first add the future costs (i.e., the scaled internal LM costs) of that nonterminal node
	  // this node could potentially have BOS costs built into it - which havent been converted into
	  // real LM costs since they are in nested BP structures. We have to ensure that we dont call
	  // BOS again on this node, so we pass !pNode->m_BOSDone 
	  pNewHyp->m_scaledFutureCostEstimate += pNode->m_scaledFutureCostEstimate;
	  // process m_firstWords of the backpointer node corresponding to this nonterminal
	  for (CWordSequence::iterator iter4= pNode->m_firstWords.begin(); iter4 != pNode->m_firstWords.end(); iter4++) {
	    //new way - if the pHyp does not have m_BOSDone then there is no way a back pointer of pHyp has m_BOSDone
	    //but it is possible for pHyp->m_BOSDone == true and pHyp->m_BOSDone == false
	    if (!pHyp->m_BOSDone){
	      assert(pHyp->m_BOSDone==pNode->m_BOSDone);
	    }
	    //if pHyp is for a sub cell, then we know pHyp->m_BOSDone == false, and pNode->m_BOSDone == false, which would
	    //cause you to then ALLOW BOS calculations. But we only want this if pHyp->m_BOSDone
	    //old way
	    //futureLMCost += UnwindHyp_LMProcessWord(historyContext, *iter4, !pNode->m_BOSDone);
	    futureLMCost += CChartParserInfo::UnwindHyp_LMProcessWord(historyContext, *iter4, (!pNode->m_BOSDone)&&(pHyp->m_BOSDone));
	  }
	  // do the backpointer node's m_lastWords encompass a whole history context?
	  if (pNode->m_lastWords.size() == CChartParserInfo::C_HistoryLength) {
	    // yes - simply copy over m_lastWords as our new history context
	    historyContext = pNode->m_lastWords;
	  } else {
	    // otherwise, we can just leave our context as it is since the m_firstWords we last added are identical to
	    // the m_lastWords
	    assert(pNode->m_firstWords == pNode->m_lastWords);
	  }
	}
      }

      //EOS cost ========================================================================================
      //need to add the EOS cost ONLy if we are working with a full sentence spanning node in retrieval 
      //this condition was added in to make sure that we could use RetrieveKBestParses for nodes that 
      //are not the sentence spanning node
      if (pHyp->I()==CChartParserInfo::m_sourceLength && pHyp->J()==1){
	//CWordSequence lastContext;
	//int contextSize = 0;
	//cout<<"Future EOS for: "; CChartParserInfo::PrintVocabSequence(pNewHyp->m_firstWords); cout<<endl;
	CWordSequence workingContext;
	if (historyContext.size()!=0){
	  workingContext = historyContext;
	}else{
	  workingContext = pNewHyp->m_firstWords;
	}
	//cout<<"Future EOS historyContext: "; CChartParserInfo::PrintVocabSequence(historyContext); cout<<endl;
	//cout<<"Future EOS workingContext: "; CChartParserInfo::PrintVocabSequence(workingContext); cout<<endl;
	futureLMCost += CChartParserInfo::UnwindHyp_EOS(workingContext);
      }
      //EOS cost ========================================================================================

      //add pBackHyp's neg log probs to pNewHyp
      for (int m=0; m<CRule::C_NumModelsTotal; m++){
	pNewHyp->m_negLogProbs[m] += pBackHyp->m_negLogProbs[m];
      }

      // update hyp's LM cost and total cost
      pNewHyp->m_negLogProbs[CRule::C_LMLocation] += pNewHypLMCost;
      pNewHyp->m_totalCost += m_scalingFactors[CRule::C_LMLocation] * pNewHypLMCost;

      //adjust m_totalCost variable by the difference between pBackHyp and best hyp in pEquivalenceClass
      pNewHyp->m_totalCost += pBackHyp->m_totalCost - pEquivalenceClass->m_bestTotalCost;

      //add this hyp's explicit future LM cost to m_scaledFutureCostEstimate
      //(the implicit LM costs from the backpointer nodes have been added above already)
      pNewHyp->m_scaledFutureCostEstimate += m_scalingFactors[CRule::C_LMLocation] * futureLMCost;
	
#ifndef NDEBUG
      if (debug) {
	cout << "Adding hyp to beam:";
	pNewHyp->DisplayRecursive();
      }
#endif
      
      //when a new hyp gets formed, track the current TM, LM cost
      //make sure we do this NOW, not AFTER the FutureCostImpact has been multiplied in
      pNewHyp->AddParentHypCost(pNewHyp->ScaledTMLMCost());


      // now lessen the impact of m_scaledFutureCostEstimate to make the estimated cost for the not yet
      // LM-processed words more optimistic (same as is done in beginning of RetrieveKBestParses)
      pNewHyp->m_scaledFutureCostEstimate *= CChartParserInfo::FutureCostImpact;


      //if this is the first hyp in the back pointer node currently being unwound, then instantiating the backpointer with this hyp
      //should simply replace the backpointer with a real hyp choice, and keep all total cost the same. This is a very important
      //asssertion, if this is unclear, post to the forum
      if (CChartParserInfo::m_compareTargetWords){
	if (isFirst){
	  
	  //you cant do this assertion if the FutureCostImpact!=1.0
	  assert(CChartParserInfo::FutureCostImpact==1.0);
#ifndef NDEBUG
	  if (debug){
	    cout<<"Assertion checking savedAllCosts:"<<savedAllCosts<<" pNewHyp->m_totalCost: "<<pNewHyp->m_totalCost<<" scaled: "<<pNewHyp->m_scaledFutureCostEstimate<<endl;
	  }
#endif
	  if (!(abs(savedAllCosts-(pNewHyp->m_totalCost + pNewHyp->m_scaledFutureCostEstimate))<0.01)){
	    cout<<"Assertion checking savedAllCosts:"<<savedAllCosts<<" pNewHyp->m_totalCost: "<<pNewHyp->m_totalCost<<" scaled: "<<pNewHyp->m_scaledFutureCostEstimate<<endl;
	    exit(-1);
	  }
	  assert(fabs(savedAllCosts-(pNewHyp->m_totalCost + pNewHyp->m_scaledFutureCostEstimate))<0.01);
	  isFirst=false;
	}else{
	  //make sure that the LM costs are higher
	}
      }

      //add pNewHyp to the beam, deleting it if its m_totalCost+m_scaledFutureCostEstimate is worse than all K hyps on the beam
      //this deletion IS safe in cmp=1 mode (since all spawned off hyps are of lower cost), but in cmp=0 mode, a hyp that has
      //a relatively high m_scaledFutureCostEstimate, could result in spawned hyps of lower real LM cost.
      pNewHyp->SetBackInfo(beam.LookupProto(pHyp), pHyp);

      //if the new hyp has a lower cost than the current hyp - this can ONLY occur in cmp0 mode
      if (pNewHyp->ScaledTMLMCost()<pHyp->ScaledTMLMCost()){
	//cout<<"Unwind: lowered: "<<pNewHyp->ScaledTMLMCost()<<endl;
	CChartParserInfo::m_childBetter++;
      }

      //insert could potentially file this beam onto the elements to be expanded OR it could file it into a recomb position
      //cout<<"\t\t"; pNewHyp->PrintTranslationOutput(); cout<<" FW: "; CChartParserInfo::PrintVocabSequence(pNewHyp->m_firstWords); cout<<" Cost: "<<pNewHyp->ScaledTMLMCost()<<endl;
      beam.InsertHyp(pNewHyp);
    }
    //track all the hyps that you plan to insert as new


    //ARV 10/03 dont do this anymore
    //delete the original hyp that we just unwound - remember we erased it already from the beam above
    //delete pHyp;
    //now this pHyp is DEF stored in at least prototype back pointer, so we dont delete it, we let that
    //prototype backpoiner delete it!

  }
}
*/


//! Standard K-best retrieval based on the full-sentence node
vector<CChartParserHypComplete*> CChartParser::RetrieveKBestParses(int K){
  // make sure we have an _S hyp spanning the full sentence
  if (GetCompleteCell(CChartParserInfo::m_sourceLength,1).m_productionNodes.count(CChartParserInfo::m_FullSentenceNT)==0) {
    cout << "Error: Sentence has no translation (i.e., _S hyp). Sentence spanning chart cell contents:" << endl;
    DisplayCell(CChartParserInfo::m_sourceLength,1);
    exit(1);
  }

  // there should be only the full-sentence _S production node...
  assert(!CChartParserInfo::m_GlueAndEOSAssumptions || GetCompleteCell(CChartParserInfo::m_sourceLength,1).m_productionNodes.size() == 1);
  // ... , which only should contain the [<s> <s> <*> </s> </s>] fine-grained EC
  assert(!CChartParserInfo::m_GlueAndEOSAssumptions || GetCompleteCell(CChartParserInfo::m_sourceLength,1).m_productionNodes[CChartParserInfo::m_FullSentenceNT]->m_fineGrainedNodes.size() == 1);

  return RetrieveKBestParsesSpan(K, GetCompleteCell(CChartParserInfo::m_sourceLength,1).m_productionNodes[CChartParserInfo::m_FullSentenceNT]);
}

//!Returns the k best hyps under a particular cell - on all production nodes separately
/*
 * This method should not normally be used.
 *  There are two ways to do this 
 *   -- call k/#num prods for each of prod result then add boundary word cost and put them all together in one beam
 *   -- calculate boundary word cost for each hyp on the beam and update the model cost before putting it on the the beam
 *      and then operate as if we have a full spanning hyp during k-best - this is more involved...
 */
vector<CChartParserHypComplete*> CChartParser::RetrieveKBestParsesCell(int K, int i, int j){
 
  //pCompleteProductionNode is NOT in the chartCell
  if (i==CChartParserInfo::m_sourceLength && j==1){
    cerr << "Warning: Retrieving not only full-sentence span _S hyps, but also other full-sentence hyps. Hope you know what you're doing..." << endl;
    //    return RetrieveKBestParsesSpan(K, m_pCompleteTranslationsNode);
  }
  CCompleteHypChartCell& chartCell = m_completeHypsChart[i][j];
  vector<CChartParserHypComplete*> cellNBestList;
 
  //go thru each production node
  int fineGrainedK = MAX((int)((double)K/(double)chartCell.m_productionNodes.size()), 1);
  for (CCoarseGrainedEq::const_iterator iter = chartCell.m_productionNodes.begin(); 
       iter!= chartCell.m_productionNodes.end();
       ++iter
       ){
    CProductionNode* pNode = (*iter).second;
    vector<CChartParserHypComplete*> prodNBestList = RetrieveKBestParsesSpan(fineGrainedK, pNode);
    cellNBestList.insert(cellNBestList.end(), prodNBestList.begin(), prodNBestList.end());
  }
  return cellNBestList;
}

//!Returns the k best hyps under a particular production node
/*
 * There are two ways to do this 
 *   -- call k/#num classes for each of the equivalence classes then add boundary word cost and put them all together in one beam
 *   -- calculate boundary word cost for each hyp on the beam and update the model cost before putting it on the the beam
 *      and then operate as if we have a full spanning hyp during k-best - this is more involved...
 */
//CHypBeam* CChartParser::RetrieveKBestParsesSpan(int K, CProductionNode* pTopNode){
vector<CChartParserHypComplete*> CChartParser::RetrieveKBestParsesSpan(int K, CProductionNode* pTopNode){

  //CHypBeam* pBeam = new CHypBeam(K * CChartParserInfo::BeamSizeFactor);
  //the top level beam is K best and will handle the introduction of these hyps appropriately calling ScaledTMLMCost(), which is why
  //we need to update the costs directly before filing them into the top beam. But its very expensive to compute upto K best for
  //each fine-grained class...

  vector<CChartParserHypComplete*> prodNBestList;

  //ask for upto fineGrainedK per fine-grained equivalence class
  //note that this is only relevant if there are more than one fine grained EC (which is not the
  //case for the full-sentence _S node, which only has the [<s> <s> <*> </s> </s>] EC
  int fineGrainedK = MAX((int)((double)K/(double)pTopNode->m_fineGrainedNodes.size()), 1);
  for (CChartParserNodeContainer::iterator fineIter = pTopNode->m_fineGrainedNodes.begin(); fineIter!= pTopNode->m_fineGrainedNodes.end(); ++fineIter){
 
    CChartParserNode* pFine = fineIter->second;
    if (CChartParserInfo::m_verbosity>0){
      cout<<"Retrieving upto: "<<fineGrainedK<<" for: "; pFine->Display(); cout<<endl;
    }

    // Boundary word cost
    // When you are calculating sub-sentence spans, you need to account for the first CChartParserInfo::C_HistoryLength words
    // This cost will be zero if this is the full-sentence node or any _S subspan with j==1 since [<s> <s>] has 
    // no costs, so we don't have to worry about it.
    // If you want to use other kinds of subspans, you perhaps should add the boundary cost estimate for the first
    // CChartParserInfo::C_HistoryLength words to the LM neglob probs (don't do it as in the next line though as this assumes
    // all hyps to have the same first words, which is not true in cmp0)
    //    float boundaryCostUnscaled = CUtilities::ComputeLMCost(pFine->m_lmContext, true);
    assert(pFine->m_scaledBoundaryCost==0.0);

    vector<CChartParserHypComplete*> fineNBestList; //will contain fresh hyps that need to be deleted
    cout<<"Using m_nbestStyle: "<<CChartParserInfo::m_nbestStyle<<endl;

    if (CChartParserInfo::m_nbestStyle==3){ 

      time_t time_start, time_end;
      clock_t time_start_CPU, time_end_CPU;
      time(&time_start);
      time_start_CPU = clock();

      //CHypBeam* pFineBeam = RetrieveKBestParses(fineGrainedK, pFine);
      CHypBeam* pFineBeam;
      if (CChartParserInfo::m_breadthAStar){
	cout<<"Using A* Recombine Breadth ";	
	pFineBeam = RetrieveByAgenda(fineGrainedK, pFine);
      }else{
	cout<<"Using A* 1st best ";	
	pFineBeam = RetrieveKBestParses(fineGrainedK, pFine);
      }
      time(&time_end);
      time_end_CPU = clock();

      double NumSecsAStar = difftime(time_start, time_end);
      double NumSecsAStar_CPU = ((double) (-time_start_CPU + time_end_CPU)) / CLOCKS_PER_SEC;
      

      //dont handle the unknown words just yet
      time_start_CPU = clock();
      fineNBestList = HC05onBeam(fineGrainedK, *pFineBeam);
      time_end_CPU = clock();
      double NumSecsHC05_CPU = ((double) (-time_start_CPU + time_end_CPU)) / CLOCKS_PER_SEC;

      //print this calculation to calculate which component was spent on the A* segment 
      cout<<"TIME-Sent: NumSecsAStar: "<<NumSecsAStar<<" NumSecsAStar_CPU: "<<NumSecsAStar_CPU<<" NumSecsHC05_CPU: "<<NumSecsHC05_CPU<<endl;

      //this will perform the following operations
      // - enter the beam and delete each of the equivalence classes stored within
      // - before deleting the equivalence class, they call ClearRecombined to remove all the recombined intermediate hyps
      // - delete the eq class ALSO delete all hyps generated during the HC05 process
      //
      // arv : we dont want to do this in here.. the hc05 method gets to do all its clearing OUTSIDE      
      // delete pFineBeam;
      m_pBeamToDelete = pFineBeam;

      
    }else if (CChartParserInfo::m_nbestStyle==2){      
      cout<<"Using Huang/Chiang(2005) K-best"<<endl;
      vector<CChartParserHypComplete*> hcNBestList = pFine->RetrieveKBestHC(fineGrainedK);
      //the memory associated with these hyps will be deleted when the node is deleted, clone it here
      for (vector<CChartParserHypComplete*>::iterator iter = hcNBestList.begin(); iter != hcNBestList.end(); ++iter){
	CChartParserHypComplete* pHyp = (*iter);
	fineNBestList.push_back(new CChartParserHypComplete(*pHyp));
      }
      //memory for HC05 will get released when you delete the nodes in the parse
      
    }else if (CChartParserInfo::m_nbestStyle==1){
      cout<<"USING A* normal: (not currently supported"<<CChartParserInfo::m_extractUnique<<endl;
      exit(-1);
      //      CHypBeam* pFineBeam = RetrieveKBestParses(fineGrainedK, pFine);
      //      for (CSortedHypNBestContainer::iterator iter = pFineBeam->m_store.begin(); iter != pFineBeam->m_store.end(); ++iter){
      //	//just cast it down for now, we are only cloning a non-nest item
      //	CChartParserHypComplete* pHyp = (CChartParserHypComplete*)(*iter).second;
      //	fineNBestList.push_back(new CChartParserHypComplete(*pHyp));
      //      }
      //      delete pFineBeam; //clearing all memory associated with the n-best list process
    }else{
      cout<<"ERROR: Unknown n-best extration style, exitting"<<endl;
      cerr<<"ERROR: Unknown n-best extration style, exitting"<<endl;
      exit(-1);
    }

    //Now do all the post processing for hyps after n-best extraction


    //if we haven't done any LM calculations or we want to verify that the piecewise LM calculations 
    //were correct - do it now before calling HandleUnknownAndTags messes up the translationOutput
    if (CChartParserInfo::m_rescoreLM){
      RescoreLM(fineNBestList);
    }

    // - Call HandleUnknownAndTags which modifies the final m_translationOutput
    for (vector<CChartParserHypComplete*>::iterator iter = fineNBestList.begin(); iter != fineNBestList.end(); ++iter){
      CChartParserHypComplete* pHyp = (*iter);
      //DO IT HERE, its safe! since we know these hyps will never be referred to again in the search
      pHyp->HandleUnknownAndTags();
    }

    //now add all hyps to this production result n-best list
    prodNBestList.insert(prodNBestList.end(), fineNBestList.begin(), fineNBestList.end());
    
  }
  //now this beam should contain hyps across all the equivalence classes with the correct boundary costs
  return prodNBestList;

}


vector<CChartParserHypComplete*> CChartParser::HC05onBeam(int K, CHypBeam& beam){

  //SANITY check to make sure all hyps got filed into the same class
  //int iterationCount = 0;
  //for (; iter!= beam.m_store.end() && iter->second->m_completelyUnwound && iterationCount<K; ++iter, ++iterationCount){
  //}
  //at this point you should only have 1 element, the top most one
  //assert(iterationCount==1);
  //END SANITY check

  CSortedHypNBestContainer::iterator iter = beam.m_store.begin();    
  iter = beam.m_store.begin();
  

  CChartParserHypCompleteNBest* pTopHyp;
  //the beam might actually be empty in the m_dontReloadBeam case and the m_breadthAStar case
  if (CChartParserInfo::m_dontReloadBeam || CChartParserInfo::m_breadthAStar){
    pTopHyp = m_pFullSentenceHyp;
    assert(pTopHyp!=NULL);
  }else{
    assert(iter!=beam.m_store.end());
    pTopHyp= iter->second;
  }
  //the equivalence class for all the hyps
  CChartParserHypCompleteEqClass* pAllEq = beam.LookupProto(pTopHyp);
  assert(pAllEq!=NULL);
  CChartParserHypCompleteNBest* pTopHC05 = pAllEq->GetKthBest(K-1,true);
  //pAllEq->PrintNBest();
  
  vector<CChartParserHypComplete*> toRet;
  float lastElementCost = 9999;
  for (vector<CChartParserHypComplete*>::iterator iter= pAllEq->m_candidateVector.begin(); iter!= pAllEq->m_candidateVector.end(); ++iter){
    CChartParserHypComplete* pHyp = (*iter);
    assert(pHyp->m_scaledFutureCostEstimate==0.0);
    lastElementCost = pHyp->ScaledTMLMCost();
    //give cloned memory
    toRet.push_back(new CChartParserHypComplete(*pHyp));
  }
  //track the cost of the very last element coming out of HC05
  cout<<"Log: A* Rec + HC05 Kth Cost: "<<lastElementCost<<endl;
  return toRet;
}



void CChartParser::RescoreLM(vector<CChartParserHypComplete*>& nBestList){

  bool debug = false;

  if (CChartParserInfo::m_rescoreLM==2) {
    cout << "Sanity-test ";
  }
  cout << "LM-Rescoring" << endl;

  for (vector<CChartParserHypComplete*>::iterator iter = nBestList.begin(); iter!= nBestList.end(); ++iter){
    //CChartParserHypComplete* pHyp = (*iter).second;
    CChartParserHypComplete* pHyp = (*iter);
   

    assert(pHyp->m_scaledFutureCostEstimate == 0.0); // there are no future costs anymore, everything is settled


    /*
      cout<<"Rescoring: ";
      for (CWordSequence::iterator iter3=pHyp->m_translationOutput.begin(); iter3!=pHyp->m_translationOutput.end(); ++iter3){
      cout<<CChartParser::m_pVocabulary->GetWord(*iter3)<<" ";
      }
      cout<<endl;
    */

    //CachedSriLM::s_debugLM = true;
    //    pHyp->Display(); cout << endl; 
    if (debug){      cout<<"BeforeRescoring: "; pHyp->Display(); cout<<endl;    }

    //compute LM costs, including (in case this is not a full sentence hyp starting with <s> <s>), an estimate for
    //the first CChartParserInfo::C_HistoryLength words
    for (int i=0; i<CRule::C_NumLMs; i++) {
      float c = CUtilities::ComputeLMCost(*CLanguageModel::m_languageModels[i], pHyp->m_translationOutput, true, true);
      if (CChartParserInfo::m_rescoreLM==2 && fabs(c - pHyp->m_negLogProbs[CRule::C_LMLocation+i]) > 0.001) {
	cerr << "Sanity check error - check STDOUT for more info." << endl;
	cout << "Sanity check error: --RescoreLM=2 was passed, but for the " << (i+1) << 
	  "-th LM recomputing LM costs during Rescoring yielded a different cost (recomputedCost=" << c << " != " <<
	  pHyp->m_negLogProbs[CRule::C_LMLocation+i] << "=originalCost)! Hyp concerned:" << endl;
	pHyp->DisplayRecursive();
	exit(2);
      }
      pHyp->m_negLogProbs[CRule::C_LMLocation+i] = c;
    }

    // recompute total cost
    pHyp->m_totalCost = 0.0;
    for (int i=0; i<pHyp->m_negLogProbs.size(); i++){
      pHyp->m_totalCost += pHyp->m_negLogProbs[i]*CChartParserInfo::m_scalingFactors[i];
    }
    if (debug){
      cout<<"AfterRecoring: "; pHyp->Display(); cout<<endl;
    }
  }
  //re-sort after rescoring
  sort(nBestList.begin(), nBestList.end(), KBestCandidateMinCompare());
  reverse(nBestList.begin(), nBestList.end());
}

void CChartParser::InitializeBeam(CHypBeam& beam, CSortedHypContainer* sortedHyps){
  int counter = 0;
  cout<<"\nLog: A* Rec: TopHyps: "<<sortedHyps->size()<<endl;
  for (CSortedHypContainer::const_iterator iter = sortedHyps->begin(); iter!= sortedHyps->end(); ++iter){
    counter++;

    CChartParserHypComplete* pOldHyp = (*iter).second;
    CChartParserHypCompleteNBest* pNewHyp = new CChartParserHypCompleteNBest(*((*iter).second));

    //this is a very special case of a full-span full-sentence category complete hyp that came into existence via the
    //unknown-rule - this rule must thus have span-length 1 and thus our source sentence must also have only one word
    //(the unknown word)
    if (pNewHyp->m_translationOutput.size() == 1 && pNewHyp->m_translationOutput.front() == CChartParserInfo::m_unknownWI) {
      // save the corresponding source word
      // since this must come from an unknown-rule of the form _UNKNOWN#_UNKNOWN#@X#weights, we expect this rule to have 
      // no backpointers and a span of one
      if ((pNewHyp->m_backPointers.size() != 0) || (pNewHyp->I() != 1)) {
	cerr << "In UnwindHyp: wrong format of unknown-rule employed. Unknown-rule's hyp is: ";
	pNewHyp->DisplayRecursive(CChartParserInfo::m_pVocabulary, 0);
      }

      //parsing is indexed from 1 while the m_pSentence is indexed from 0
      assert(CChartParserInfo::m_pSentence.size()>pNewHyp->J()-1);
      pNewHyp->m_targetOrderedUnknownSourceWords.push_back(CChartParserInfo::m_pSentence[pNewHyp->J()-1]);
      assert(pNewHyp->J() == 1);
    }

    //initialize the parse output - THIS IS BEFORE ANY UNWINDING IS DONE
    CWordSequence newParsingOutput; 
    CChartParserHypComplete::AddParseOutputSubTreeStart(newParsingOutput, pNewHyp);
    
    newParsingOutput.insert(newParsingOutput.end(), pNewHyp->m_translationOutput.begin(), pNewHyp->m_translationOutput.end());
    CChartParserHypComplete::AddParseOutputSubTreeEnd(newParsingOutput);
    pNewHyp->m_parsingOutput = newParsingOutput;
    
    //make it very clear that these are the purely unwound hyps (even tho they might be all terminals)
    pNewHyp->SetBackInfo(NULL, NULL);

    //the minute these hyps are inserted they might create equivalance classes, remember these are classes based on 
    //the contet before m_posNT (which is at 0 right now), followed by remaining content
    //cout<<" Before insertion: m_active"<<pNewHyp->m_active<<endl;
    beam.InsertHyp((pNewHyp));
  }
}

CHypBeam* CChartParser::RetrieveByAgenda(int K, CChartParserNode* pTopFineNode){
#ifndef NDEBUG
  bool debug = true;
#endif
   //clear the stats for this retrieval round
  CChartParserInfo::ResetNBestStatistics();  
  CSortedHypContainer* sortedHyps = &(pTopFineNode->m_hypContainer);
  CHypBeam* pBeam = new CHypBeam(K * CChartParserInfo::BeamSizeFactor);
  CHypBeam& beam = *pBeam;
  InitializeBeam(beam, sortedHyps); //beam now has all the starter elements in it
  m_pFullSentenceHyp = NULL; //we will convey the first best thru this
#ifndef NDEBUG
  if (debug){
    cout<<"Initialized beam size:"<<beam.m_store.size()<<endl;
  }
#endif  
  bool done = false;
  while (!done){
    vector<CChartParserHypCompleteNBest*> agenda = beam.GetAgenda(CChartParserInfo::m_aStarBeam);
    vector<CChartParserHypCompleteNBest*>::iterator iter = agenda.begin();
    //skip past all the elements that are commpletely unwound!
    int iterationCount = 0;
#ifndef NDEBUG
    if (debug){
      cout<<"Agenda size:"<<agenda.size()<<endl;
    }
#endif      
    if (agenda.size()==0){
      assert(m_pFullSentenceHyp!=NULL);
      return pBeam;
    }
    CChartParserHypCompleteNBest* pBestHyp = (*iter);
    float lowestCost = pBestHyp->ScaledTMLMCost(); //this is the lowest cost on the agenda
#ifndef NDEBUG
    if (debug){
      cout<<"LowestCost: "<<lowestCost<<endl;
    }
#endif  

    //move thru the completely unwound elements
    for (; iter!= agenda.end() && (*iter)->m_completelyUnwound; ++iter, ++iterationCount){
    }

    //this means all the elements on the agenda are totally unwound
    if (iter==agenda.end()){
      //the agenda is stocked with a set of completely unwound items. Files they guys onto the real
      //beam, they will all go into the same class
      //for (vector<CChartParserHypCompleteNBest*>::iterator it = agenda.begin(); it!=agenda.end(); ++it){
      //assert((*it)->m_completelyUnwound);
      //beam.InsertHyp();
      //} 
      assert(agenda.size()>0);
      m_pFullSentenceHyp = agenda[0]; //setting m_pFullSentenceHyp allows the HC05onBeam to start
      done = true;
    }else{
    //now the iter is at the first hyp that we should unwind..lets unwind until the end!
      for (; iter!= agenda.end(); ++iter){
	//children within the lowestCost + m_agendaBeam get a chance on the beam
	//children above the lowestCost + m_astarbeam terminates the unwinding process
	CChartParserHypCompleteNBest* pCurrentHyp = (*iter);
	if (!pCurrentHyp->m_completelyUnwound){
	  CChartParserHypCompleteNBest* pFullTmp = UnwindHyp(pCurrentHyp, beam, lowestCost); 
	  if (pFullTmp!=NULL){
	    m_pFullSentenceHyp = pFullTmp;
	  }
	  pCurrentHyp->m_active = false; // this ensures that this hyp wont get back on the agenda		  
	}
	//if the hyp is unwound and  on the agenda, it would not be bumped off
	//to get online updates, you have to look at the real beam now, not the agenda
	//lowestCost = ((beam.m_store.begin()))->second->ScaledTMLMCost();
	//arv: we no longer store anything sorted... so we need a variable instead
	lowestCost = beam.GetLowestCost();
      }
    }
#ifndef NDEBUG
    if (debug) {
      cout<<"\n\nBeam after expansion ==================== "<<endl;  
      beam.PrintBeam();
    }
#endif  
  }
  
  
  return pBeam;
}

//! returns a pointer to a newly created hyp beam of the K best completely unwound translation hypotheses. 
/*! 
  An A* style lazy unwinding algorithm
  Don't forget to delete the returned beam! The beam contains only
  freshly created hyps. These will all be deleted by the beam's destructor.
*/
CHypBeam* CChartParser::RetrieveKBestParses(int K, CChartParserNode* pTopFineNode) {
#ifndef NDEBUG
  bool debug = true;
#endif
  //clear the stats for this retrieval round
  CChartParserInfo::ResetNBestStatistics();
  
  CSortedHypContainer* sortedHyps = &(pTopFineNode->m_hypContainer);

  //our beam: a multimap of complete hyps sorted by their totalcosts with an overridden insert function that only keeps upto K elements in the beam
  // if we correctly separate equivalence classes by first words and last words, we only need to keep track of the best
  // K hyps on the beam at any time.
  // If we are in CompareTargetWords==false though, the 'best cost' value for each EC is only an estimate, and so it is 
  // better to keep track of the best K * BeamSizeFactor hyps
  // the same applies when we do uniquing while K-best extracting (ExtractUnique=1) because then we're constantly deleting
  // undesired hyps from the beam, thus making it smaller than K
  CHypBeam* pBeam = new CHypBeam(K * CChartParserInfo::BeamSizeFactor);
  CHypBeam& beam = *pBeam;

  //the full sentence spanning hyp
  CChartParserHypCompleteNBest* pFullHyp = NULL;

  int counter = 0;
  //the set of hyps that are sentence spanning
  //we used to restrict to only pull up the first K elements, but this is not the point of the astar search, we want to keep
  //pulling until we are happy with the result
  //for (CSortedHypContainer::const_iterator iter = sortedHyps->begin(); iter!= sortedHyps->end() && counter < K; ++iter){

  cout<<"\nLog: A* Rec: TopHyps: "<<sortedHyps->size()<<endl;
  for (CSortedHypContainer::const_iterator iter = sortedHyps->begin(); iter!= sortedHyps->end(); ++iter){
    counter++;

    CChartParserHypComplete* pOldHyp = (*iter).second;

#ifndef NDEBUG
    if (debug){
      cout<<"Copying from old: "; pOldHyp->PrintTranslationOutput(); cout<<" m_taggedTargets: "; CChartParserInfo::PrintVocabSequence(pOldHyp->m_taggedTargets); cout<<endl;
    }
#endif
    //clone retrieved hyp (b/c we will delete all the elements on the beam and don't want to try delete things twice)
    //CChartParserHypComplete* pNewHyp = new CChartParserHypComplete((*iter).second);

    CChartParserHypCompleteNBest* pNewHyp = new CChartParserHypCompleteNBest(*((*iter).second));

    //pNewHyp has its m_posNT correctly set, but its the job of unwind hyp to clone this hyp and make it completely unwound
    //if (pNewHyp->m_posNT==pNewHyp->m_translationOutput.size()){
    //pNewHyp->m_completelyUnwound=true;
    //}


    //arv: this parent cost tracking doesnt quite work
    //make sure to do this BEFORE multiplying in the FutureCostImpact
    //this is just for tracking the hyp as it unrolls
    //pNewHyp->AddParentHypCost(pNewHyp->ScaledTMLMCost());

    //arv: this was the old style. now the cost is kept purely in the m_totalCost
    // the current RetrieveKBestParses and UnwindHyp will re-calculate the LM from scratch and keep adding 
    // the so-far-LM-processed words' LM costs to m_negLogProbs[CRule::C_LMLocation]
    // m_scaledFutureCostEstimate will contain the optimistic estimate of LMCost*FutureCostImpact for the not-yet
    // LM-recalculated words
    //pNewHyp->m_scaledFutureCostEstimate *= CChartParserInfo::FutureCostImpact;
    //assert(pNewHyp->m_negLogProbs[CRule::C_LMLocation] == 0.0);
    

    //arv: no such thing as first words
    // initialize m_firstWords for use in UnwindHyp
    // m_firstWords will now be abused to contain all words of the translationOutput that have been processed so far
    //pNewHyp->m_firstWords.clear();

    //this is a very special case of a full-span full-sentence category complete hyp that came into existence via the
    //unknown-rule - this rule must thus have span-length 1 and thus our source sentence must also have only one word
    //(the unknown word)
    if (pNewHyp->m_translationOutput.size() == 1 && pNewHyp->m_translationOutput.front() == CChartParserInfo::m_unknownWI) {
      // save the corresponding source word
      // since this must come from an unknown-rule of the form _UNKNOWN#_UNKNOWN#@X#weights, we expect this rule to have 
      // no backpointers and a span of one
      if ((pNewHyp->m_backPointers.size() != 0) || (pNewHyp->I() != 1)) {
	cerr << "In UnwindHyp: wrong format of unknown-rule employed. Unknown-rule's hyp is: ";
	pNewHyp->DisplayRecursive(CChartParserInfo::m_pVocabulary, 0);
      }

      //parsing is indexed from 1 while the m_pSentence is indexed from 0
      assert(CChartParserInfo::m_pSentence.size()>pNewHyp->J()-1);
      pNewHyp->m_targetOrderedUnknownSourceWords.push_back(CChartParserInfo::m_pSentence[pNewHyp->J()-1]);
      assert(pNewHyp->J() == 1);
    }

    //add the hyp to the beam
#ifndef NDEBUG
    if (debug) {
      cout << "Adding to beam the full-sentence spanning hyp: \n";
      pNewHyp->DisplayRecursive();
      cout << endl;
    }
#endif
    
    //initialize the parse output - THIS IS BEFORE ANY UNWINDING IS DONE
    CWordSequence newParsingOutput; 
    CChartParserHypComplete::AddParseOutputSubTreeStart(newParsingOutput, pNewHyp);
    
    newParsingOutput.insert(newParsingOutput.end(), pNewHyp->m_translationOutput.begin(), pNewHyp->m_translationOutput.end());
    CChartParserHypComplete::AddParseOutputSubTreeEnd(newParsingOutput);
    pNewHyp->m_parsingOutput = newParsingOutput;
    
    //adding to beam
    //cout<<"Adding to beam: "; pNewHyp->Display(); cout<<endl;
    //CChartParserInfo::PrintVocabSequence(pNewHyp->m_parsingOutput); cout<<endl;

    //make it very clear that these are the purely unwound hyps (even tho they might be all terminals)
    pNewHyp->SetBackInfo(NULL, NULL);
    //the minute these hyps are inserted they might create equivalance classes, remember these are classes based on 
    //the contet before m_posNT (which is at 0 right now), followed by remaining content
    //cout<<" Before insertion: m_active"<<pNewHyp->m_active<<endl;
    beam.InsertHyp((pNewHyp));
    //cout<<" After insertion: m_active"<<pNewHyp->m_active<<endl;
  }

  //cout<<"Unwinding hyps..."<<endl;
  bool done = false;
  cout<<"Log: A* Rec: Init: "<<beam.m_store.size()<<endl;
#ifndef NDEBUG
    if (debug) {
      cout<<"\n\nBeam INIT ==================== "<<endl;  
      beam.PrintBeam();
    }
#endif  

  vector<float> lowestUnwoundCosts;
  vector<float> lowestAllCosts;
  while (!done) {
    
    //iterate through the beam through increasingly costly hyps and find the best hyp that is still not completely unwound
    //if all hyps are completely unwound, we are done
    //CSortedHypContainer::iterator iter = beam.m_store.begin();

    CSortedHypNBestContainer::iterator iter = beam.m_store.begin();
    if (iter!=beam.m_store.end()){
      CChartParserHypCompleteNBest* pTopHyp = iter->second;
      lowestAllCosts.push_back(pTopHyp->ScaledTMLMCost());
    }


    int iterationCount = 0;
    //look for the lowest cost unwound hyp
    float worstCompleteHyp = 9999;
    float lowestCostHyp = 9999;
    //for (; iter!= beam.m_store.end() && iter->second->m_completelyUnwound && iterationCount<K; ++iter, ++iterationCount){

    //remove the restriction the we only go past the first K elements, we want to end up at the next
    //unwound hyp regardlesss if the first K have been fully un-wound
    for (; iter!= beam.m_store.end() && iter->second->m_completelyUnwound; ++iter, ++iterationCount){
      if (iterationCount==0){
	lowestUnwoundCosts.push_back(iter->second->ScaledTMLMCost());
	lowestCostHyp = iter->second->ScaledTMLMCost();
      }
      assert(iter->second->m_scaledFutureCostEstimate==0.0);
      worstCompleteHyp = iter->second->ScaledTMLMCost();
    }
    //post-condition: worstCompleteHyp holds the cost 

#ifndef NDEBUG
    if (debug){
      cout<<"Popping beam iterationCount: "<<iterationCount<<" size: "<<beam.m_store.size()<<endl;
    }
#endif


    //if the beam is empty, lets terminate. 
    if (iter==beam.m_store.end()) {
      //yes, so we're done
      done = true;

      //we should indicate the termination reason
      cout<<"Log: A* Rec: ";
      if (iter==beam.m_store.end()){
	cout<<" beam empty : iterationCount:"<<iterationCount;
      }
      if (iterationCount == K){
	cout<<" first k unwound ";
      }
      cout<<endl;

      //remove the hyps beyond the first K from the beam
      //only reason to do this is to save memory before running hc05
      //beam.Shrink(K);
    } else {

      //pop the top element of
      CChartParserInfo::m_topCallsKBest++;
      //no, so unwind hypothesis iter->second
      CChartParserHypCompleteNBest* pTopHyp = iter->second;
      //cout<<"Unwinding: "; pTopHyp->PrintTranslationOutput(); cout<<" Cost: "<<pTopHyp->ScaledTMLMCost()<<endl;
      
      //quit if the next hyps cost is much higher
      if (pTopHyp->ScaledTMLMCost() > CChartParserInfo::GetPruningCutoff(lowestCostHyp, CChartParserInfo::m_aStarBeam) ){

	//display the cost of the rejected item to compare it to HC05's n-best extraction
	//we want the K'th best 
	cout<<"Log: A* Rec: Rejected Item Cost: "<<pTopHyp->ScaledTMLMCost()<<endl;
	//so we're done
	done = true;
      }else{	

	//the beam could be made smaller here, because you could have found a not complete hyp that has been deactivated and delete him
	//we use the MIN because lowestCostHyp is potentially undefined (since its defined by the existence of m_completelyUnwound hyps)
	//at the top of the beam. 
	CChartParserHypCompleteNBest* toUnwind = iter->second;
	beam.m_store.erase(iter);
	CChartParserHypCompleteNBest* pFullHypTmp = UnwindHyp(toUnwind, beam, MIN(pTopHyp->ScaledTMLMCost(), lowestCostHyp));
	if (pFullHypTmp!=NULL){
#ifndef NDEBUG
    if (debug) {
      cout<<"Setting COMPLETE: "; pFullHypTmp->PrintTranslationOutput(); cout<<endl;
    }
#endif
	  pFullHyp = pFullHypTmp;
	}

      }

    }

#ifndef NDEBUG
    //if (debug) {
    //cout<<"\n\nBeam after ==================== "<<endl;
    //beam.PrintBeam();
      
    //cout<<"\n\nEQ after ==================== "<<endl;
    //beam.PrintClasses();
    //}
#endif
  }
#ifndef NDEBUG
  if (debug) {
    cout<<"\n\nBeam astar end ==================== "<<endl;
    beam.PrintBeam();
    
    cout<<"\n\nEQ astart end  ==================== "<<endl;
    beam.PrintClasses();
  }
#endif
   
  CChartParserInfo::m_numUnique = pBeam->m_uniqueChecker.Size();
  cout<<"Log: NBest: "; CChartParserInfo::PrintNBestStatistics(); cout<<endl;
  //cout<<"Log: NBestLowestUnwound: "; copy(lowestUnwoundCosts.begin(), lowestUnwoundCosts.end(), ostream_iterator<float>(cout, " ")); cout<<endl;
  //cout<<"Log: NBestLowestAll: "; copy(lowestAllCosts.begin(), lowestAllCosts.end(), ostream_iterator<float>(cout, " ")); cout<<endl;

  
  //set the hyp that was sentences spanning
  m_pFullSentenceHyp=NULL;
  m_pFullSentenceHyp = pFullHyp;
  assert(m_pFullSentenceHyp!=NULL);
  //return the created upto-K-best beam of completely-unwound hyps
  return pBeam;
}


//outputs the Huang and Chiang N-Best list
/*
  void CChartParser::OutputHC(vector<CChartParserHypComplete*> candidateSet, int translationCount){
  
  int count = 0;
  for (vector<CChartParserHypComplete*>::iterator iter=pCompleteProductionNodeFine->m_candidateVector.begin(); count<translationCount && iter!=pCompleteProductionNodeFine->m_candidateVector.end(); ++count, ++iter){
  CChartParserHypComplete* pTopHyp = (*iter);
  //cout<<"RETRIEVED: "; pTopHyp->DisplaySimpleKBest(); cout<<endl;
    
  }
  }
*/

//! iterate through the beam from best hyp to worst (assuming all hyps
//are completely unwound and unknown-handled) and output the hyp's translations
//void CChartParser::OutputBeam(CHypBeam* pBeam, int translationCount) {
void CChartParser::OutputNBest(vector<CChartParserHypComplete*>::iterator first, vector<CChartParserHypComplete*>::iterator last, int translationCount) {

  for (vector<CChartParserHypComplete*>::iterator iter = first; translationCount-- != 0 && iter!= last; ++iter){
    //CChartParserHypComplete* pHyp = (*iter).second;
    CChartParserHypComplete* pHyp = (*iter);
    cout<<endl;
    cout<<"ScaledTMLMCost: "<<pHyp->ScaledTMLMCost()<<" m_totalCost: "<<pHyp->m_totalCost<<" raw:";
    for (int i=0; i<pHyp->m_negLogProbs.size(); i++){
      cout<<" "<<pHyp->m_negLogProbs[i];
    }
    cout << " UNKNOWN-TAGGED-handled translation: ";
    //output the parse tree too
    CChartParserInfo::PrintVocabSequence(pHyp->m_translationOutput);
    cout<<" PARSE: "; CChartParserInfo::PrintParsingOutput(pHyp->m_parsingOutput); 

  }
  cout<<endl;

}







void CChartParser::PruneCell(const int i, const int j){
  
  bool debug = false;


  //CCompleteHypChartCell& chartCell = m_completeHypsChart.GetRow(i)[j];
  CCompleteHypChartCell& chartCell = m_completeHypsChart[i][j];
  if (debug){
    cout<<"====== Before Pruning ==== "<<endl;
    DisplayCell(i, j);
  }
  vector<CProductionNode*> allProductionNodes; 
  float lowestNTNodeCost = FLT_MAX;
  //fill the allProductionNodes and also determine lowest All-Component-Cost across all noterminal production nodes
  for (CCoarseGrainedEq::const_iterator iter = chartCell.m_productionNodes.begin(); 
       iter!= chartCell.m_productionNodes.end();
       ++iter
       ){
    CProductionNode* node = (*iter).second;
    allProductionNodes.push_back(node);
    if (!node->m_lexical) {
      lowestNTNodeCost = MIN(lowestNTNodeCost,node->m_pLowestCostHyp->AllComponentCost());
    }
  }
  //az: sorting unnecessary, we'll iterate through all of them anyway
  //  sort(allProductionNodes.begin(), allProductionNodes.end(), CompareProductionNode());

  if (allProductionNodes.size()==0){
    return;
  }


  double safeCutoffLogProb = CChartParserInfo::GetPruningCutoff(lowestNTNodeCost, CChartParserInfo::m_maxCostDifferencePerCell);


  //clear out what was left this these nodes
  chartCell.m_productionNodes.clear();

  assert(safeCutoffLogProb!=0.0);
  pair<int, int> prodKeptAll(0, 0);
  //  vector<CProductionNode*> toDelete;
  for (vector<CProductionNode*>::iterator iter= allProductionNodes.begin(); iter!= allProductionNodes.end(); ++iter){
    CProductionNode* pNode = (*iter);

    if (pNode->m_pLowestCostHyp->AllComponentCost() > safeCutoffLogProb ){
      //as long as you are not a glue or full sentnence node
      if (!pNode->m_lexical && pNode->GetResult()!=CChartParserInfo::m_GlueNT &&  pNode->GetResult()!=CChartParserInfo::m_FullSentenceNT){
	if (CChartParserInfo::m_verbosity>1){
	  cout<<"Deleting: "; pNode->Display(); cout<<" lowest: ";pNode->m_pLowestCostHyp->AllComponentCost(); cout<<endl;
	}
	//	toDelete.push_back(pNode);
	//delete the entry from the m_newlyCreatedNodes
	CCoarseGrainedEq::iterator toEraseIter = m_newlyCreatedNodes.find(pNode->GetResult());
	//if its also a newly created node, then erase it from that map
	//az: if it's not newly created, we shouldn't delete it b/c the Self rule might already have created hyps
	//that depend on it
	if (toEraseIter!=m_newlyCreatedNodes.end()){
	  m_newlyCreatedNodes.erase(toEraseIter);
	  delete pNode;
	} else {
	  chartCell.m_productionNodes[pNode->GetResult()] = pNode;
	  prodKeptAll.first++;
	}
      }else{
	//this means you have lowest cost but are a @_G or @_S
	//cout<<"Tried to delete glue or s, but we need to keep it: "; pNode->Display(); cout<<endl;
	chartCell.m_productionNodes[pNode->GetResult()] = pNode;
	prodKeptAll.first++;
      }
    }else{
      chartCell.m_productionNodes[pNode->GetResult()] = pNode;
      prodKeptAll.first++;
    }
    prodKeptAll.second++;
  }
  //   //now go thru and delete the ones that were just bad
  //   for (vector<CProductionNode*>::iterator iter= toDelete.begin(); iter!= toDelete.end(); ++iter){
  //     CProductionNode* pNode = (*iter);
  //     assert(pNode->GetResult()!=CChartParserInfo::m_GlueNT && pNode->GetResult()!=CChartParserInfo::m_FullSentenceNT);
  //     //delete the entry from the m_newlyCreatedNodes
  //     CCoarseGrainedEq::iterator toEraseIter = m_newlyCreatedNodes.find(pNode->GetResult());
  //     //if its also a newly created node, then erase it from that map
  //     //az: if it's not newly created, we shouldn't delete it b/c the Self rule might already have created hyps
  //     //that depend on it
  //     if (toEraseIter!=m_newlyCreatedNodes.end()){
  //       m_newlyCreatedNodes.erase(toEraseIter);
  //       delete pNode;
  //       (*iter) = NULL;
  //     }
  //     //    delete pNode;
  //     //    (*iter) = NULL;
  //   }

  cout<<"Log: PruneCell production nodes kept: "<<prodKeptAll.first<<" / "<<prodKeptAll.second<<" safeCutoffLogProb: "<<safeCutoffLogProb<<endl;
}


int CChartParser::FinalizeAllNodes(const CCoarseGrainedEq& productionNodes, bool addAll){
  int numEqClasses = 0;
  for (CCoarseGrainedEq::const_iterator iter = productionNodes.begin(); iter!= productionNodes.end(); ++iter){
    CProductionNode* node = (*iter).second;
    node->FinalizeProductionNode(addAll);
    //cout<<"Finalized: "; node->Display();cout<<endl;
    numEqClasses += node->m_fineGrainedNodes.size();
  }
  //cout<<"Finalizing: "; m_pCompleteTranslationsNode->Display();cout<<endl;
  //  m_pCompleteTranslationsNode->FinalizeProductionNode(addAll);
  //note, there should only be 1 eq class here
  //  numEqClasses += m_pCompleteTranslationsNode->m_fineGrainedNodes.size();
  return numEqClasses;
}




void CChartParser::ParseChart(){
  // main CKY+ loop
  //int i,j; //i is the length of the produced rule, and j is the starting word index of that rule
  for (int i=2; i<=CChartParserInfo::m_sourceLength; i++){
    cout << "Log: ParseChart(): Building completed hyps of size i=" << i << "... " << endl;
    int skippingCount = 0;
    int total_count_EC_IH_pairs = 0;

    //track the span - limit rules based on this span
    m_combinationCount = i;

    //bool maxCombinationCountExceeded = m_combinationCount > CChartParserInfo::m_maxCombinationCount;
    //now do the actual combinations
    if (m_combinationCount == CChartParserInfo::m_maxCombinationCount+1){
      cout << endl << endl << "MaxCombinationCount " << CChartParserInfo::m_maxCombinationCount << " exceeded." << endl << endl;
      //CChartParserInfo::s_maxCombinationCountExceeded = true;
    }
    //left side index
    //az: we used to have an optimization here that creates only cells with j=1 once the max combi count has been exceeded
    //this opti ruled out combining lexial rules spanning more than max-combi-count words, so I commented it out again
    //but made sure that in fundamental hyp only cells with j=1 are explored for which complete hyp and incomplete hyp
    //are purely lexical, so basically the opti has been moved into ApplyFundamentalRule
    //    int jIndex = m_combinationCount > CChartParserInfo::m_maxCombinationCount ? 1 : CChartParserInfo::m_sourceLength-i+1;
    for (int j=1; j<=CChartParserInfo::m_sourceLength-i+1; j++){
      //    for (int j=1; j<=jIndex; j++){

      //try to apply the fundamental rule to fill up cell i,j with new
      //complete hyps


      //but first pre-compute the best rule app count we will achieve over
      //all our new complete hyps combinations if we just glue them together
      // We base our computation on adjacent complete hyps, as we don't
      // know yet which incomplete hyps will combine with which complete
      // hyps. 
      // bestCombinedRuleAppCount will always be achieved, since for
      // CHs X and Y, gluing to an _S hyp can take place 
      // costing only one additional rule app count

      /* here's how rule app count works now for an example:

      Example:

      <s> the ass </s>
      the#DT, ass#NN, the ass#NP, DT NN#NP

      i\j         1        2           3    4
      1          <s>      the          ass   </s>
      _S(1)   DT(1)        NN(1)
      _S.(1)  NP.(0) (from 'the')
      NP.(1) (from 'DT')

      2:RAC_min= 1+1+1    1+1+1        1+0+1
      _S(3)    NP(1) 
      _S.(3)   NP(3)
           
      3:RAC_min= 1+1+1
      _S(3) (combining _S.[1,1] with NP[2,2])
      _S.(3)
                    
      4:RAC_min= 3+0+1
      _S(4)

      */

      float bestCombinedRuleAppCount = FLT_MAX/2; // divide by 2 to
      // avoid overflow
      // when adding
      // stuff
      int k_max=-1;
      for (int k=1; k<i; k++) { // k: length of source (left) cell
	// the ECs (hypergraph nodes) in the source cell
	CRuleAppCountSortedNodeIndex& eqClasses1 = m_completeHypsChart[k][j].ruleAppCountSortedNodeIndex;
	//	int totalEC1Count = eqClasses1.size();

	// the ECs (hypergraph nodes) in the target cell
	CRuleAppCountSortedNodeIndex& eqClasses2 = m_completeHypsChart[i-k][j+k].ruleAppCountSortedNodeIndex;
	//	int totalEC2Count = eqClasses2.size();

	
	// is there anything to combine at all?
	//	if (totalEC1Count>0 && totalEC2Count>0) {
	//yes

	//az: we used to check only for _S as the left-side cell's hyps. Now, we don't have _S hyps anymore
	//except in cells with j=1, so we have to check all hyps and add 1 b/c of the transition to glueing,
	//which costs 1 rule-app-count

	// determine best value of m_bestTotalRuleApplicationCount amongst all
	// hyps in the left cell
	float bestRuleAppCountEC1Iteration;

	//	  if (GetCompleteCell(k,j).m_cellType == 1) {
	//	    // commented out as for that case we have a complete _S hyp on top of the j==1 BOS hyp
	//	    // this means we have a BOS cell k=1, j=1
	//	    assert(k==1); assert(j==1);
	//	    // the rule app count for this cell will be 0 as it contains the terminal hyp <s> which will be subsumed by
	//	    // the glue rule
	//	    bestRuleAppCountEC1Iteration = 0;
	//	  } else 
	{
	  //cout<<"Checking cell: "<<k<<","<<j<<" with "<<i-k<<","<<j+k<<endl;
	  // in case of j==1, check for complete _S hyp
	  CProductionNode* pClass1;
	  if (j==1 && CChartParserInfo::m_GlueAndEOSAssumptions) {
	    CRuleAppCountSortedNodeIndex::iterator itClasses1;
	    for (	itClasses1 = eqClasses1.begin(); itClasses1!=eqClasses1.end(); ++itClasses1){
	      pClass1 = itClasses1->second;
	      //cout<<"c1 option: "; pClass1->Display(); cout<<endl;
	      if (pClass1->GetResult()==CChartParserInfo::m_FullSentenceNT) {
		break;
	      }
	    }
	    if (itClasses1 == eqClasses1.end()) {
	      cerr << "Grammar's Glue rule is not working. Since j==1, the _S hyp should have been in the cell" << endl;
	      cout << "Grammar's Glue rule is not working. Since j==1, the _S hyp should have been in the cell" << endl;
	      assert(false);
	    } else {
	      bestRuleAppCountEC1Iteration = pClass1->m_bestTotalRuleApplicationCount;
	    }
	  } else {
	    // our grammar has direct glue rules: @_S @N # @1 @2 # @_S
	    // thus we'll use the very first (i.e., minimum-rule-app-count) complete hyp 
	    // that is not a terminal symbol hyp
	    CRuleAppCountSortedNodeIndex::iterator itClasses1;
	    for (	itClasses1 = eqClasses1.begin(); itClasses1!=eqClasses1.end(); ++itClasses1){
	      pClass1 = itClasses1->second;
	      //cout<<"c1 option: "; pClass1->Display(); cout<<endl;
	      if (!pClass1->m_lexical) {
		break;
	      }
	    }
	    if (itClasses1 == eqClasses1.end()) {
	      // found no non-lexical hyp
	      assert(eqClasses1.begin()==eqClasses1.end()); // then there should be no hyps at all
	      // set best possible RAC to the best RAC that we can get by gluing subcells together to get 
	      // a translation spanning this cell k,j
	      bestRuleAppCountEC1Iteration = GetCompleteCell(k,j).m_bestGluedRuleAppCount;
	    } else {
	      // MIN of: 
	      // best value of m_bestTotalRuleApplicationCount amongst all hyps in the left cell
	      // and best RAC that we can get by gluing subcells together to get 
	      // a translation spanning this cell k,j
	      bestRuleAppCountEC1Iteration = MIN(GetCompleteCell(k,j).m_bestGluedRuleAppCount,
						 pClass1->m_bestTotalRuleApplicationCount);
	    }
	  }
	}
	  
	// now determine best value of m_bestTotalRuleApplicationCount amongst all
	// hyps in the right cell
	float bestRuleAppCountEC2Iteration;

	if (GetCompleteCell(i-k,j+k).m_cellType == 2) {
	  // this means we have a EOS cell
	  assert(i-k==1); assert(j+k==CChartParserInfo::m_sourceLength);
	  // the rule app count for this cell will be 0 as it contains the terminal hyp </s> which will be subsumed by
	  // the glue rule
	  bestRuleAppCountEC2Iteration = 0;
	} else {
	  // our grammar has direct glue rules <s> @_S @_N # <s> <s> @1 @2 # @_S
	  // thus we'll use the very first (i.e., minimum-rule-app-count) complete hyp 
	  // that is not a terminal symbol hyp
	  CProductionNode* pClass2;
	  CRuleAppCountSortedNodeIndex::iterator itClasses2;
	  for (	itClasses2 = eqClasses2.begin(); itClasses2!=eqClasses2.end(); ++itClasses2){
	    pClass2 = itClasses2->second;
	    //cout<<"c2 option: "; pClass2->Display(); cout<<endl;
	    if (!pClass2->m_lexical) {
	      break;
	    }
	  }
	  float minRAC2 = (itClasses2 == eqClasses2.end()) ?
	    // found no non-lexical hyp
	    FLT_MAX/10
	    :
	    pClass2->m_bestTotalRuleApplicationCount;
	  bestRuleAppCountEC2Iteration = (j==1)?
	    // left cell used for the RAC calculation is a glue _S node
	    // so for this right cell always use the actual bestRuleAppCount since
	    // m_bestGluedRuleAppCount (if better) cannot be achieved due to pruning
	    minRAC2
	    :
	    // MIN of: 
	    // best value of m_bestTotalRuleApplicationCount amongst all hyps in the right cell
	    // and best RAC that we can get by gluing subcells together to get 
	    // a translation spanning this cell i-k,j+k
	    bestRuleAppCountEC2Iteration = MIN(minRAC2,GetCompleteCell(i-k,j+k).m_bestGluedRuleAppCount);
	}
	//cout<<"checking *** "<<endl;
	//cout<<"long cell: "; pClass1->Display();
	//cout<<"short cell: "; pClass2->Display();
	  
	if (bestRuleAppCountEC1Iteration + bestRuleAppCountEC2Iteration + 1 < bestCombinedRuleAppCount) {
	  // gluing left-cell EC1 and right-cell EC2 together (+1 RAC b/c of the use of the glue rule)
	  // sets a new record RAC for this combined cell
	  bestCombinedRuleAppCount = bestRuleAppCountEC1Iteration + bestRuleAppCountEC2Iteration + 1;
	  //cout<<" *** re-setting? **** ";
	  //cout<<endl;
	  k_max=k;
	}

      }//go thru each k
      
      if (k_max==-1){
	//if we are completely unable to find any matching hyps
	cout<<"ERROR: Unable to find any glue rule style matching for cell: "<<i<<","<<j<<endl;
	exit(-1);
      }
      // set this cell's best RAC achievable by gluing subcells
      GetCompleteCell(i,j).m_bestGluedRuleAppCount = bestCombinedRuleAppCount;


      //now that we know what the bestCombinedRuleAppCount is, do the actual ApplyFundamentalRule application

      //if (CChartParserInfo::m_verbosity>0){
      cout<<"Building Cell: i="<<i<<" j="<<j<<" bestCombinedRuleAppCount: "<<bestCombinedRuleAppCount<<" at: ("<<k_max<<","<<j<<") ("<<i-k_max<<","<<j+k_max<<") "<<endl;
      //}
      //Reset the expansion statistics
      CChartParserInfo::ResetCellStatistics();

      //combo pruning across values of k
      RulesByProductionResult hypRuleCombinations;
      vector<CChartParserHypIncomplete*> incompleteHypsToDelete;
      
      assert(CChartParserInfo::m_hypsAdded==0);
      for (int k=1; k<i; k++){ // k: length of source (left) cell; i-k: length of target (right) cell
	if (CChartParserInfo::m_verbosity>0){
	  cout<<"Calling ApplyFundamentalRule src: ["<<k<<","<<j<<"] ["<<i-k<<","<<j+k<<"] bestCombinedRuleAppCount: "<<bestCombinedRuleAppCount<<endl;
	}

	pair<int,int> tmppair = 
	  ApplyFundamentalRule(k,j, (i-k), (j+k),bestCombinedRuleAppCount, hypRuleCombinations, incompleteHypsToDelete);
	// just some statistics
	skippingCount += tmppair.first;
	total_count_EC_IH_pairs += tmppair.second;
	if (CChartParserInfo::m_verbosity>0){
	  cout<<"Skipped: "<<tmppair.first<<" out of "<<tmppair.second<<endl;
	}
      }

      //spawn out any complete hyps in global mode
      if (CChartParserInfo::m_comboPruningGlobal){
	if (hypRuleCombinations.size()>0){
	  //now we have built up a series of CChartParserHypIncomplete - rule vector pairs, that we can process from the top down
	  pair<int, int> expandedAndPossible = SpawnKBestCompleteHyps(hypRuleCombinations, false);
	  //if (CChartParserInfo::m_verbosity>0){
	  //cout<<"Log: ExpPossGlobal: "<<expandedAndPossible.first<<" / "<<expandedAndPossible.second<<endl;
	  //}
	  hypRuleCombinations.clear();
	}
      } else {
	// not in global mode - this means we have done the Spawning already in ApplyFundamentalRule
      }
      //and delete the incomplete hyps that had no more matches
      for (vector<CChartParserHypIncomplete*>::iterator dIter = incompleteHypsToDelete.begin(); dIter!= incompleteHypsToDelete.end(); ++dIter){
	CChartParserHypIncomplete* pNewIncompleteHypPrototype = (*dIter);
	delete pNewIncompleteHypPrototype;
	(*dIter)=NULL;
      }
      incompleteHypsToDelete.clear();

      // now sort all created nodes in cell i,j according to best
      // rule-app-count and more importantly, call SelfRule
      IndexNodesAccordingToRuleAppCount(i,j);
    } // end j loop
    
    cout << "Log: " << skippingCount << " out of " << total_count_EC_IH_pairs << 
      " possible combinations of incomplete hyps and hypergraph nodes were skipped due to too high combined Rule-Application-Counts."
	 << endl;
  } // end i loop
       
  //we don't need the incomplete hyps anymore - clear them
  //when clearing a container, first clear all its elements
  for (int i=1; i<=CChartParserInfo::m_sourceLength; i++){
    for (int j=1; j<=CChartParserInfo::m_sourceLength; j++){
      ClearIHCCell(i, j);
    }
    m_incompleteHypsChart[i].clear();
  }
  m_incompleteHypsChart.clear();
}




//! returns: first: n.o. source IH/target node pairs rule-app-diff skipped;
//second: total n.o. source/target pairs
pair<int,int> CChartParser::ApplyFundamentalRule(int sourceCellRow, int sourceCellColumn, int targetCellRow, int targetCellColumn, float bestResultingRuleAppCountEstimate, RulesByProductionResult& hypRuleCombinations, vector<CChartParserHypIncomplete*>& incompleteHypsToDelete){

  // the bestResultingRuleAppCountEstimate estimate of the minimum-rule-app-count rule created by this ApplyFundamentalRule method
  // will change if we create a lower-rule-app-count rule

  // correspondingly, maxResultingRuleAppCountAllowed will also adapt
  float maxResultingRuleAppCountAllowed
    = CChartParserInfo::m_multiplicativeRA ? (int)(ceil(bestResultingRuleAppCountEstimate*m_maxRuleAppCountDifference)) : (int)(bestResultingRuleAppCountEstimate + m_maxRuleAppCountDifference);


  int skippingOccurred = 0;

  //iterate through the completed items in the target side, and for each production, check to
  //see if any are available on the incomplete side
  bool s_debugApplyFund = false;
  //allow all lexical only elements to pass thru the m_maxCombinationCount filter which only allows S, G combination thru
  bool allowLexicalsThru = true;

  // the (coarse-grained) ECs (hypergraph nodes) in the target cell
  CRuleAppCountSortedNodeIndex& eqClasses = m_completeHypsChart[targetCellRow][targetCellColumn].ruleAppCountSortedNodeIndex;
  int totalECCount = eqClasses.size();
  int pairsExplored = 0; //count the actual number of pair explored (less than totalECCount*totalIncompleteHypCount) due to MCC

  // the incomplete hyps in the source cell
  CSortedIncompleteHypContainer& incompleteHyps = m_incompleteHypsChart[sourceCellRow][sourceCellColumn];
  int totalIncompleteHypCount = incompleteHyps.size();

  if (s_debugApplyFund){cout<<"totalECCount: "<<totalECCount<<" totalIncompleteHypCount: "<<totalIncompleteHypCount<<endl;}
  // is there anything to combine at all?
  if ((totalECCount>0 && totalIncompleteHypCount>0) &&
      // optimization: if we are supposed to create cells spanning over the last source word (the </s> word),
      // that can only be possible if we are creating full-sentence-cells since all rules in the grammar 
      // involving </s> are assumed to be either the glue rule (which is left-most) or a rule with source side <s> ... </s>
      (!CChartParserInfo::m_GlueAndEOSAssumptions || // don't do the opti
       targetCellRow+targetCellColumn-1 < CChartParserInfo::m_sourceLength || // not reaching over the </s> word
       sourceCellColumn == 1 // starting with <s> word (which, when reaching over the </s> word, means, we are spanning the whole sentence
       )      ) {
    //yes

    // best value of m_bestTotalRuleApplicationCount amongst all nodes
    // hyps in the target cell
    float bestRuleAppCountECIteration = (*eqClasses.begin()).first;

    // best value of m_totalRuleApplicationCount amongst all incomplete
    // hyps in the source cell
    float bestRuleAppCountIncompleteHypIteration = (*incompleteHyps.begin()).first;
  
    if (s_debugApplyFund){
      //  cout<<"Num of eq classes: "<<eqClasses.size()<<endl;
    }

    //iterate through the incompleted hyps of current IC chart cell
    //(sourceCellRow,sourceCellColumn)
    bool continueIncompleteHypIteration = true;
    int iterationCountIH = 0;
    for (CSortedIncompleteHypContainer::const_iterator iterHyps =  incompleteHyps.begin(); continueIncompleteHypIteration != 0 && iterHyps != incompleteHyps.end(); ++iterHyps, ++iterationCountIH){

      float ruleAppCountIH = (*iterHyps).first;

      if (ruleAppCountIH + bestRuleAppCountECIteration + 1 > maxResultingRuleAppCountAllowed) {
        // no chance of getting an allowed combined rule app count anymore
	// don't consider this and any subsequent IHs of
	// this cell anymore
	continueIncompleteHypIteration = false;
	skippingOccurred += (totalIncompleteHypCount - iterationCountIH)*totalECCount;

      } else {

	      
	CChartParserHypIncomplete* pIncompleteHyp = (*iterHyps).second; // the current incomplete hyp: our candidate for a match with pNode
	bool incompleteIsLexical = pIncompleteHyp->m_productionBackPointers.size() == 0;

	// we will only do the combination if the source span hasn't exceeded the panic threshold
	// or if the inc. hyp is lexical (since we don't want to miss out on long phrase-pair rules)
	// or if this incomplete hyp starts at source word 1, i.e., the corresponding rule's source side starts 
	// with <s> (e.g., glue hypotheses)
	if (m_combinationCount <= CChartParserInfo::m_maxCombinationCount || 
	    (sourceCellColumn==1) ||
	    (incompleteIsLexical&& allowLexicalsThru)) {

	  //do the expansion for each equivalence class, until the EC's rule app
	  //counts get too bad
	  bool continueECIteration = true;
	  int iterationCountEC = 0;
	  for (CRuleAppCountSortedNodeIndex::const_iterator iter = eqClasses.begin();  continueECIteration && iter!= eqClasses.end(); ++iter, ++iterationCountEC){

	    float ruleAppCountEC = (*iter).first;

	    if (ruleAppCountIH + ruleAppCountEC + 1 > maxResultingRuleAppCountAllowed) {
	      // don't consider this and any subsequent incomplete hyps of
	      // this cell anymore; their rule app count is too high
	      continueECIteration = false;
	      skippingOccurred += totalECCount - iterationCountEC;
	    } else {

	      // our current node that we want to combine with pIncompleteHyp
	      CProductionNode* pNode = (*iter).second;
	      bool completeIsLexical = pNode->m_lexical;
	

	      // ...and check that we're still below the max source span
	      if (m_combinationCount <= CChartParserInfo::m_maxCombinationCount || 
		  (sourceCellColumn==1) ||
		  //		    (sourceCellColumn==1 && pIncompleteHyp->m_prefix[0] == CChartParserInfo::m_FullSentenceNT && !completeIsLexical) ||
		  (incompleteIsLexical&&completeIsLexical&&allowLexicalsThru)) {
		//def going to check out this pair
		pairsExplored++;
		  

		vector<CWordIndex> newPrefix = pIncompleteHyp->m_prefix; // copy the current prefix from the selected incomplete hyp
		newPrefix.push_back(pNode->GetResult()); // append the production result of the candidate-matching completed node to the prefix
		  
		//if (incompleteIsLexical && completeIsLexical && m_combinationCount > CChartParserInfo::m_maxCombinationCount){

		if (CChartParserInfo::m_verbosity>5){
		  cout<<"      Try FundRule-Combining: "; pIncompleteHyp->Display(); cout<<endl;
		  cout<<"      with: "; pNode->Display(); cout<<endl;
		}
	    
		CBTreeResult bTreeQueryResult = m_bTree.QueryBTree(newPrefix); // look up newPrefix in the rule set

		if (bTreeQueryResult.first != 0) { // are there any matches at all?
		  // good, so it's worth expanding pIncompleteHyp to a prototypical incomplete hyp that already encompasses the
		  // backpointer to and the cost info of the hypergraph node pNode to be subsumed

		  if (s_debugApplyFund || CChartParserInfo::m_verbosity>5){
		    cout<<"        matched: ";
		    CChartParserInfo::PrintVocabSequence(newPrefix);
		    cout<<endl;
		  }

		  //clone the current incomplete hyp
		  CChartParserHypIncomplete* pNewIncompleteHypPrototype = new CChartParserHypIncomplete(*pIncompleteHyp);

		  // expand the incomplete hyp's tail with pNode, meaning: extending the span length m_i, extending 
		  // prefix and backpointer info, updating cost variables and LM stuff
		  if (s_debugApplyFund){
		    cout<<"\t Creating new incomplete hyp PROTOTYPE by tail-expanding incompl. hyp:\t";
		    pNewIncompleteHypPrototype->Display(CChartParserInfo::m_pVocabulary);
		    cout<<"\t with hypergraph node:\t";
		    pNode->Display(CChartParserInfo::m_pVocabulary);
		  }
		  pNewIncompleteHypPrototype->ExpandTail(pNode); //adds another backpointer + increments the m_totalRuleApplicationCount for pNewIncompleteHypPrototype
		  assert(pNewIncompleteHypPrototype->I() == sourceCellRow+targetCellRow);
		  assert(pNewIncompleteHypPrototype->J() == sourceCellColumn);
	  
		  // now go through the exact matches and the possible prefix match and file one complete hyp to the chart for each exact match
		  // further, in case of prefix match, file pNewIncompleteHypPrototype, otherwise, delete pNewIncompleteHypPrototype

		  if ((bTreeQueryResult.first & C_ExactMatch) != 0  // at least one exact match
		      && (ruleAppCountIH + ruleAppCountEC + 1 < bestResultingRuleAppCountEstimate)
		      ) {
		    // created a new hyp achieving a new rule-app-count record low
		    bestResultingRuleAppCountEstimate = ruleAppCountIH + ruleAppCountEC + 1;
		    // correspondingly, maxResultingRuleAppCountAllowed will also adapt
		    maxResultingRuleAppCountAllowed = CChartParserInfo::m_multiplicativeRA ? 
		      (ceil(bestResultingRuleAppCountEstimate*m_maxRuleAppCountDifference)) : 
		      (bestResultingRuleAppCountEstimate + m_maxRuleAppCountDifference);
		  }


		  // call ProcessBTreeResult
		  //ProcessBTreeResult is liable to call AddCompleteHyp, reset the stack level
		  CChartParserInfo::m_stackLevel = 0;

		  //pNewIncompleteHypPrototype = ProcessBTreeResult(bTreeQueryResult,pNewIncompleteHypPrototype, false);
		  if (!ProcessBTreeResult(hypRuleCombinations, bTreeQueryResult,pNewIncompleteHypPrototype)){
		    incompleteHypsToDelete.push_back(pNewIncompleteHypPrototype);
		  }else{
		    if (s_debugApplyFund || CChartParserInfo::m_verbosity>5){
		      cout << "    prefix-match, so calling AddIncompleteHyp." << endl;
		    }
		    //we can just file the pNewIncompleteHypPrototype with its current m_totalRuleApplicationCount right here
		    //because its back pointers are to nodes that arent going to have changing elements
		    AddIncompleteHyp(pNewIncompleteHypPrototype);
		  }
		  //cout<<"Processed Complete result running:"<<hypRuleCombinations.size()<<endl;
		    
		  //if we are doing local pruning - local pruning at this level is not supported anymore
		  /*
		    if (!CChartParserInfo::m_comboPruningGlobal){
		    if (hypRuleCombinations.size()>0){
		    //cout<<"PrefixNormal: "; CChartParserInfo::PrintVocabSequence(newPrefix); cout<<endl;
		    //now we have built up a series of CChartParserHypIncomplete - rule vector pairs, that we can process from the top down
		    pair<int, int> expandedAndPossible = SpawnKBestCompleteHyps(hypRuleCombinations, false);
		    if (CChartParserInfo::m_verbosity>0){
		    cout<<"Log: ExpPossLocal: "<<expandedAndPossible.first<<" / "<<expandedAndPossible.second<<endl;
		    }
		    }
		    hypRuleCombinations.clear();
		    }
		  */
		    
		  //if we found something in the btree	
		}else{
		  //cout<<" no match "<<endl;
		}
		//as long as we abide by the MaxCombinationCount conditions
	      }else{
		//cout<<" prefix[0]!=@_S "<<endl;
	      }
	    }//rule app count difference has not been exceeded
	  }//go thru each incomplete hyp
	}//MCC count check
      }//rule app count check
    }//gone thru each equivalence class
  }

  // back at the main level of ApplyFund

  //local pruning means doing it within ApplyFundamentalRule
  if (!CChartParserInfo::m_comboPruningGlobal){
    if (hypRuleCombinations.size()>0){
      //now we have built up a series of CChartParserHypIncomplete - rule vector pairs, that we can process from the top down
      pair<int, int> expandedAndPossible = SpawnKBestCompleteHyps(hypRuleCombinations, false);
      if (CChartParserInfo::m_verbosity>0){
	cout<<"Log: ExpPossGlobal: "<<expandedAndPossible.first<<" / "<<expandedAndPossible.second<<endl;
      }
    }
    hypRuleCombinations.clear();
  }
    
  //delete all the incompleted hyps *do this outside ApplyFundamentalRule*
  /*
    for (vector<CChartParserHypIncomplete*>::iterator dIter = incompleteHypsToDelete.begin(); dIter!= incompleteHypsToDelete.end(); ++dIter){
    CChartParserHypIncomplete* pNewIncompleteHypPrototype = (*dIter);
    delete pNewIncompleteHypPrototype;
    (*dIter)=NULL;
    }
    incompleteHypsToDelete.clear();
  */
  
  if (CChartParserInfo::m_verbosity>0){
    cout<<"Log: ApplyFundamentalRule skipped: "<<skippingOccurred<<" explored: "<<pairsExplored<<" possible: "<<totalIncompleteHypCount*totalECCount<<endl;
  }
  return pair<int,int>(skippingOccurred,totalIncompleteHypCount*totalECCount);
} // end of ApplyFundamentalRule





void CChartParser::IndexNodesAccordingToRuleAppCount(int i, int j) {

  bool debug = false;
  //remove several of the fine grained eq classes
  // NO MORE notion of prune cell
  //if (CChartParserInfo::m_compareTargetWords){
  
  cout<<"Entering IndexNodesAccordingToRuleAppCount i: "<<i<<" j: "<<j<<endl;

  //here we prune the m_productionNodes (i.e., coarse-grained pruning), removing elements from the m_newlyCreatedNodes
  if (CChartParserInfo::m_maxCostDifferencePerCell<FLT_MAX/10){
    PruneCell(i, j);
  }
  
  // current chart cell in the complete hyp chart
  CCompleteHypChartCell& chartCell = m_completeHypsChart[i][j];
  int numEqClassesBeforeSelf = 0;
  numEqClassesBeforeSelf += FinalizeAllNodes(chartCell.m_productionNodes);
  //if (CChartParserInfo::m_compareTargetWords){
  //  PruneCell(i, j);
  //}
  //cause we have this stupid PruneCell in operation we need to RankFineGrainedClasses again
  //FinalizeAllNodes(chartCell.m_productionNodes);

  //a vector of all hyps that have need to have their m_totalRuleApplicationCount updated
  //all these incomplete hyps will have only matched their very first backpointer
  vector<CChartParserHypIncomplete*> selfRuleIncompleteHyps;
  
  
  int waveCount = 0; // counts the waves of self-rule applications
  bool unknownHandling = false; // just for info purposes
  while (!m_newlyCreatedNodes.empty()) {
    // next wave of self-rule applications
    waveCount++;
    //copy the set of nodes newly created in last wave
    //vector<CProductionNode*> newlyCreatedNodesCopy = m_newlyCreatedNodes;
    CCoarseGrainedEq newlyCreatedNodesCopy = m_newlyCreatedNodes;
    //empty m_newlyCreatedNodes
    m_newlyCreatedNodes.clear();
    
    if (debug){
      cout<<"*** Beginning self rule round: "<<endl;
      for (CCoarseGrainedEq::iterator iter=newlyCreatedNodesCopy.begin(); iter!= newlyCreatedNodesCopy.end(); ++iter){
	(*iter).second->Display();
	cout<<endl;
      }
      cout<<" Starting now..."<<endl;
    }
    
    // now apply self-rule to all newly created nodes
    //for (vector<CProductionNode*>::const_iterator it = newlyCreatedNodesCopy.begin(); it != newlyCreatedNodesCopy.end(); ++it) {
    for (CCoarseGrainedEq::const_iterator it = newlyCreatedNodesCopy.begin(); it != newlyCreatedNodesCopy.end(); ++it) {

      //OLD style, now is a single finalization step
      //compute the sorted structure - we have to do this here because a) PruneCell must have been called 
      //(*it)->RankFineGrainedClasses();
      //(*it)->FinalizeProductionNode();

      // according to the assumptions we make in section "Self Rule" at the comments in the 
      // beginning of FastTranslateChart.cc, we never have a 'shortcut' case such as NN->NP->_S and NN->_S in the
      // rule file. There is, however, one problematic case: _S _G -> _S (which will fire an _S complete hyp
      // within Fundamental Rule execution), and with IndexNodesAccordingToRuleAppCount again as self-rule for _S:
      // _S _G -> _S (which will trigger an IH for _S based on _S's current hyp's rule-app-count), but also within
      // IndexNodesAccordingToRuleAppCount the self-rule for e.g. NN:
      // NN -> _S
      // which will add hyps to the _S EC possibly after _S's self rule has already been fired.
      // As a remedy, call Self Rule on _S at the very last.
      // Note: Strictly this is not necessary anymore now, since we re-calculate IH's rule-app-counts
      // based on the latest backpointer's ECs within ReCalculateRuleAppCount
      
      if ((*it).first == CChartParserInfo::m_FullSentenceNT && newlyCreatedNodesCopy.size() != 1) {
       	// make sure that m_FullSentenceNT is expanded at the very last
       	//it ensures that glue rule node gets correct m_bestTotalRuleApplicationCount
	//(this shouldn't be necessary anymore)
	
       	//don't treat it in this loop yet but later
       	//m_newlyCreatedNodes.push_back(*it);
	m_newlyCreatedNodes.insert(std::pair<CWordIndex, CProductionNode*>((*it).first, (*it).second));
      } else {
	
	//if ((*it)->GetResult() == CChartParserInfo::m_FullSentenceNT && newlyCreatedNodesCopy.size() == 1) {
	//cout<<"ALONE ---- "<<endl;
	//}
	CChartParserInfo::m_stackLevel++;
	//output settings
	if (CChartParserInfo::m_verbosity>0){
	  CChartParserInfo::TabStack(); cout<<"SelfFor: "; cout<<CChartParserInfo::m_pVocabulary->GetWord((*it).second->GetResult()); cout<<" "; 
	}
	p_applySelfRuleRecursive((*it).second, selfRuleIncompleteHyps);
	CChartParserInfo::m_stackLevel--;
	//cout<<"Updating done......"<<endl;
      }
    }

    if (i==1) {
      // we're on the lexical level
      if (waveCount == 1) {
	// first self-rule wave is over: check if the lexical hyp triggered any self-rules filing completed hyps in the cell
	if (m_newlyCreatedNodes.empty() && 
	    // exception: BOS or EOS b/c for these no unknown rules are used 
	    // but rather we rely on at least one complete-sentence rule making use of BOS and EOS
	    GetCompleteCell(i,j).m_cellType == 0) {

	  // add _UNKNOWN lexical hyp edge as the actual word was unmatched by any rules
	  // all we now do is add a second terminal hyp, this time with result m_unknownWI ("_UNKNOWN")
	  // this hyp can then be expanded by rules in the rule file such as: _UNKNOWN#_UNKNOWN#@NP#1 1 1
	  cout << "Inserting _UNKNOWN at i,j=" << i << "," << j << endl;
	  CChartParserHypComplete* pUNKHyp = new CChartParserHypComplete(CChartParserInfo::m_unknownWI, i, j);
	  bool wasAdded = AddCompletedHyp(pUNKHyp);
	  assert(wasAdded==true);
	  // 	  if (!wasAdded){
	  // 	    cout<<"ERROR: pUNKHyp was rejected by the online beam criteria";
	  // 	    exit(-1);
	  // 	  }
	  unknownHandling = true;
	  assert(!m_newlyCreatedNodes.empty()); // should contain the new node
	}
      } else if (waveCount == 2 && unknownHandling && m_newlyCreatedNodes.empty()) {
	// even the _UNKNOWN hyp didn't trigger any new completed hyps to be filed - seems there 
	// are no unknown-rules in the DB
	cerr << "Warning: Unknown-hyp for word " << j << 
	  " didn't trigger any new hyps. Seems you don't have any _UNKNOWN rules in the rule database." 
	     << endl;
	cout << "Warning: Unknown-hyp for word " << j << 
	  " didn't trigger any new hyps. Seems you don't have any _UNKNOWN rules in the rule database." 
	     << endl;
      }
    }


    //re-prune the m_productionNodes (i.e., coarse-grained pruning), removing elements from the m_newlyCreatedNodes
    if (CChartParserInfo::m_maxCostDifferencePerCell<FLT_MAX/10){
      PruneCell(i, j);
    }

    //re-finalize at the end of the round - with the addAll param set to true
    //no - dont set this addAll to true, we want all hyps to be compared on equal ground
    //FinalizeAllNodes(chartCell.m_productionNodes, true);
    FinalizeAllNodes(chartCell.m_productionNodes, false);
  }


  int numEqClasses = 0;
  int numHyps = 0;
  // index the nodes according to rule app count
  if (CChartParserInfo::m_verbosity>0){
    cout<<"\nLog: IndexCell: ";
  }
  for (CCoarseGrainedEq::const_iterator iter = chartCell.m_productionNodes.begin(); 
       iter!= chartCell.m_productionNodes.end();
       ++iter
       ){
    CProductionNode* node = (*iter).second;
    int initialSize = node->m_fineGrainedNodes.size();
    if (CChartParserInfo::m_verbosity>0){
      cout<<CChartParserInfo::m_pVocabulary->GetWord(node->GetResult())<<" ("<<initialSize<<" eq "<<node->NumHyps()<<" hyps ) ";
    }
    numEqClasses +=initialSize;
    numHyps += node->NumHyps();
    // add node to the rule-app-count sorted index
    chartCell.ruleAppCountSortedNodeIndex.insert(pair<float,CProductionNode*>(node->m_bestTotalRuleApplicationCount, node));
  }
  if (CChartParserInfo::m_verbosity>0){
    cout<<endl;
  }

  //  if (CChartParserInfo::m_verbosity>0){
  //    if (i==CChartParserInfo::m_sourceLength && j==1){
  //      cout<<" FullSent: "<<CChartParserInfo::m_pVocabulary->GetWord(m_pCompleteTranslationsNode->GetResult())<<" ("<<m_pCompleteTranslationsNode->m_fineGrainedNodes.size()<<" eq "<<m_pCompleteTranslationsNode->NumHyps()<<" ) "<<endl;
  //    }
  //  }



  //index all of the self rule incomplete hyps after asking them to recalculate their backpointers
  for (vector<CChartParserHypIncomplete*>::iterator iter = selfRuleIncompleteHyps.begin(); iter!= selfRuleIncompleteHyps.end(); ++iter){
    CChartParserHypIncomplete* pIncompleteHyp = (*iter);
    pIncompleteHyp->ReCalculateRuleAppCount();
    AddIncompleteHyp(pIncompleteHyp);
  }

  cout<<"Log: Production Nodes: "<<chartCell.m_productionNodes.size()<<" Eq Classes: before self "<<numEqClassesBeforeSelf<<" after: "<<numEqClasses<<endl;
  //numHyps is the actual number of hyps left in this cell after FinalizeAllNodes has been called
  //m_hypsDeleted counts those that have been added due to cube pruning, but then deleted cause they didnt fit in the histogram or threshold beam
  //m_hypsAdded is the total number of hyps added due to the cube pruning expansion step. As you increase the fuzz, this value should increase too
  //m_totalCalls is the number of calls to request additional elements from the cube pruning process
  cout<<"Log: numHyps: "<<numHyps<<" m_hypsDeleted: "<<CChartParserInfo::m_hypsDeleted<<" m_hypsAdded: "<<CChartParserInfo::m_hypsAdded<<" m_numTopCalls: "<<CChartParserInfo::m_numTopCalls<<" m_allPossibleExpansions: "<<CChartParserInfo::m_allPossibleExpansions<<" numAddedPostFailure: "<<CChartParserInfo::m_numAddedPostFailure<<" m_numTuplesProposed: "<<CChartParserInfo::m_numTuplesProposed<<" m_hypsDeletedHistogram: "<<CChartParserInfo::m_hypsDeletedHistogram<<" m_hypsDeletedThreshold: "<<CChartParserInfo::m_hypsDeletedThreshold<<endl;

  
  cout <<"Log(end-of IndexNodesAccordingToRuleAppCount): " << chartCell.m_productionNodes.size() << " nodes, " << m_incompleteHypsChart[i][j].size() << " incomplete hyps in cell "<< i << "," << j << ". " << endl << endl;

  if (!CChartParserInfo::m_GlueAndEOSAssumptions && CChartParserInfo::m_lowestRuleAppCount && j==1) {
    // test whether the estimated best-possible RAC achievable by gluing ended up being achieved by a complete _S hyp filed
    if (chartCell.m_bestGluedRuleAppCount < chartCell.m_productionNodes[CChartParserInfo::m_FullSentenceNT]->m_bestTotalRuleApplicationCount) {
      cout << "Minor Warning: For _S, best-possible RAC achievable by gluing m_bestGluedRuleAppCount=" 
	   << chartCell.m_bestGluedRuleAppCount << " < " 
	   <<  chartCell.m_productionNodes[CChartParserInfo::m_FullSentenceNT]->m_bestTotalRuleApplicationCount
	   << "=m_bestTotalRuleApplicationCount (best RAC actually achieved). You might want to use less severe pruning." 
	   << endl;
    }
  }

  //in very verbose mode
  if (debug || CChartParserInfo::m_verbosity>1){
    cout<<"====== Cell "<<i<<","<<j<<" ==== "<<endl;
    DisplayCell(i, j);
    cout<<"================================="<<endl;
  }
  
}




/*!
  After a chart cell has been filled with all lexical hyps or the hyps resulting from the Fundamental rule,
  p_applySelfRuleRecursive is called on each coarse-grained EC in the cell, i.e., on each result nonterminal,
  in order to look up rules whose source side starts with that NT. In case of exact matches, new complete
  hyps might be fired in the same cell. In case new result NTs arise this way, p_applySelfRuleRecursive 
  has to be called for these also. All this is done in IndexNodesAccordingToRuleAppCount().
  The self rule queries the B-tree for rules
  that begin with that production result pNode. In case of a strict prefix match, it 
  adds these rules as incomplete hyps into 
  the IH chart. In case of an exact match, it files them as complete hyps to the CH chart.
*/
void CChartParser::p_applySelfRuleRecursive(CProductionNode* pNode, vector<CChartParserHypIncomplete*>& scheduleForAdding){

  bool s_debugApplySelf = false;
  if (s_debugApplySelf){
    //only display information if successful
    cout<<"APPLYING SELF RULE***********"<<endl;
    cout<<"Recursing for: "<<CChartParserInfo::m_pVocabulary->GetWord(pNode->GetResult())<<endl;
    cout<<"Recursing node: ";
    pNode->Display(CChartParserInfo::m_pVocabulary);
    cout<<endl;
  }

  vector<CWordIndex> newPrefix; // create empty vector of source symbols ...
  newPrefix.push_back(pNode->GetResult()); // ... and append the production result of the node
  if (s_debugApplySelf){
    cout<<"querying: "; CChartParserInfo::PrintVocabSequence(newPrefix); cout<<endl;
  }
  CBTreeResult bTreeQueryResult = m_bTree.QueryBTree(newPrefix); // look up newPrefix in the rule set
  if (s_debugApplySelf){
    cout<<"returned bTreeQueryResult"<<endl;
  }
  if (bTreeQueryResult.first != 0) { // are there any matches at all?
    // good, so it's worth expanding pIncompleteHyp to a prototypical incomplete hyp that already encompasses the
    // backpointer to and the cost info of the hypergraph node pNode to be subsumed

    if (s_debugApplySelf){
      cout<<"***APPLY_SELF_RULE(";
      pNode->Display();
      cout<<"Found rules exact- or prefix-matching with: "<<CChartParserInfo::m_pVocabulary->GetWord(pNode->GetResult())<<endl;
      cout<<"Creating new incomplete hyp prototype by tail-expanding a new empty incompl. hyp with hypergraph node:\t";
      pNode->Display(CChartParserInfo::m_pVocabulary);
      cout<<endl;
    }

    //create a new incomplete hyp to be used as prototype for all the B-tree matches
    CChartParserHypIncomplete* pNewIncompleteHypPrototype = new CChartParserHypIncomplete(pNode->J());

    // expand the incomplete hyp's (still empty) tail with pNode, meaning: extending the span length m_i, extending 
    // prefix and backpointer info
    pNewIncompleteHypPrototype->ExpandTail(pNode);
	
    // now go through the exact matches and the possible prefix match and file one complete hyp to the chart for each exact match
    // further, in case of prefix match, file pNewIncompleteHypPrototype, otherwise, delete pNewIncompleteHypPrototype
    //get back the value of the pNewIncompleteHypPrototype to see if it has been deleted

    RulesByProductionResult hypRuleCombinations;
    //if ProcessBTreeResult return true, that means this pNewIncompleteHypPrototype was stored, otherwise delete it
    //cout<<"PrefixSelf: "; CChartParserInfo::PrintVocabSequence(newPrefix); cout<<endl;
    bool incompleteUsed = ProcessBTreeResult(hypRuleCombinations, bTreeQueryResult,pNewIncompleteHypPrototype);
    if (hypRuleCombinations.size()>0){
      
      if (CChartParserInfo::m_verbosity>0){cout<<"Self-Hyps found: ";}

      //Curently check that new rules dont enter into existing hyps, but this is not correct
      //assert(SelfRuleCompliance(pNode, hypRuleCombinations)==true);

      //az: changed to false to also prune in SelfRule mode
      bool addAll = false; //make sure everything is added in addAll==true mode
      SpawnKBestCompleteHyps(hypRuleCombinations, addAll);
    }else{
      if (CChartParserInfo::m_verbosity>0){cout<<"NoCompleteFound ";}
    }
    //delete the pNewIncompleteHypPrototype if it turns out that we arent filing it anywhere
    if (!incompleteUsed){
      delete pNewIncompleteHypPrototype;
      pNewIncompleteHypPrototype = NULL;
    }else{
      //need to call AddIncompleteHyp on  pNewIncompleteHypPrototype but it should be with the finalized backpointer nodes
      if (CChartParserInfo::m_verbosity>0){cout << "Scheduling for adding: "; pNewIncompleteHypPrototype->Display(); cout << endl;}
      scheduleForAdding.push_back(pNewIncompleteHypPrototype);
    }
  }else{
    if (CChartParserInfo::m_verbosity>0){cout<<"NothingFound: ";}
    if (s_debugApplySelf){
      cout<<"no matches at all"<<endl;
    }
  }
}






// void CChartParser::TestKBest(RulesByProductionResult::iterator begin, RulesByProductionResult::iterator end){
//   CCombinationTupleBeam beamAll(CChartParserInfo::m_comboPruningBeamSize, 1.0);
//   CCombinationTupleBeam beamLess(CChartParserInfo::m_comboPruningBeamSize, 0.9);
  
//   //load up for the all case
//   cout<<"********** Load 1.0 ******* "<<endl;
//   int terminallyAddedAll, allPossibleExpansionsAll;
//   InitializeComboBeam(&beamAll, begin, end, terminallyAddedAll, allPossibleExpansionsAll, false);
//   SortedCombinations* pAll = beamAll.FillNBestReturnPtr();

//   //load up for the less case
//   cout<<"********** Load 0.9 ******* "<<endl;
//   int terminallyAddedLess, allPossibleExpansionsLess;
//   InitializeComboBeam(&beamLess, begin, end, terminallyAddedLess, allPossibleExpansionsLess, false);
//   SortedCombinations* pLess = beamLess.FillNBestReturnPtr();

//   //now, verify that every element in the pLess case also exists in the pAll case
//   UniqueCombinations allUnique, lessUnique;
//   CCombinationTupleBeam::ConvertSortedToUnique(pAll, &allUnique);
  

//   cout<<"Testing: pAll: "<<pAll->size()<<" pLess: "<<pLess->size()<<endl;
//   if (pAll->begin()!=pAll->end() && pLess->begin()!=pLess->end()){
//     double lowAll = pAll->begin()->second->m_cost;
//     double highAll = (--(pAll->end()))->second->m_cost;

//     double lowLess = pLess->begin()->second->m_cost;
//     double highLess = (--(pLess->end()))->second->m_cost;
//     cout<<"lowAll: "<<lowAll<<" highAll: "<<highAll<<" lowLess: "<<lowLess<<" highLess: "<<highLess<<endl;
//     assert(lowLess==lowAll);
//     assert(highAll>=highLess);
//   }


//   //go thru every element in pLess and check to make sure that it is also in allUnique
//   int count = 0;
//   for (SortedCombinations::iterator tmpIter = pLess->begin(); tmpIter!=pLess->end(); ++tmpIter){
//     cout<<"Checking: "; (*tmpIter).second->Display();
//     CCombinationTuple* toAdd = (*tmpIter).second;
//     if (allUnique.count(toAdd)==0){
//       cout<<"ERROR: Missing No: "<<count<<" "; toAdd->Display();
//       assert(false);
//     }
//     count++;
//   }
  
//   //ConvertSortedToUnique(pLess, &lessUnique);
//   beamAll.Clear();
//   beamLess.Clear();
//   allUnique.clear();
// }







void CChartParser::InitializeComboBeam(CCombinationTupleBeam* pBeam, RulesByProductionResult::iterator begin, RulesByProductionResult::iterator end, int& allPossibleExpansions, bool addAll){

  bool debug = false;
  //double fakeExternalCost = 0.0;
  
  CWordIndex productionResultProcessed;
  //add all rules' best-instantiations
  for (RulesByProductionResult::iterator match_iter=begin; match_iter!=end; ++match_iter){

    if (match_iter==begin){
      productionResultProcessed = match_iter->first;
      if (debug){
	cout<<"Spawning for: "<<CChartParserInfo::m_pVocabulary->GetWord(productionResultProcessed)<<endl;
      }
    }else{
      assert(productionResultProcessed == match_iter->first);
    }
    CChartParserHypIncomplete* pNewIncompleteHypPrototype = match_iter->second.first; 
    //vector<CRule>* matchedRules = &(match_iter->second.second);

    //for ( vector<CRule>::iterator iter = matchedRules->begin(); iter!=matchedRules->end(); ++iter){
    CRule* pOneItem = &(match_iter->second.second); // the rule found
    

    // lets put this hyp's best instantiation onto the beam
    // Using the coarse-grained-EC back pointers provided from pIncomplete, instantiate the <e, 1> element, 
    // where the e refers to the rule being used 
    // and the 1 refers to the vector of 1st best derivations from each of the back pointer nodes
    vector< std::vector<CChartParserNode*>::iterator > bestDerivationVector;
    //add all the backpointers - using the underlying BP nodes' best cost which 
    //is m_bestTotalCost AND LM costs AND boundary LM cost
    // go thru each one and pick out the first best pChartParserNode
    for (
	 vector<CProductionNode*>::const_iterator iter = pNewIncompleteHypPrototype->m_productionBackPointers.begin();
	 iter!= pNewIncompleteHypPrototype->m_productionBackPointers.end();
	 ++iter ){

      CProductionNode* pProdNode = (*iter);

      //cout<<"Rule: "; pRule->Display(); cout<<endl;
      //cout<<"Incomplete: "; m_pIncomplete->Display(); cout<<endl;
      //cout<<"Prod: "<<CChartParserInfo::m_pVocabulary->GetWord(pProdNode->GetResult())<<"="<<pProdNode->m_fineGrainedNodes.size()<<" m_lowestToHighestCostFineGrained: "<<pProdNode->m_lowestToHighestCostFineGrained.size()<<endl;

      assert(pProdNode->m_fineGrainedNodes.size()==pProdNode->m_lowestToHighestCostFineGrained.size());
      assert(pProdNode->m_lowestToHighestCostFineGrained.size()>0);

      // initialize the derivation vector with (1,...,1) means
      // using for each element an iterator pointing to the best backpointed-node for each NT substitution site
      std::vector<CChartParserNode*>::iterator firstBest = pProdNode->m_lowestToHighestCostFineGrained.begin();
      bestDerivationVector.push_back(firstBest);
    }
    
    // create tuple
    CCombinationTuple* c = new CCombinationTuple(pOneItem, pNewIncompleteHypPrototype, bestDerivationVector);
#ifndef NDEBUG
    if (debug){
      if (pNewIncompleteHypPrototype->m_productionBackPointers.size()==0){
	cout<<"Add phrasal rule to beam: "; c->Display();
      } else {
	cout<<"Add to beam: "; c->Display(); 
      }
    }
#endif
    // add it to the beam and before that instantiate its corresponding complete hyp, calculating LM costs etc.
    bool wasInserted = pBeam->InsertAndCompleteBaseCombination(c);
    assert(wasInserted==true); //the InsertBaseCombination never does any pruning and no duplicates should have occurred yet
    
    //===========================================
    // for diagnotic purposes, lets see what the total number expanded would have been
    //===========================================
    int possibleExpansions = 1;
    for (vector<CProductionNode*>::iterator iter = pNewIncompleteHypPrototype->m_productionBackPointers.begin(); iter!= pNewIncompleteHypPrototype->m_productionBackPointers.end(); ++iter){
      CProductionNode* pPNode = (*iter);
      //cout<<CChartParserInfo::m_pVocabulary->GetWord(pPNode->GetResult())<<"-"<<pPNode->m_fineGrainedNodes.size()<<" ";
      possibleExpansions *= pPNode->m_fineGrainedNodes.size();
    }
    //cout<<endl;
    allPossibleExpansions += possibleExpansions;

  }//all rules processed 
}

/*
bool CChartParser::IsTupleFuzzCompliant(double hypCost, double lastOfficialCost){
//do this in exp scale like everything else
if (firstFailedCost==FLT_MAX){
return true;
}
//double firstFailedProb = exp(-firstFailedCost);
//double cutoffLogProb = -log(firstFailedProb - firstFailedProb*CChartParserInfo::m_comboPruningBeamPerc);
double cutoffLogProb = CChartParserInfo::GetPruningCutoff(firstFailedCost, CChartParserInfo::m_comboPruningFuzzCostDifference);
return (hypCost <= cutoffLogProb);
}
*/

pair<int, int> CChartParser::SpawnForProductionResult(RulesByProductionResult::iterator begin, RulesByProductionResult::iterator end, bool addAll){
  //now the interesting move. we need to add all the top level rules, then unwind to identify the top k hyps
  CCombinationTupleBeam beam;
  //initialize the beam with the pairs of incomplete hyps and the completing rules
  int allPossibleExpansions = 0;
  //  double fakeExternalCost = 0.0;
    
  bool debug = false;

  //, CChartParserInfo::m_comboPruningBeamSize, CChartParserInfo::m_comboPruningFuzzCostDifference);

  //cout<<"Beginning Testing ============="<<endl;
  //TestKBest(begin, end);
  //cout<<"Testing passed ================"<<endl;

  if (debug){
    cout<<"*** BEGIN addAll: "<<addAll<<" *** "<<endl;
  }  
  //fill the beam with the initial n-best list based on the top translations 
  InitializeComboBeam(&beam, begin, end, allPossibleExpansions, addAll);
  clock_t spawnTime = clock(); //accumulate these values over the whole sentence
  int initialNumTuples  = beam.NumTuplesProposed();

  //  double firstFailedCost = FLT_MAX;
  //keep going retrieving tuples and expand these tuples' neighbors until there are no more elements on the beam or you hit a break out condtion:
  //a) hit the limit m_comboPruningBeamSize of elements that you are allowed to expand
  //b) more than m_comboPruningMinBeamSize tuples have been considered and the tuple is worse than bestHyp+prodCategoryPruningMapCostDifference+m_comboPruningFuzzCostDifference
  // If a) occurs, we still retrieve what we have on the beam (since that doesn't cost anything as the LM costs have already been calculated for these tuples), but we don't expand the retrieved tuples' neighbors anymore
  // If b) occurs, it would be stupid to still retrieve the rest from the beam since these tuples must all be worse
  // than the tuple causing the cutoff based break-out, therefore none of these would be accepted by AddCompletedHyp
  bool expandNeighbors = true; // whether to expand retrieved tuples' neighbors and put these on the beam
  int count = 0; // n.o. tuples retrieved from beam
  int numAdded = 0; // n.o. tuples retrieved that were actually added (not pruned away immediately) by AddCompletedHyp
  int numAddedPostFailure = 0; // n.o. good (i.e., added) tuples that were saved because of an avoided m_comboPruningFuzzCostDifference break-out attempt that was averted due to m_comboPruningMinBeamSize not having been exceeded
  bool breakOutAttemptAverted = false; // after this gets true, added tuples are counted towards numAddedPostFailure
  //  int numTopCalls = 0;
  //see call to beam.NextTuple

#ifndef NDEBUG
  float lastCostDebug = -(FLT_MAX)/10.0;
#endif

  //  float worstOfficialCost = -FLT_MAX/10.0; // worst cost encountered by hyps within the official beam
  //  float fuzzMaxAllowedCost;
  //  bool stillFuzzing = true;

  for (CCombinationTuple* pTuple = beam.NextTuple(expandNeighbors); 
       pTuple != NULL;
       count++){

#ifndef NDEBUG
    //in cmp=0 we can be sure that each successive cost is higher than the last one
    //same holds when ParserInfo::m_comboPruningUseLM==false
    if (!CChartParserInfo::m_compareTargetWords || !CChartParserInfo::m_comboPruningUseLM){
      assert(pTuple->Cost()>=lastCostDebug);
      lastCostDebug = pTuple->Cost();
    }
    if (debug){
      cout<<"NextTuple: "; pTuple->Display(); cout<<endl;
    }
#endif

    // hyp to be added
    CChartParserHypComplete* pNewCompleteHyp = pTuple->m_pComplete;

    // check whether the hyps cost satisfy m_comboPruningFuzzCostDifference
    float cutoffFromBestCost = 
      GetCompleteCell(pNewCompleteHyp->I(),pNewCompleteHyp->J()).m_productionNodes.count(pNewCompleteHyp->GetResult())==0 ?
      FLT_MAX/10
      :
      GetCompleteCell(pNewCompleteHyp->I(),pNewCompleteHyp->J()).m_productionNodes[pNewCompleteHyp->GetResult()]->m_cutoffLogProb;
    if (!addAll && pTuple->Cost() > CChartParserInfo::GetPruningCutoff(cutoffFromBestCost, CChartParserInfo::m_comboPruningFuzzCostDifference)) {
      // the only thing that can save this tuple and subequent ones now is m_comboPruningMinBeamSize
      if (count >= CChartParserInfo::m_comboPruningMinBeamSize) {
	break; // stop retrieving hyps
      } else {
	// got lucky
	breakOutAttemptAverted = true;
      }
    }	

#ifndef NDEBUG
    if (debug){cout<<"Was compliant -> ";}
    if (debug){
      cout<<"Created: "; pNewCompleteHyp->Display();
    }
    if (debug){cout<<" HypCost: "<<pNewCompleteHyp->m_totalCost + pNewCompleteHyp->m_scaledFutureCostEstimate<<endl;}
#endif

    // now try to add the complete hyp corresponding to this tuple
    // AddCompletedHyp is responsible for deleting the hyp ultimately, no matter whether it accepts it or prunes it
    // that means we have to set its pointer in pTuple to NULL so that it won't get deleted when destructing pTuple
    pTuple->m_pComplete = NULL;
    assert((*(beam.m_uniqeTuples.find(pTuple)))->m_pComplete == NULL);
    
    if (!AddCompletedHyp(pNewCompleteHyp, addAll)){
      // failure - the hyp was pruned away

#ifndef NDEBUG
      assert(!addAll);
      if (debug){cout<<"Failed beaming -> ";}
#endif

    }else{
      // success of adding the hyp
      if (debug){cout<<"Added "<<endl;}
      //track how many hyps you are actually getting thru AddCompletedHyp
      numAdded++;
      if (breakOutAttemptAverted) numAddedPostFailure++;
    }
  
    // should we still expand retrieved tuples' neighbors or only retrieve the remaining tuples from the beam?
    expandNeighbors = (count < CChartParserInfo::m_comboPruningBeamSize); 
    pTuple = beam.NextTuple(expandNeighbors); // retrieve next tuple
  }

  if ( //CChartParserInfo::m_comboPruningBeamSize>=FLT_MAX/10 && 
      CChartParserInfo::m_comboPruningFuzzCostDifference>=FLT_MAX/10){
    //cout<<"Checking to make sure beam is empty"<<endl;
    if (beam.Size()!=0){
      cout<<"BeamSize: "<<beam.Size()<<" count: "<<count<<endl;
    }
    assert(beam.Size()==0);
  }

  // some statistics
  pair<int, int> toReturn(numAdded, allPossibleExpansions);
  //cout<<"Spawn Added: "<<terminallyAdded+numAdded<<" / numTopCalls: "<<numTopCalls<<" initialNumTuples: "<<initialNumTuples<<" numTuplesProposedNonTerminal: "<<beam.NumTuplesProposed()<<" / allPossibleExpansions: "<<allPossibleExpansions<<" post-failure: "<<numAddedPostFailure<<endl;
  CChartParserInfo::m_hypsAdded += numAdded;
  CChartParserInfo::m_numTopCalls += count;
  //this variable is a good measure of how much work was done during combo pruning
  CChartParserInfo::m_numTuplesProposed += beam.NumTuplesProposed(); 
  CChartParserInfo::m_allPossibleExpansions += allPossibleExpansions;
  CChartParserInfo::m_numAddedPostFailure += numAddedPostFailure;

  //remove all memory associated with this beam
  //az: is now done by beam's destructor// beam.Clear();  
  if (debug){
    cout<<"*** End ***"<<endl;
  }
  clock_t spawnTimeEnd = clock(); //accumulate these values over the whole data set
  double spawncpu = ((double)  (-spawnTime + spawnTimeEnd)) / CLOCKS_PER_SEC;
  CChartParserInfo::m_spawntimecpu += spawncpu;
  return toReturn;
}





//return the K best (without LM) complete hyps that result from application of the rules inside this rule set
//void CChartParser::SpawnKBestCompleteHyps(vector<CRule>& matchedRules, CChartParserHypIncomplete* pNewIncompleteHypPrototype, int K, bool selfRuleMode){
//void CChartParser::SpawnKBestCompleteHyps(vector< pair<CChartParserHypIncomplete*, vector<CRule> > >& hypRuleCombinations, bool selfRuleMode){
pair<int, int> CChartParser::SpawnKBestCompleteHyps(RulesByProductionResult& hypRuleCombinations, bool addAll){

  bool debug = false;
  
  //do pruning and spawning for each production result that this k split is producing
  //this pruning happens before the BeamPruning call is made on all production results
  if (debug){
    cout<<"Start Spawning === "<<endl;
  }
  pair<int, int> expandedAndPossible;

  for (RulesByProductionResult::iterator iter=hypRuleCombinations.begin(); iter!=hypRuleCombinations.end(); ){
    CWordIndex productionResult = (*iter).first;
    RulesByProductionResult::iterator end = hypRuleCombinations.upper_bound(productionResult);
    
    //write out the rules for this cell filing
    if (CChartParserInfo::m_verbosity>0){
      for (RulesByProductionResult::iterator tmp = iter; tmp!=end; ++tmp){
	CChartParserHypIncomplete* pNewIncompleteHypPrototype = tmp->second.first; 
	CRule* pOneItem = &(tmp->second.second); // the rule found
	cout<<"RULE: "; DisplayInstantiatedRule(pOneItem, pNewIncompleteHypPrototype); cout<<endl;
      }
    }
    //iter < end defines the range
    pair<int, int> ep = SpawnForProductionResult(iter, end, addAll);
    expandedAndPossible.first+=ep.first;
    expandedAndPossible.second+=ep.second;
    //at the end, reset to the next production result
    iter=end;
  }
  if (debug){
    cout<<"End Spawning === "<<endl;
  }
  return  expandedAndPossible;

}

//display a rule, with its backpointer production results as instantiated by the pIncompleteHyp
void CChartParser::DisplayInstantiatedRule(CRule* pRule, CChartParserHypIncomplete* pIncompleteHyp){
  cout<<CChartParserInfo::m_pVocabulary->GetWord(pRule->m_productionResult)<<" -> ";
  for (int i=0; i<pRule->m_translationOutput.size(); i++){
    if (pRule->m_translationOutput[i]<=0){
      //cout<<" BP="<<m_translationOutput[i]<<" ";
      assert(pIncompleteHyp->m_productionBackPointers.size()>-1*pRule->m_translationOutput[i]);
      CProductionNode* pNode = pIncompleteHyp->m_productionBackPointers[-1*pRule->m_translationOutput[i]];
      cout<<" "<<CChartParserInfo::m_pVocabulary->GetWord(pNode->GetResult())<<"["<<pNode->I()<<","<<pNode->J()<<"] "; 
    }else{
      cout<<CChartParserInfo::m_pVocabulary->GetWord(pRule->m_translationOutput[i])<<" ";
    }
  }
}



// called by ApplyFundamentalRule and p_ApplySelfRule to create and file new complete hyps from pNewIncompleteHypPrototype
// and the matching transducer rule for each exact match returned and possibly file pNewIncompleteHypPrototype in the IH chart if exists(strict prefix match)
// if there was no strict prefix match, pNewIncompleteHypPrototype is deleted at the end of this method
// arv: return the ptr to the prototypical incomplete hyp, so that ProcessBTreeResult can indicate whether it delete the incomplete hyp or not
//CChartParserHypIncomplete* CChartParser::ProcessBTreeResult(CBTreeResult& bTreeQueryResult, CChartParserHypIncomplete* pNewIncompleteHypPrototype, bool selfRuleMode) {

//if it returns false you should delete the pNewIncompleteHypPrototype
//bool CChartParser::ProcessBTreeResult( vector< pair<CChartParserHypIncomplete*, vector<CRule> > >& hypRuleCombinations, CBTreeResult& bTreeQueryResult, CChartParserHypIncomplete* pNewIncompleteHypPrototype) {
bool CChartParser::ProcessBTreeResult( RulesByProductionResult& hypRuleCombinations, CBTreeResult& bTreeQueryResult, CChartParserHypIncomplete* pNewIncompleteHypPrototype) {
  bool s_debugApplyFund = false;

  // go through the exact matches and the possible prefix match
  if ((bTreeQueryResult.first & C_ExactMatch) != 0) { // at least one exact match

    //get the production result from the rule - and register the incomplete hyps (for the backpointers) and the generating rule
    //go thru each rule that was matched
    for ( vector<CRule>::iterator iter = bTreeQueryResult.second.begin(); iter!=bTreeQueryResult.second.end(); ++iter){
      CRule& oneItem = (*iter); // the rule found
      CWordIndex productionResult = oneItem.m_productionResult;
      // 	  if (productionResult == CChartParserInfo::m_FullSentenceNT && pNewIncompleteHypPrototype->J() != 1) {
      // 	    // Optimization:
      // 	    // we're about to fire an _S complete hyp into a cell with j > 1. Suppress this,
      // 	    // since _S hyps are only of use for j=1, i.e., for spans starting with source word 1.
      // 	    // This suppression results automatically in suppression of self rule application for _S
      // 	    // and thus of firing the incomplete hyp _S for the glue rule(s) @_S -> @_S @X / @1 @2 for each NT @X
	  
      // 	    // zero out the ExactMatch bit from the result info
      // 	    bTreeQueryResult.first &= ~C_ExactMatch;
      // 	  } else {
      // schedule the pair to be added as a complete hyp
      pair<CChartParserHypIncomplete*, CRule> payLoad(pNewIncompleteHypPrototype, oneItem);
      hypRuleCombinations.insert(pair<CWordIndex, pair<CChartParserHypIncomplete*, CRule> >(productionResult, payLoad));
      //	  }
    }

  }

  //checking the second least significant bits... we ONLY delete the incomplete hyp
  //if there are NO strict prefix matches. If there ARE strict prefix matches,
  //we add the incomplete hyp to the IC chart to account for this
  if (
      (bTreeQueryResult.first & C_StrictPrefixMatch) != 0 &&
      // optimization: don't create incomplete hyps if we've reached the complete-sentence span
      !(pNewIncompleteHypPrototype->I()==CChartParserInfo::m_sourceLength)
      ) { 
    // strict prefix match
    // we merely need to file pNewIncompleteHypPrototype in the IH chart
    if (s_debugApplyFund){
      cout<<"Found at least one strict prefix-match. ";
      cout	  <<"Therefore *scheduling* filing the new incomplete hyp prototype in the IH chart."<<endl;
    }
    //AddIncompleteHyp(pNewIncompleteHypPrototype); //files the hyp with its final m_totalRuleApplicationCount 
    
    return true;
  } else {
    //return false, to make it clear that you should delete the pNewIncompleteHypPrototype but ONLY after you have done any spawning that hypRuleCombinations has listed
    return false;
  }
  
}

//! A compliance check, make sure that you dont have situations where term -> @1 term -> @2 and @1->@2
/*
  bool CChartParser::SelfRuleCompliance(CProductionNode* pNode, const RulesByProductionResult& hypRuleCombinations){

  //we should verify that we are NEVER creating any rules whose production result is a not-new production node
  //ie term -> @1, term ->@2 and then in IndexNodesAccordingToRuleAppCount @1->@2
  //the result should either be in A) m_newlyCreatedNodes
  for (RulesByProductionResult::const_iterator iter=hypRuleCombinations.begin(); iter!=hypRuleCombinations.end(); ){
  CWordIndex productionResult = (*iter).first;
  
  CCompleteHypChartCell& chartCell = m_completeHypsChart[pNode->I()][pNode->J()];
  if (chartCell.productionResults.count(productionResult)==0){
  // then its fine
  }else{
  //look for this production result in the pCurrentNewNodes
  bool wasFound = false;
  //search for it in m_newlyCreatedNodes - this search is linear and its slow... but there should really be very few of these
  for (vector<CProductionNode*>::iterator prodIter = m_newlyCreatedNodes.begin(); prodIter!=m_newlyCreatedNodes.end(); ++prodIter){
  if ((*prodIter)->GetResult()==productionResult){
  wasFound = true;
  break;
  }
  }
  //or its now in the currently see m_productionNodes
  if (!wasFound){
  cout<<"m_newlyCreatedNodes: ";
  for (vector<CProductionNode*>::iterator prodIter = m_newlyCreatedNodes.begin(); prodIter!=m_newlyCreatedNodes.end(); ++prodIter){
  cout<<CChartParserInfo::m_pVocabulary->GetWord((*prodIter)->GetResult())<<" ";
  }
  cout<<endl;
  const CRule* pOneItem = &(iter->second.second); // the rule found
  cout<<"Rule: "; pOneItem->Display(); cout<<endl;
  cout<<"Generated by: "; pNode->Display(); cout<<endl;
  }

  //immediately return
  if (!wasFound){
  return wasFound;
  }
  }
  RulesByProductionResult::const_iterator end = hypRuleCombinations.upper_bound(productionResult);
  //at the end, reset to the next production result
  iter=end;
  }
  cout<<"Verified"<<endl;
  return true;
  
  }
*/

