#include <ParserInfo.hh>
#include <limits.h>
#include <assert.h>

using namespace std;
//============================================================================
// static initialization modified by FastTranslateChart
//============================================================================
//! For CChartParserHyp
vector<double> CChartParserInfo::m_scalingFactors = vector<double>();

//! For CChartParser
bool CChartParserInfo::m_GlueAndEOSAssumptions = true;
double CChartParserInfo::FutureCostImpact = 0.0;

int CChartParserInfo::m_sourceLength = 0;
CVocabulary* CChartParserInfo::m_pVocabulary = NULL;


int CChartParserInfo::BeamSizeFactor = 10;
int CChartParserInfo::m_recomputeLMCostsDuringKBest=0;
int CChartParserInfo::m_rescoreLM = 0;
int CChartParserInfo::C_HistoryLength = 2;
int CChartParserInfo::C_SRI_HistoryLength = 2;

bool CChartParserInfo::m_compareTargetWords = true;
bool CChartParserInfo::m_lmInCoarse = true;


//CSentence* CChartParserInfo::m_pSentence = NULL;
CWordSequence CChartParserInfo::m_pSentence = CWordSequence();

int CChartParserInfo::MAX_HYP_COUNT = 100;
//int CChartParserInfo::m_combinationCount = 0; //!< this one is initialized in InitSearch
int CChartParserInfo::m_maxCombinationCount = 0;
CWordIndex CChartParserInfo::m_unknownWI = -1;
CWordIndex CChartParserInfo::m_bosWI = -1;
CWordIndex CChartParserInfo::m_eosWI = -1;
CWordIndex CChartParserInfo::m_lmBoundaryWI = -1;
CWordIndex CChartParserInfo::m_FullSentenceNT = 0;
CWordIndex CChartParserInfo::m_GlueNT = 0;
int CChartParserInfo::m_stackLevel = 0;
bool CChartParserInfo::m_noTranslationOutputOfUnknownWords = false;
//nbest options
bool CChartParserInfo::m_extractUnique = true;
int CChartParserInfo::m_extractRecombine = 1; //!< Re-combine items as you do the n-best extraction, only applicable in cmp=0
int CChartParserInfo::m_nbestStyle = 2; //!< By default use the HC05 search

double CChartParserInfo::m_aStarBeam = 0.8; //! fuzz factor for termination of A* searhc
double CChartParserInfo::m_astarFuzz = 0.0; //! fuzz factor to terminate search

//for Pruning
NTPruningMap CChartParserInfo::m_pruningMap = NTPruningMap();
TagSet CChartParserInfo::m_tagsForLM = TagSet();
double CChartParserInfo::m_maxCostDifferencePerCell; //beam perc for production nodes
bool CChartParserInfo::m_pruningMapUseBoundary = false;


//combo pruning
float CChartParserInfo::m_comboPruningBeamSize; //!< used in the spawning process
float CChartParserInfo::m_comboPruningMinBeamSize; //!< used in the spawning process
float CChartParserInfo::m_comboPruningFuzzCostDifference; //!< used in the spawning process

bool CChartParserInfo::m_comboPruningGlobal=1; //!< does combo pruning across all incomplete hyps
bool CChartParserInfo::m_comboPruningUseLM=1; //!< use the LM during combo pruning
int CChartParserInfo::m_updateBeamSize=0; //!< update the beam size to allow BeamSizeFactor active hyps
//bool CChartParserInfo::m_comboPruningUseLexical; //!< use the Lexical hyps in the same beam during combo pruning
//rule app count operations to simulate how it used to work before (these are really bools, but the params handles stuff strangely)
int CChartParserInfo::m_updateHypRuleAppCount=1; //update the hyp's rule app count with the fine-grained eq class counts
int CChartParserInfo::m_lowestRuleAppCount=0; //use the lowest rule app count instead of the count of the lowest cost hyp
int CChartParserInfo::m_multiplicativeRA=0; //use the rule app count as a multiplicative factor
//system operation
int CChartParserInfo::m_verbosity=0; //!< how verbose is the output (0-1-2) 

int CChartParserInfo::m_saveAllAlternatives=1; //! if true, then we save all alternatives in the eq class (uses much more memory)
int CChartParserInfo::m_breadthAStar=1; //run the fast way
int CChartParserInfo::SAMTLM_UseBERFormat=0;
int CChartParserInfo::m_useSpanFeatures=0;
int CChartParserInfo::m_normalizeReferences = 0; //! Normalize references with the normalizing script


//decoding statistics -- see SpawnKBestCompleteHyps for details
int CChartParserInfo::m_hypsAdded=0;
double CChartParserInfo::m_spawntimecpu = 0;

int CChartParserInfo::m_hypsDeleted=0; //deleted after completion in PruneUnfiledHyps
int CChartParserInfo::m_hypsDeletedThreshold = 0; //total number of hyps deleted when finalizing the node due to threshold
int CChartParserInfo::m_hypsDeletedHistogram = 0; //total number of hyps deleted when finalizing the node due to histogram

int CChartParserInfo::m_numTopCalls =0; 
int CChartParserInfo::m_numTuplesProposed =0;  //number of tuples that the combo beam had to calculate costs for
int CChartParserInfo::m_allPossibleExpansions=0;
int CChartParserInfo::m_numAddedPostFailure = 0;
int CChartParserInfo::m_childBetter =0; 

//kbest stats
int CChartParserInfo::m_noCollisions=0;
int CChartParserInfo::m_uniqueCollisions=0;
int CChartParserInfo::m_betteredCollisions=0;
int CChartParserInfo::m_worseCollisions=0;
int CChartParserInfo::m_proposedAddedKBest=0;
int CChartParserInfo::m_topCallsKBest=0;
int CChartParserInfo::m_deletedKBest=0;
int CChartParserInfo::m_numUnique=0;
int CChartParserInfo::m_globalLMCache=0;
int CChartParserInfo::m_dontReloadBeam=0;

uint64_t CChartParserInfo::m_lmCall=0;
uint64_t CChartParserInfo::m_lmCacheHit=0;



NTOccurrenceStatisticsMap CChartParserInfo::m_nonterminalOccurrenceStatsMap = NTOccurrenceStatisticsMap();

//arv - dont need this anymore
//bool CChartParserInfo::s_maxCombinationCountExceeded = false;

//float CChartParserInfo::GetPruningCutoff(float lowestCost, double beamValue){
  //}
//}

//float CChartParserInfo::GetPruningCutoffPerWord(float lowestCost, double beamValue, int numWords ){
//}

void CChartParserInfo::LoadTagsForLM(const char* options){
  cout<<"Processing TagSet: "<<options<<endl;
  string optionsString = options;
  char* s = strtok((char*)optionsString.c_str(), ",");
  while (s!=NULL){
    CWordIndex nt = CChartParserInfo::m_pVocabulary->GetIndex(s);
    CChartParserInfo::m_tagsForLM.insert(nt);
    cout<<"AddingTag: "<<s<<" ";
    s = strtok(NULL, ",");
  }
  cout<<endl;
  
}

float CChartParserInfo::StringToFloat(string s){
  return StringToFloat(s.c_str());
}

float CChartParserInfo::StringToFloat(const char* s){

    double beta;
    if (strcmp(s, "inf")==0){
      beta = FLT_MAX/10; // very high number for infinity
    } else {
      beta = atof(s);
    }
    return beta;
}


void CChartParserInfo::LoadPruningMap(const char* options){
  cout<<"Processing PruningMap: "<<options<<endl;
  string optionsString = options;
  char* s = strtok((char*)optionsString.c_str(), "-");
  while (s!=NULL){
    //get the non-terminal - unless its -1
    CWordIndex nt;
    if (strcmp(s, "0")==0){
      //the generic non-terminal
      nt=0;
    }else{
      //cout<<"Getting Index: "<<s<<endl;
      //cout<<"Vocab:"<<CChartParserInfo::m_pVocabulary<<endl;
      nt = CChartParserInfo::m_pVocabulary->GetIndex(s);
      //az: the following will never happen as GetIndex creates indices for words on the fly
      //      if (nt==CVocabulary::UNK){
      //	cout<<"Unknown Non-Terminal in LoadPruningMap: "<<s<<endl;
      //	exit(-1);
      //      }
    }

    cout<<"AddingPruningOption: "<<s;
    //load the next 2 elements
    s = strtok(NULL, "-");
    double b = atof(s);

    s = strtok(NULL, "-");
    double beta;
    if (strcmp(s, "inf")==0){
      beta = FLT_MAX/10; // very high number for infinity
    } else {
      beta = atof(s);
    }
    cout<<" b="<<b<<" beta="<<beta<<endl;

    //    if (beta>1.0){
    //      cout<<"ERROR: Pruning is done as -log(1.0-beta) + minValue, which means beta <1.0"<<endl;
    //      cerr<<"ERROR: Pruning is done as -log(1.0-beta) + minValue, which means beta <1.0"<<endl;
    //      exit(-1);
    //    }

    //m_pruningMap.insert(pair<CWordIndex, pair<double, double> >(nt, pair<double, double>(b, beta)));
    m_pruningMap[nt] = pair<double, double>(b, beta);

    //move it forward
    s = strtok(NULL, "-");
  }
}




// /*
// float CChartParserInfo::UnwindHyp_LMProcessWord(CWordSequence& processedWords, CWordIndex w, bool allowBOS) {
//   float pHypLMCost = 0.0;
//   if (processedWords.size() < C_HistoryLength) {
//       // our history context is still too short - simply push the word in
//       processedWords.push_back(w);
//       if (processedWords.size() == C_HistoryLength) {
// 	// it just filled up - compute costs of first words given beginning of sentence
// 	// UNLESS the BOS has already been computed as part of the future cost estimate
// 	if (allowBOS){
// 	  pHypLMCost = CChartParserInfo::GetBeginOfSentenceLMCost(processedWords);
// 	}
//       }
//   } else {
//     // compute prob of *item given history (which we stored in firstWords)
//     pHypLMCost = CChartParserInfo::GetLMCost(processedWords, w);
//     // add the processed word to the history context
//     processedWords.push_back(w);
//   }
//   return pHypLMCost;
// }
// //calculates your EOS cost with your workingContext (series of words at the end of the sentence)
// float CChartParserInfo::UnwindHyp_EOS(const CWordSequence& workingContext){
//   CWordSequence lastContext;
//   int contextSize = 0;
//   for (CWordSequence::const_reverse_iterator riter = workingContext.rbegin(); riter!=workingContext.rend() && (contextSize<C_HistoryLength); riter++){
//     lastContext.insert(lastContext.begin(), *riter);
//     contextSize++;
//   }
  
//   return CChartParserInfo::GetEndOfSentenceLMCost(lastContext);
// }
// */


// float CChartParserInfo::GetEndOfSentenceLMCost(const CWordSequence& words){
  
//   if (CChartParserInfo::m_rescoreLM==1){
//     return 0.0;
//   }


//   if (CChartParserInfo::m_pLM==NULL){
//     return 1.0;
//   }
  
//   //cout<<"EOS LM: "; PrintVocabSequence(words); cout<<endl;
//   //pass the context in reverse
//   vector<CWordIndex> wordsCopy;
//   wordsCopy.insert(wordsCopy.end(), words.rbegin(), words.rend());

//   //get the EOS log-prob - if wordsCopy.size()==1 you will be a reduced gram estimate ie p(eos|w), NOT p(eos | w, bos)
//   float firstPart = -1.0*(CChartParserInfo::m_pLM->EOSLogProb(wordsCopy));
//   //cout<<"firstPart: "<<firstPart<<endl;
//   return firstPart;
// }

// float CChartParserInfo::GetBeginOfSentenceLMCost(const CWordSequence& words) {
  
//   if (CChartParserInfo::m_rescoreLM==1){
//       return 0.0;
//   }


//   if (CChartParserInfo::m_pLM==NULL){
//     return words.size();
//   }else{
//     //cout<<"BOS LM:"<<CChartParserInfo::m_pLM<<endl;
//   }

//   //What we want in the end is the cost of the first word given the beginning-of-sentence, plus the cost of the 
//   //second word given BOS,firstword
//   float firstWordsLMCost = 0.0;
//   //  CWordSequence tmpUnprocessedWords; //irrelevant to us
//   //  CWordSequence tmpMostRecentWords(C_HistoryLength, CChartParserInfo::m_unknownWI); // history context is unknown, but complete
//   //  //  words to be sucessively LM-processed, therafter pushed into tmpMostRecentWords
//   //  /*
//   //  cout<<"BOS Calc =============="<<words.size()<<endl;
//   //  CChartParserInfo::PrintVocabSequence(words);
//   //  cout<<endl;
//   //  */
  
//   //CachedSriLM::s_debugLM = true;
//   //ProcessLMWords(tmpUnprocessedWords,tmpMostRecentWords,words,firstWordsLMCost);

//   //ARV style
//   vector<CWordIndex> wordsCopy;
//   wordsCopy.insert(wordsCopy.end(), words.rbegin(), words.rend());
//   float firstPart = -1.0*(CChartParserInfo::m_pLM->BOSLogProb(wordsCopy));
//   wordsCopy.erase(wordsCopy.begin());

//   //if there was only 1 word of context available, then only calculate for the one word
//   while(wordsCopy.size()!=0){
//     float secondPart = -1.0*(CChartParserInfo::m_pLM->BOSLogProb(wordsCopy));
//     //CachedSriLM::s_debugLM = false;
//     //assert(tmpUnprocessedWords.size() == 0); // all words should have been processed as the history context was complete
//     firstPart += secondPart;
//     wordsCopy.erase(wordsCopy.begin());
//   }
//   return firstPart;

// }

// float CChartParserInfo::ComputeBoundaryWordsLM(const CWordSequence& firstWords){

//   if (CChartParserInfo::m_rescoreLM==1){
//     return 0.0;
//   }

//   CWordSequence firstWordsRev = firstWords;
//   reverse(firstWordsRev.begin(), firstWordsRev.end());
//   //firstWords: the dog was
//   //firstWordsRev: was dog the (in wordMostNext style)

//   //cout<<"DEBUG: "; PrintVocabSequence(firstWords); cout<<endl; 
//   float lmcost = 0.0;
//   while (firstWordsRev.size()>0){
//     //cout<<"entering.."<<endl;
//     float logProb = -1.0*(CChartParserInfo::m_pLM->LogProb(firstWordsRev));
//     //cout<<"logProb: "<<logProb<<endl;
//     lmcost += logProb; 
//     //and remove the first element
//     //firstWordsRev: dog the
//     firstWordsRev.erase(firstWordsRev.begin());
//   }
//   return lmcost;
// }








void CChartParserInfo::PrintParsingOutput(const CWordSequence& v, ostream& os){
  //!Static utility method that avoids having to write this loop every time
  for (CWordSequence::const_iterator iter=v.begin(); iter!= v.end(); ++iter){
      if ((*iter)<=0){
	os <<*iter<<" ";
      }else{
	//      cout<<" ("<<m_pVocabulary->GetWord(*iter)<<") ";
	  //	  string wd = CChartParserInfo::m_pVocabulary->GetWord(*iter);
	  //os<<wd;
	  
	
	//  //if we see a NT, conjoin the next two elements, this is the start end span
	//  if (wd[0]=='@'){

	// if we see the start-subtree symbol '#(', conjoin all following elements upto '#'
	if (CChartParserInfo::m_pVocabulary->GetIndex("#(") == *iter) {
	  os << "( ";
	  assert(iter!=v.end());
	  ++iter;
	  PrintParsingOutput_SafePrint(*iter, os);
	  // also, treat the next two elements as numbers, not word indices (because they represent the source span)
	  assert(iter!=v.end());
	  ++iter;
	  os<<"_" << *iter;
	  assert(iter!=v.end());
	  ++iter;
	  os<< "-" << *iter;
	  assert(iter!=v.end());
	  ++iter;
	  while (*iter != CChartParserInfo::m_pVocabulary->GetIndex("#") ) {
	    os << "_";
	    PrintParsingOutput_SafePrint(*iter, os);
	    ++iter;
	  }
	} else {
	  PrintParsingOutput_SafePrint(*iter, os);
	}	  
	os<<" ";
      }
  }
}

//! Sub of PrintParsingOutput
void CChartParserInfo::PrintParsingOutput_SafePrint(const int i, ostream& os){
  if (i == CChartParserInfo::m_pVocabulary->GetIndex("#(") ) {
    os << "(";
  } else if (i == CChartParserInfo::m_pVocabulary->GetIndex("#)") ) {
    os << ")";
  } else if (i == CChartParserInfo::m_pVocabulary->GetIndex("(") ) {
    os << "#LeftParen";
  } else if (i == CChartParserInfo::m_pVocabulary->GetIndex(")") ) {
    os << "#RightParen";
  } else if (i == CChartParserInfo::m_pVocabulary->GetIndex("\"") ) {
    os << "#QuotationMark";
  } else {
    os << CChartParserInfo::m_pVocabulary->GetWord(i);
  }
}



//!Static utility method that avoids having to write this loop every time
void CChartParserInfo::PrintVocabSequence(const CWordSequence& v, ostream& os){
  for (CWordSequence::const_iterator iter=v.begin(); iter!= v.end(); ++iter){
    if ((*iter)<=0){
      os << *iter;
    }else{
	//      cout<<" ("<<m_pVocabulary->GetWord(*iter)<<") ";
      if (m_pVocabulary!=NULL){
	os<<m_pVocabulary->GetWord(*iter);
      }else{
	os<<*iter;
	}
    }
    //if not end
    if (iter!=v.end()-1){
	os<<" ";
    }
  }
}



