#ifndef __UniqueCompleteHyps__
#define __UniqueCompleteHyps__

#include <ParserInfo.hh>
#include <ParsingHeader.hh>

//Returns a hash code based on the words translated of the partially unwound pHyp and the address of the back pointers remaining
struct HashUniqueTargetsHyp{

  //add to basic hash code for collapsing sequences
  void LMContextAndBackpointers(const CChartParserHypComplete* pHyp, size_t& hashCode)const;
  //adds to basic hash code for uniquing
  void TerminalsAndBackpointers(const CChartParserHypComplete* pHyp, size_t& hashCode)const;
  size_t operator()(const CChartParserHypComplete* pHyp)const;
};

//!Track unique hyps (only track the hashCodes, no way of recovering the original hyps)
typedef SGI_HASH_NAMESPACE::hash_map<unsigned long, double> TUniqueHyps ; //!< maintains just hash entries of each hyp to save memory (hash entry + cost)

//!Utility class that keeps track of which hyps represent the same target sequence
/*!
 * Comparison is done based on target sequence AND memory addresses of remaining backpointers
 */
class CUniqueCompleteHyps{
  
public:

  CUniqueCompleteHyps(){}
  //! Check to see if a hyp is unique, if it is, then save its hash, return the hyp, otherwise return NULL
  CChartParserHypComplete* CheckUnique(CChartParserHypComplete* pHyp);
  int Size(){ return m_uniqueHyps.size(); }
private:

  TUniqueHyps m_uniqueHyps;   //! Internal store for hashes of hyps
};


#endif
