#include <iostream>
#include <fstream>
#include <OptimizeMER.hh>
#include <Vocabulary.hh>
#include <BleuSearch.hh>
#include <TERSearch.hh>
#include <ParserInfo.hh>
//!classes to take in command line options
//#include <param.hh>
//#include <paramdef.hh>
#include <Parameter.h>
#include <NBestList.hh>

#define MAXBUF_T 1024

using namespace std;

void LoadInit(const char* fileName, vector< Lambdas >& rangeAndInit);
//void LoadFeats(const char* fileName, int numModels, vector< Features*>& allFeats, vector<RawCounts*>& allScores, int linelimit=-1);
void LoadFeats(ifstream& fp, int numModels, vector< Features*>& allFeats, vector<RawCounts*>& allScores, int linelimit=-1); //use the numlines to indicate how many lines to read, by default all
void LoadFeatsViaNBest(ifstream& fp, int numModels,  vector< Features*>& allFeats, vector<RawCounts*>& allScores, vector<int>& candPairsOld,  vector<int>& candPairsNew, vector<NBestList*>& nBestLists);
void LoadCandidatePairsAsNumCands(const char* fileName, vector<int >& candPairs);
void ParseDoubleSequence(const char* floatSequenceStr, vector<double>& doubleVector) ;

int main(int argc, char** argv){


  //CParam Param;
  string FeatsFile = "";
  string CandsFile = "";
  string InitFile = "";
  string ScoringMetric = "IBMBLEU";
  int OptiViaNBest=1; //pass everything via the n-best list to unique it

  int OptiNumStarts=1;
  int OptiNumPermutes=1; 
  int OptiNumModify=4;
  int OptiFirst=0;
  int OptiDoNorm=1;
  int OptiTwoStage=1;
  double OptiEpsilon=0.0;
  double OptiEpsilonScale=1.0;
  
  //Param.AddParamDef("FeatsFile", '\0' &FeatsFile, "", "Feature file including feature values and metric scores", True);
  //Param.AddParamDef("CandsFile ", '\0' &CandsFile, NULL, "Cand file indicating number of candidates per source sentences", True);
  //Param.AddParamDef("ScoringMetric ", '\0' &ScoringMetric, NULL, "Scoring metric to use in optimization IBMBLEU, NISTBLEU or NIST", True);

  //Param.AddParamDef("FeatsFile", '\0', &FeatsFile, NULL, "Features for each candidate translation", True);
  //Param.AddParamDef("CandsFile", '\0', &CandsFile, NULL, "Number of candidates for each sentence ", True);
  //Param.AddParamDef("InitFile", '\0', &InitFile, NULL, "3 lines - with the initial scaling factors and the min max ranges", True);
  //Param.AddParamDef("ScoringMetric", '\0', &ScoringMetric, NULL, "ScoringMetric used to run MER", True);


 //==========================================================================================
   // Optimization MER parameters
   //==========================================================================================
  getGlobalParSet().insert(new Parameter<string>("FeatsFile",ParameterChangedFlag, "Features for each candidate translation",FeatsFile,0));
  getGlobalParSet().insert(new Parameter<string>("CandsFile",ParameterChangedFlag, "Number of candidates for each sentence ",CandsFile,0));
  getGlobalParSet().insert(new Parameter<string>("InitFile",ParameterChangedFlag, "3 lines - with the initial scaling factors and the min max ranges",InitFile,0));
  getGlobalParSet().insert(new Parameter<string>("ScoringMetric",ParameterChangedFlag, "ScoringMetric used to run MER",ScoringMetric,0));
  getGlobalParSet().insert(new Parameter<int>("Opti_NumStarts",ParameterChangedFlag,"Number of starts of f the optimizer 1==just use the initial values ",OptiNumStarts,0));
  getGlobalParSet().insert(new Parameter<int>("Opti_NumPermutes",ParameterChangedFlag,    "Number of times you change a few dimension and try to reconverg 1==just use what the initial values ",OptiNumPermutes,0));
  getGlobalParSet().insert(new Parameter<int>("Opti_NumModify",ParameterChangedFlag,"Number of features that are zeroed out in each iteration ",OptiNumModify,0));
  getGlobalParSet().insert(new Parameter<int>("Opti_First",ParameterChangedFlag,  "Parameter index that the optimization should start blinking from ",OptiFirst,0));
  getGlobalParSet().insert(new Parameter<int>("Opti_DoNorm",ParameterChangedFlag,"Normalize scaling factors after each iteration ",OptiDoNorm,0));
  getGlobalParSet().insert(new Parameter<int>("Opti_TwoStage",ParameterChangedFlag, "Run 2 stage optimization - nulling out certain fields and re-optimizing - Dont use this option AND Opti_Num... >1 ",OptiTwoStage,0));
  getGlobalParSet().insert(new Parameter<double>("Opti_Epsilon",ParameterChangedFlag, "Optimize until the last dimension's error gain is within Opti_Epsilon of the previous value",OptiEpsilon,0));
  getGlobalParSet().insert(new Parameter<double>("Opti_EpsilonScale",ParameterChangedFlag, "Scale up the threshold when using two stage optimization to save time",OptiEpsilonScale,0));
  getGlobalParSet().insert(new Parameter<int>("Opti_ViaNBest",ParameterChangedFlag, "Pass the feats thru the nbest lists",OptiViaNBest,0));

  parseArguments(argc, argv);

  /*
  Param.AddParamDef("Opti_NumStarts", '\0', &OptiNumStarts, 1,"Number of starts of f the optimizer 1==just use the initial values ",True);

  Param.AddParamDef("Opti_NumPermutes", '\0', &OptiNumPermutes, 1,	"Number of times you change a few dimension and try to reconverg 1==just use what the initial values ",True);

  Param.AddParamDef("Opti_NumModify", '\0', &OptiNumModify, 4,"Number of features that are zeroed out in each iteration ",True);

  Param.AddParamDef("Opti_First", '\0', &OptiFirst, 0,	"Parameter index that the optimization should start blinking from ",	True);

  Param.AddParamDef("Opti_DoNorm", '\0', &OptiDoNorm, 1,"Normalize scaling factors after each iteration ",True);

  Param.AddParamDef("Opti_TwoStage", '\0', &OptiTwoStage, 1, "Run 2 stage optimization - nulling out certain fields and re-optimizing - Dont use this option AND Opti_Num... >1 ",True);

  Param.AddParamDef("Opti_Epsilon", '\0', &OptiEpsilon, 0.0, "Optimize until the last dimension's error gain is within Opti_Epsilon of the previous value",True);

  Param.AddParamDef("Opti_EpsilonScale", '\0', &OptiEpsilonScale, 1.0, "Scale up the threshold when using two stage optimization to save time", True);

  Param.AddParamDef("Opti_ViaNBest", '\0', &OptiViaNBest, 1, "Pass the feats thru the nbest lists", True);
 
  //interpret the parameters
  if (! Param.Interpret( argv, argc) ) {
    cerr << "\nParameters could not be interpreted correctly\n ... exiting!\n";
    exit(-1); 
  }
  */

  if ((FeatsFile=="") || (CandsFile=="") || (InitFile=="") || (ScoringMetric=="")){
    cout<<"Usage: MER --FeatsFile --CandsFiles --InitFile --ScoringMetric (must be supplied)"<<endl;
    exit(-1);
  }
  if ((ScoringMetric == "IBMBLEU") || ( ScoringMetric == "NISTBLEU") || (ScoringMetric == "NIST") || (ScoringMetric=="TER")){
    cout<<"Scoring with:"<<ScoringMetric<<endl;
  }else{
    cout<<"Unsupported scoring metric:"<<ScoringMetric<<endl;
    exit(-1);
  }

  //scaling factors for the parser
  
  //vector<double> ScalingFactors;
  //cout << "Scaling factors (feature weights) for parsing:";
  //ParseDoubleSequence(FeatureWeightsParsing, ScalingFactors);
  CVocabulary *pVoc = new CVocabulary;
  CChartParserInfo::m_pVocabulary = pVoc;
  

  
  vector<int> candPairs;
  vector< Lambdas > rangeAndInit;
  vector< Features*> allFeats;
  vector<RawCounts*> allScores;
  vector<NBestList*> nBestLists; //n-best lists get filled - we dont bother deleting them because the program is going to end


  LoadInit(InitFile.c_str(), rangeAndInit);
  LoadCandidatePairsAsNumCands(CandsFile.c_str(), candPairs);
  assert(rangeAndInit.size()==3);
  cout<<"About to load feats"<<endl;
  int numModels = rangeAndInit[0].size();
  cout<<"numModels="<<numModels<<endl;

  ifstream fp_feats(FeatsFile.c_str());
  if (!fp_feats.good()){
    cerr<<"Unable to open:"<<FeatsFile<<endl;
  }

  //LoadFeats(fp_feats, numModels, allFeats, allScores);
  if (OptiViaNBest){
    vector<int> candPairsNew;
    LoadFeatsViaNBest(fp_feats, numModels, allFeats, allScores, candPairs, candPairsNew, nBestLists);
    candPairs = candPairsNew;
  }else{
    LoadFeats(fp_feats, numModels, allFeats, allScores);
  }

  MetricSearch* pMetric;
  if (ScoringMetric == "TER"){
    pMetric = new TERSearch();
  }else{
    pMetric = new BleuSearch(ScoringMetric);
    ((BleuSearch*)pMetric)->SetNGramConsidered(4);
  }


  COptimizeMER test(pMetric);

  test.Init(rangeAndInit[0], rangeAndInit[1], OptiNumPermutes, OptiNumModify, OptiNumStarts, OptiTwoStage, OptiDoNorm, OptiEpsilon, OptiEpsilonScale);
  test.SetStart(OptiFirst); 
  //set this in the MER.cc (and not FastTranslateChart) because this MER.cc only runs opti once - thus never dealing
  //with the remaining unblanked features. Callers of this MER.cc must manage the OptiFirst variable by themselves


  cout<<"Running opti init with:";
  copy(rangeAndInit[2].begin(), rangeAndInit[2].end(), ostream_iterator<float>(cout, " "));
  cout<<endl;
  Lambdas returnLambdas = test.minModelCostMaxMetricScore(allFeats, allScores, candPairs, rangeAndInit[2], allFeats.size());
  cout<<"MER Completed:\nGot Params: "; copy(returnLambdas.begin(), returnLambdas.end(), ostream_iterator<double>(cout, "_"));   cout<<endl;

}

//the cand pairs can change now after going thru the nbest list 
void LoadFeatsViaNBest(ifstream& fp, int numModels,  vector< Features*>& allFeats, vector<RawCounts*>& allScores, vector<int>& candPairsOld,  vector<int>& candPairsNew, vector<NBestList*>& nBestLists ){

  //we are going to read thru the file, process the right amount of features to creat 
      
  for (int i=0; i<candPairsOld.size(); i++){
    NBestList* pNB = new NBestList;
    //dont keep any translations
    pNB->Init(false, true);
    nBestLists.push_back(pNB);
    
    vector< Features*> allFeatsOneSent;
    vector<RawCounts*> allScoresOneSent;

    int numcands = candPairsOld[i];
    LoadFeats(fp, numModels, allFeatsOneSent, allScoresOneSent, numcands); //just pull out the number that u need
    
    //load up the nbest list
    if (allFeatsOneSent.size() != allScoresOneSent.size()){
      cerr<<" feats and scores not the same size for nbest list: "<<i<<" feats: "<<allFeatsOneSent.size()<<" scores: "<<allScoresOneSent.size()<<endl;
      exit(-1);
    }

    //when you get stuff into the nbest list, it has a memory of its own
    string translation = "<s>";
    for (int d = 0; d<allFeatsOneSent.size(); d++){
      pNB->InsertCandidate(*(allFeatsOneSent[d]), CChartParserInfo::makeSequence<CWordSequenceVector>(translation), (*allScoresOneSent[d]));
    }

    //delete all this one sentence stuff that got created
    for (int d = 0; d<allFeatsOneSent.size(); d++){
      delete allFeatsOneSent[d];
      delete allScoresOneSent[d];
    }
  }


  //now harvest the pointer from the n-best list stored data
  for (vector<NBestList*>::iterator iter = nBestLists.begin(); iter!= nBestLists.end(); ++iter){
    NBestList* oneSent = (*iter);
    oneSent->LoadOptimizationData(allFeats, allScores, candPairsNew);
  }
}

//void LoadFeats(const char* fileName, int numModels, vector< Features*>& allFeats, vector<RawCounts*>& allScores, int linelimit){
void LoadFeats(ifstream& fp, int numModels, vector< Features*>& allFeats, vector<RawCounts*>& allScores, int linelimit){

  char buf[MAXBUF_T];
  buf[0] = '\0';
  int numLines = 0;
  for (; !fp.eof(); ){
    fp.getline(buf, MAXBUF_T);
    if (strlen(buf)>0){
      vector<double> l;
      const char* first = strtok(buf, " ");
      while (first!=NULL){
	l.push_back(atof(first));
	first = strtok(NULL," ");
      }

      //cout<<"Read feats:"; copy(l.begin(), l.end(), ostream_iterator<double>(cout, " ")); cout<<" numModels: "<<numModels<<endl;
      
      Features* f = new Features();
      f->reserve(numModels);
      f->insert(f->end(), l.begin(), l.begin()+numModels);
      //cout<<"size: "<<f->size()<<endl;
      //cout<<"Read f:"; copy(f->begin(), f->end(), ostream_iterator<double>(cout, " ")); cout<<endl;
      assert(f->size()==numModels);
      
      vector<float> scoreVector;
      scoreVector.reserve(NSZ*2+1);
      scoreVector.insert(scoreVector.end(), l.begin()+numModels, l.end());
      //cout<<"Read scoreVector:"; copy(scoreVector.begin(), scoreVector.end(), ostream_iterator<double>(cout, " ")); cout<<endl;
      
      RawCounts* r = new RawCounts();
      for (int i=0; i<NSZ; i++){
	int locInVector = i*2;
	assert(scoreVector.size()>locInVector+1);
	r->numCorrect[i] = scoreVector[locInVector];
	r->numSuggested[i] = scoreVector[locInVector+1];
      }
      //cout<<"done counts:"<<scoreVector.size()<<endl;
      assert(scoreVector.size()==NSZ*2+1);
      r->referenceLengths = scoreVector[scoreVector.size()-1];

      /*
      cout<<"Features: ";copy(f->begin(), f->end(), ostream_iterator<double>(cout, " "));
      cout<<"Counts: ";
      for (int i=0; i<NSZ; i++){
	cout<<i<<"="<<r->numCorrect[i]<<"/"<<r->numSuggested[i]<<" ";
      }
      cout<<r->referenceLengths;
      cout<<endl;
      */

      //this way of doing it does not handle uniquing - use other version of this function
      
      allFeats.push_back(f);
      allScores.push_back(r);

      numLines++;
      //cout<<"numLines: "<<numLines<<" linelimit: "<<linelimit<<endl;

      //if you hit the number of lines that u were supposed to read... then just go home now
      if (numLines==linelimit){
	return;
      }
      if (numLines%10000==0){
	cout<<"loading feats: "<<numLines<<endl;
      }
    }
  } 
}

void LoadInit(const char* fileName, vector< Lambdas >& rangeAndInit){
  ifstream fp(fileName);
  if (!fp.good()){
    cerr<<"Unable to open:"<<fileName<<endl;
  }

  char buf[MAXBUF_T];
  buf[0] = '\0';
  for (; !fp.eof(); ){
    fp.getline(buf, MAXBUF_T);
    if (strlen(buf)>0){
      string candsLine(buf);
      Lambdas l;
      const char* first = strtok(buf, "_");
      while (first!=NULL){
	l.push_back(atof(first));
	first = strtok(NULL,"_");
      }
      cout<<"Init: "; copy(l.begin(), l.end(), ostream_iterator<double>(cout, " ")); cout<<endl;
      rangeAndInit.push_back(l);
      if (rangeAndInit.size()>1){
	assert(rangeAndInit[rangeAndInit.size()-1].size() == rangeAndInit[rangeAndInit.size()-2].size());
      }
    }
  }
  //if there was only one line initialized - ie the scaling factor line
  if (rangeAndInit.size()==1){
    Lambdas minl, maxl;
    for (int i=0;i<rangeAndInit[0].size();i++) {
      minl.push_back(-100);
      maxl.push_back(100);
    }
    rangeAndInit.insert(rangeAndInit.begin(), maxl);
    rangeAndInit.insert(rangeAndInit.begin(), minl);

  }
  assert(rangeAndInit.size()==3);
}


void LoadCandidatePairsAsNumCands(const char* fileName, vector< int >& candPairs){

  ifstream fp(fileName);
  if (!fp.good()){
    cerr<<"Unable to open:"<<fileName<<endl;
  }

  char buf[MAXBUF_T];
  buf[0] = '\0';
  for (; !fp.eof(); ){
    fp.getline(buf, MAXBUF_T);
    if (strlen(buf)>0){
      string candsLine(buf);
      const char* first = strtok(buf, " ");
      const char* second = strtok(NULL," ");


      //int snum;
      //int numCands;
      //ss>>snum>>numCands;
      int snum = atoi(first);
      int numCands = atoi(second);
      cout<<"Sentence: "<<snum<<" numCands: "<<numCands<<endl;
      candPairs.push_back(numCands);
    }
  }
  cout<<"Cands loaded ========== "<<endl;
}

void ParseDoubleSequence(const char* floatSequenceStr, vector<double>& doubleVector) {
  doubleVector.clear();
  const char* element = strtok( (char*)floatSequenceStr, "_" );
  while ( element != NULL ) {
    doubleVector.push_back( atof( element ) );
    element = strtok(NULL, "_");
  }
}
