#include <iostream>
#include <vector>
#include <map>
#include <string>
#include <sstream>
#include <fstream>
#include <iterator>
#include <iomanip>
#include <math.h>

//====================================================================
// version of merge counts that works on sorted rules. it assumes that
// all rules with the same src,tgt,prod are listed sequentially
//====================================================================

using namespace std;


typedef pair< vector<string>, vector<double> > AccRule; //data for accumulated rules

vector<double> g_rcosts; //global rule costs
vector<double> g_pcosts; //global phrase count costs
vector<double> g_ccosts; //global phrase count costs

vector<double> g_rulecostvector; //global vector for rules
vector<double> g_countcostvector; //global vector for rules
vector<double> g_phrasecostvector; //global vector for rules

vector<string> g_parts;
vector<string> g_pureparts;

time_t startTime;

int numfeatures_count = 4;
int numfeatures_phrase = 1;


//all set from the cmd line
int lexminfreq;
int nonlexminfreq;
int totalnumfeatures;
int originalfeaturecount;

int doaverage;

//tracking progress
int numlines=0;
int numunique=0;

//track for each of types
int numunique_phrase = 0;
int numwritten_phrase = 0;

int numunique_rule = 0;
int numwritten_rule = 0;

int numunique_rule_nonlex = 0;
int numwritten_rule_nonlex = 0;

//void Tokenize(char* str, vector<string>& tokens,const char* delim)
void Tokenize(string str, vector<string>& tokens,const string delimiters )
{
  
  // Skip delimiters at beginning.
  string::size_type lastPos = str.find_first_not_of(delimiters, 0);
  // Find first "non-delimiter".
  string::size_type pos     = str.find_first_of(delimiters, lastPos);

  while (string::npos != pos || string::npos != lastPos)
    {
      // Found a token, add it to the vector.
      tokens.push_back(string(str.substr(lastPos, pos - lastPos)));
      // Skip delimiters.  Note the "not_of"
      lastPos = str.find_first_not_of(delimiters, pos);
      // Find next "non-delimiter"
      pos = str.find_first_of(delimiters, lastPos);
    }
  
  /*
  char* pch = strtok(str, delim);
  while (pch != NULL){
    tokens.push_back(string(pch));
    pch = strtok(NULL, delim);
  }
  */

}

//Format: @_COUNT#@1#result#v1 v2
void handleCounts(vector<string>& parts, vector<double>& ccosts){
  string counttag = parts[0];
  string at1 = parts[1];
  string result = parts[2];
  string countvals = parts[3];
  //get the count data, store it in rulecostvector

  stringstream ss(countvals);
  double c;
  int i = 0;
  while (ss >> c){
    g_countcostvector[i++] = c;
  }
  
  //accululate into ccosts
  for (int i=0; i<ccosts.size(); i++){
    ccosts[i] += g_countcostvector[i];
  }

}

bool ispurelexical(string& source){

  g_pureparts.clear();
  //char* source_str = (char*) source.c_str();
  //Tokenize(source_str, g_pureparts, " ");
  Tokenize(source, g_pureparts, " ");
  for (vector<string>::iterator iter=g_pureparts.begin(); iter!= g_pureparts.end(); ++iter){
    char c= (*iter)[0];
    if (c=='@'){
      return false;
    }
  }
  return true;
}


  //New: we dont see any phrase counts any more

//note: in this case the key has its trailing # on it
void printPhraseVector(string& key, vector<string>& parts, vector<double>& pcosts){
  //the first non label must have the count
  bool islex = ispurelexical(parts[1]);

  int cutoff = (islex) ? lexminfreq : nonlexminfreq;
  cutoff = max(2, cutoff);

  if (pcosts[0]+0.1>=cutoff){
    cout<<key;
    copy(pcosts.begin(), pcosts.end(), ostream_iterator<double>(cout, " ")); 
    cout<<endl;
    numwritten_phrase++;
  }
  numunique_phrase++;
}

//Format: @_PHRASE#.*#.*#v1
void handlePhraseCounts(string& key, string countvals, vector<double>& pcosts){

  stringstream ss(countvals);
  double c;
  int i =0;
  while (ss >> c){
    g_phrasecostvector[i++]= c;
  }
  
  for (int i=0; i<pcosts.size(); i++){
    pcosts[i] += g_phrasecostvector[i];
  }
}



//void handleRules(vector<string>& parts, RuleMap& rmap, int originalfeaturecount, int totalnumfeatures){
void handleRules(string& key, vector<string>& parts, vector<double>& rcosts){
  //string src = parts[0];
  //string tgt = parts[1];
  //string result = parts[2];
  string costs = parts[3];

  //set up this rule's cost vector, we will need it whatever happens. all the value in this vector
  //REMAIN the way they come in (log space) EXCEPT the freq rules that we call exp on. Any features
  //that are after the original count will be maxed
  stringstream ss(costs);
  double c;
  int i = 0;
  while (ss >> c){
    g_rulecostvector[i++] = c;   
  }

  //cout<<"Rulecostvector: "<<costs<<endl;
  //in all cases, convert the 6,7th indices to be be back in prob space, ie in freq counts
  g_rulecostvector[6] = exp(-g_rulecostvector[6]); //the real freq
  g_rulecostvector[7] = round(exp(-g_rulecostvector[7])); //the simple freq
  
  if (rcosts[7] == 0){
    rcosts = g_rulecostvector;
  }else{
    //vector<double>* tmpcostvectorPtr = &(rcosts);
    // we are going to do this all based on the simple frequency (like in extractrules) based on having a simple count

    if (doaverage==1){
      double existingfrequency = rcosts[7];
      double newruleFrequency = g_rulecostvector[7];
      double normalization_constant = log(existingfrequency + newruleFrequency);
      
    //the first 3 fields are binary values (glue, rule app count, tgt word count)
      for (int i = 3; i < originalfeaturecount; i++) {
	//dont update the freq field or fields where the value is the same
	if ( (i != 6) && (i !=7) && (g_rulecostvector[i] != rcosts[i]) ) {
	  rcosts[i] = -log(
			   exp(-rcosts[i]) * existingfrequency + exp(-g_rulecostvector[i]) * newruleFrequency
			   ) + normalization_constant;
	}      
      }//processed all fields except 6,7
    }

    //now max the rest of the features (these features are left in their prob space)
    for (int i= originalfeaturecount; i<totalnumfeatures; i++){
      rcosts[i] = (rcosts[i] > g_rulecostvector[i]) ? rcosts[i] : g_rulecostvector[i];
    }
    
    //increment the prob
    rcosts[6] += g_rulecostvector[6];
    rcosts[7] += g_rulecostvector[7];
    //cout<<"saving: "<<(*tmpcostvectorPtr)[6]<<" "<<(*tmpcostvectorPtr)[7]<<endl;
  }
}



void printCountVector(string& key, vector<double>& ccosts){

  cout<<"@_COUNT"<<"#@1#"<<key<<"#";
  copy(ccosts.begin(), ccosts.end(), ostream_iterator<double>(cout, " ")); 
  cout<<endl;
  
}


//Prints out the rule based on the rcosts
void printRuleVector(string& key, vector<string>& parts, vector<double>& rcosts){

  //print out the cost vector for a particular key , if it meets requirements
  //vector<string> parts;
  //Tokenize(key, parts, "#");
  bool islex = ispurelexical(parts[0]);  
  int cutoff = (islex) ? lexminfreq : nonlexminfreq;
  cout<<setprecision(8);
  //this checks simple frequency rather than the posterior
  if (rcosts[7]+0.1>=cutoff){
    //cout<<"initial: "<<rcosts[6]<<" "<<rcosts[7]<<endl;
    //cout<<"islex: "<<islex<<" "<<(*iter).first<<"#";
    cout<<key<<"#";
    //make sure to convert the freqs back into log space!!
    rcosts[6] = -log(rcosts[6]);
    rcosts[7] = -log(rcosts[7]);
    //cout << ios::setprecision(8); //8 points after decimal 
    copy(rcosts.begin(), rcosts.end(), ostream_iterator<double>(cout, " ")); 
    cout<<endl;
    numwritten_rule++;
    if (!islex){
      numwritten_rule_nonlex++;
    }
  }
  if (!islex){
    numunique_rule_nonlex++;
  }
  numunique_rule++;

}

void printAndReset(string& key, vector<string>& parts, int type){
  if (type==1){
    printRuleVector(key, parts, g_rcosts); //converts the 6,7 fields back
    //g_rcosts = vector<double>(totalnumfeatures, 0);
    g_rcosts[6] = 0;
    g_rcosts[7] = 0;

  }else if (type==2){
    printPhraseVector(key, parts, g_pcosts); //drops phrase counts 
    //g_pcosts = vector<double>(numfeatures_phrase, 0); //there are 1 costs associated with each phrase      
    for (int i=0; i<g_pcosts.size(); i++){
      g_pcosts[i] = 0;
    }
  }else if (type==3){
    printCountVector(key, g_ccosts); //drops the counts
    for (int i=0; i<g_ccosts.size(); i++){
      g_ccosts[i] = 0;
    }
  }
  numunique++;
  if (numlines % 100000 ==0 ){
    time_t nowtime = time(NULL);
    cerr<<"Processed "<<numlines<<" lines " <<" unique: "<<numunique<<" u_rule: "<<numunique_rule<<" w_rule: "<<numwritten_rule<<" u_rule_non: "<<numunique_rule_nonlex<<" w_rule_non: "<<numwritten_rule_nonlex<<" u_phrase: "<<numunique_phrase<<" w_phrase: "<<numwritten_phrase<<" key: "<<key<<" totalsec: "<<nowtime-startTime<<endl;
  }
  
}

pair<string, string> replaceWithSymbols(string str, string cats){
  vector<string> str_words;
  Tokenize(str, str_words, " ");    

  vector<string> cats_words;
  Tokenize(cats, cats_words, " ");
  
  string final = "";
  for(int i=0; i<str_words.size(); i++){
    if (str_words[i][0] == '@'){
      int index = int(str_words[i][1])-48-1;
      final += string("@")+cats_words[index];
    }else{
      final += str_words[i];
    }
    if (i!=str_words.size()-1){
      final += " ";
    }
  }
  return pair<string,string>(final,cats_words[cats_words.size()-1]);
}

//all of these rules have the same source side - so we can go thru them and accumulate counts for
//each of the following statistics for both the posterior and simple counts
// src frequency - with @1s 
// src and arg frequency - with @NPs 
// rule body - src and tgt with @1s
void processSameSrcBank(vector<AccRule>& rules){

  //since they all share the same src, no map here
  double src_simple=0;
  double src_posterior=0;

  map<string, double> srcandarg_simple;
  map<string, double> srcandarg_posterior;

  map<string, double> rulebody_simple;
  map<string, double> rulebody_posterior;

  bool debug = false;

  //go thru all rules with the same src side
  if (rules.size()>2 && debug){
    cout<<"SeeMe===="<<endl;
  }

  for (int i=0; i<rules.size(); i++){
    AccRule& rule = rules[i];

    //debug output
    if (debug){
      cout<<"ProcRule: "<<rule.first[0]<<"#"<<rule.first[1]<<"#"<<rule.first[2]<<" acc: ";
      copy(rule.second.begin(), rule.second.end(), ostream_iterator<double>(cout, " ")); 
      cout<<endl;
    }
    //increment the simple counts
    src_simple += rule.second[7];
    src_posterior += rule.second[6];

    vector<string> cats_andarg;
    Tokenize(rule.first[2], cats_andarg, " ");
    string cats = "";
    //we dont want the category
    for (int i=0; i<cats_andarg.size()-1; i++){
      cats = cats + cats_andarg[i];
      if (i!=cats_andarg.size()-2){
	cats = cats + " ";
      }
    }

    string srcandarg = rule.first[0]+string("#")+cats;
    string rulebody = rule.first[0]+string("#")+rule.first[1];

    //hash the freq counts
    srcandarg_simple[srcandarg] += rule.second[7];
    srcandarg_posterior[srcandarg] += rule.second[6];
    
    rulebody_simple[rulebody] += rule.second[7];
    rulebody_posterior[rulebody] += rule.second[6];
    if (debug){
      cout<<"\t\tIncremented: "<<"src: "<<src_posterior<<" srcandarg: "<<" key: "<<srcandarg<<" value: "<<srcandarg_posterior[srcandarg]<<" rulebody: "<<rulebody_posterior[rulebody]<<endl;
    }
  }
  
  //go thru all the rules
  for (int i=0; i<rules.size(); i++){
    AccRule& rule = rules[i];

    
    bool islex = ispurelexical(rule.first[0]);  
    int cutoff = (islex) ? lexminfreq : nonlexminfreq;
    cout<<setprecision(8);
    //we want to check the posterior count when making the cutoff decisions
    if (rule.second[7]+0.1>=cutoff){
      
      if (debug){
	cout<<"OutputRule: ";
      }
      
      
      //cout<<"Replacing: "<<rule.first[0]<<" WITH: "<<rule.first[2]<<endl;
      pair<string,string> srcAndResult = replaceWithSymbols(rule.first[0], rule.first[2]);
      //cout<<"Produced: "<<srcAndResult.first<<" CAT: "<<srcAndResult.second<<endl;
      //cout<<rule.first[0]<<"#"<<rule.first[1]<<"#"<<rule.first[2];
      cout<<srcAndResult.first<<"#"<<rule.first[1]<<"#@"<<srcAndResult.second<<"#";
      
      if (debug){
	cout<<" acc: ";    
	cout<<"simple: ";
      }
      cout<<rule.second[0]<<" "<<rule.second[1]<<" "<<rule.second[2]<<" ";
      
      if (debug){
	cout<<"result: ";
      }
      cout<<rule.second[3]<<" ";

      //in the future there should just be one pass of such calculations
      vector<string> cats_andarg;
      Tokenize(rule.first[2], cats_andarg, " ");
      string cats = "";
      //we dont want the category
      for (int i=0; i<cats_andarg.size()-1; i++){
	cats = cats + cats_andarg[i];
	if (i!=cats_andarg.size()-2){
	  cats = cats + " ";
	}
      }
      
      string srcandarg = rule.first[0]+string("#")+cats;
      string rulebody = rule.first[0]+string("#")+rule.first[1];
      
      
      double exclarg = rulebody_posterior[rulebody]/src_posterior;
      double inclarg = rule.second[6]/srcandarg_posterior[srcandarg]; //posterior
      if (debug){
      cout<<" prob: ";
      }
      cout<<-log(exclarg)<<" "<<-log(inclarg)<<" ";
      
      if (debug){
	cout<<" freq: ";
      }
      cout<<-log(rule.second[6])<<" "<<-log(rule.second[7]); //posterior and simple
      if (debug){
	cout<<" rbody: "<<rulebody_posterior[rulebody]<<" srcpost: "<<src_posterior<<" srcandarg: "<<srcandarg_posterior[srcandarg];
      }
      cout<<endl;
      //statistics to track output percentages
      numwritten_rule++;
      if (!islex){
	numwritten_rule_nonlex++;
      }
    }else{
      //cout<<"Dropping: "<<rule.first[0]<<"#"<<rule.first[1]<<"#"<<rule.first[2]<<" costs: ";
      //copy(rule.second.begin(), rule.second.end(), ostream_iterator<double>(cout, " ")); 
      //cout<<endl;

    }
    if (!islex){
      numunique_rule_nonlex++;
    }
    numunique_rule++;

  }


  
}

int main(int argc, char** argv){


  numlines = 0; //just tracking how many lines have been processed
  numunique = 0; //num print calls made (ie num numunique)
  numunique_rule = 0; numwritten_rule = 0;
  numunique_rule_nonlex = 0; numwritten_rule_nonlex = 0;
  numunique_phrase = 0; numwritten_phrase = 0;

  int numcountlines = 0;

  lexminfreq = atoi(argv[1]);
  nonlexminfreq = atoi(argv[2]);
  originalfeaturecount = atoi(argv[3]);
  totalnumfeatures = atoi(argv[4]);

  //do averaging of weights, dont do it if you have phrase counts
  doaverage = atoi(argv[5]);

  cerr<<"lexminfreq: "<<lexminfreq<<" nonlexminfreq: "<<nonlexminfreq<<" originalfeaturecount: "<<originalfeaturecount<<" totalnumfeatures: "<<totalnumfeatures<<" doaverage: "<<doaverage<<endl;

  //initialize the rcosts vector with 0's in all fields
  g_rcosts = vector<double>(totalnumfeatures, 0);
  g_rulecostvector = vector<double>(totalnumfeatures, 0);

  g_ccosts = vector<double>(numfeatures_count, 0); //there are 4 costs associated with each count
  g_countcostvector = vector<double>(numfeatures_count, 0); 

  g_pcosts = vector<double>(numfeatures_phrase, 0); //there are 2 costs associated with each phrase
  g_phrasecostvector = vector<double>(numfeatures_phrase, 0);


  
  vector< AccRule > g_samesrc; //keeps rules that have the same src side (keep in in already split up form)


  int lasttype = 0; //1: rules, 2: counts 3: phrase counts
  int dataset = 0; //marker indicating which part we have processed
  startTime =time(NULL);


  //structures to track the rule data
  string key, srckey;
  string lastkey = ""; //this lastkey variable is used to track the last type encountered
  string lastsrckey = "";

  //structures to track the count data
  //technically, we could have re-calculated all this count info in this script... but if you want
  //to pass thru some 
  string countkey;
  string lastcountkey;
  vector<string> lastcountparts;

  vector<string> lastparts; //parts data for the last key

  while (cin){
    string input_line;
    getline(cin, input_line);
    //cout<<"InputLine: "<<input_line<<endl;

    g_parts.clear();
    //char* input_line_str =(char*) input_line.c_str();
    //Tokenize(input_line_str, g_parts, "#");    
    Tokenize(input_line, g_parts, "#");    
    if (input_line == ""){
      continue;
    }
    numlines++;
    //cout<<"Read: "<<input_line<<endl;
    //===================================================================
    // Handle count data
    //===================================================================
    //@_COUNT#@1#@_S#val1 val2 val3 val4
    if (g_parts[0].find("@_COUNT") != string::npos){
      //now we start tracking based on 
      countkey = g_parts[2]; //get the NT category, this is what designates different count structures
      if (countkey != lastcountkey && numcountlines>1 ){
	printCountVector(lastcountkey, g_ccosts); //prints the lastcountkey and the accumulated counts
	//clear the count vector
	for (int i=0; i<g_ccosts.size(); i++){
	  g_ccosts[i] = 0;
	}
      }
      lastcountkey = countkey;
      lastcountparts = g_parts;
      lasttype = 3; //3 is the code for NTs
      handleCounts(g_parts, g_ccosts);
      numcountlines++;

    }else if (g_parts[0].find("@_PHRASE") != string::npos){

      //There is no more issue of handling all these phrase counts now
      //these will be estimated directly when the rules have been seen      
      //this block is only in there for backwards compat
      /*
      key = "";
      for (int i=0; i<g_parts.size()-1; i++){
	key = key + g_parts[i] + "#";
      }
      string countvals = g_parts[g_parts.size()-1];

      if (key != lastkey && numlines>1 ){
	printAndReset(lastkey, lastparts, lasttype);
      }

      lastkey = key;
      lastparts = g_parts;
      lasttype = 2;//2 is the code for phrase counts
      handlePhraseCounts(key, countvals, g_pcosts);      
      */
      
    }
    else{
      //rule mode - 
      if (g_parts.size()==4){

	//if this rule is new - src # tgt # result+categories #
        key = string(g_parts[0]+"#"+g_parts[1]+"#"+g_parts[2]);
	srckey = string(g_parts[0]);
	//cout<<"\tsrckey: "<<srckey<<"\tkey: "<<key<<endl;

	//if we see a new rule
	if (key != lastkey && numlines>1 ){

	  //now you have the uniqe last rule with lastparts and lastkey and g_rcosts
	  g_samesrc.push_back(AccRule(lastparts, g_rcosts));
	  //we might have seen a new src key as well
	  if (srckey != lastsrckey && numlines>1){
	    
	    //process the last src vector bank
	    /*
	    cout<<"Printing same src bank === "<<endl;
	    for (int i=0; i<g_samesrc.size(); i++){
	      AccRule& rule = g_samesrc[i];
	      cout<<"Rule: "<<rule.first[0]<<"#"<<rule.first[1]<<"#"<<rule.first[2]<<" acc: ";
	      copy(rule.second.begin(), rule.second.end(), ostream_iterator<double>(cout, " ")); 
	      cout<<endl;
	    }
	    cout<<endl;	
	    */
	    processSameSrcBank(g_samesrc);
	    
	    //clear the same src vector
	    g_samesrc.clear();
	    
	  }
	  
	  //reset the frequency costs
	  g_rcosts[6] = 0;
	  g_rcosts[7] = 0;
	}else{
	  //same exact rule form - we dont want to add it to any bank
	}
	lastkey = key;	
	lastsrckey = srckey;
	lastparts = g_parts;

	lasttype = 1;


	//accumulate and average the rules
	// the additional parameters to this handle function deal with averaging and maxing
	handleRules(key, g_parts, g_rcosts); 

      }else{
	cerr<<"WARNING: rule doesnt have 4 parts: "<<input_line<<endl;
	for (int i=0; i<g_parts.size(); i++){
	  cerr<<" Part: "<<i<<" "<<g_parts[i]<<endl;
	}	
      }
    }
  }

  //need to take care of printing the last count block - since we only print when we see a new one
  if (numcountlines>0){
    printCountVector(lastcountkey, g_ccosts);
  }

  //need to take care of printing the last same src block - since we only print when we see the next one
  //add the last rule that was seen into the same src block that is running
  //if the last rule was the same as the rules before it, g_rcosts have been accumulating
  g_samesrc.push_back(AccRule(lastparts, g_rcosts));
  processSameSrcBank(g_samesrc);
  
}


