

#ifndef __CString_h__
#define __CString_h__

//========================= Kopf ====================================
//
// Die Klasse CString bietet einen Ersatz fr die C-Stringroutinen

// Diese Klasse ist eine Modifikation der Klasse String aus:
//     Martin Aupperle
//     Programmierhandbuch Visual C++
//     Vieweg, Braunschweig 1994

// 92.08.01 : Version 1 (Au)
// 92.08.01 : Formulierung als Ableitung von CContainerItem (Au)
// 92.08.01 : Version 1.2: Mit Transferfunktionalit"at (Au)

// 97.09.05 : Version 2 (Vogel)
//            Nicht mehr von CContainerItem abgeleitet
//            Funktionsnamen ge"andert
//            Stream-Operatoren hinzugef"ugt



//========================= Definitionsabh"angigkeiten ===============
//
#include <iostream>
#include <iosfwd>

#include <assert.h>


#include <Types.h>



//========================= Export ==================================


//--------------------------------------------------------------------
//       class CString
//

class CString {

public:

  CString(); // erzeugt leeren CString
  ~CString();

  CString( char );
  CString( const char * );
  CString( const CString & );
  CString &operator = ( const CString & );
  CString &operator = ( const char * );

  //-- Diese Konstruktoren wandeln eine Zahl in einen CString
  //   Der Formatstring ist optional. Ist er nicht angegeben oder
  //   NULL, wird die Standardkonvertierung durchgef"uhrt.
  //
  //   value: die zu konvertierende Zahl
  //   fmt  : optionaler Formatstring

  CString( int value,       const char *fmt = NULL );
  CString( long value,      const char *fmt = NULL );
  CString( float value,     const char *fmt = NULL );
  CString( double value,    const char *fmt = NULL );


  // -------------------------------------------------------
  // Test, ob String g"ultig ist
  //-- liefert true, wenn das Objekt g"ultig ist
  bool IsValid() const;


  //-- liefert die aktuelle L"ange des Strings. F"ur ung"ultiges Objekt
  //   wird 0 geliefert.
  int GetLength() const;


  /////////////////////////////////////////////////////////////////////
  // Funktionen, die testen, ob ein String in eine Zahl konvertiert
  // werden kann, sowie Funktionen, die diese Konvertierung 
  // durchf"uhren
  /////////////////////////////////////////////////////////////////////

  //-- liefern true, wenn Objekt zu einer Zahl konvertiert
  //   werden kann. 
  //   value: Konvertierte Zahl, falls Konvertierung mglich, sonst undef.
  //   pos  : offset des ersten nicht mehr konvertierten Zeichens

  bool ToDouble( double &value, int &pos ) const;
  bool ToLong  ( long   &value, int &pos ) const;

  //-- Wandeln das Objekt in eine Zahl. Falls nicht m"oglich:
  //   liefern die Werte strnaL, strnaD zur"uck.

  operator long() const;
  operator double() const;

  //-- Diese Werte werden von den Konvertierungsoperatoren
  //   zur"uckgeliefert, wenn der CString nicht konvertiert werden kann
  static long strnaL;  //   str not a long
  static float strnaD; //   str not a double

  //-- liefern true, wenn Objekt zu einer Zahl (isNumeric) 
  //   bzw. zu einer Ganzzahl (isInteger) konvertiert werden kann. 
  bool IsNumeric() const;
  bool IsInteger() const;


  /////////////////////////////////////////////////////////////////////
  // Zugriff auf einzelne Zeichen innerhalb des Strings
  /////////////////////////////////////////////////////////////////////

  //-- erm"oglicht lesenden Zugriff auf interne Stringdarstellung
  /*virtual*/ operator const char *() const;

  //-- liefert true, wenn ofs g"ultig ist (d.h. CString[ ofs ] ein
  //   echtes Stringzeichen referenziert)
  bool IsValidIndex( int ofs ) const;

  //-- F"ur ung"ultiges Objekt oder ung"ultiges ofs in operator [] wird 
  //   Referenz auf einen Puffer mit diesem Wert geliefert.
  //   Puffer wird jedesmal neu mit noData initialisiert
  static char NoData;

  //-- Operator [] liefert Referenz auf Zeichen an pos ofs (oder Referenz
  //   auf noData falls Objekt oder ofs ung"ultig sind). Erlaubt schreibenden
  //   und lesenden Zugriff. Pr"uft G"utigkeit von ofs nur, wenn 
  //   VALIDITYCHECK definiert ist.
  char &operator [] ( int ofs );

  //-- Operator [] liefert Zeichen an Position ofs (oder noData
  //   falls Objekt oder ofs ung"ultig sind. Erlaubt jedoch 
  //   nur lesenden Zugriff, um Operator [] f"ur konstante Objekte verwenden
  //   zu k"onnen. Pr"uft G"ultigkeit von ofs nur, wenn 
  //   VALIDITYCHECK definiert ist.
  char operator [] ( int ofs ) const;


  /////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////

  //-- Der Vergleichsmodus bestimmt, wie der Vergleich zweier Strings
  //   durchgef"uhrt wird. Da mehrere Optionen gleichzeitig m"oglich sind,
  //   werden sie als Bits in einem int angegeben
  enum TCompareMode {
    cmpStandard     = 0x00,      // Normaler Vergleich
    cmpNoCase       = 0x01,      // ignorieren Gro-Kleinschreibung
    cmpNoLeading    = 0x02,      // ignorieren f"uhrende Leerzeichen
    cmpNoTrailing   = 0x04,      // ignorieren anh"angende Leerzeichen

    cmpNoBlanks     = cmpNoLeading + cmpNoTrailing,
    cmpNothing      = cmpNoBlanks +  cmpNoCase
  };

  //-- diese Variable bestimmt den aktuellen Vergleichsmodus
  //   gilt f"ur alle Strings im Programm, da static.
  static TCompareMode CompareMode;


  //-- compare vergleicht den Parameter mit dem eigenen Objekt,
  //   beachtet Vergleichsmodus. Optional kann Vergleichsmodus
  //   mit angegeben werden. Ergebnis wie Bibliotheksfunktion strcmp.
  //   ist ein Objekt ung"ultig oder kann Vergleich nicht durchgef"uhrt
  //   werden: notValid zur"uck.
  int Compare( const CString & ) const;
  int Compare( const CString &, TCompareMode ) const;

  //-- dieser Wert wird von compare zur"uckgegeben, wenn Vergleich
  //   nicht m"oglich war.
  static int NotValid;



  /////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////

  //-- pos liefert den offset von arg im eigenen Objekt.
  //   falls nicht gefunden: notFound zur"uck
  //   falls Objekt(e) fehlerhaft oder Suche nicht m"oglich: notValid zur"uck
  //   Beachtet Vergleichsmodus. Optional kann der Vergleichsmodus
  //   angegeben werden
  int Pos( const CString & ) const;
  int Pos( const CString &, TCompareMode ) const;

  //-- Der Wert dieser Variablen wird zur"uckgegeben, wenn nichts
  //   gefunden wird. (Note: Falls pos fehlschl"agt, weil Objekte 
  //   ung"ultig sind oder nicht durchgef"uhrt werden kann: notValid)
  static int NotFound;

  //-- subString extrahiert einen Teilstring der L"ange l ab Position
  //   ofs im eigenen Objekt. ofs, l k"onnen auerhalb der Grenzen
  //   liegen, Ergebnis ist dann der Leerstring.
  //   Die L"ange des extrahierten Strings kann kleiner als l sein.
  CString SubString( int ofs, int l ) const;

  //-- operator () ist ein alias f"ur subString
  CString operator () ( int ofs, int l ) const;



  /////////////////////////////////////////////////////////////////////
  // Manipulatoren f"ur Strings
  /////////////////////////////////////////////////////////////////////

  //-- append h"angt das Argument an die eigene Instanz an.
  //   liefert true wenn Operation erfolgreich war, false wenn nicht
  //   bzw. f"ur ung"ultige Objekte
  bool Append( const CString & );

  //-- Operatoren += h"angt Argument an die eigene Instanz an.
  //   liefern eigene Instanz zur"uck
  CString &operator += ( const CString & );

  //-- TrimTo erzeugt einen CString mit maximal der L"ange l,
  //   indem der CString hinten abgeschnitten wird, wenn er zu lang ist.
  //   Zu kurze Strings bleiben wie sie sind.
  bool TrimTo( int l );
  
  //-- FillTo erzeugt einen CString mit minimal der L"ange l, 
  //   indem der CString mit char aufgef"ullt wird, wenn er zu kurz ist. 
  //   Zu lange Strings bleiben wie sie sind.
  bool FillTo( int l, char = ' ' );

  //-- Insert f"ugt das Argument an offset ofs ein. 
  //   Falls ofs auerhalb der Stringgrenzen: false zur"uck.
  bool Insert( int ofs, const CString & );

  //-- wie Insert, jedoch wird bei einem Einf"ugen rechts vom Rand
  //   der CString zuerst auf die erforderliche L"ange gebracht. 
  //   Einf"ugen links vom Rand ist nicht erlaubt  
  bool InsertWithFill( int ofs, const CString &, char = ' ' );

  //-- Push f"ugt CString vor bestehendem CString ein, Pop extrahiert
  //   die ersten l Zeichen und gibt sie in statischem Puffer zur"uck
  bool Push( const CString & );
  CString Pop( int l );

  //-- Delete l"oscht l Zeichen ab Position ofs,
  //   l wird evtl. angepat, wenn ofs/l zu gro ist.
  //   bei ung"ultigem ofs (auerhalb der Grenzen) wird nichts gel"oscht,
  //   aber trotzdem true zur"uck
  bool Delete( int ofs, int l );

  //-- L"oscht f"uhrende bzw. anh"angende Leerzeichen
  bool TrimLeadingBlanks();
  bool TrimTrailingBlanks();

  //-- Konvertiert alle Zeichen in Gro/Kleinbuchstaben
  bool ToUpperCase();
  bool ToLowerCase();



  /////////////////////////////////////////////////////////////////////
  // Eingabe von Tastatur, Ausgabe an Stream
  /////////////////////////////////////////////////////////////////////

  //-- Einlesen der Instanz von der Tastatur
  //   Der Prompt ist optional. Ist er angegeben, wird er vor dem
  //   Einlesen ausgegeben
  void Read( const char *prompt = NULL );

  bool ToStream( std::ostream & ) const;


private:
  //-- Verwaltungsfunktionen -----------------------------------------

  //-- Init erzeugt den Grundzustand aus dem Zusatand "uninitialisiert"
  //   wird ausschlielich in den Konstruktoren verwendet
  void Init();


  //-- Die folgenden Funktionen gehen alle von initialisiertem Objekt aus.
  //   Objekt kann jedoch ung"ultig sein.

  //-- Assurem_Size stellt sicher, da mindestens newm_Size bytes 
  //   allokiert sind. 
  //   Vergr"oert Speicherbereich, falls notwendig. 
  //   Im Fehlerfall (kein Speicher mehr) wird Objekt ung"ultig.
  bool AssureSize( int newSize );


  //-- Set kopiert einen neuen CString ins Objekt. Pr"uft Parameter auf
  //   G"ultigkeit. Ist Parameter NULL, wird Objekt ung"ultig
  bool Set( const char *newStr );

  //-- versetzt das Objekt in den Zustand "ung"ultig". Evtl. allokierter 
  //   Speicher wird freigegeben.
  void Invalidate();

  //-- aus Effizienzgesichtspunkten wird der leere CString ohne allokiertes
  //   Byte implementiert.
  bool IsEmptyString() const;
  void SetEmptyString();

  static char EmptyBuf;

  //-- checkIndex pr"uft den Index ofs auf G"ultigkeit. Gibt Meldung
  //   aus, wenn Objekt ung"ultig oder ofs zu gro/klein ist
  bool CheckIndex( int ofs ) const;

  char *m_pData;  //-- Zeigt auf Zeichenkette auf Heap oder ist NULL
  int m_Size; //-- Gr"osse des zugewiesenen Speicherbereiches

  //-- Zeichenkette zur Repr"asentation eines ung"ultigen Objekts

  static const char *InvalidString;
};



/////////////////////////////////////////////////////////////////////
// Vergleichs-Operatoren
/////////////////////////////////////////////////////////////////////

//-- Diese Operatoren vergleichen zwei Strings und beachten dabei
//   den Vergleichsmodus

bool operator == ( const CString &, const CString & );
bool operator != ( const CString &, const CString & );
bool operator <  ( const CString &, const CString & );
bool operator <= ( const CString &, const CString & );
bool operator >  ( const CString &, const CString & );
bool operator >= ( const CString &, const CString & );



/////////////////////////////////////////////////////////////////////
// Additions-Operator
/////////////////////////////////////////////////////////////////////

//-- Operator + h"angt arg2 an arg1 an und liefert Ergebnis zur"uck.

CString operator + ( const CString &, const CString & );



/////////////////////////////////////////////////////////////////////
// Stream-Operatoren
/////////////////////////////////////////////////////////////////////

std::istream& operator>>(std::istream &InStream, CString &String);
std::ostream& operator<<(std::ostream &OutStream, const CString &String);






//========================= Inlines ===============================
//

//--------------------------------------------------------------------
//       CString:: IsValid
//
inline bool
CString::IsValid() const
{
  return m_pData != &NoData;
} // isValid



#ifndef VALIDITYCHECK

//--------------------------------------------------------------------
//       CString:: operator []
//

inline char &
CString::operator [] ( int ofs )
{
  return m_pData[ ofs ];
} // op []

inline char
CString::operator [] ( int ofs ) const
{
  return m_pData[ ofs ];
} // op []

#endif



//--------------------------------------------------------------------
//       CString:: ~CString
//

inline
CString::~CString()
{
  Invalidate();
} // dtor


//--------------------------------------------------------------------
//       CString:: IsValidIndex
//
inline bool
CString::IsValidIndex( int ofs ) const
{
  return IsValid() && ofs >= 0 && ofs < m_Size-1;
} // isValidIndex


//--------------------------------------------------------------------
//       CString:: Compare
//
inline int
CString::Compare( const CString &arg, TCompareMode CmpMode ) const
{
  CompareMode = CmpMode;
  return Compare( arg );
} // compare


//--------------------------------------------------------------------
//       CString:: Pos
//
inline int
CString::Pos( const CString &arg, TCompareMode CmpMode ) const
{
  CompareMode = CmpMode;
  return Pos( arg );
} // pos


//--------------------------------------------------------------------
//       CString:: operator ()
//
inline CString
CString::operator () (  int ofs, int l ) const
{
  return SubString( ofs, l );
} // op ()


//--------------------------------------------------------------------
//       CString:: operator +=
//
inline CString &
CString::operator += ( const CString &arg )
{
  Append( arg );
  return *this;
} // op +=


//--------------------------------------------------------------------
//       CString:: InsertWithFill
//
inline bool
CString::InsertWithFill( int ofs, const CString &arg, char c )
{
  return FillTo( ofs, c ) && Insert( ofs, arg );
} // insertWithFill


//--------------------------------------------------------------------
//       CString:: Push
//
inline bool
CString::Push( const CString &arg )
{
  return Insert( 0, arg );
} // push


//--------------------------------------------------------------------
//       CString:: Pop
//
inline CString
CString::Pop( int l )
{
  CString s = SubString( 0, l );
  Delete( 0, l ); 
  return s; 
} // pop


//--------------------------------------------------------------------
//       CString:: getUsedMem
//
/*
inline long
CString::GetUsedMem() const
{

  if ( IsValid() ) 
    return sizeof( CString ) + m_Size;
  else
    //-- wir nehmen (nicht ganz korrekt) an, da ein ung"ultiges Objekt
    //   auch keinen Speicher braucht
    return 0; 

} // getUsedMem
*/



//--------------------------------------------------------------------
//       CString:: Init
//
inline void
CString::Init()
{
  m_pData = &NoData;
  m_Size = 0;
} // init


//--------------------------------------------------------------------
//       CString:: IsEmptyString
//
inline bool
CString::IsEmptyString() const
{
  return m_pData == &EmptyBuf;
} // isEmptyString


//--------------------------------------------------------------------
//       CString:: SetEmptyString
//
inline void
CString::SetEmptyString()
{
  Invalidate(); // evtl. vorher allokierten Speicher l"oschen
  m_pData = &EmptyBuf;
  m_Size = 1;
} // setEmptyString



//--------------------------------------------------------------------
// Vergleichs-Operatoren
//
inline bool
operator == ( const CString &lhs, const CString &rhs )
{
  int result = lhs.Compare( rhs );
  return result == CString::NotValid ? false : bool( result == 0 );
} // op ==


inline bool
operator != ( const CString &lhs, const CString &rhs )
{
  int result = lhs.Compare( rhs );
  return result == CString::NotValid ? false : bool( result != 0 );  
} // op !=


inline bool
operator <  ( const CString &lhs, const CString &rhs )
{
  int result = lhs.Compare( rhs );
  return result == CString::NotValid ? false : bool( result < 0 );
} // op <


inline bool
operator <= ( const CString &lhs, const CString &rhs )
{
  int result = lhs.Compare( rhs );
  return result == CString::NotValid ? false : bool( result <= 0 ); 
} // op <=


inline bool
operator >  ( const CString &lhs, const CString &rhs )
{
  int result = lhs.Compare( rhs );
  return result == CString::NotValid ? false : bool( result > 0 ); 
} // op >


inline bool
operator >= ( const CString &lhs, const CString &rhs )
{
  int result = lhs.Compare( rhs );
  return result == CString::NotValid ? false : bool( result >= 0 ); 
} // op >=



//--------------------------------------------------------------------
// Stream-Operatoren
//

inline std::istream&
operator>>(std::istream& InStream, CString& /*String*/)
{
  assert( 0 );
  //char *pBuffer;
  //InStream >> pBuffer;
  //String = CString(pBuffer);
  return InStream;
}


inline std::ostream&
operator<<(std::ostream& OutStream, const CString& String)
{
  OutStream << (const char *)String;
  return OutStream;
}

#endif
