/**
   @file DefaultAllocator.hh

   Provides an allocator that simply calls new and delete.

   CVS id: $Id: DefaultAllocator.hh,v 1.6 2005/02/11 01:18:43 dietmar Exp $
 */
#ifndef DefaultAllocator_hh
#define DefaultAllocator_hh

// autoconf's symbols
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ObjectAllocator.hh>  // for base class

/**
   @brief A simple allocator that just calls new and delete.

   The allocator calls the class's default constructor to construct an object of class T
   during Create().

   class T must define a default constructor (that is, a constructor without any parameters)
   and a copy constructor (for Copy()).

   If your class T needs parameters to initialize corretctly, you might want to provide
   some kind of init method or derive a more specialized allocator class from CObjectAllocator,
   or something like that.
 */
template<class T>
class CDefaultAllocator : public CObjectAllocator<T> {
public:
  //! Allocate object, using new and default constructor.
  virtual T* Create()             { ++m_cObjects; ++m_cCreateCalls; return new T; }
  //! Copy object, using new and copy constructor.
  virtual T* Copy( const T& obj ) { ++m_cObjects; ++m_cCopyCalls;   return new T(obj); }
  //! Delete object.
  virtual void Del(T* deletable)  { assert(m_cObjects > 0); --m_cObjects; ++m_cDelCalls; delete deletable; }
}; // CDefaultAllocator

#endif // header already included?
