/*

Copyright (C) 1997,1998,1999,2000,2001  Franz Josef Och (RWTH Aachen - Lehrstuhl fuer Informatik VI)

This file is part of GIZA++ ( extension of GIZA ).

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
#include "Parameter.h"
#include <fstream>
//#include "unistd.h"
#include <strstream>


bool absolutePathNames=0;
string ParameterPathPrefix;
bool ParameterChangedFlag=0;

bool writeParameters(ofstream&of,const ParSet&parset,int level)
{
  if(!of)return 0;
  for(ParSet::const_iterator i=parset.begin();i!=parset.end();++i)
    {
      if(((*i)->getLevel()==level||level==-1)&&(*i)->onlyCopy==0)
	{
	  ostrstream os;
	  (*i)->printValue(os);
	  os << ends;
	  string s(os.str());
	  of << (*i)->getString() << " ";
	  if( absolutePathNames&&(*i)->isFilename()&&s.length()&&s[0]!='/' )
	    {
	      char path[1024];
	      getcwd(path,1024);
	      of << path << '/';
	    }
	  if( ParameterPathPrefix.length()&&(*i)->isFilename()&&s.length()&&s[0]!='/' )
	    of << ParameterPathPrefix << '/'; 
	  (*i)->printValue(of);
	  of << endl;
	}
    }
  return 1;
}

bool readParameters(ifstream&f,const ParSet&parset,int verb,int level)
{
  string s;
  if(!f)return 0;
  while(getline(f,s))
    {
      istrstream eingabe(s.c_str());
      string s1,s2;
      eingabe>>s1>>s2;
      if(makeSetCommand(s1,s2,parset,verb,level)==0)
	cerr << "ERROR: could not set: (C) " << s1 << " " << s2 << endl;
    }
  return 1;
}

 
bool makeSetCommand(string _s1,string s2,const ParSet&parset,int verb,int level)
{
  ParPtr anf;
  int anfset=0;
  string s1=simpleString(_s1);
  for(ParSet::const_iterator i=parset.begin();i!=parset.end();++i)
    {
      if( *(*i)==s1 )
	{
	  if( level==-1 || level==(*i)->getLevel() )
	    (*i)->setParameter(s2,verb);
	  else if(verb>1)
	    cerr << "ERROR: Could not set: (A) " << s1 << " " << s2 << " " << level << " " << (*i)->getLevel() << endl;
	  return 1;
	}
      else if( (*i)->getString().substr(0,s1.length())==s1 )
	{
	  anf=(*i);anfset++;
	}
    }
  if(anfset==1)
    {
      if( level==-1 || level==anf->getLevel() )
	anf->setParameter(s2,verb);
      else if( verb>1 )
	cerr << "ERROR: Could not set: (B) " << s1 << " " << s2 << " " << level << " " << anf->getLevel() << endl;
      return 1;
    }
  if( anfset>1 )
    cerr << "ERROR: ambiguous parameter '" << s1 << "'.\n";
  if( anfset==0 ){
    cerr << "ERROR: parameter '" << s1 << "' does not exist.\n";
    exit(-1);
  }
  return 0;
}

ostream& printPars(ostream&of,const ParSet&parset,int level)
{
  if(!of)return of;
  for(ParSet::const_iterator i=parset.begin();i!=parset.end();++i)
    {
      if(((*i)->getLevel()==level||level==-1)&&(*i)->onlyCopy==0)
	{
	  (*i)->printAt(of);
	  of << endl;
	}
    }
  return of;
}

//lowercases each character in the string
//removes all elements that are not a-z 0-9
string simpleString(const string s)
{
  string k;
  for(unsigned int i=0;i<s.length();++i)
    {
      char c[2];
      c[0]=tolower(s[i]);
      c[1]=0;
      if( (c[0]>='a'&&c[0]<='z')||(c[0]>='0'&&c[0]<='9')||(c[0]=='_'))
	k += c;
    }
  return k;
}


ParSet&getGlobalParSet()
{
  static ParSet x;
  return x;
}

void printGIZAPars(ostream&out)
{
  out << "general parameters:\n"
         "-------------------\n";
  printPars(out,getGlobalParSet(),-1);
  out << '\n';

  /*
  out << "No. of iterations:\n-"
         "------------------\n";
  printPars(out,getGlobalParSet(),PARLEV_ITER);
  out << '\n';

  out << "parameter for various heuristics in GIZA++ for efficient training:\n"
         "------------------------------------------------------------------\n";
  printPars(out,getGlobalParSet(),PARLEV_OPTHEUR);
  out << '\n';

  out << "parameters for describing the type and amount of output:\n"
         "-----------------------------------------------------------\n";
  printPars(out,getGlobalParSet(),PARLEV_OUTPUT);
  out << '\n';

  out << "parameters describing input files:\n"
         "----------------------------------\n";
  printPars(out,getGlobalParSet(),PARLEV_INPUT);
  out << '\n';

  out << "smoothing parameters:\n"
         "---------------------\n";
  printPars(out,getGlobalParSet(),PARLEV_SMOOTH);
  out << '\n';

  out << "parameters modifying the models:\n"
         "--------------------------------\n";
  printPars(out,getGlobalParSet(),PARLEV_MODELS);
  out << '\n';

  out << "parameters modifying the EM-algorithm:\n"
         "--------------------------------------\n";
  printPars(out,getGlobalParSet(),PARLEV_EM);
  */
  out << '\n';
}

void parseConfigFile (char * fname )
  // This functions reads in the configuration file to set up some run-time
  // parameters. The parameters are global variables that are defined in 
  // main.cc and used all over the place in the program
  // The format of the configuration file can be explained in the following way
  // FORMAT:
  // the character '\n' separates lines ..
  // lines that start with "//" (skipping over white spaces are considered 
  // as comments and will be ignored.
  // Any other line is considered as an attribute setting instruction and it 
  // is divided into haves (separated by a colon ":"). The first half is the
  // attribute value which consists of the concatenation of all non-white space
  // tokens before the colon. These tokens will have spaces eseparating them.
  // The attribute vlue is the first token after the colon (any thing after 
  // it will be ignored ;
  // For example :
  // if the configuration file has the following entry:
  //
  // NO.   ITERATIONS   MODEL 2 :	10
  //
  // then the attribute is "NO. ITERATIONS MODEL 2" , and the attribute value
  // is "10"  (these do not include the quotation marks).

{

  string line, word, attrib, attribval ;
  ifstream Config_File(fname);
  if(!Config_File){
    cerr << "ERROR:  Cannot open configuration file " << fname << "!\n" ;
    exit(1);
  }

  cout << "The following options are from the config file and will be overwritten by any command line options.\n";
  
  while(getline(Config_File, line)){

    istrstream buffer(line.c_str());
    word = attrib = attribval = "" ;
    buffer >> word  ;
    if (word != "//"){ // if line does not start with "//" (i.e. not a comment)
      attrib = word ;
      while((buffer >> word) && (word != ":")){
	attrib += " " + word ;
      }      
      if(!(buffer >> attribval))
	{
	  istrstream buffer2(line.c_str());
	  buffer2>>attrib;
	  buffer2>>attribval;
	}

      // This# is where (1) the configuration file is defined and
      //               (2) parsing of its attributes occurs.
      if ( line == ""){}
      else if(  !makeSetCommand(attrib,attribval,getGlobalParSet(),-1) )
	cerr << "ERROR: Unrecognized attribute :" << attrib << '\n';
    }
  }
}


void parseArguments(int argc, char *argv[])
{
  int arg = 1;

  if (argc==1){
    cout << "Usage (Minimal):" << " FastTranslateChart --SentenceList X --RuleDB --FeatureWeightsParsing" << '\n';
    printGIZAPars(cout);
    exit(0);
  }

  if(!strcmp(argv[1], "--h") || !strcmp(argv[1], "--help")){
    //printHelp();
    cout << "Usage:\n\n" << " FastTranslateChart --options1...n" << '\n';
    printGIZAPars(cout);
    exit(0);
  }
  if( argv[1][0]=='-' ){
    arg=0;
  }else{
    parseConfigFile(argv[1]);
  }

  //go thru each element in the argv. We can change this to support = if we make a new 
  //argv that replaces the = with a space
  while(++arg<argc){

    //arv: in the original GIZA code, --param would mean that that a TRUE flag was set
    //     now -- is used in the same way as - to signify a parameter with an argument
    //if( strlen(argv[arg])>2 && argv[arg][0]=='-' && argv[arg][1]=='-' )
    //{
    //arg+1 means the next parameter
    //if( !makeSetCommand(argv[arg],argv[arg+1],getGlobalParSet(),-1))
    //cerr << "WARNING: ignoring unrecognized option:  "<< argv[arg] << '\n' ;  
    //}
    //else 
    if( arg+1<argc && !makeSetCommand(argv[arg],argv[arg+1],getGlobalParSet(),-1))
      cerr << "WARNING: ignoring unrecognized option:  "<< argv[arg] << '\n' ;  
    else
      {
	arg++;
      }
  }
  //if( OPath.length() )
  //OPath+="/";
  //Prefix = (OPath + Prefix);
  //LogFilename = (OPath + LogFilename);
  printGIZAPars(cout);
}

